@@ -1,11 +1,10 @@
-{-# LANGUAGE CPP #-}
-
 -- | A program which either tries to add setup code for Agda's Emacs
 -- mode to the users .emacs file, or provides information to Emacs
 -- about where the Emacs mode is installed.
 
 module Main (main) where
 
+import Control.Applicative
 import Control.Exception
 import Control.Monad
 import Data.Char
@@ -17,9 +16,6 @@ import System.Environment
 import System.Exit
 import System.FilePath
 import System.IO
-#if !(MIN_VERSION_base(4,2,0))
-import qualified System.IO.UTF8 as UTF8
-#endif
 import System.Process
 
 import Paths_Agda (getDataDir, version)
@@ -37,37 +33,42 @@ main = do
              setupDotEmacs (Files { thisProgram = prog
                                   , dotEmacs    = dotEmacs
                                   })
+          | arg == compileFlag ->
+             compileElispFiles
     _  -> do inform usage
              exitFailure
 
 -- Command line options.
 
-setupFlag  = "setup"
-locateFlag = "locate"
+setupFlag   = "setup"
+locateFlag  = "locate"
+compileFlag = "compile"
 
 -- | Usage information.
 
 usage :: String
 usage = unlines
   [ "This program, which is part of Agda version " ++ ver ++ ", can be run"
-  , "in two modes, depending on which option it is invoked with:"
+  , "in three modes, depending on which option it is invoked with:"
   , ""
   , setupFlag
   , ""
   , "  The program tries to add setup code for Agda's Emacs mode to the"
   , "  current user's .emacs file. It is assumed that the .emacs file"
-#if MIN_VERSION_base(4,2,0)
   , "  uses the character encoding specified by the locale."
-#else
-  , "  uses ASCII or some other character encoding which ASCII is"
-  , "  compatible with (like Latin-1 or UTF-8)."
-#endif
   , ""
   , locateFlag
   , ""
   , "  The path to the Emacs mode's main file is printed on standard"
   , "  output (using the UTF-8 character encoding and no trailing"
   , "  newline)."
+  , ""
+  , compileFlag
+  , ""
+  , "  The program tries to compile Agda's Emacs mode's source files."
+  , ""
+  , "  WARNING: If you reinstall the Agda mode without recompiling the Emacs"
+  , "  Lisp files, then Emacs may continue using the old, compiled files."
   ]
 
 -- | The current version of Agda.
@@ -86,12 +87,8 @@ printEmacsModeFile :: IO ()
 printEmacsModeFile = do
   dataDir <- getDataDir
   let path = dataDir </> "emacs-mode" </> "agda2.el"
-#if MIN_VERSION_base(4,2,0)
   hSetEncoding stdout utf8
   putStr path
-#else
-  UTF8.putStr path
-#endif
 
 ------------------------------------------------------------------------
 -- Setting up the .emacs file
@@ -192,6 +189,41 @@ escape s = "\"" ++ concatMap esc s ++ "\""
         | isAscii c && isPrint c = [c]
         | otherwise              = "\\x" ++ showHex (fromEnum c) "\\ "
 
+------------------------------------------------------------------------
+-- Compiling Emacs Lisp files
+
+-- | The Agda mode's Emacs Lisp files, given in the order in which
+-- they should be compiled.
+
+emacsLispFiles :: [FilePath]
+emacsLispFiles =
+  [ "agda2-abbrevs.el"
+  , "annotation.el"
+  , "agda2-queue.el"
+  , "eri.el"
+  , "agda2.el"
+  , "agda-input.el"
+  , "agda2-highlight.el"
+  , "agda2-mode.el"
+  ]
+
+-- | Tries to compile the Agda mode's Emacs Lisp files.
+
+compileElispFiles :: IO ()
+compileElispFiles = do
+  dataDir <- (</> "emacs-mode") <$> getDataDir
+  let elFiles = map (dataDir </>) emacsLispFiles
+  elFiles <- filterM doesFileExist elFiles
+  exit <- rawSystem "emacs" $
+                    [ "--no-init-file", "--no-site-file"
+                    , "--directory", dataDir
+                    , "--batch"
+                    , "--funcall", "batch-byte-compile"
+                    ] ++ elFiles
+  unless (exit == ExitSuccess) $ do
+    informLn "Unable to compile Emacs Lisp files."
+    exitFailure
+
 ------------------------------------------------------------------------
 -- Helper functions
 
