@@ -20,7 +20,6 @@
 
 (require 'quail)
 (require 'cl)
-
 ;; Quail is quite stateful, so be careful when editing this code.  Note
 ;; that with-temp-buffer is used below whenever buffer-local state is
 ;; modified.
@@ -570,7 +569,7 @@ order for the change to take effect."
   ("Ge"  . ("ε"))  ("GE"  . ("Ε"))
   ("Gz"  . ("ζ"))  ("GZ"  . ("Ζ"))
   ;; \eta \Eta
-  ("Gth" . ("θ"))  ("GTH" . ("θ"))
+  ("Gth" . ("θ"))  ("GTH" . ("Θ"))
   ("Gi"  . ("ι"))  ("GI"  . ("Ι"))
   ("Gk"  . ("κ"))  ("GK"  . ("Κ"))
   ("Gl"  . ("λ"))  ("GL"  . ("Λ"))  ("Gl-" . ("ƛ"))
