@@ -57,10 +57,19 @@ position."
 
 (defun annotation-annotate (start end anns &optional info goto)
   "Annotate text between START and END in the current buffer.
-ANNS are the annotations to apply. All the symbols in ANNS are
-looked up in `annotation-bindings', and the font-lock-face text
-property for the given character range is set to the resulting
-list of faces. If the string INFO is non-nil, the mouse-face
+
+Nothing happens if either START or END are out of bounds for the
+current (possibly narrowed) buffer, or END <= START.
+
+If ANNS is nil, then those text properties between START and END
+that have been set by this function are deleted. Otherwise the
+following happens.
+
+All the symbols in ANNS are looked up in `annotation-bindings',
+and the font-lock-face text property for the given character
+range is set to the resulting list of faces.
+
+If the string INFO is non-nil, the mouse-face
 property is set to highlight, and INFO is used as the help-echo
 string. If GOTO has the form (FILENAME . POSITION), then the
 mouse-face property is set to highlight, and the given
@@ -75,93 +84,94 @@ annotation-annotated property set to t, and
 annotation-annotations is set to a list with all the properties
 that have been set; this ensures that the text properties can
 later be removed (if the annotation-* properties are not tampered
-with).
-
-Note finally that nothing happens if either START or END are out of
-bounds for the current (possibly narrowed) buffer, or END < START."
+with)."
   (incf start annotations-offset)
   (incf end annotations-offset)
   (when (and (<= (point-min) start)
              (< start end)
              (<= end (point-max)))
-    (let ((faces (delq nil
-                       (mapcar (lambda (ann)
-                                 (cdr (assoc ann annotation-bindings)))
-                               anns)))
-          (props nil))
-      (when faces
-        (put-text-property start end 'font-lock-face faces)
-        (add-to-list 'props 'font-lock-face))
-      (when (consp goto)
-        (add-text-properties start end
-                             `(annotation-goto ,goto
-                               mouse-face highlight))
-        (add-to-list 'props 'annotation-goto)
-        (add-to-list 'props 'mouse-face))
-      (when info
-        (add-text-properties start end
-                             `(mouse-face highlight help-echo ,info))
-        (add-to-list 'props 'mouse-face)
-        (add-to-list 'props 'help-echo))
-      (when props
-        (let ((pos start)
-              mid)
-          (while (< pos end)
-            (setq mid (next-single-property-change pos
-                         'annotation-annotations nil end))
-            (let* ((old-props (get-text-property pos 'annotation-annotations))
-                   (all-props (union old-props props)))
-              (add-text-properties pos mid
-                 `(annotation-annotated t annotation-annotations ,all-props))
-              (setq pos mid))))))))
+    (if (null anns)
+        (annotation-remove-annotations start end)
+      (let ((faces (delq nil
+                         (mapcar (lambda (ann)
+                                   (cdr (assoc ann annotation-bindings)))
+                                 anns)))
+            (props nil))
+        (when faces
+          (put-text-property start end 'font-lock-face faces)
+          (add-to-list 'props 'font-lock-face))
+        (when (consp goto)
+          (add-text-properties start end
+                               `(annotation-goto ,goto
+                                 mouse-face highlight))
+          (add-to-list 'props 'annotation-goto)
+          (add-to-list 'props 'mouse-face))
+        (when info
+          (add-text-properties start end
+                               `(mouse-face highlight help-echo ,info))
+          (add-to-list 'props 'mouse-face)
+          (add-to-list 'props 'help-echo))
+        (when props
+          (let ((pos start)
+                mid)
+            (while (< pos end)
+              (setq mid (next-single-property-change pos
+                           'annotation-annotations nil end))
+              (let* ((old-props (get-text-property pos 'annotation-annotations))
+                     (all-props (union old-props props)))
+                (add-text-properties pos mid
+                   `(annotation-annotated t annotation-annotations ,all-props))
+                (setq pos mid)))))))))
 
 (defmacro annotation-preserve-mod-p-and-undo (&rest code)
-  "Run CODE preserving both the undo data and the modification bit."
+  "Run CODE preserving both the undo data and the modification bit.
+Modification hooks are also disabled."
   (let ((modp (make-symbol "modp")))
   `(let ((,modp (buffer-modified-p))
          ;; Don't check if the file is being modified by some other process.
          (buffer-file-name nil)
          ;; Don't record those changes on the undo-log.
-         (buffer-undo-list t))
+         (buffer-undo-list t)
+         ;; Don't run modification hooks.
+         (inhibit-modification-hooks t))
      (unwind-protect
          (progn ,@code)
        (restore-buffer-modified-p ,modp)))))
 
-(defun annotation-remove-annotations ()
-  "Remove all text properties set by `annotation-annotate' in the current buffer.
-This function preserves the file modification stamp of the current buffer
-and does not modify the undo list.
+(defun annotation-remove-annotations (&optional start end)
+  "Remove all text properties set by `annotation-annotate'.
 
-Note: This function may fail if there is read-only text in the buffer."
+In the current buffer. If START and END are given, then
+properties are only removed between these positions.
+
+This function preserves the file modification stamp of the
+current buffer, does not modify the undo list, and temporarily
+disables all modification hooks.
+
+Note: This function may fail if there is read-only text in the
+buffer."
 
   ;; remove-text-properties fails for read-only text.
 
   (annotation-preserve-mod-p-and-undo
-   (let ((pos (point-min))
+   (let ((pos (or start (point-min)))
          pos2)
      (while pos
-       (setq pos2 (next-single-property-change pos 'annotation-annotated))
+       (setq pos2 (next-single-property-change pos 'annotation-annotated
+                                               nil end))
        (let ((props (get-text-property pos 'annotation-annotations)))
          (when props
            (remove-text-properties pos (or pos2 (point-max))
               (mapcan (lambda (prop) (list prop nil))
                       (append '(annotation-annotated annotation-annotations)
                               props)))))
-       (setq pos pos2)))))
+       (setq pos (unless (equal pos2 end) pos2))))))
 
-(defun annotation-load-file (file removep &optional goto-help)
-  "Apply the annotations in FILE.
-If FILE is empty, then this function does nothing; otherwise the
-following comments apply.
+(defun annotation-load (goto-help &rest cmds)
+  "Apply highlighting annotations in CMDS in the current buffer.
 
-If (`funcall' REMOVEP anns) is non-nil, then all existing text
-properties set by `annotation-annotate' in the current buffer are
-first removed. Here anns is a list containing all the
-annotations (third argument to `annotation-annotate') to be
-applied (in some order, with duplicates removed).
-
-FILE, if non-empty, should contain a list of lists (start end
-anns &optional info goto). Text between start and end will be
+The argument CMDS should be a list of lists (start end anns
+&optional info goto). Text between start and end will be
 annotated with the annotations in the list anns (using
 `annotation-annotate'). If info and/or goto are present they will
 be used as the corresponding arguments to `annotation-annotate'.
@@ -172,28 +182,19 @@ GOTO-HELP. The intention is that the default help text should
 inform the user about the \"goto\" facility.
 
 This function preserves the file modification stamp of the
-current buffer and does not modify the undo list.
+current buffer, does not modify the undo list, and temporarily
+disables all modification hooks.
 
-Note: This function may fail if there is read-only text in the buffer."
+Note: This function may fail if there is read-only text in the
+buffer."
   (annotation-preserve-mod-p-and-undo
-   (when (file-readable-p file)
-     (let ((cmds (with-temp-buffer
-                    (insert-file-contents file)
-                    (if (eq (point-min) (point-max))
-                        'empty-file
-                      (goto-char (point-min))
-                      (read (current-buffer))))))
-       (when (listp cmds)
-         (let ((anns (delete-dups
-                      (apply 'append (mapcar (lambda (x) (nth 2 x)) cmds)))))
-           (if (funcall removep anns)
-               (annotation-remove-annotations))
-           (dolist (cmd cmds)
-             (destructuring-bind (start end anns &optional info goto) cmd
-               (let ((info (if (and (not info) (consp goto))
-                               goto-help
-                             info)))
-                 (annotation-annotate start end anns info goto))))))))))
+    (when (listp cmds)
+      (dolist (cmd cmds)
+        (destructuring-bind (start end anns &optional info goto) cmd
+          (let ((info (if (and (not info) (consp goto))
+                          goto-help
+                        info)))
+            (annotation-annotate start end anns info goto)))))))
 
 (provide 'annotation)
 ;;; annotation.el ends here
