@@ -39,6 +39,7 @@ import Agda.Syntax.Scope.Monad (withCurrentModule)
 import qualified Agda.Syntax.Abstract.Name as AN
 import qualified Agda.TypeChecking.Monad.Base as MB
 import Agda.TypeChecking.EtaContract (etaContract)
+import qualified Agda.Utils.HashMap as HMap
 
 import Agda.Auto.Convert
 import Agda.Auto.NarrowingSearch
@@ -163,7 +164,7 @@ auto ii rng argstr = liftTCM $ do
                   [] -> case Map.elems tccons of
                    (m, mytype, mylocalVars, _) : [] -> do
                        defdfv <- case thisdefinfo of
-                                  Just (def, _, _) -> fromIntegral `liftM` getdfv mi def
+                                  Just (def, _, _) -> getdfv mi def
                                   Nothing -> return 0
                        ee <- liftIO $ newIORef $ ConstDef {cdname = "T", cdorigin = __IMPOSSIBLE__, cdtype = NotM $ Sort (Set 0), cdcont = Postulate, cddeffreevars = 0}
                        let modargs = drop (length mylocalVars - defdfv) mylocalVars
@@ -196,9 +197,9 @@ auto ii rng argstr = liftTCM $ do
                    Just (def, clause, _) -> do
                     let [rectyp'] = mymrectyp
                     defdfv <- getdfv mi def
-                    myrecdef <- liftIO $ newIORef $ ConstDef {cdname = "", cdorigin = (Nothing, def), cdtype = rectyp', cdcont = Postulate, cddeffreevars = fromIntegral defdfv}
+                    myrecdef <- liftIO $ newIORef $ ConstDef {cdname = "", cdorigin = (Nothing, def), cdtype = rectyp', cdcont = Postulate, cddeffreevars = defdfv}
                     (_, pats) <- constructPats cmap mi clause
-                    defdfv <- fromIntegral `liftM` getdfv mi def
+                    defdfv <- getdfv mi def
                     return $ if contains_constructor pats then
                       (Just (pats, myrecdef), defdfv)
                      else
@@ -285,7 +286,7 @@ auto ii rng argstr = liftTCM $ do
                  ticks <- liftIO $ newIORef 0
                  let [rectyp'] = mymrectyp
                  defdfv <- getdfv mi def
-                 myrecdef <- liftIO $ newIORef $ ConstDef {cdname = "", cdorigin = (Nothing, def), cdtype = rectyp', cdcont = Postulate, cddeffreevars = fromIntegral defdfv}
+                 myrecdef <- liftIO $ newIORef $ ConstDef {cdname = "", cdorigin = (Nothing, def), cdtype = rectyp', cdcont = Postulate, cddeffreevars = defdfv}
                  sols <- liftIO $ System.Timeout.timeout (timeout * 1000000) (
                     let r d = do
                          sols <- liftIO $ caseSplitSearch ticks __IMPOSSIBLE__ myhints meqr __IMPOSSIBLE__ d myrecdef ctx mytype pats
@@ -304,7 +305,7 @@ auto ii rng argstr = liftTCM $ do
                         -- Normalise the dot patterns
                         ps <- addCtxTel tel $ normalise ps
                         body <- etaContractBody body
-                        liftM modifyAbstractClause $ inContext [] $ reify $ NamedClause def $ I.Clause noRange tel perm ps body
+                        liftM modifyAbstractClause $ inContext [] $ reify $ AN.QNamed def $ I.Clause noRange tel perm ps body
                       ) cls'
                      pcs <- withInteractionId ii $ mapM prettyA cls''
                      ticks <- liftIO $ readIORef ticks
@@ -319,11 +320,11 @@ auto ii rng argstr = liftTCM $ do
 
             MRefine listmode -> do
              mv <- lookupMeta mi
-             let HasType _ tt = mvJudgement mv
+             let tt = jMetaType $ mvJudgement mv
                  minfo = getMetaInfo mv
              targettyp <- withMetaInfo minfo $ do
               vs <- getContextArgs
-              let targettype = tt `piApply` permute (takeP (fromIntegral $ length vs) $ mvPermutation mv) vs
+              let targettype = tt `piApply` permute (takeP (length vs) $ mvPermutation mv) vs
               normalise targettype
              let tctx = length $ envContext $ clEnv minfo
 
@@ -331,7 +332,7 @@ auto ii rng argstr = liftTCM $ do
                st <- liftTCM $ join $ pureTCM $ \st _ -> return st
                let defs = sigDefinitions $ stSignature st
                    idefs = sigDefinitions $ stImports st
-                   alldefs = Map.keys defs ++ Map.keys idefs
+                   alldefs = HMap.keys defs ++ HMap.keys idefs
                liftM catMaybes $ mapM (\n ->
                  case thisdefinfo of
                   Just (def, _, _) | def == n -> return Nothing
@@ -343,7 +344,7 @@ auto ii rng argstr = liftTCM $ do
                      c <- getConstInfo n
                      ctyp <- normalise $ defType c
                      cdfv <- withMetaInfo minfo $ getDefFreeVars n
-                     return $ case matchType cdfv (fromIntegral tctx) ctyp targettyp of
+                     return $ case matchType cdfv tctx ctyp targettyp of
                       Nothing -> Nothing
                       Just score -> Just (show cn, score)
                 ) alldefs
@@ -359,7 +360,7 @@ auto ii rng argstr = liftTCM $ do
                  c <- getConstInfo n
                  ctyp <- normalise $ defType c
                  cdfv <- withMetaInfo minfo $ getDefFreeVars n
-                 return $ case matchType cdfv (fromIntegral tctx) ctyp targettyp of
+                 return $ case matchType cdfv tctx ctyp targettyp of
                   Nothing -> Nothing
                   Just score -> Just (show cn, score)
                 ) modnames
