@@ -23,7 +23,7 @@ import Agda.TypeChecking.Monad.Signature (getConstInfo, getDefFreeVars, getImpor
 import Agda.Utils.Permutation (Permutation(Perm), idP, permute, takeP)
 import Agda.Interaction.BasicOps (rewrite, Rewrite(..))
 import Agda.TypeChecking.Level (reallyUnLevelView)
-import Agda.TypeChecking.Monad.Base (mvJudgement, mvPermutation, getMetaInfo, ctxEntry, envContext, clEnv, Judgement(HasType))
+import Agda.TypeChecking.Monad.Base (mvJudgement, mvPermutation, getMetaInfo, ctxEntry, envContext, clEnv)
 import Agda.TypeChecking.Monad.MetaVars (lookupMeta, withMetaInfo)
 import Agda.TypeChecking.Monad.Context (getContextArgs)
 import Agda.TypeChecking.Monad.Constraints (getAllConstraints)
@@ -33,6 +33,7 @@ import Agda.TypeChecking.Reduce (Normalise, normalise, instantiate)
 import Agda.TypeChecking.EtaContract (etaContract)
 import Agda.TypeChecking.Primitive (constructorForm)
 import Agda.TypeChecking.Free (freeIn)
+import qualified Agda.Utils.HashMap as HMap
 
 import Agda.Auto.NarrowingSearch
 import Agda.Auto.Syntax
@@ -98,7 +99,8 @@ tomy imi icns typs = do
        cons2 <- mapM (\con -> getConst True con TMAll) cons
        return (Datatype cons2 [], [])
       MB.Record {MB.recFields = fields, MB.recTel = tel} -> do -- the value of recPars seems unreliable or don't know what it signifies
-       let pars n (I.El _ (I.Pi it typ)) = C.Arg (C.argHiding it) (C.argRelevance it) (I.Var n []) : pars (n - 1) (I.unAbs typ)
+       let pars n (I.El _ (I.Pi it typ)) = C.Arg (C.domHiding it) (C.domRelevance it) (I.var n) : pars (n - 1) (I.unAbs typ)
+           pars n (I.El s (I.Shared p))  = pars n (I.El s (I.derefPtr p))
            pars _ (I.El _ _) = []
            contyp npar I.EmptyTel = I.El (I.mkType 0 {- arbitrary -}) (I.Def cn (pars (npar - 1) typ))
            contyp npar (I.ExtendTel it (I.Abs v tel)) = I.El (I.mkType 0 {- arbitrary -}) (I.Pi it (I.Abs v (contyp (npar + 1) tel)))
@@ -143,12 +145,12 @@ tomy imi icns typs = do
          m <- getMeta mi
          sol' <- tomyExp sol
          modify $ \s -> s {sEqs = (Map.insert (Map.size (fst $ sEqs s)) (Just (False, Meta m, sol')) (fst $ sEqs s), snd $ sEqs s)}
-       let HasType _ tt = mvJudgement mv
+       let tt = MB.jMetaType $ mvJudgement mv
            minfo = getMetaInfo mv
-           localVars = map (snd . C.unArg . ctxEntry) . envContext . clEnv $ minfo
+           localVars = map (snd . C.unDom . ctxEntry) . envContext . clEnv $ minfo
        (targettype, localVars) <- lift $ withMetaInfo minfo $ do
         vs <- getContextArgs
-        let targettype = tt `piApply` permute (takeP (fromIntegral $ length vs) $ mvPermutation mv) vs
+        let targettype = tt `piApply` permute (takeP (length vs) $ mvPermutation mv) vs
         targettype <- norm targettype
         localVars <- mapM norm localVars
         return (targettype, localVars)
@@ -204,9 +206,9 @@ getConst iscon name mode = do
     Nothing -> do
      mainm <- gets sMainMeta
      dfv <- lift $ getdfv mainm name
-     let nomi = fromIntegral $ I.arity (MB.defType def)
-     ccon <- lift $ liftIO $ newIORef (ConstDef {cdname = show name ++ ".CONS", cdorigin = (Just nomi, conname), cdtype = __IMPOSSIBLE__, cdcont = Constructor (nomi - fromIntegral dfv), cddeffreevars = fromIntegral dfv}) -- ?? correct value of deffreevars for records?
-     c <- lift $ liftIO $ newIORef (ConstDef {cdname = show name, cdorigin = (Nothing, name), cdtype = __IMPOSSIBLE__, cdcont = Datatype [ccon] [], cddeffreevars = fromIntegral dfv}) -- ?? correct value of deffreevars for records?
+     let nomi = I.arity (MB.defType def)
+     ccon <- lift $ liftIO $ newIORef (ConstDef {cdname = show name ++ ".CONS", cdorigin = (Just nomi, conname), cdtype = __IMPOSSIBLE__, cdcont = Constructor (nomi - dfv), cddeffreevars = dfv}) -- ?? correct value of deffreevars for records?
+     c <- lift $ liftIO $ newIORef (ConstDef {cdname = show name, cdorigin = (Nothing, name), cdtype = __IMPOSSIBLE__, cdcont = Datatype [ccon] [], cddeffreevars = dfv}) -- ?? correct value of deffreevars for records?
      modify (\s -> s {sConsts = (Map.insert name (mode, c) cmap, name : snd (sConsts s))})
      return $ if iscon then ccon else c
   _ -> do
@@ -217,12 +219,12 @@ getConst iscon name mode = do
     Nothing -> do
      (miscon, sname) <- if iscon then do
        let MB.Constructor {MB.conPars = npar, MB.conData = dname} = MB.theDef def
-       return (Just (fromIntegral npar), show dname ++ "." ++ show (I.qnameName name))
+       return (Just npar, show dname ++ "." ++ show (I.qnameName name))
       else
        return (Nothing, show name)
      mainm <- gets sMainMeta
      dfv <- lift $ getdfv mainm name
-     c <- lift $ liftIO $ newIORef (ConstDef {cdname = sname, cdorigin = (miscon, name), cdtype = __IMPOSSIBLE__, cdcont = __IMPOSSIBLE__, cddeffreevars = fromIntegral dfv})
+     c <- lift $ liftIO $ newIORef (ConstDef {cdname = sname, cdorigin = (miscon, name), cdtype = __IMPOSSIBLE__, cdcont = __IMPOSSIBLE__, cddeffreevars = dfv})
      modify (\s -> s {sConsts = (Map.insert name (mode, c) cmap, name : snd (sConsts s))})
      return c
 
@@ -346,7 +348,7 @@ tomyType (I.El _ t) = tomyExp t -- sort info is thrown away
 tomyExp :: I.Term -> TOM (MExp O)
 tomyExp (I.Var v as) = do
  as' <- tomyExps as
- return $ NotM $ App Nothing (NotM OKVal) (Var $ fromIntegral v) as'
+ return $ NotM $ App Nothing (NotM OKVal) (Var v) as'
 tomyExp (I.Lam hid b) = do
  b' <- tomyExp (I.absBody b)
  return $ NotM $ Lam (cnvh hid) (Abs (Id $ I.absName b) b')
@@ -367,7 +369,7 @@ tomyExp (I.Con name as) = do
  cc <- lift $ liftIO $ readIORef c
  let Just npar = fst $ cdorigin cc
  return $ NotM $ App Nothing (NotM OKVal) (Const c) (foldl (\x _ -> NotM $ ALConPar x) as' [1..npar])
-tomyExp (I.Pi (C.Arg hid _ x) b) = do
+tomyExp (I.Pi (C.Dom hid _ x) b) = do
  let y    = I.absBody b
      name = I.absName b
  x' <- tomyType x
@@ -388,6 +390,7 @@ tomyExp t@I.MetaV{} = do
    return $ Meta m
   _ -> tomyExp t
 tomyExp (I.DontCare _) = return $ NotM $ dontCare
+tomyExp (I.Shared p) = tomyExp $ I.derefPtr p
 
 tomyExps [] = return $ NotM ALNil
 tomyExps (C.Arg hid _ a : as) = do
@@ -410,10 +413,11 @@ fmExp m (I.Lit _) = False
 fmExp m (I.Level (I.Max as)) = any (fmLevel m) as
 fmExp m (I.Def _ as) = fmExps m as
 fmExp m (I.Con _ as) = fmExps m as
-fmExp m (I.Pi x y)  = fmType m (C.unArg x) || fmType m (I.unAbs y)
+fmExp m (I.Pi x y)  = fmType m (C.unDom x) || fmType m (I.unAbs y)
 fmExp m (I.Sort _) = False
 fmExp m (I.MetaV mid _) = mid == m
 fmExp m (I.DontCare _) = False
+fmExp m (I.Shared p) = fmExp m $ I.derefPtr p
 
 fmExps m [] = False
 fmExps m (a : as) = fmExp m (C.unArg a) || fmExps m as
@@ -453,7 +457,7 @@ frommyExp (Meta m) = do
 frommyExp (NotM e) =
  case e of
   App _ _ (Var v) as ->
-   frommyExps 0 as (I.Var (fromIntegral v) [])
+   frommyExps 0 as (I.Var v [])
   App _ _ (Const c) as -> do
    cdef <- lift $ readIORef c
    let (iscon, name) = cdorigin cdef
@@ -465,7 +469,7 @@ frommyExp (NotM e) =
   Pi _ hid _ x (Abs mid y) -> do
    x' <- frommyType x
    y' <- frommyType y
-   return $ I.Pi (C.Arg (icnvh hid) C.Relevant x') (I.Abs (case mid of {NoId -> "x"; Id id -> id}) y')
+   return $ I.Pi (C.Dom (icnvh hid) C.Relevant x') (I.Abs (case mid of {NoId -> "x"; Id id -> id}) y')
    -- maybe have case for Pi where possdep is False which produces Fun (and has to unweaken y), return $ I.Fun (C.Arg (icnvh hid) x') y'
   Sort (Set l) ->
    return $ I.Sort (I.mkType (fromIntegral l))
@@ -506,6 +510,7 @@ frommyExps ndrop (NotM as) trm =
   addend x (I.Var h xs) = I.Var h (xs ++ [x])
   addend x (I.Con h xs) = I.Con h (xs ++ [x])
   addend x (I.Def h xs) = I.Def h (xs ++ [x])
+  addend x (I.Shared p) = addend x (I.derefPtr p)
   addend _ _ = __IMPOSSIBLE__
 
 -- --------------------------------
@@ -562,7 +567,7 @@ frommyClause (ids, pats, mrhs) = do
       let Id id = mid
       tel <- ctel ctx
       t' <- frommyType t
-      return $ I.ExtendTel (C.Arg (icnvh hid) C.Relevant t') (I.Abs id tel)
+      return $ I.ExtendTel (C.Dom (icnvh hid) C.Relevant t') (I.Abs id tel)
  tel <- ctel $ reverse ids
  let getperms 0 [] perm nv = return (perm, nv)
      getperms n [] _ _ = __IMPOSSIBLE__
@@ -677,7 +682,7 @@ findClauseDeep :: I.MetaId -> MB.TCM (Maybe (AN.QName, I.Clause, Bool))
 findClauseDeep m = do
   sig <- getImportedSignature
   let res = do
-        def <- Map.elems $ MB.sigDefinitions sig
+        def <- HMap.elems $ MB.sigDefinitions sig
         MB.Function{MB.funClauses = cs} <- [MB.theDef def]
         c <- cs
         unless (peelbinds False findMeta $ I.clauseBody c) []
@@ -692,43 +697,44 @@ findClauseDeep m = do
                   I.NoBody -> d
                   I.Body e -> f e
     findMeta e =
-     case e of
+     case I.ignoreSharing e of
       I.Var _ as -> findMetas as
       I.Lam _ b -> findMeta (I.absBody b)
       I.Lit{} -> False
       I.Level (I.Max as) -> any (fmLevel m) as
       I.Def _ as -> findMetas as
       I.Con _ as -> findMetas as
-      I.Pi it ot -> findMetat (C.unArg it) || findMetat (I.unAbs ot)
+      I.Pi it ot -> findMetat (C.unDom it) || findMetat (I.unAbs ot)
       I.Sort{} -> False
       I.MetaV m' _  -> m == m'
       I.DontCare _ -> False
+      I.Shared{} -> __IMPOSSIBLE__
     findMetas = any (findMeta . C.unArg)
     findMetat (I.El _ e) = findMeta e
     toplevel e =
-     case e of
+     case I.ignoreSharing e of
       I.MetaV{} -> True
       _ -> False
 
 -- ---------------------------------------
 
-matchType :: Integer -> Integer -> I.Type -> I.Type -> Maybe (Nat, Nat) -- Nat is deffreevars of const, Nat is ctx length of target type, left arg is const type, right is target type
+matchType :: Int -> Int -> I.Type -> I.Type -> Maybe (Nat, Nat) -- Nat is deffreevars of const, Nat is ctx length of target type, left arg is const type, right is target type
 matchType cdfv tctx ctyp ttyp = trmodps cdfv ctyp
  where
   trmodps 0 ctyp = tr 0 0 ctyp
-  trmodps n ctyp = case ctyp of
-   I.El _ (I.Pi _ ot) -> trmodps (n - 1) (I.absBody ot)
+  trmodps n ctyp = case I.ignoreSharing $ I.unEl ctyp of
+   I.Pi _ ot -> trmodps (n - 1) (I.absBody ot)
    _ -> __IMPOSSIBLE__
   tr narg na ctyp =
    case ft 0 0 Just ctyp ttyp of
     Just n -> Just (n, narg)
-    Nothing -> case ctyp of
-     I.El _ (I.Pi _ (I.Abs _ ot)) -> tr (narg + 1) (na + 1) ot
-     I.El _ (I.Pi _ (I.NoAbs _ ot)) -> tr (narg + 1) na ot
+    Nothing -> case I.ignoreSharing $ I.unEl ctyp of
+     I.Pi _ (I.Abs _ ot) -> tr (narg + 1) (na + 1) ot
+     I.Pi _ (I.NoAbs _ ot) -> tr (narg + 1) na ot
      _ -> Nothing
    where
     ft nl n c (I.El _ e1) (I.El _ e2) = f nl n c e1 e2
-    f nl n c e1 e2 = case e1 of
+    f nl n c e1 e2 = case I.ignoreSharing e1 of
      I.Var v1 as1 | v1 < nl -> case e2 of
       I.Var v2 as2 | v1 == v2 -> fs nl (n + 1) c as1 as2
       _ -> Nothing
@@ -736,14 +742,14 @@ matchType cdfv tctx ctyp ttyp = trmodps cdfv ctyp
      I.Var v1 as1 -> case e2 of
       I.Var v2 as2 | cdfv + na + nl - v1 == tctx + nl - v2 -> fs nl (n + 1) c as1 as2
       _ -> Nothing
-     _ -> case (e1, e2) of
+     _ -> case (I.ignoreSharing e1, I.ignoreSharing e2) of
       (I.MetaV{}, _) -> c n
       (_, I.MetaV{}) -> c n
       (I.Lam hid1 b1, I.Lam hid2 b2) | hid1 == hid2 -> f (nl + 1) n c (I.absBody b1) (I.absBody b2)
       (I.Lit lit1, I.Lit lit2) | lit1 == lit2 -> c (n + 1)
       (I.Def n1 as1, I.Def n2 as2) | n1 == n2 -> fs nl (n + 1) c as1 as2
       (I.Con n1 as1, I.Con n2 as2) | n1 == n2 -> fs nl (n + 1) c as1 as2
-      (I.Pi (C.Arg hid1 rel1 it1) ot1, I.Pi (C.Arg hid2 rel2 it2) ot2) | hid1 == hid2 -> ft nl n (\n -> ft (nl + 1) n c (I.absBody ot1) (I.absBody ot2)) it1 it2
+      (I.Pi (C.Dom hid1 rel1 it1) ot1, I.Pi (C.Dom hid2 rel2 it2) ot2) | hid1 == hid2 -> ft nl n (\n -> ft (nl + 1) n c (I.absBody ot1) (I.absBody ot2)) it1 it2
       (I.Sort{}, I.Sort{}) -> c n -- sloppy
       _ -> Nothing
     fs nl n c es1 es2 = case (es1, es2) of
