@@ -13,23 +13,28 @@ import qualified Data.Set as S
 import Data.Maybe
 import Data.Monoid
 import qualified Language.Haskell.Exts.Syntax as Haskell
-import System.Directory
+import System.Directory ( canonicalizePath, createDirectoryIfMissing
+                        , getCurrentDirectory, setCurrentDirectory
+                        )
 import System.Exit
 import System.FilePath hiding (normalise)
 import System.Process hiding (env)
 
 import Paths_Agda
+import Agda.Compiler.CallCompiler
 import Agda.Compiler.MAlonzo.Pretty -- TODO: Move shared code to Compiler.*
 import Agda.Compiler.MAlonzo.Primitives (checkTypeOfMain)
 import Agda.Interaction.FindFile
 import Agda.Interaction.Options
 import Agda.Interaction.Imports
+import Agda.Syntax.Common (Delayed(..))
 import qualified Agda.Syntax.Concrete.Name as CN
 import Agda.Syntax.Internal hiding (Term(..))
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Monad.Builtin
 import Agda.TypeChecking.Serialise
 import Agda.Utils.FileName
+import qualified Agda.Utils.HashMap as HMap
 
 import qualified Agda.TypeChecking.Pretty as P
 
@@ -120,7 +125,7 @@ compileModule i = do
                         "Compiling: " ++ show (iModuleName i)
                     resetNameSupply
                     initialAnalysis i
-                    let defns = M.toList $ sigDefinitions $ iSignature i
+                    let defns = HMap.toList $ sigDefinitions $ iSignature i
                     -- Epic cannot parse files with no definitions
                     if (not $ null defns) then do
                         code <- compileDefns defns
@@ -142,7 +147,7 @@ initialAnalysis :: Interface -> Compile TCM ()
 initialAnalysis inter = do
   Prim.initialPrims
   modify $ \s -> s {curModule = mempty}
-  let defs = M.toList $ sigDefinitions $ iSignature inter
+  let defs = HMap.toList $ sigDefinitions $ iSignature inter
   forM_ defs $ \(q, def) -> do
     addDefName q
     case theDef def of
@@ -154,7 +159,7 @@ initialAnalysis inter = do
                 putConstrTag suc (PrimTag "primSuc")
             _ -> return ()
       Constructor {conPars = np} -> do
-        putForcedArgs q . drop (fromIntegral np) . ForceC.makeForcedArgs $ defType def
+        putForcedArgs q . drop np . ForceC.makeForcedArgs $ defType def
         putConArity q =<< lift (constructorArity q)
       f@(Function{}) -> do
         when ("main" == show (qnameName q)) $ do
@@ -256,10 +261,5 @@ callEpic' flags = do
           -- , "-trace"
           , "-i", dataDir </> "stdagda" <.> "c"
           ] ++ flags epicFlags
-    lift $ reportSLn "" 1 $
-      "calling: " ++ unwords (epic : epicCommand)
-    res <- liftIO $ system (unwords (epic : epicCommand))
-                  -- rawSystem epic epicCommand
-    case res of
-        ExitSuccess   -> return ()
-        ExitFailure i -> epicError $ "Epic returned " ++ show res ++ ", which should not happen. Please report this as a bug."
+
+    lift $ callCompiler epic epicCommand
