@@ -1,3 +1,4 @@
+{-# LANGUAGE CPP #-}
 module Agda.Compiler.MAlonzo.Primitives where
 
 import Control.Monad.Reader
@@ -18,29 +19,64 @@ import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Pretty
 import Agda.Utils.Monad
+import Agda.Utils.Impossible
+import qualified Agda.Utils.HashMap as HMap
 
+#include "../../undefined.h"
+
+{- OLD
 -- | Check that the main function has type IO a, for some a.
 checkTypeOfMain :: QName -> Type -> TCM ()
 checkTypeOfMain q ty
   | show (qnameName q) /= "main" = return ()
   | otherwise = do
-    Def io _ <- primIO
+    Def io _ <- ignoreSharing <$> primIO
     ty <- normalise ty
-    case unEl ty of
+    case ignoreSharing $ unEl ty of
       Def d _ | d == io -> return ()
       _                 -> do
         err <- fsep $
           pwords "The type of main should be" ++
           [prettyTCM io] ++ pwords " A, for some A. The given type is" ++ [prettyTCM ty]
         typeError $ GenericError $ show err
+-}
+
+-- | Check that the main function has type IO a, for some a.
+checkTypeOfMain :: QName -> Type -> TCM [HS.Decl] -> TCM [HS.Decl]
+checkTypeOfMain q ty ret
+  | show (qnameName q) /= "main" = ret
+  | otherwise = do
+    Def io _ <- ignoreSharing <$> primIO
+    ty <- normalise ty
+    case ignoreSharing $ unEl ty of
+      Def d _ | d == io -> (mainAlias :) <$> ret
+      _                 -> do
+        err <- fsep $
+          pwords "The type of main should be" ++
+          [prettyTCM io] ++ pwords " A, for some A. The given type is" ++ [prettyTCM ty]
+        typeError $ GenericError $ show err
+  where
+    mainAlias = HS.FunBind [HS.Match dummy mainLHS [] Nothing mainRHS $ HS.BDecls [] ]
+    mainLHS   = HS.Ident "main"
+    mainRHS   = HS.UnGuardedRhs $ HS.Var $ HS.UnQual $ unqhname "d" q
 
 -- Haskell modules to be imported for BUILT-INs
 importsForPrim :: TCM [HS.ModuleName]
 importsForPrim =
   xForPrim $
   L.map (\(s, ms) -> (s, return (L.map HS.ModuleName ms))) $
-  [ "CHAR" |-> ["Data.Char"]
-  -- , "IO" |-> ["System.IO"]
+  [ "CHAR"           |-> ["Data.Char"]
+  , "primIsDigit"    |-> ["Data.Char"]
+  , "primIsLower"    |-> ["Data.Char"]
+  , "primIsDigit"    |-> ["Data.Char"]
+  , "primIsAlpha"    |-> ["Data.Char"]
+  , "primIsSpace"    |-> ["Data.Char"]
+  , "primIsAscii"    |-> ["Data.Char"]
+  , "primIsLatin1"   |-> ["Data.Char"]
+  , "primIsPrint"    |-> ["Data.Char"]
+  , "primIsHexDigit" |-> ["Data.Char"]
+  , "primToUpper"    |-> ["Data.Char"]
+  , "primToLower"    |-> ["Data.Char"]
   ]
   where (|->) = (,)
 
@@ -107,10 +143,15 @@ mazHBoolToBool   = "mazHBoolToBool"
 
 xForPrim :: [(String, TCM [a])] -> TCM [a]
 xForPrim table = do
-  qs <- keys   <$> curDefs
+  qs <- HMap.keys <$> curDefs
   bs <- toList <$> gets stBuiltinThings
+  let getName (Builtin (Def q _))    = q
+      getName (Builtin (Con q _))    = q
+      getName (Builtin (Shared p))   = getName (Builtin $ derefPtr p)
+      getName (Builtin _)            = __IMPOSSIBLE__
+      getName (Prim (PrimFun q _ _)) = q
   concat <$> sequence [ maybe (return []) id $ L.lookup s table
-                        | (s, Builtin (Def q _)) <- bs, q `elem` qs ]
+                        | (s, def) <- bs, getName def `elem` qs ]
 
 
 -- Definition bodies for primitive functions
@@ -165,7 +206,7 @@ primBody s = maybe unimplemented (either (hsVarUQ . HS.Ident) id <$>) $
   , "primIsAscii"        |-> pred "Data.Char.isAscii"
   , "primIsLatin1"       |-> pred "Data.Char.isLatin1"
   , "primIsPrint"        |-> pred "Data.Char.isPrint"
-  , "primIsHExDigit"     |-> pred "Data.Char.isHexDigit"
+  , "primIsHexDigit"     |-> pred "Data.Char.isHexDigit"
   , "primToUpper"        |-> return "Data.Char.toUpper"
   , "primToLower"        |-> return "Data.Char.toLower"
   , "primCharToNat" |-> do toN <- bltQual' "NATURAL" mazIntToNat
@@ -218,7 +259,9 @@ primBody s = maybe unimplemented (either (hsVarUQ . HS.Ident) id <$>) $
   unimplemented = typeError $ NotImplemented s
 
   lam x t = Lam Hidden (Abs x t)
+{- UNUSED
   var x   = Arg Hidden Relevant (Var x [])
+-}
 
 ----------------------
 
