@@ -17,7 +17,7 @@ import qualified Agda.Syntax.Concrete as C -- ToDo: Remove with instance of ToCo
 import Agda.Syntax.Position
 import Agda.Syntax.Abstract as A hiding (Open)
 import Agda.Syntax.Common
-import Agda.Syntax.Info(ExprInfo(..),MetaInfo(..))
+import Agda.Syntax.Info(ExprInfo(..),MetaInfo(..),emptyMetaInfo)
 import Agda.Syntax.Internal as I
 import Agda.Syntax.Translation.InternalToAbstract
 import Agda.Syntax.Translation.AbstractToConcrete
@@ -29,7 +29,7 @@ import Agda.Syntax.Parser
 
 import Agda.TypeChecker
 import Agda.TypeChecking.Conversion
-import Agda.TypeChecking.Monad as M
+import Agda.TypeChecking.Monad as M hiding (MetaInfo)
 import Agda.TypeChecking.MetaVars
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Substitute
@@ -39,7 +39,7 @@ import Agda.TypeChecking.Coverage
 import Agda.TypeChecking.Records
 import Agda.TypeChecking.Irrelevance (wakeIrrelevantVars)
 import Agda.TypeChecking.Pretty (prettyTCM)
-import Agda.TypeChecking.Eliminators (unElim)
+-- UNUSED: import Agda.TypeChecking.Eliminators (unElim)
 import qualified Agda.TypeChecking.Pretty as TP
 
 import Agda.Utils.List
@@ -82,7 +82,8 @@ giveExpr mi e =
 		    let t' = t `piApply` ctx
 		    v	<- checkExpr e t'
 		    case mvInstantiation mv of
-			InstV v' -> equalTerm t' v (v' `apply` ctx)
+			InstV v' -> whenM ((Irrelevant /=) <$> asks envRelevance) $
+                                      equalTerm t' v (v' `apply` ctx)
 			_	 -> updateMeta mi v
 		    reify v
 		 IsSort{} -> __IMPOSSIBLE__
@@ -93,7 +94,7 @@ give ii mr e = liftTCM $ do
   mis <- getInteractionPoints
   r <- getInteractionRange ii
   updateMetaVarRange mi $ maybe r id mr
-  giveExpr mi e `catchError` \err -> case errError err of
+  giveExpr mi e `catchError` \err -> case err of
     PatternErr _ -> do
       err <- withInteractionId ii $ TP.text "Failed to give" TP.<+> prettyTCM e
       typeError $ GenericError $ show err
@@ -129,8 +130,9 @@ refine ii mr e =
 				    { Agda.Syntax.Info.metaRange = r
                                     , Agda.Syntax.Info.metaScope = scope { scopePrecedence = ArgumentCtx }
 				    , metaNumber = Nothing
+                                    , metaNameSuggestion = ""
 				    }
-                      in App (ExprRange $ r) e (defaultArg $ unnamed metaVar)
+                      in App (ExprRange $ r) e (defaultNamedArg metaVar)
                  --ToDo: The position of metaVar is not correct
                  --ToDo: The fixity of metavars is not correct -- fixed? MT
 
@@ -151,6 +153,7 @@ evalInMeta ii e =
 
 
 data Rewrite =  AsIs | Instantiated | HeadNormal | Normalised
+    deriving (Read)
 
 --rewrite :: Rewrite -> Term -> TCM Term
 rewrite AsIs	     t = return t
@@ -171,7 +174,7 @@ data OutputConstraint a b
       | JustSort b | CmpSorts Comparison b b
       | Guard (OutputConstraint a b) ProblemId
       | Assign b a | TypedAssign b a a
-      | IsEmptyType a | FindInScopeOF b
+      | IsEmptyType a | FindInScopeOF b a [(a,a)]
   deriving (Functor)
 
 -- | A subset of 'OutputConstraint'.
@@ -198,7 +201,7 @@ outputFormId (OutputForm _ o) = out o
       Assign i _           -> i
       TypedAssign i _ _    -> i
       IsEmptyType _        -> __IMPOSSIBLE__   -- Should never be used on IsEmpty constraints
-      FindInScopeOF _      -> __IMPOSSIBLE__
+      FindInScopeOF _ _ _  -> __IMPOSSIBLE__
 
 instance Reify ProblemConstraint (Closure (OutputForm Expr Expr)) where
   reify (PConstr pid cl) = enterClosure cl $ \c -> buildClosure =<< (OutputForm pid <$> reify c)
@@ -230,10 +233,14 @@ instance Reify Constraint (OutputConstraint Expr Expr) where
           OpenIFS{}  -> __IMPOSSIBLE__
           InstS{} -> __IMPOSSIBLE__
           InstV{} -> __IMPOSSIBLE__
-    reify (FindInScope m) = do
+    reify (FindInScope m cands) = do
       m' <- reify (MetaV m [])
-      return $ FindInScopeOF m' -- IFSTODO
-    reify (IsEmpty a) = IsEmptyType <$> reify a
+      ctxArgs <- getContextArgs
+      t <- getMetaType m
+      t' <- reify t
+      cands' <- mapM (\(tm,ty) -> (,) <$> reify tm <*> reify ty) cands
+      return $ FindInScopeOF m' t' cands' -- IFSTODO
+    reify (IsEmpty r a) = IsEmptyType <$> reify a
 
 showComparison :: Comparison -> String
 showComparison CmpEq  = " = "
@@ -257,7 +264,9 @@ instance (Show a,Show b) => Show (OutputConstraint a b) where
     show (Assign m e)           = show m ++ " := " ++ show e
     show (TypedAssign m e a)    = show m ++ " := " ++ show e ++ " :? " ++ show a
     show (IsEmptyType a)        = "Is empty: " ++ show a
-    show (FindInScopeOF s)      = "Find in Scope: " ++ show s
+    show (FindInScopeOF s t cs) = "Resolve implicit argument " ++ showCand (s,t) ++ ". Candidates: [" ++
+                                    intercalate ", " (map showCand cs) ++ "]"
+      where showCand (tm,ty) = show tm ++ " : " ++ show ty
 
 instance (ToConcrete a c, ToConcrete b d) =>
          ToConcrete (OutputForm a b) (OutputForm c d) where
@@ -285,7 +294,9 @@ instance (ToConcrete a c, ToConcrete b d) =>
     toConcrete (TypedAssign m e a) = TypedAssign <$> toConcrete m <*> toConcreteCtx TopCtx e
                                                                   <*> toConcreteCtx TopCtx a
     toConcrete (IsEmptyType a) = IsEmptyType <$> toConcreteCtx TopCtx a
-    toConcrete (FindInScopeOF s) = FindInScopeOF <$> toConcrete s
+    toConcrete (FindInScopeOF s t cs) =
+      FindInScopeOF <$> toConcrete s <*> toConcrete t
+                    <*> mapM (\(tm,ty) -> (,) <$> toConcrete tm <*> toConcrete ty) cs
 
 instance (Pretty a, Pretty b) => Pretty (OutputConstraint' a b) where
   pretty (OfType' e t) = pretty e <+> text ":" <+> pretty t
@@ -297,8 +308,14 @@ instance (ToConcrete a c, ToConcrete b d) =>
 --ToDo: Move somewhere else
 instance ToConcrete InteractionId C.Expr where
     toConcrete (InteractionId i) = return $ C.QuestionMark noRange (Just i)
+{- UNUSED
 instance ToConcrete MetaId C.Expr where
-    toConcrete (MetaId i) = return $ C.Underscore noRange (Just i)
+    toConcrete x@(MetaId i) = do
+      return $ C.Underscore noRange (Just $ "_" ++ show i)
+-}
+instance ToConcrete NamedMeta C.Expr where
+    toConcrete i = do
+      return $ C.Underscore noRange (Just $ show i)
 
 judgToOutputForm :: Judgement a c -> OutputConstraint a c
 judgToOutputForm (HasType e t) = OfType e t
@@ -316,7 +333,7 @@ getConstraints = liftTCM $ do
     toOutputForm (ii, mi, e) = do
       mv <- getMetaInfo <$> lookupMeta mi
       withMetaInfo mv $ do
-        let m = QuestionMark $ MetaInfo noRange emptyScopeInfo (Just $ fromIntegral ii)
+        let m = QuestionMark $ emptyMetaInfo { metaNumber = Just $ fromIntegral ii }
         abstractToConcrete_ $ OutputForm 0 $ Assign m e
 
 getSolvedInteractionPoints :: TCM [(InteractionId, MetaId, Expr)]
@@ -340,25 +357,30 @@ getSolvedInteractionPoints = do
           BlockedConst{}                 -> unsol
           PostponedTypeCheckingProblem{} -> unsol
 
-typeOfMetaMI :: Rewrite -> MetaId -> TCM (OutputConstraint Expr MetaId)
+typeOfMetaMI :: Rewrite -> MetaId -> TCM (OutputConstraint Expr NamedMeta)
 typeOfMetaMI norm mi =
      do mv <- lookupMeta mi
 	withMetaInfo (getMetaInfo mv) $
 	  rewriteJudg mv (mvJudgement mv)
    where
     rewriteJudg mv (HasType i t) = do
+      ms <- getMetaNameSuggestion i
       t <- rewrite norm t
       vs <- getContextArgs
+      let x = NamedMeta ms i
       reportSDoc "interactive.meta" 10 $ TP.vcat
         [ TP.text $ unwords ["permuting", show i, "with", show $ mvPermutation mv]
         , TP.nest 2 $ TP.vcat
           [ TP.text "len  =" TP.<+> TP.text (show $ length vs)
           , TP.text "args =" TP.<+> prettyTCM vs
           , TP.text "t    =" TP.<+> prettyTCM t
+          , TP.text "x    =" TP.<+> TP.text (show x)
           ]
         ]
-      OfType i <$> reify (t `piApply` permute (takeP (size vs) $ mvPermutation mv) vs)
-    rewriteJudg mv (IsSort i t) = return $ JustSort i
+      OfType x <$> reify (t `piApply` permute (takeP (size vs) $ mvPermutation mv) vs)
+    rewriteJudg mv (IsSort i t) = do
+      ms <- getMetaNameSuggestion i
+      return $ JustSort $ NamedMeta ms i
 
 
 typeOfMeta :: Rewrite -> InteractionId -> TCM (OutputConstraint Expr InteractionId)
@@ -371,7 +393,7 @@ typesOfVisibleMetas :: Rewrite -> TCM [OutputConstraint Expr InteractionId]
 typesOfVisibleMetas norm =
   liftTCM $ mapM (typeOfMeta norm) =<< getInteractionPoints
 
-typesOfHiddenMetas :: Rewrite -> TCM [OutputConstraint Expr MetaId]
+typesOfHiddenMetas :: Rewrite -> TCM [OutputConstraint Expr NamedMeta]
 typesOfHiddenMetas norm = liftTCM $ do
   is    <- getInteractionMetas
   store <- Map.filterWithKey (openAndImplicit is) <$> getMetaStore
@@ -394,7 +416,7 @@ contextOfMeta ii norm = do
 	visible _	     = __IMPOSSIBLE__
         reifyContext xs = reverse <$> zipWithM out [1..] xs
 
-        out i (Arg h _ (x, t)) = escapeContext i $ do
+        out i (Dom h _ (x, t)) = escapeContext i $ do
           t' <- reify =<< rewrite norm t
           return $ OfType x t'
 
@@ -424,22 +446,22 @@ withInteractionId i ret = do
 
 withMetaId :: MetaId -> TCM a -> TCM a
 withMetaId m ret = do
-  info <- lookupMeta m
-  withMetaInfo (mvInfo info) ret
+  mv <- lookupMeta m
+  withMetaInfo' mv ret
 
 -- The intro tactic
 
 -- Returns the terms (as strings) that can be
 -- used to refine the goal. Uses the coverage checker
 -- to find out which constructors are possible.
-introTactic :: InteractionId -> TCM [String]
-introTactic ii = do
+introTactic :: Bool -> InteractionId -> TCM [String]
+introTactic pmLambda ii = do
   mi <- lookupInteractionId ii
   mv <- lookupMeta mi
   withMetaInfo (getMetaInfo mv) $ case mvJudgement mv of
     HasType _ t -> do
         t <- reduce =<< piApply t <$> getContextArgs
-        case unEl t of
+        case ignoreSharing $ unEl t of
           I.Def d _ -> do
             def <- getConstInfo d
             case theDef def of
@@ -464,27 +486,34 @@ introTactic ii = do
 
     introFun tel = addCtxTel tel' $ do
         imp <- showImplicitArguments
-        let okHiding h = imp || h == NotHidden
-        vars <- mapM showTCM [ Arg h Relevant (I.Var i [])
-                             | (h, i) <- zip hs $ reverse [0..n - 1]
-                             , okHiding h
-                             ]
-        return [ unwords $ ["λ"] ++ vars ++ ["→", "?"] ]
+        let okHiding0 h = imp || h == NotHidden
+            -- if none of the vars were displayed, we would get a parse error
+            -- thus, we switch to displaying all
+            allHidden   = null (filter okHiding0 hs)
+            okHiding    = if allHidden then const True else okHiding0
+        vars <- -- setShowImplicitArguments (imp || allHidden) $
+                (if allHidden then withShowAllArguments else id) $
+                  mapM showTCM [ Arg h Relevant (var i)
+                               | (h, i) <- zip hs $ downFrom n
+                               , okHiding h
+                               ]
+        if pmLambda
+           then return [ unwords $ ["λ", "{"] ++ vars ++ ["→", "?", "}"] ]
+           else return [ unwords $ ["λ"]      ++ vars ++ ["→", "?"] ]
       where
         n = size tel
-        hs   = map argHiding $ telToList tel
+        hs   = map domHiding $ telToList tel
         tel' = telFromList [ fmap makeName b | b <- telToList tel ]
         makeName ("_", t) = ("x", t)
         makeName (x, t)   = (x, t)
 
     introData t = do
-      let tel  = telFromList [defaultArg ("_", t)]
-          perm = idP 1
+      let tel  = telFromList [domFromArg $ defaultArg ("_", t)]
           pat  = [defaultArg (I.VarP "c")]
-      r <- split CoInductive tel perm pat 0
+      r <- splitLast CoInductive tel pat
       case r of
         Left err -> return []
-        Right cs -> mapM showTCM $ concatMap (conName . scPats) cs
+        Right cov -> mapM showTCM $ concatMap (conName . scPats) $ splitClauses cov
 
     introRec d = do
       hfs <- getRecordFieldNames d
@@ -517,7 +546,7 @@ atTopLevel m = inConcreteMode $ do
           M.withCurrentModule current $
             withScope_ scope $
               addContext (zipWith' (fmap . (,))
-                                   (map snd $ scopeLocals scope)
+                                   (reverse $ map snd $ scopeLocals scope)
                                    (map (fmap snd) $ telToList tel)) $
                 m
 
