@@ -4,6 +4,7 @@ module Agda.Interaction.Highlighting.Vim where
 
 import Control.Monad.Trans
 import Data.Char
+import Data.Function ( on )
 import Data.Set ( Set )
 import Data.Map ( Map )
 import System.FilePath
@@ -23,7 +24,8 @@ import Agda.Utils.Tuple
 #include "../../undefined.h"
 import Agda.Utils.Impossible
 
-on f g x y = f (g x) (g y)
+-- defined in Data.Function
+-- on f g x y = f (g x) (g y)
 
 vimFile :: FilePath -> FilePath
 vimFile file =
@@ -46,8 +48,8 @@ match _ [] = ""
 match cat ws	= "syn match " ++ cat ++ " \"" ++
 		    concat (List.intersperse "\\|" $ map escape ws) ++ "\""
 
-matches :: [String] -> [String] -> [String] -> [String] -> [String]
-matches cons icons defs idefs =
+matches :: [String] -> [String] -> [String] -> [String] -> [String] -> [String] -> [String]
+matches cons icons defs idefs flds iflds =
     map snd
     $ List.sortBy (compare `on` fst)
     $ cons' ++ defs' ++ icons' ++ idefs'
@@ -56,6 +58,8 @@ matches cons icons defs idefs =
 	icons' = foo "agdaInfixConstructor" $ classify length icons
 	defs'  = foo "agdaFunction"	    $ classify length defs
 	idefs' = foo "agdaInfixFunction"    $ classify length idefs
+	flds'  = foo "agdaProjection"	    $ classify length flds
+	iflds' = foo "agdaInfixProjection"  $ classify length iflds
 
 	classify f = List.groupBy ((==) `on` f)
 		     . List.sortBy (compare `on` f)
@@ -64,16 +68,19 @@ matches cons icons defs idefs =
 	foo cat = map (length . head /\ match cat)
 
 toVim :: NamesInScope -> String
-toVim ns = unlines $ matches mcons micons mdefs midefs
+toVim ns = unlines $ matches mcons micons mdefs midefs mflds miflds
     where
 	cons = [ x | (x, def:_) <- Map.toList ns, anameKind def == ConName ]
 	defs = [ x | (x, def:_) <- Map.toList ns, anameKind def == DefName ]
+	flds = [ x | (x, fld:_) <- Map.toList ns, anameKind fld == FldName ]
 
 	mcons = map show cons
 	mdefs = map show defs
+	mflds = map show flds
 
 	micons = concatMap parts cons
 	midefs = concatMap parts defs
+	miflds = concatMap parts flds
 
 	parts (NoName _ _) = []
 	parts (Name _ [_]) = []
