@@ -4,23 +4,26 @@
 -}
 module Agda.Interaction.Imports where
 
-import Prelude hiding (catch)
+import Prelude
 
 import Control.Arrow
 import Control.Monad.Error
+import Control.Monad.Reader
 import Control.Monad.State
 import qualified Control.Exception as E
+
+import Data.Function (on)
 import qualified Data.Map as Map
 import qualified Data.List as List
 import qualified Data.Set as Set
+import qualified Data.Foldable as Fold (toList)
 import qualified Data.ByteString.Lazy as BS
 import Data.List
 import Data.Maybe
+import Data.Monoid (mempty, mappend)
 import Data.Map (Map)
 import Data.Set (Set)
-import System.Directory
-import System.Time
-import qualified Agda.Utils.IO.Locale as LocIO
+import System.Directory (doesFileExist, getModificationTime, removeFile)
 import System.FilePath hiding (splitPath)
 
 import Agda.Syntax.Position
@@ -35,6 +38,7 @@ import Agda.Syntax.Internal
 
 import Agda.Termination.TermCheck
 
+import Agda.TypeChecking.Errors
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Monad.Builtin
@@ -42,8 +46,10 @@ import Agda.TypeChecking.Serialise
 import Agda.TypeChecking.Primitive
 import Agda.TypeChecker
 
+import Agda.Interaction.EmacsCommand
 import Agda.Interaction.FindFile
 import Agda.Interaction.Options
+import Agda.Interaction.Response (Response(Resp_RunningInfo))
 import Agda.Interaction.Highlighting.Precise (HighlightingInfo)
 import Agda.Interaction.Highlighting.Generate
 import Agda.Interaction.Highlighting.Vim
@@ -54,6 +60,7 @@ import Agda.Utils.Monad
 import Agda.Utils.IO.Binary
 import Agda.Utils.Pretty
 import Agda.Utils.Fresh
+import Agda.Utils.Time
 
 import Agda.Utils.Impossible
 #include "../undefined.h"
@@ -79,24 +86,25 @@ mergeInterface i = do
             Just b1 = Map.lookup b bs
             Just b2 = Map.lookup b bi
     mapM_ check (map fst $ Map.toList $ Map.intersection bs bi)
-    addImportedThings sig bi (iHaskellImports i)
+    addImportedThings sig bi (iHaskellImports i) (iPatternSyns i)
     reportSLn "import.iface.merge" 20 $
       "  Rebinding primitives " ++ show prim
     prim <- Map.fromList <$> mapM rebind prim
     modify $ \st -> st { stImportedBuiltins = stImportedBuiltins st `Map.union` prim
 		       }
     where
-	rebind x = do
+	rebind (x, q) = do
 	    PrimImpl _ pf <- lookupPrimitiveFunction x
-	    return (x, Prim pf)
+	    return (x, Prim $ pf { primFunName = q })
 
 addImportedThings ::
-  Signature -> BuiltinThings PrimFun -> Set String -> TCM ()
-addImportedThings isig ibuiltin hsImports =
+  Signature -> BuiltinThings PrimFun -> Set String -> A.PatternSynDefns -> TCM ()
+addImportedThings isig ibuiltin hsImports patsyns =
   modify $ \st -> st
     { stImports          = unionSignatures [stImports st, isig]
     , stImportedBuiltins = Map.union (stImportedBuiltins st) ibuiltin
     , stHaskellImports   = Set.union (stHaskellImports st) hsImports
+    , stPatternSynImports = Map.union (stPatternSynImports st) patsyns
     }
 
 -- | Scope checks the given module. A proper version of the module
@@ -107,7 +115,7 @@ scopeCheckImport x = do
     reportSLn "import.scope" 5 $ "Scope checking " ++ show x
     verboseS "import.scope" 10 $ do
       visited <- Map.keys <$> getVisitedModules
-      liftIO $ LocIO.putStrLn $
+      reportSLn "" 0 $
         "  visited: " ++ intercalate ", " (map (render . pretty) visited)
     i <- fst <$> getInterface x
     addImport x
@@ -149,30 +157,6 @@ alreadyVisited x getIface = do
                 }
           return r
 
--- | Warnings.
---
--- Invariant: The fields are never empty at the same time.
-
-data Warnings = Warnings
-  { terminationProblems   :: [TerminationError]
-    -- ^ Termination checking problems are not reported if
-    -- 'optTerminationCheck' is 'False'.
-  , unsolvedMetaVariables :: [Range]
-    -- ^ Meta-variable problems are reported as type errors unless
-    -- 'optAllowUnsolved' is 'True'.
-  , unsolvedConstraints   :: Constraints
-    -- ^ Same as 'unsolvedMetaVariables'.
-  }
-
--- | Turns warnings into an error. Even if several errors are possible
--- only one is raised.
-
-warningsToError :: Warnings -> TypeError
-warningsToError (Warnings [] [] [])    = __IMPOSSIBLE__
-warningsToError (Warnings _ w@(_:_) _) = UnsolvedMetas w
-warningsToError (Warnings _ _ w@(_:_)) = UnsolvedConstraints w
-warningsToError (Warnings w@(_:_) _ _) = TerminationCheckFailed w
-
 -- | Type checks the given module (if necessary).
 
 typeCheck :: AbsolutePath -> TCM (Interface, Maybe Warnings)
@@ -206,9 +190,10 @@ getInterface' :: C.TopLevelModuleName
                        -- be preserved?
               -> TCM (Interface, Either Warnings ClockTime)
 getInterface' x includeStateChanges =
+  withIncreasedModuleNestingLevel $
   -- Preserve the pragma options unless includeStateChanges is True.
-  bracket (stPragmaOptions <$> get)
-          (unless includeStateChanges . setPragmaOptions) $ \_ -> do
+  bracket_ (stPragmaOptions <$> get)
+           (unless includeStateChanges . setPragmaOptions) $ do
    -- Forget the pragma options (locally).
    setCommandLineOptions . stPersistentOptions . stPersistent =<< get
 
@@ -241,7 +226,10 @@ getInterface' x includeStateChanges =
     visited <- isVisited x
     reportSLn "import.iface" 5 $ if visited then "  We've been here. Don't merge."
                                  else "  New module. Let's check it out."
-    unless (visited || stateChangesIncluded) $ mergeInterface i
+    unless (visited || stateChangesIncluded) $ do
+      mergeInterface i
+      ifTopLevelAndHighlightingLevelIs NonInteractive $
+        highlightFromInterface i file
 
     modify (\s -> s { stCurrentModule = Just $ iModuleName i })
 
@@ -255,7 +243,7 @@ getInterface' x includeStateChanges =
     where
       isCached file = do
         let ifile = filePath $ toIFile file
-        exist <- liftIO $ doesFileExist ifile
+        exist <- liftIO $ doesFileExistCaseSensitive ifile
         if not exist
           then return False
           else do
@@ -265,6 +253,18 @@ getInterface' x includeStateChanges =
               Just (mi, mt) | mt >= t -> True
               _                       -> False
 
+      -- Formats the "Checking", "Finished" and "Skipping" messages.
+      chaseMsg kind file = do
+        nesting <- envModuleNestingLevel <$> ask
+        let s = genericReplicate nesting ' ' ++ kind ++
+                " " ++ render (pretty x) ++
+                case file of
+                  Nothing -> "."
+                  Just f  -> " (" ++ f ++ ")."
+        ifM (envEmacs <$> ask)
+            (get >>= \st -> stInteractionOutputCallback st $ Resp_RunningInfo s)
+            (reportSLn "" 1 s)
+
       skip file = do
         -- Examine the mtime of the interface file. If it is newer than the
         -- stored version (in stDecodedModules), or if there is no stored version,
@@ -301,52 +301,63 @@ getInterface' x includeStateChanges =
                 -- liftIO close	-- Close the interface file. See above.
                 typeCheck file
               else do
-                unless cached $ reportSLn "" 1 $
-                  "Skipping " ++ render (pretty x) ++
-                    " (" ++ ifile ++ ")."
+                unless cached $ chaseMsg "Skipping" (Just ifile)
                 -- We set the pragma options of the skipped file here,
                 -- because if the top-level file is skipped we want the
                 -- pragmas to apply to interactive commands in the UI.
                 mapM_ setOptionsFromPragma (iPragmaOptions i)
                 return (False, (i, Right t))
 
-      typeCheck file =
-        let ret a = do
-              reportSLn "" 1 $ "Finished " ++ render (pretty x) ++ "."
-              return a
-        in do
+      typeCheck file = do
+          let withMsgs m = do
+                chaseMsg "Checking" (Just $ filePath file)
+                x <- m
+                chaseMsg "Finished" Nothing
+                return x
+
           -- Do the type checking.
-          reportSLn "" 1 $ "Checking " ++ render (pretty x) ++ " (" ++ filePath file ++ ")."
+
           if includeStateChanges then do
-            r <- createInterface file x
+            r <- withMsgs $ createInterface file x
 
             -- Merge the signature with the signature for imported
             -- things.
             sig <- getSignature
-            addImportedThings sig Map.empty Set.empty
+            patsyns <- getPatternSyns
+            addImportedThings sig Map.empty Set.empty patsyns
             setSignature emptySignature
+            setPatternSyns Map.empty
 
-            ret (True, r)
+            return (True, r)
            else do
             ms       <- getImportPath
+            emacs    <- envEmacs <$> ask
+            nesting  <- envModuleNestingLevel <$> ask
             mf       <- stModuleToSource <$> get
             vs       <- getVisitedModules
             ds       <- getDecodedModules
             opts     <- stPersistentOptions . stPersistent <$> get
             isig     <- getImportedSignature
             ibuiltin <- gets stImportedBuiltins
+            ipatsyns <- getPatternSynImports
+            ho       <- stInteractionOutputCallback <$> get
             -- Every interface is treated in isolation. Note: Changes
             -- to stDecodedModules are not preserved if an error is
             -- encountered in an imported module.
             r <- liftIO $ runTCM $
-                   withImportPath ms $ do
+                   withImportPath ms $
+                   local (\e -> e { envEmacs              = emacs
+                                  , envModuleNestingLevel = nesting
+                                  }) $ do
                      setDecodedModules ds
                      setCommandLineOptions opts
-                     modify $ \s -> s { stModuleToSource = mf }
+                     modify $ \s -> s { stModuleToSource     = mf
+                                      , stInteractionOutputCallback = ho
+                                      }
                      setVisitedModules vs
-                     addImportedThings isig ibuiltin Set.empty
+                     addImportedThings isig ibuiltin Set.empty ipatsyns
 
-                     r <- createInterface file x
+                     r <- withMsgs $ createInterface file x
 
                      mf        <- stModuleToSource <$> get
                      ds        <- getDecodedModules
@@ -363,12 +374,28 @@ getInterface' x includeStateChanges =
                 Right (r, update) -> do
                   update
                   case r of
-                    (_, Right _) -> do
-                      r <- skip file
-                      ret r
-                    _ ->
-                      ret (False, r)
+                    (_, Right _) ->
+                      -- We skip the file which has just been type-checked to
+                      -- be able to forget some of the local state from
+                      -- checking the module.
+                      -- Note that this doesn't actually read the interface
+                      -- file, only the cached interface.
+                      skip file
+                    _ -> return (False, r)
 
+-- | Print the highlighting information contained in the given
+-- interface.
+
+highlightFromInterface
+  :: Interface
+  -> AbsolutePath
+     -- ^ The corresponding file.
+  -> TCM ()
+highlightFromInterface i file = do
+  reportSLn "import.iface" 5 $
+    "Generating syntax info for " ++ filePath file ++
+    " (read from interface)."
+  printHighlightingInfo (iHighlighting i)
 
 readInterface :: FilePath -> TCM (Maybe Interface)
 readInterface file = do
@@ -389,9 +416,9 @@ readInterface file = do
   -- Catch exceptions
   `catchError` handler
   where
-    handler e = case errError e of
+    handler e = case e of
       IOException _ e -> do
-        liftIO $ LocIO.putStrLn $ "IO exception: " ++ show e
+        reportSLn "" 0 $ "IO exception: " ++ show e
         return Nothing   -- Work-around for file locking bug.
                          -- TODO: What does this refer to? Please
                          -- document.
@@ -424,12 +451,17 @@ createInterface
   :: AbsolutePath          -- ^ The file to type check.
   -> C.TopLevelModuleName  -- ^ The expected module name.
   -> TCM (Interface, Either Warnings ClockTime)
-createInterface file mname = do
-    reportSLn "import.iface.create" 5  $
+createInterface file mname =
+  local (\e -> e { envCurrentPath = file }) $ do
+    modFile       <- stModuleToSource <$> get
+    fileTokenInfo <- generateTokenInfo file
+    modify $ \st -> st { stTokens = fileTokenInfo }
+
+    reportSLn "import.iface.create" 5 $
       "Creating interface for " ++ render (pretty mname) ++ "."
     verboseS "import.iface.create" 10 $ do
       visited <- Map.keys <$> getVisitedModules
-      liftIO $ LocIO.putStrLn $
+      reportSLn "" 0 $
         "  visited: " ++ intercalate ", " (map (render . pretty) visited)
 
     previousHsImports <- getHaskellImports
@@ -444,47 +476,63 @@ createInterface file mname = do
     mapM_ setOptionsFromPragma options
     topLevel <- concreteToAbstract_ (TopLevel top)
 
-    termErrs <- catchError (do
-      -- Type checking.
-      checkDecls (topLevelDecls topLevel)
-      unfreezeMetas
+    let ds = topLevelDecls topLevel
 
-      -- Count number of metas
-      verboseS "profile.metas" 10 $ do
-        MetaId n <- fresh
-        tickN "metas" n
+    ifTopLevelAndHighlightingLevelIs NonInteractive $ do
+      -- Generate and print approximate syntax highlighting info.
+      printHighlightingInfo fileTokenInfo
+      mapM_ (\d -> generateAndPrintSyntaxInfo d Partial) ds
 
-      -- Termination checking.
-      termErrs <- ifM (optTerminationCheck <$> pragmaOptions)
-                      (termDecls $ topLevelDecls topLevel)
-                      (return [])
-      mapM_ (\e -> reportSLn "term.warn.no" 2
-                     (show (termErrFunctions e) ++
-                      " do(es) NOT pass the termination checker."))
-            termErrs
-      return termErrs
-      ) (\e -> do
-        -- If there is an error syntax highlighting info can still be
-        -- generated.
-        case rStart $ getRange e of
-          Just (Pn { srcFile = Just f }) | f == file -> do
-            syntaxInfo <- generateSyntaxInfo file (Just e) topLevel []
-            modFile    <- stModuleToSource <$> get
-            -- The highlighting info is included with the error.
-            case errHighlighting e of
-              Just _  -> __IMPOSSIBLE__
-              Nothing ->
-                throwError $ e { errHighlighting =
-                                   Just (syntaxInfo, modFile) }
-          _ -> throwError e
-      )
+    catchError (checkDecls ds) $ \e -> do
+      ifTopLevelAndHighlightingLevelIs NonInteractive $
+        printErrorInfo e
+      throwError e
 
-    -- Generate syntax highlighting info.
-    syntaxInfo <- generateSyntaxInfo file Nothing topLevel termErrs
+    unfreezeMetas
+
+    -- Count number of metas
+    verboseS "profile.metas" 10 $ do
+      MetaId n <- fresh
+      tickN "metas" (fromIntegral n)
+
+    -- Move any remaining token highlighting to stSyntaxInfo.
+    ifTopLevelAndHighlightingLevelIs NonInteractive $
+      printHighlightingInfo . stTokens =<< get
+    modify $ \st ->
+      st { stTokens     = mempty
+         , stSyntaxInfo = stSyntaxInfo st `mappend` stTokens st
+         }
 
-    -- Generate Vim file.
     whenM (optGenerateVimFile <$> commandLineOptions) $
-	withScope_ (insideScope topLevel) $ generateVimFile $ filePath file
+      -- Generate Vim file.
+      withScope_ (insideScope topLevel) $ generateVimFile $ filePath file
+
+    setScope $ outsideScope topLevel
+
+    reportSLn "scope.top" 50 $ "SCOPE " ++ show (insideScope topLevel)
+
+    syntaxInfo <- stSyntaxInfo <$> get
+    i <- buildInterface topLevel syntaxInfo previousHsImports options
+
+    -- TODO: It would be nice if unsolved things were highlighted
+    -- after every mutual block.
+
+    termErrs            <- Fold.toList <$> stTermErrs <$> get
+    unsolvedMetas       <- List.nub <$> (mapM getMetaRange =<< getOpenMetas)
+    unsolvedConstraints <- getAllConstraints
+
+    ifTopLevelAndHighlightingLevelIs NonInteractive $
+      printUnsolvedInfo
+
+    r <- if and [ null termErrs, null unsolvedMetas, null unsolvedConstraints ]
+     then do
+      -- The file was successfully type-checked (and no warnings were
+      -- encountered), so the interface should be written out.
+      let ifile = filePath $ toIFile file
+      t  <- writeInterface ifile i
+      return (i, Right t)
+     else
+      return (i, Left $ Warnings termErrs unsolvedMetas unsolvedConstraints)
 
     -- Print stats
     stats <- Map.toList <$> getStatistics
@@ -495,32 +543,7 @@ createInterface file mname = do
         [ "  " ++ s ++ " = " ++ show n
         | (s, n) <- sortBy (compare `on` snd) stats ]
 
-    -- Check if there are unsolved meta-variables...
-    unsolvedOK    <- optAllowUnsolved <$> pragmaOptions
-    unsolvedMetas <- List.nub <$> (mapM getMetaRange =<< getOpenMetas)
-    unless (null unsolvedMetas || unsolvedOK) $
-      typeError $ UnsolvedMetas unsolvedMetas
-
-    -- ...or unsolved constraints
-    unsolvedConstraints <- getAllConstraints
-    unless (null unsolvedConstraints || unsolvedOK) $
-      typeError $ UnsolvedConstraints unsolvedConstraints
-
-    setScope $ outsideScope topLevel
-
-    reportSLn "scope.top" 50 $ "SCOPE " ++ show (insideScope topLevel)
-
-    i <- buildInterface topLevel syntaxInfo previousHsImports options
-
-    if and [ null termErrs, null unsolvedMetas, null unsolvedConstraints ]
-     then do
-      -- The file was successfully type-checked (and no warnings were
-      -- encountered), so the interface should be written out.
-      let ifile = filePath $ toIFile file
-      t  <- writeInterface ifile i
-      return (i, Right t)
-     else
-      return (i, Left $ Warnings termErrs unsolvedMetas unsolvedConstraints)
+    return r
 
 -- | Builds an interface for the current module, which should already
 -- have been successfully type checked.
@@ -543,7 +566,8 @@ buildInterface topLevel syntaxInfo previousHsImports pragmas = do
     builtin <- gets stLocalBuiltins
     ms      <- getImports
     hsImps  <- getHaskellImports
-    let	builtin' = Map.mapWithKey (\x b -> fmap (const x) b) builtin
+    patsyns <- getPatternSyns
+    let	builtin' = Map.mapWithKey (\x b -> fmap (\pf -> (x, primFunName pf)) b) builtin
     reportSLn "import.iface" 7 "  instantiating all meta variables"
     i <- instantiateFull $ Interface
 			{ iImportedModules = Set.toList ms
@@ -556,6 +580,7 @@ buildInterface topLevel syntaxInfo previousHsImports pragmas = do
                                                             previousHsImports
                         , iHighlighting    = syntaxInfo
                         , iPragmaOptions   = pragmas
+                        , iPatternSyns     = patsyns
 			}
     reportSLn "import.iface" 7 "  interface complete"
     return i
