@@ -11,6 +11,7 @@ module Agda.Interaction.Options
     , parsePragmaOptions
     , parsePluginOptions
     , defaultOptions
+    , defaultInteractionOptions
     , defaultVerbosity
     , standardOptions_
     , unsafePragmaOptions
@@ -49,24 +50,27 @@ type Verbosity = Trie String Int
 
 data CommandLineOptions =
     Options { optProgramName          :: String
-	    , optInputFile            :: Maybe FilePath
-	    , optIncludeDirs          :: Either [FilePath] [AbsolutePath]
+            , optInputFile            :: Maybe FilePath
+            , optIncludeDirs          :: Either [FilePath] [AbsolutePath]
               -- ^ 'Left' is used temporarily, before the paths have
               -- been made absolute. An empty 'Left' list is
               -- interpreted as @["."]@ (see
               -- 'Agda.TypeChecking.Monad.Options.makeIncludeDirsAbsolute').
-	    , optShowVersion          :: Bool
-	    , optShowHelp             :: Bool
-	    , optInteractive          :: Bool
-	    , optRunTests             :: Bool
-	    , optCompile              :: Bool
-	    , optEpicCompile          :: Bool
-	    , optJSCompile            :: Bool
+            , optShowVersion          :: Bool
+            , optShowHelp             :: Bool
+            , optInteractive          :: Bool
+            , optRunTests             :: Bool
+            , optGHCiInteraction      :: Bool
+            , optCompile              :: Bool
+            , optEpicCompile          :: Bool
+            , optJSCompile            :: Bool
             , optCompileDir           :: Maybe FilePath
               -- ^ In the absence of a path the project root is used.
 	    , optGenerateVimFile      :: Bool
+            , optGenerateLaTeX        :: Bool
 	    , optGenerateHTML         :: Bool
 	    , optDependencyGraph      :: Maybe FilePath
+	    , optLaTeXDir             :: FilePath
 	    , optHTMLDir              :: FilePath
 	    , optCSSFile              :: Maybe FilePath
 	    , optIgnoreInterfaces     :: Bool
@@ -82,6 +86,7 @@ data CommandLineOptions =
 
 data PragmaOptions = PragmaOptions
   { optShowImplicit              :: Bool
+  , optShowIrrelevant            :: Bool
   , optVerbose                   :: Verbosity
   , optProofIrrelevance          :: Bool
   , optAllowUnsolved             :: Bool
@@ -97,6 +102,7 @@ data PragmaOptions = PragmaOptions
   , optIrrelevantProjections     :: Bool
   , optExperimentalIrrelevance   :: Bool  -- ^ irrelevant levels, irrelevant data matching
   , optWithoutK                  :: Bool
+  , optCopatterns                :: Bool  -- ^ definitions by copattern matching
   }
   deriving Show
 
@@ -112,25 +118,38 @@ type OptionsPragma = [String]
 mapFlag :: (String -> String) -> OptDescr a -> OptDescr a
 mapFlag f (Option _ long arg descr) = Option [] (map f long) arg descr
 
+-- | For batch usage.
 defaultVerbosity :: Verbosity
 defaultVerbosity = Trie.singleton [] 1
 
+-- | For interactive usage, do not print any debug messages
+--   by default.
+defaultInteractionVerbosity :: Verbosity
+defaultInteractionVerbosity = Trie.singleton [] 0
+
+defaultInteractionOptions :: PragmaOptions
+defaultInteractionOptions = defaultPragmaOptions
+  { optVerbose = defaultInteractionVerbosity }
+
 defaultOptions :: CommandLineOptions
 defaultOptions =
     Options { optProgramName          = "agda"
-	    , optInputFile            = Nothing
-	    , optIncludeDirs          = Left []
-	    , optShowVersion          = False
-	    , optShowHelp             = False
-	    , optInteractive          = False
-	    , optRunTests             = False
-	    , optCompile              = False
-	    , optEpicCompile          = False
-	    , optJSCompile            = False
+            , optInputFile            = Nothing
+            , optIncludeDirs          = Left []
+            , optShowVersion          = False
+            , optShowHelp             = False
+            , optInteractive          = False
+            , optRunTests             = False
+            , optGHCiInteraction      = False
+            , optCompile              = False
+            , optEpicCompile          = False
+            , optJSCompile            = False
             , optCompileDir           = Nothing
 	    , optGenerateVimFile      = False
+            , optGenerateLaTeX        = False
 	    , optGenerateHTML         = False
 	    , optDependencyGraph      = Nothing
+	    , optLaTeXDir             = defaultLaTeXDir
 	    , optHTMLDir              = defaultHTMLDir
 	    , optCSSFile              = Nothing
 	    , optIgnoreInterfaces     = False
@@ -139,11 +158,12 @@ defaultOptions =
             , optPragmaOptions        = defaultPragmaOptions
             , optEpicFlags            = []
             , optSafe                 = False
-	    }
+            }
 
 defaultPragmaOptions :: PragmaOptions
 defaultPragmaOptions = PragmaOptions
   { optShowImplicit              = False
+  , optShowIrrelevant            = False
   , optVerbose                   = defaultVerbosity
   , optProofIrrelevance          = False
   , optExperimentalIrrelevance   = False
@@ -159,8 +179,13 @@ defaultPragmaOptions = PragmaOptions
   , optGuardingTypeConstructors  = False
   , optUniversePolymorphism      = True
   , optWithoutK                  = False
+  , optCopatterns                = False
   }
 
+-- | The default output directory for LaTeX.
+
+defaultLaTeXDir = "latex"
+
 -- | The default output directory for HTML.
 
 defaultHTMLDir = "html"
@@ -181,16 +206,20 @@ checkOpts :: Flag CommandLineOptions
 checkOpts opts
   | not (atMostOne [optAllowUnsolved . p, optCompile]) = Left
       "Unsolved meta variables are not allowed when compiling.\n"
-  | not (atMostOne [optInteractive, optCompile, optEpicCompile, optJSCompile]) =
-      Left "Choose at most one: compiler/interactive mode.\n"
-  | not (atMostOne [optGenerateHTML, optInteractive]) =
-      Left "Choose at most one: HTML generator or interactive mode.\n"
-  | not (atMostOne [isJust . optDependencyGraph, optInteractive]) =
-      Left "Choose at most one: Dependency graph generator or interactive mode.\n"
+  | not (atMostOne [optGHCiInteraction, isJust . optInputFile]) =
+      Left "Choose at most one: input file or --interaction.\n"
+  | not (atMostOne $ interactive ++ [optCompile, optEpicCompile, optJSCompile]) =
+      Left "Choose at most one: compilers/--interactive/--interaction.\n"
+  | not (atMostOne $ interactive ++ [optGenerateHTML]) =
+      Left "Choose at most one: --html/--interactive/--interaction.\n"
+  | not (atMostOne $ interactive ++ [isJust . optDependencyGraph]) =
+      Left "Choose at most one: --dependency-graph/--interactive/--interaction.\n"
   | not (atMostOne [ optUniversePolymorphism . p
                    , not . optUniverseCheck . p
                    ]) =
       Left "Cannot have both universe polymorphism and type in type.\n"
+  | not (atMostOne $ interactive ++ [optGenerateLaTeX]) =
+      Left "Choose at most one: --latex/--interactive/--interaction.\n"
   | (not . null . optEpicFlags $ opts)
       && not (optEpicCompile opts) =
       Left "Cannot set Epic flags without using the Epic backend.\n"
@@ -200,6 +229,8 @@ checkOpts opts
 
   p = optPragmaOptions
 
+  interactive = [optInteractive, optGHCiInteraction]
+
 -- Check for unsafe pramas. Gives a list of used unsafe flags.
 
 unsafePragmaOptions :: PragmaOptions -> [String]
@@ -212,7 +243,8 @@ unsafePragmaOptions opts =
   [ "--sized-types"                              | optSizedTypes opts                ] ++
   [ "--injective-type-constructors"              | optInjectiveTypeConstructors opts ] ++
   [ "--guardedness-preserving-type-constructors" | optGuardingTypeConstructors opts  ] ++
-  [ "--experimental-irrelevance"                 | optExperimentalIrrelevance opts   ]
+  [ "--experimental-irrelevance"                 | optExperimentalIrrelevance opts   ] ++
+  [ "--copatterns"                               | optCopatterns opts   ]
 
 -- The default pragma options should be considered safe
 
@@ -220,15 +252,15 @@ defaultPragmaOptionsSafe :: IO Bool
 defaultPragmaOptionsSafe
     | null unsafe = return True
     | otherwise   = do putStrLn $ "Following pragmas are default but not safe: "
-                                  ++ intercalate ", " unsafe
+                                        ++ intercalate ", " unsafe
                        return False
   where unsafe = unsafePragmaOptions defaultPragmaOptions
 
 inputFlag :: FilePath -> Flag CommandLineOptions
 inputFlag f o =
     case optInputFile o of
-	Nothing  -> return $ o { optInputFile = Just f }
-	Just _	 -> throwError "only one input file allowed"
+        Nothing  -> return $ o { optInputFile = Just f }
+        Just _   -> throwError "only one input file allowed"
 
 versionFlag                  o = return $ o { optShowVersion               = True  }
 helpFlag                     o = return $ o { optShowHelp                  = True  }
@@ -239,8 +271,12 @@ noIrrelevantProjectionsFlag  o = return $ o { optIrrelevantProjections     = Fal
 ignoreInterfacesFlag         o = return $ o { optIgnoreInterfaces          = True  }
 allowUnsolvedFlag            o = return $ o { optAllowUnsolved             = True  }
 showImplicitFlag             o = return $ o { optShowImplicit              = True  }
+showIrrelevantFlag           o = return $ o { optShowIrrelevant            = True  }
 runTestsFlag                 o = return $ o { optRunTests                  = True  }
+ghciInteractionFlag          o = return $ o { optGHCiInteraction           = True  }
 vimFlag                      o = return $ o { optGenerateVimFile           = True  }
+latexFlag                    o = return $ o { optGenerateLaTeX             = True  }
+latexDirFlag               d o = return $ o { optLaTeXDir                  = d     }
 noPositivityFlag             o = return $ o { optDisablePositivity         = True  }
 dontTerminationCheckFlag     o = return $ o { optTerminationCheck          = False }
 dontCompletenessCheckFlag    o = return $ o { optCompletenessCheck         = False }
@@ -253,11 +289,12 @@ universePolymorphismFlag     o = return $ o { optUniversePolymorphism      = Tru
 noUniversePolymorphismFlag   o = return $ o { optUniversePolymorphism      = False }
 noForcingFlag                o = return $ o { optForcing                   = False }
 withoutKFlag                 o = return $ o { optWithoutK                  = True  }
+copatternsFlag               o = return $ o { optCopatterns                = True  }
 
 interactiveFlag  o = return $ o { optInteractive    = True
                                 , optPragmaOptions  = (optPragmaOptions o)
                                                         { optAllowUnsolved = True }
-		                }
+                                }
 compileFlag      o = return $ o { optCompile    = True }
 compileEpicFlag  o = return $ o { optEpicCompile = True}
 compileJSFlag    o = return $ o { optJSCompile = True}
@@ -274,8 +311,8 @@ includeFlag d o = return $ o { optIncludeDirs = Left (d : ds) }
   where ds = either id (const []) $ optIncludeDirs o
 
 verboseFlag s o =
-    do	(k,n) <- parseVerbose s
-	return $ o { optVerbose = Trie.insert k n $ optVerbose o }
+    do  (k,n) <- parseVerbose s
+        return $ o { optVerbose = Trie.insert k n $ optVerbose o }
   where
     parseVerbose s = case wordsBy (`elem` ":.") s of
       []  -> usage
@@ -293,14 +330,16 @@ terminationDepthFlag s o =
 integerArgument :: String -> String -> Either String Int
 integerArgument flag s =
     readM s `catchError` \_ ->
-	throwError $ "option '" ++ flag ++ "' requires an integer argument"
+        throwError $ "option '" ++ flag ++ "' requires an integer argument"
 
 standardOptions :: [OptDescr (Flag CommandLineOptions)]
 standardOptions =
-    [ Option ['V']  ["version"]	(NoArg versionFlag) "show version number"
-    , Option ['?']  ["help"]	(NoArg helpFlag)    "show this help"
+    [ Option ['V']  ["version"] (NoArg versionFlag) "show version number"
+    , Option ['?']  ["help"]    (NoArg helpFlag)    "show this help"
     , Option ['I']  ["interactive"] (NoArg interactiveFlag)
-		    "start in interactive mode"
+                    "start in interactive mode"
+    , Option []     ["interaction"] (NoArg ghciInteractionFlag)
+                    "for use with the Emacs mode"
     , Option ['c']  ["compile"] (NoArg compileFlag)
                     "compile program using the MAlonzo backend (experimental)"
     , Option []     ["epic"] (NoArg compileEpicFlag) "compile program using the Epic backend"
@@ -315,6 +354,11 @@ standardOptions =
 		    "run internal test suite"
     , Option []	    ["vim"] (NoArg vimFlag)
 		    "generate Vim highlighting files"
+    , Option []	    ["latex"] (NoArg latexFlag)
+                    "generate LaTeX with highlighted source code"
+    , Option []	    ["latex-dir"] (ReqArg latexDirFlag "DIR")
+                    ("directory in which LaTeX files are placed (default: " ++
+                     defaultLaTeXDir ++ ")")
     , Option []	    ["html"] (NoArg htmlFlag)
 		    "generate HTML files with highlighted source code"
     , Option []	    ["dependency-graph"] (ReqArg dependencyGraphFlag "FILE")
@@ -343,6 +387,8 @@ pragmaOptions :: [OptDescr (Flag PragmaOptions)]
 pragmaOptions =
     [ Option []	    ["show-implicit"] (NoArg showImplicitFlag)
 		    "show implicit arguments when printing"
+    , Option []	    ["show-irrelevant"] (NoArg showIrrelevantFlag)
+		    "show irrelevant arguments when printing"
     , Option ['v']  ["verbose"]	(ReqArg verboseFlag "N")
                     "set verbosity level to N"
     -- , Option []	    ["proof-irrelevance"] (NoArg proofIrrelevanceFlag)
@@ -375,6 +421,8 @@ pragmaOptions =
                     "enable potentially unsound irrelevance features (irrelevant levels, irrelevant data matching)"
     , Option []     ["without-K"] (NoArg withoutKFlag)
                     "disable the K rule (maybe)"
+    , Option []     ["copatterns"] (NoArg copatternsFlag)
+                    "enable definitions by copattern matching"
     ]
 
 -- | Used for printing usage info.
@@ -427,7 +475,7 @@ usage options pluginInfos progName =
     where
 	header progName = unlines [ "Agda"
 				  , ""
-				  , "Usage: " ++ progName ++ " [OPTIONS...] FILE"
+				  , "Usage: " ++ progName ++ " [OPTIONS...] [FILE]"
 				  ]
 
         pluginMsg (name, help, inherited, opts)
