@@ -1,4 +1,5 @@
 {-# LANGUAGE CPP #-}
+{-# LANGUAGE PatternGuards #-}
 
 {-| Agda main module.
 -}
@@ -16,8 +17,6 @@ import Data.Maybe
 import System.Environment
 import System.Exit
 import System.FilePath
-import qualified Agda.Utils.IO.Locale as LocIO
-import System.Time
 
 import Agda.Syntax.Position
 import Agda.Syntax.Parser
@@ -28,16 +27,16 @@ import Agda.Syntax.Translation.ConcreteToAbstract
 import Agda.Syntax.Translation.AbstractToConcrete
 import Agda.Syntax.Translation.InternalToAbstract
 import Agda.Syntax.Abstract.Name
-import Agda.Syntax.Strict
 import Agda.Syntax.Scope.Base
 
 import Agda.Interaction.Exceptions
 import Agda.Interaction.CommandLine.CommandLine
 import Agda.Interaction.Options
 import Agda.Interaction.Monad
-import Agda.Interaction.GhciTop ()      -- to make sure it compiles
+import Agda.Interaction.GhcTop (mimicGHCi)
 import qualified Agda.Interaction.Imports as Imp
 import qualified Agda.Interaction.Highlighting.Dot as Dot
+import qualified Agda.Interaction.Highlighting.LaTeX as LaTeX
 import Agda.Interaction.Highlighting.HTML
 
 import Agda.TypeChecker
@@ -82,6 +81,7 @@ runAgda = do
           unless ok exitFailure
       | isNothing (optInputFile opts)
           && not (optInteractive opts)
+          && not (optGHCiInteraction opts)
                             -> liftIO printUsage
       | otherwise           -> do
           setCommandLineOptions opts
@@ -89,18 +89,23 @@ runAgda = do
   where
     checkFile :: TCM ()
     checkFile = do
-      i       <- optInteractive <$> liftTCM commandLineOptions
-      compile <- optCompile     <$> liftTCM commandLineOptions
-      epic    <- optEpicCompile <$> liftTCM commandLineOptions
-      js      <- optJSCompile   <$> liftTCM commandLineOptions
-      when i $ liftIO $ LocIO.putStr splashScreen
+      i       <- optInteractive     <$> liftTCM commandLineOptions
+      ghci    <- optGHCiInteraction <$> liftTCM commandLineOptions
+      compile <- optCompile         <$> liftTCM commandLineOptions
+      epic    <- optEpicCompile     <$> liftTCM commandLineOptions
+      js      <- optJSCompile       <$> liftTCM commandLineOptions
+      when i $ liftIO $ putStr splashScreen
       let failIfNoInt (Just i) = return i
           -- The allowed combinations of command-line
           -- options should rule out Nothing here.
           failIfNoInt Nothing  = __IMPOSSIBLE__
 
+          failIfInt x Nothing  = x
+          failIfInt _ (Just _) = __IMPOSSIBLE__
+
           interaction :: TCM (Maybe Interface) -> TCM ()
           interaction | i         = runIM . interactionLoop
+                      | ghci      = (failIfInt mimicGHCi =<<)
                       | compile   = (MAlonzo.compilerMain =<<) . (failIfNoInt =<<)
                       | epic      = (Epic.compilerMain    =<<) . (failIfNoInt =<<)
                       | js        = (JS.compilerMain      =<<) . (failIfNoInt =<<)
@@ -115,12 +120,12 @@ runAgda = do
           unsolvedOK <- optAllowUnsolved <$> pragmaOptions
 
           result <- case mw of
-            Just (Imp.Warnings [] [] []) -> __IMPOSSIBLE__
-            Just (Imp.Warnings _ unsolved@(_:_) _)
+            Just (Warnings [] [] []) -> __IMPOSSIBLE__
+            Just (Warnings _ unsolved@(_:_) _)
               | not unsolvedOK -> typeError $ UnsolvedMetas unsolved
-            Just (Imp.Warnings _ _ unsolved@(_:_))
+            Just (Warnings _ _ unsolved@(_:_))
               | not unsolvedOK -> typeError $ UnsolvedConstraints unsolved
-            Just (Imp.Warnings termErrs@(_:_) _ _) ->
+            Just (Warnings termErrs@(_:_) _ _) ->
               typeError $ TerminationCheckFailed termErrs
             Just _  -> return Nothing
             Nothing -> return $ Just i
@@ -131,40 +136,39 @@ runAgda = do
           whenM (isJust . optDependencyGraph <$> commandLineOptions) $
             Dot.generateDot $ i
 
+          whenM (optGenerateLaTeX <$> commandLineOptions) $
+            LaTeX.generateLaTeX (iModuleName i) (iHighlighting i)
+
           return result
 
 -- | Print usage information.
 printUsage :: IO ()
 printUsage = do
   progName <- getProgName
-  LocIO.putStr $ usage standardOptions_ [] progName
+  putStr $ usage standardOptions_ [] progName
 
 -- | Print version information.
 printVersion :: IO ()
 printVersion =
-  LocIO.putStrLn $ "Agda version " ++ version
+  putStrLn $ "Agda version " ++ version
 
 -- | What to do for bad options.
 optionError :: String -> IO ()
 optionError err = do
-  LocIO.putStrLn $ "Error: " ++ err
+  putStrLn $ "Error: " ++ err
   printUsage
   exitFailure
 
 -- | Main
 main :: IO ()
 main = do
-#if MIN_VERSION_base(4,2,0)
-    -- Ensure that UTF-8 is used for functions in Agda.Utils.IO.Locale
-    IO.hSetEncoding IO.stdout IO.utf8
-#endif
     r <- runTCM $ runAgda `catchError` \err -> do
       s <- prettyError err
-      liftIO $ LocIO.putStrLn s
+      liftIO $ putStrLn s
       throwError err
     case r of
       Right _ -> exitSuccess
       Left _  -> exitFailure
   `catchImpossible` \e -> do
-    LocIO.putStr $ show e
+    putStr $ show e
     exitFailure
