@@ -239,8 +239,8 @@ instance Show Name where
   show x = show (nameConcrete x) -- ++ "|" ++ show (nameId x)
 
 instance Hashable Name where
-  {-# INLINE hash #-}
-  hash = hash . nameId
+  {-# INLINE hashWithSalt #-}
+  hashWithSalt salt = (hashWithSalt salt) . nameId
 
 instance Show QName where
   show q = concat $ intersperse "." $ map show $ qnameToList q
@@ -255,8 +255,8 @@ instance Ord QName where
   compare = compare `on` qnameName
 
 instance Hashable QName where
-  {-# INLINE hash #-}
-  hash = hash . qnameName
+  {-# INLINE hashWithSalt #-}
+  hashWithSalt salt = (hashWithSalt salt) . qnameName
 
 instance HasRange Name where
   getRange = getRange . nameConcrete
