@@ -22,12 +22,11 @@ module Agda.Syntax.Concrete
     , ModuleApplication(..)
     , TypeSignature
     , Constructor
-    , Field
     , ImportDirective(..), UsingOrHiding(..), ImportedName(..)
     , Renaming(..), AsName(..)
     , defaultImportDir
-    , OpenShortHand(..)
-    , LHS(..), Pattern(..)
+    , OpenShortHand(..), RewriteEqn, WithExpr
+    , LHS(..), Pattern(..), LHSCore(..)
     , RHS(..), WhereClause(..)
     , Pragma(..)
     , Module
@@ -38,7 +37,7 @@ module Agda.Syntax.Concrete
     )
     where
 
-import Data.Generics (Typeable, Data)
+import Data.Typeable (Typeable)
 import Data.Foldable hiding (concatMap)
 import Data.Traversable
 import Agda.Syntax.Position
@@ -55,7 +54,7 @@ import Agda.Utils.Impossible
 data OpApp e
         = SyntaxBindingLambda !Range [LamBinding] e -- ^ an abstraction inside a special syntax declaration (see Issue 358 why we introduce this).
         | Ordinary e
-    deriving (Typeable, Data, Functor)
+    deriving (Typeable, Functor)
 
 fromOrdinary :: e -> OpApp e -> e
 fromOrdinary d (Ordinary e) = e
@@ -66,10 +65,10 @@ data Expr
 	= Ident QName			       -- ^ ex: @x@
 	| Lit Literal			       -- ^ ex: @1@ or @\"foo\"@
 	| QuestionMark !Range (Maybe Nat)      -- ^ ex: @?@ or @{! ... !}@
-	| Underscore !Range (Maybe Nat)	       -- ^ ex: @_@
+	| Underscore !Range (Maybe String)     -- ^ ex: @_@ or @_A_5@
 	| RawApp !Range [Expr]		       -- ^ before parsing operators
 	| App !Range Expr (NamedArg Expr)      -- ^ ex: @e e@, @e {e}@, or @e {x = e}@
-	| OpApp !Range Name [OpApp Expr]       -- ^ ex: @e + e@
+	| OpApp !Range QName [OpApp Expr]      -- ^ ex: @e + e@
         | WithApp !Range Expr [Expr]           -- ^ ex: @e | e1 | .. | en@
 	| HiddenArg !Range (Named String Expr) -- ^ ex: @{e}@ or @{x=e}@
 	| InstanceArg !Range (Named String Expr) -- ^ ex: @{{e}}@ or @{{x=e}}@
@@ -80,7 +79,7 @@ data Expr
 	| Pi Telescope Expr		       -- ^ ex: @(xs:e) -> e@ or @{xs:e} -> e@
 	| Set !Range			       -- ^ ex: @Set@
 	| Prop !Range			       -- ^ ex: @Prop@
-	| SetN !Range Nat		       -- ^ ex: @Set0, Set1, ..@
+	| SetN !Range Integer                  -- ^ ex: @Set0, Set1, ..@
 	| Rec !Range [(Name, Expr)]	       -- ^ ex: @record {x = a; y = b}@
 	| RecUpdate !Range Expr [(Name, Expr)] -- ^ ex: @record e {x = a; y = b}@
 	| Let !Range [Declaration] Expr	       -- ^ ex: @let Ds in e@
@@ -94,44 +93,44 @@ data Expr
         | QuoteTerm !Range                     -- ^ ex: @quoteTerm@, should be applied to a term
         | Unquote !Range                       -- ^ ex: @unquote@, should be applied to a term of type @Term@
         | DontCare Expr                        -- ^ to print irrelevant things
-    deriving (Typeable, Data)
+    deriving (Typeable)
 
 
 -- | Concrete patterns. No literals in patterns at the moment.
 data Pattern
-	= IdentP QName
-	| AppP Pattern (NamedArg Pattern)
-	| RawAppP !Range [Pattern]
-	| OpAppP !Range Name [Pattern]
-	| HiddenP !Range (Named String Pattern)
-	| InstanceP !Range (Named String Pattern)
-	| ParenP !Range Pattern
-	| WildP !Range
-	| AbsurdP !Range
-	| AsP !Range Name Pattern
-	| DotP !Range Expr
-	| LitP Literal
-    deriving (Typeable, Data)
+	= IdentP QName                            -- ^ @c@ or @x@
+	| AppP Pattern (NamedArg Pattern)         -- ^ @p p'@ or @p {x = p'}@
+	| RawAppP !Range [Pattern]                -- ^ @p1..pn@ before parsing operators
+	| OpAppP !Range QName [Pattern]           -- ^ eg: @p => p'@ for operator @_=>_@
+	| HiddenP !Range (Named String Pattern)   -- ^ @{p}@ or @{x = p}@
+	| InstanceP !Range (Named String Pattern) -- ^ @{{p}}@ or @{{x = p}}@
+	| ParenP !Range Pattern                   -- ^ @(p)@
+	| WildP !Range                            -- ^ @_@
+	| AbsurdP !Range                          -- ^ @()@
+	| AsP !Range Name Pattern                 -- ^ @x\@p@ unused
+	| DotP !Range Expr                        -- ^ @.e@
+	| LitP Literal                            -- ^ @0@, @1@, etc.
+    deriving (Typeable)
 
 
 -- | A lambda binding is either domain free or typed.
 data LamBinding
 	= DomainFree Hiding Relevance BoundName -- ^ . @x@ or @{x}@ or @.x@ or @.{x}@ or @{.x}@
 	| DomainFull TypedBindings              -- ^ . @(xs : e)@ or @{xs : e}@
-    deriving (Typeable, Data)
+    deriving (Typeable)
 
 
 -- | A sequence of typed bindings with hiding information. Appears in dependent
 --   function spaces, typed lambdas, and telescopes.
 data TypedBindings = TypedBindings !Range (Arg TypedBinding)
 	-- ^ . @(xs : e)@ or @{xs : e}@
-    deriving (Typeable, Data)
+    deriving (Typeable)
 
 
 data BoundName = BName { boundName   :: Name
                        , bnameFixity :: Fixity'
                        }
-    deriving (Typeable, Data)
+    deriving (Typeable)
 
 mkBoundName_ :: Name -> BoundName
 mkBoundName_ x = BName x defaultFixity'
@@ -140,7 +139,7 @@ mkBoundName_ x = BName x defaultFixity'
 data TypedBinding
 	= TBind !Range [BoundName] Expr   -- Binding @x1,..,xn:A@
 	| TNoBind Expr		    -- No binding @A@, equivalent to @_ : A@.
-    deriving (Typeable, Data)
+    deriving (Typeable)
 
 
 -- | A telescope is a sequence of typed bindings. Bound variables are in scope
@@ -154,25 +153,46 @@ type Telescope = [TypedBindings]
 
    We use fixity information to see which name is actually defined.
 -}
-data LHS = LHS { lhsOriginalPattern :: Pattern
-               , lhsWithPattern     :: [Pattern]
-               , lhsRewriteEqn      :: [RewriteEqn]
-               , lhsWithExpr        :: [WithExpr]
+data LHS = LHS { lhsOriginalPattern :: Pattern       -- ^ @f ps@
+               , lhsWithPattern     :: [Pattern]     -- ^ @| p@ (many)
+               , lhsRewriteEqn      :: [RewriteEqn]  -- ^ @rewrite e@ (many)
+               , lhsWithExpr        :: [WithExpr]    -- ^ @with e@ (many)
                }
          -- ^ original pattern, with-patterns, rewrite equations and with-expressions
          | Ellipsis Range [Pattern] [RewriteEqn] [WithExpr]
          -- ^ new with-patterns, rewrite equations and with-expressions
-  deriving (Typeable, Data)
+  deriving (Typeable)
 
 type RewriteEqn = Expr
 type WithExpr   = Expr
 
+-- | Processed (scope-checked) intermediate form of the core @f ps@ of 'LHS'.
+--   Corresponds to 'lhsOriginalPattern'.
+data LHSCore
+  = LHSHead  { lhsDefName  :: Name                -- ^ @f@
+             , lhsPats     :: [NamedArg Pattern]  -- ^ @ps@
+             }
+  | LHSProj  { lhsDestructor :: QName      -- ^ record projection identifier
+             , lhsPatsLeft   :: [NamedArg Pattern]  -- ^ side patterns
+             , lhsFocus      :: NamedArg LHSCore    -- ^ main branch
+             , lhsPatsRight  :: [NamedArg Pattern]  -- ^ side patterns
+             }
+  deriving (Typeable)
+
+{- TRASH
+lhsCoreToPattern :: LHSCore -> Pattern
+lhsCoreToPattern (LHSHead f args) = OpAppP (fuseRange f args) (unqualify f) args
+lhsCoreToPattern (LHSProj d ps1 lhscore ps2) = OpAppP (fuseRange d ps) (unqualify) ps
+  where p = lhsCoreToPattern lhscore
+        ps = ps1 ++ p : ps2
+-}
+
 data RHS = AbsurdRHS
 	 | RHS Expr
-    deriving (Typeable, Data)
+    deriving (Typeable)
 
 data WhereClause = NoWhere | AnyWhere [Declaration] | SomeWhere Name [Declaration]
-  deriving (Typeable, Data)
+  deriving (Typeable)
 
 
 -- | The things you are allowed to say when you shuffle names between name
@@ -184,20 +204,21 @@ data ImportDirective
 	    , renaming		:: [Renaming]
 	    , publicOpen	:: Bool	-- ^ Only for @open@. Exports the opened names from the current module.
 	    }
-    deriving (Typeable, Data)
+    deriving (Typeable)
 
+-- | Default is directive is @private@ (use everything, but do not export).
 defaultImportDir :: ImportDirective
 defaultImportDir = ImportDirective noRange (Hiding []) [] False
 
 data UsingOrHiding
 	= Hiding [ImportedName]
 	| Using  [ImportedName]
-    deriving (Typeable, Data)
+    deriving (Typeable)
 
 -- | An imported name can be a module or a defined name
 data ImportedName = ImportedModule  { importedName :: Name }
 		  | ImportedName    { importedName :: Name }
-    deriving (Typeable, Data, Eq, Ord)
+    deriving (Typeable, Eq, Ord)
 
 instance Show ImportedName where
     show (ImportedModule x) = "module " ++ show x
@@ -211,7 +232,7 @@ data Renaming = Renaming { renFrom    :: ImportedName
                            -- ^ The range of the \"to\" keyword. Retained
                            --   for highlighting purposes.
                          }
-    deriving (Typeable, Data)
+    deriving (Typeable)
 
 data AsName = AsName { asName  :: Name
                        -- ^ The \"as\" name.
@@ -219,7 +240,7 @@ data AsName = AsName { asName  :: Name
                        -- ^ The range of the \"as\" keyword. Retained
                        --   for highlighting purposes.
                      }
-    deriving (Typeable, Data, Show)
+    deriving (Typeable, Show)
 
 {--------------------------------------------------------------------------
     Declarations
@@ -228,9 +249,8 @@ data AsName = AsName { asName  :: Name
 -- | Just type signatures.
 type TypeSignature   = Declaration
 
--- | A constructor or field declaration is just a type signature.
+-- | A data constructor declaration is just a type signature.
 type Constructor = TypeSignature
-type Field	 = TypeSignature
 
 {-| The representation type of a declaration. The comments indicate
     which type in the intended family the constructor targets.
@@ -243,10 +263,11 @@ data Declaration
 	| DataSig     !Range Induction Name [LamBinding] Expr -- ^ lone data signature in mutual block
 	| Data        !Range Induction Name [LamBinding] (Maybe Expr) [Constructor]
 	| RecordSig   !Range Name [LamBinding] Expr -- ^ lone record signature in mutual block
-	| Record      !Range Name (Maybe Name) [LamBinding] (Maybe Expr) [Declaration]
+	| Record      !Range Name (Maybe Induction) (Maybe Name) [LamBinding] (Maybe Expr) [Declaration]
           -- ^ The optional name is a name for the record constructor.
 	| Infix Fixity [Name]
         | Syntax      Name Notation -- ^ notation declaration for a name
+        | PatternSyn  !Range Name [Name] Pattern
 	| Mutual      !Range [Declaration]
 	| Abstract    !Range [Declaration]
 	| Private     !Range [Declaration]
@@ -257,14 +278,14 @@ data Declaration
 	| ModuleMacro !Range  Name ModuleApplication OpenShortHand ImportDirective
 	| Module      !Range QName [TypedBindings] [Declaration]
 	| Pragma      Pragma
-    deriving (Typeable, Data)
+    deriving (Typeable)
 
 data ModuleApplication = SectionApp Range [TypedBindings] Expr
                        | RecordModuleIFS Range QName
-    deriving (Typeable, Data)
+    deriving (Typeable)
 
 data OpenShortHand = DoOpen | DontOpen
-    deriving (Typeable, Data, Show)
+    deriving (Typeable, Show)
 
 -- Pragmas ----------------------------------------------------------------
 
@@ -281,7 +302,8 @@ data Pragma = OptionsPragma     !Range [String]
               -- module name.
             | ImpossiblePragma !Range
             | EtaPragma !Range QName
-    deriving (Typeable, Data)
+            | NoTerminationCheckPragma !Range
+    deriving (Typeable)
 
 ---------------------------------------------------------------------------
 
@@ -312,8 +334,9 @@ appView (App r e1 e2) = vApp (appView e1) e2
 	vApp (AppView e es) arg = AppView e (es ++ [arg])
 appView (RawApp _ (e:es)) = AppView e $ map arg es
     where
-	arg (HiddenArg _ e) = Arg Hidden    Relevant e
-	arg e		    = Arg NotHidden Relevant (unnamed e)
+	arg (HiddenArg   _ e) = Arg Hidden    Relevant e
+	arg (InstanceArg _ e) = Arg Instance  Relevant e
+	arg e		      = Arg NotHidden Relevant (unnamed e)
 appView e = AppView e []
 
 {--------------------------------------------------------------------------
@@ -328,7 +351,7 @@ patternHead p =
     AppP p p'            -> patternHead p
     RawAppP _ []         -> __IMPOSSIBLE__
     RawAppP _ (p:_)      -> patternHead p
-    OpAppP _ name ps     -> return $ name
+    OpAppP _ name ps     -> return $ unqualify name
     HiddenP _ (namedPat) -> patternHead (namedThing namedPat)
     ParenP _ p           -> patternHead p
     WildP _              -> Nothing
@@ -345,9 +368,9 @@ patternNames :: Pattern -> [Name]
 patternNames p =
   case p of
     IdentP x             -> [unqualify x]
-    AppP p p'            -> concatMap patternNames [p, namedThing $ unArg p']
+    AppP p p'            -> concatMap patternNames [p, namedArg p']
     RawAppP _ ps         -> concatMap patternNames  ps
-    OpAppP _ name ps     -> name : concatMap patternNames ps
+    OpAppP _ name ps     -> unqualify name : concatMap patternNames ps
     HiddenP _ (namedPat) -> patternNames (namedThing namedPat)
     ParenP _ p           -> patternNames p
     WildP _              -> []
@@ -435,7 +458,7 @@ instance HasRange Declaration where
     getRange (DataSig r _ _ _ _)       = r
     getRange (Data r _ _ _ _ _)	       = r
     getRange (RecordSig r _ _ _)       = r
-    getRange (Record r _ _ _ _ _)      = r
+    getRange (Record r _ _ _ _ _ _)    = r
     getRange (Mutual r _)	       = r
     getRange (Abstract r _)	       = r
     getRange (Open r _ _)	       = r
@@ -447,12 +470,17 @@ instance HasRange Declaration where
     getRange (Module r _ _ _)	       = r
     getRange (Infix f _)	       = getRange f
     getRange (Syntax n _)              = getRange n
+    getRange (PatternSyn r _ _ _)      = r
     getRange (Pragma p)		       = getRange p
 
 instance HasRange LHS where
   getRange (LHS p ps eqns ws) = fuseRange p (fuseRange ps (eqns ++ ws))
   getRange (Ellipsis r _ _ _) = r
 
+instance HasRange LHSCore where
+  getRange (LHSHead f ps) = fuseRange f ps
+  getRange (LHSProj d ps1 lhscore ps2) = d `fuseRange` ps1 `fuseRange` lhscore `fuseRange` ps2
+
 instance HasRange RHS where
     getRange AbsurdRHS = noRange
     getRange (RHS e)   = getRange e
@@ -469,6 +497,7 @@ instance HasRange Pragma where
     getRange (ImportPragma r _)           = r
     getRange (ImpossiblePragma r)         = r
     getRange (EtaPragma r _)              = r
+    getRange (NoTerminationCheckPragma r) = r
 
 instance HasRange UsingOrHiding where
     getRange (Using xs)	    = getRange xs
@@ -500,3 +529,139 @@ instance HasRange Pattern where
     getRange (HiddenP r _)	= r
     getRange (InstanceP r _)	= r
     getRange (DotP r _)		= r
+
+instance KillRange AsName where
+  killRange (AsName n _) = killRange1 (flip AsName noRange) n
+
+instance KillRange BoundName where
+  killRange (BName n f) = killRange2 BName n f
+
+instance KillRange Declaration where
+  killRange (TypeSig r n e)         = killRange2 (TypeSig r) n e
+  killRange (Field n a)             = killRange2 Field n a
+  killRange (FunClause l r w)       = killRange3 FunClause l r w
+  killRange (DataSig _ i n l e)     = killRange4 (DataSig noRange) i n l e
+  killRange (Data _ i n l e c)      = killRange4 (Data noRange i) n l e c
+  killRange (RecordSig _ n l e)     = killRange3 (RecordSig noRange) n l e
+  killRange (Record _ n mi mn k e d)= killRange6 (Record noRange) n mi mn k e d
+  killRange (Infix f n)             = killRange2 Infix f n
+  killRange (Syntax n no)           = killRange1 (\n -> Syntax n no) n
+  killRange (PatternSyn _ n ns p)   = killRange3 (PatternSyn noRange) n ns p
+  killRange (Mutual _ d)            = killRange1 (Mutual noRange) d
+  killRange (Abstract _ d)          = killRange1 (Abstract noRange) d
+  killRange (Private _ d)           = killRange1 (Private noRange) d
+  killRange (Postulate _ t)         = killRange1 (Postulate noRange) t
+  killRange (Primitive _ t)         = killRange1 (Primitive noRange) t
+  killRange (Open _ q i)            = killRange2 (Open noRange) q i
+  killRange (Import _ q a o i)      = killRange3 (\q a -> Import noRange q a o) q a i
+  killRange (ModuleMacro _ n m o i) = killRange3 (\n m -> ModuleMacro noRange n m o) n m i
+  killRange (Module _ q t d)        = killRange3 (Module noRange) q t d
+  killRange (Pragma p)              = killRange1 Pragma p
+
+instance KillRange Expr where
+  killRange (Ident q)           = killRange1 Ident q
+  killRange (Lit l)             = killRange1 Lit l
+  killRange (QuestionMark _ n)  = QuestionMark noRange n
+  killRange (Underscore _ n)    = Underscore noRange n
+  killRange (RawApp _ e)        = killRange1 (RawApp noRange) e
+  killRange (App _ e a)         = killRange2 (App noRange) e a
+  killRange (OpApp _ n o)       = killRange2 (OpApp noRange) n o
+  killRange (WithApp _ e es)    = killRange2 (WithApp noRange) e es
+  killRange (HiddenArg _ n)     = killRange1 (HiddenArg noRange) n
+  killRange (InstanceArg _ n)   = killRange1 (InstanceArg noRange) n
+  killRange (Lam _ l e)         = killRange2 (Lam noRange) l e
+  killRange (AbsurdLam _ h)     = killRange1 (AbsurdLam noRange) h
+  killRange (ExtendedLam _ lrw) = killRange1 (ExtendedLam noRange) lrw
+  killRange (Fun _ e1 e2)       = killRange2 (Fun noRange) e1 e2
+  killRange (Pi t e)            = killRange2 Pi t e
+  killRange (Set _)             = Set noRange
+  killRange (Prop _)            = Prop noRange
+  killRange (SetN _ n)          = SetN noRange n
+  killRange (Rec _ ne)          = killRange1 (Rec noRange) ne
+  killRange (RecUpdate _ e ne)  = killRange2 (RecUpdate noRange) e ne
+  killRange (Let _ d e)         = killRange2 (Let noRange) d e
+  killRange (Paren _ e)         = killRange1 (Paren noRange) e
+  killRange (Absurd _)          = Absurd noRange
+  killRange (As _ n e)          = killRange2 (As noRange) n e
+  killRange (Dot _ e)           = killRange1 (Dot noRange) e
+  killRange (ETel t)            = killRange1 ETel t
+  killRange (QuoteGoal _ n e)   = killRange2 (QuoteGoal noRange) n e
+  killRange (Quote _)           = Quote noRange
+  killRange (QuoteTerm _)       = QuoteTerm noRange
+  killRange (Unquote _)         = Unquote noRange
+  killRange (DontCare e)        = killRange1 DontCare e
+
+instance KillRange ImportDirective where
+  killRange (ImportDirective _ u r p) =
+    killRange2 (\u r -> ImportDirective noRange u r p) u r
+
+instance KillRange ImportedName where
+  killRange (ImportedModule n) = killRange1 ImportedModule n
+  killRange (ImportedName   n) = killRange1 ImportedName   n
+
+instance KillRange LamBinding where
+  killRange (DomainFree h r b) = killRange2 (\h -> DomainFree h r) h b
+  killRange (DomainFull t)     = killRange1 DomainFull t
+
+instance KillRange LHS where
+  killRange (LHS p ps r w)     = killRange4 LHS p ps r w
+  killRange (Ellipsis _ p r w) = killRange3 (Ellipsis noRange) p r w
+
+instance KillRange ModuleApplication where
+  killRange (SectionApp _ t e)    = killRange2 (SectionApp noRange) t e
+  killRange (RecordModuleIFS _ q) = killRange1 (RecordModuleIFS noRange) q
+
+instance KillRange e => KillRange (OpApp e) where
+  killRange (SyntaxBindingLambda _ l e) = killRange2 (SyntaxBindingLambda noRange) l e
+  killRange (Ordinary e)                = killRange1 Ordinary e
+
+instance KillRange Pattern where
+  killRange (IdentP q)      = killRange1 IdentP q
+  killRange (AppP p n)      = killRange2 AppP p n
+  killRange (RawAppP _ p)   = killRange1 (RawAppP noRange) p
+  killRange (OpAppP _ n p)  = killRange2 (OpAppP noRange) n p
+  killRange (HiddenP _ n)   = killRange1 (HiddenP noRange) n
+  killRange (InstanceP _ n) = killRange1 (InstanceP noRange) n
+  killRange (ParenP _ p)    = killRange1 (ParenP noRange) p
+  killRange (WildP _)       = WildP noRange
+  killRange (AbsurdP _)     = AbsurdP noRange
+  killRange (AsP _ n p)     = killRange2 (AsP noRange) n p
+  killRange (DotP _ e)      = killRange1 (DotP noRange) e
+  killRange (LitP l)        = killRange1 LitP l
+
+instance KillRange Pragma where
+  killRange (OptionsPragma _ s)           = OptionsPragma noRange s
+  killRange (BuiltinPragma _ s e)         = killRange1 (BuiltinPragma noRange s) e
+  killRange (CompiledDataPragma _ q s ss) = killRange1 (\q -> CompiledDataPragma noRange q s ss) q
+  killRange (CompiledTypePragma _ q s)    = killRange1 (\q -> CompiledTypePragma noRange q s) q
+  killRange (CompiledPragma _ q s)        = killRange1 (\q -> CompiledPragma noRange q s) q
+  killRange (CompiledEpicPragma _ q s)    = killRange1 (\q -> CompiledEpicPragma noRange q s) q
+  killRange (CompiledJSPragma _ q s)      = killRange1 (\q -> CompiledJSPragma noRange q s) q
+  killRange (StaticPragma _ q)            = killRange1 (StaticPragma noRange) q
+  killRange (ImportPragma _ s)            = ImportPragma noRange s
+  killRange (ImpossiblePragma _)          = ImpossiblePragma noRange
+  killRange (EtaPragma _ q)               = killRange1 (EtaPragma noRange) q
+  killRange (NoTerminationCheckPragma _)  = NoTerminationCheckPragma noRange
+
+instance KillRange Renaming where
+  killRange (Renaming i n _) = killRange2 (\i n -> Renaming i n noRange) i n
+
+instance KillRange RHS where
+  killRange AbsurdRHS = AbsurdRHS
+  killRange (RHS e)   = killRange1 RHS e
+
+instance KillRange TypedBinding where
+  killRange (TBind _ b e) = killRange2 (TBind noRange) b e
+  killRange (TNoBind e)   = killRange1 TNoBind e
+
+instance KillRange TypedBindings where
+  killRange (TypedBindings _ t) = killRange1 (TypedBindings noRange) t
+
+instance KillRange UsingOrHiding where
+  killRange (Hiding i) = killRange1 Hiding i
+  killRange (Using  i) = killRange1 Using  i
+
+instance KillRange WhereClause where
+  killRange NoWhere         = NoWhere
+  killRange (AnyWhere d)    = killRange1 AnyWhere d
+  killRange (SomeWhere n d) = killRange2 SomeWhere n d
