@@ -3,7 +3,7 @@
 {-| The parser doesn't know about operators and parses everything as normal
     function application. This module contains the functions that parses the
     operators properly. For a stand-alone implementation of this see
-    @src\/prototyping\/mixfix@.
+    @src\/prototyping\/mixfix\/old@.
 
     It also contains the function that puts parenthesis back given the
     precedence of the context.
@@ -11,14 +11,28 @@
 module Agda.Syntax.Concrete.Operators
     ( parseApplication
     , parseLHS
+    , parsePattern
+    , parsePatternSyn
     , paren
     , mparen
+    -- exports for Copatterns
+    , validConPattern
+    , patternAppView
+    , fullParen
+    , buildParser
+    , parsePat
+    , getDefinedNames
+    , UseBoundNames(..)
+    , qualifierModules
+    , patternQNames
     ) where
 
 import Control.Applicative
+import Control.Monad
 import Control.Monad.Trans
 import Data.Typeable
 import Data.Traversable (traverse)
+import qualified Data.Traversable as Trav
 import qualified Data.Map as Map
 import qualified Data.Set as Set
 import Data.Set (Set)
@@ -27,7 +41,7 @@ import Data.Function
 
 import Agda.Syntax.Concrete.Pretty ()
 import Agda.Syntax.Common
-import Agda.Syntax.Concrete
+import Agda.Syntax.Concrete hiding (appView)
 import Agda.Syntax.Concrete.Operators.Parser
 import qualified Agda.Syntax.Abstract.Name as A
 import Agda.Syntax.Position
@@ -36,11 +50,12 @@ import Agda.Syntax.Notation
 import Agda.Syntax.Scope.Base
 import Agda.Syntax.Scope.Monad
 
-import Agda.TypeChecking.Monad.Base (typeError, TypeError(..))
+import Agda.TypeChecking.Monad.Base (typeError, TypeError(..), LHSOrPatSyn(..))
 import Agda.TypeChecking.Monad.State (getScope)
 import Agda.TypeChecking.Monad.Options
 import Agda.TypeChecking.Monad.Statistics
 
+import Agda.Utils.Either
 import Agda.Utils.ReadP
 import Agda.Utils.Monad
 import Agda.Utils.Tuple
@@ -55,47 +70,52 @@ import Agda.Utils.Impossible
 -- * Building the parser
 ---------------------------------------------------------------------------
 
-partsInScope :: ScopeM (Set Name)
-partsInScope = do
-    (names, ops) <- localNames
+partsInScope :: FlatScope -> ScopeM (Set QName)
+partsInScope flat = do
+    (names, ops) <- localNames flat
     let xs = concatMap parts names ++ concatMap notationNames ops
     return $ Set.fromList xs
     where
-        parts (NoName _ _)   = []
-        parts x@(Name _ [_]) = [x]
-        parts x@(Name _ xs)  = x : [ Name noRange [i] | i@(Id {}) <- xs ]
+        qual xs x = foldr Qual (QName x) xs
+        parts q = parts' (init $ qnameParts q) (unqualify q)
+        parts' ms (NoName _ _)   = []
+        parts' ms x@(Name _ [_]) = [qual ms x]
+                                   -- The first part should be qualified, but not the rest
+        parts' ms x@(Name _ xs)  = qual ms x : qual ms (Name noRange [first]) : [ QName $ Name noRange [i] | i <- iparts ]
+          where
+            first:iparts = [ i | i@(Id {}) <- xs ]
+
+type FlatScope = Map.Map QName [AbstractName]
 
 -- | Compute all unqualified defined names in scope and their fixities.
-getDefinedNames :: [KindOfName] -> ScopeM [(Name, Fixity')]
-getDefinedNames kinds = do
-  names <- nsNames . everythingInScope <$> getScope
-  reportSLn "scope.operators" 20 $ "everythingInScope: " ++ show names
-  return [ (x, A.nameFixity $ A.qnameName $ anameName d)
-         | (x, ds) <- Map.assocs names
-         , d       <- take 1 ds
-         , anameKind d `elem` kinds
-         ]
+getDefinedNames :: [KindOfName] -> FlatScope -> [(QName, Fixity')]
+getDefinedNames kinds names =
+  [ (x, A.nameFixity $ A.qnameName $ anameName d)
+  | (x, ds) <- Map.toList names
+  , d       <- take 1 ds
+  , anameKind d `elem` kinds
+  ]
 
 -- | Compute all names (first component) and operators (second component) in
 --   scope.
-localNames :: ScopeM ([Name], [NewNotation])
-localNames = do
-  defs   <- getDefinedNames [DefName, ConName]
+localNames :: FlatScope -> ScopeM ([QName], [NewNotation])
+localNames flat = do
+  let defs = getDefinedNames allKindsOfNames flat
   locals <- scopeLocals <$> getScope
   return $ split $ uniqBy fst $ map localOp locals ++ defs
   where
-    localOp (x, y) = (x, A.nameFixity y)
+    localOp (x, y) = (QName x, A.nameFixity y)
     split ops = ([ x | Left x <- zs], [ y | Right y <- zs ])
         where
             zs = concatMap opOrNot ops
 
-    opOrNot (x, Fixity' fx syn) = Left x
-                                :  case x of
+    opOrNot (q, Fixity' fx syn) = Left q
+                                :  case unqualify q of
                                       Name _ [_] -> []
-                                      _ -> [Right (x, fx, syntaxOf x)]
+                                      x -> [Right (q, fx, syntaxOf x)]
                                 ++ case syn of
                                     [] -> []
-                                    _ -> [Right (x, fx, syn)]
+                                    _ -> [Right (q, fx, syn)]
 
 data UseBoundNames = UseBoundNames | DontUseBoundNames
 
@@ -144,13 +164,20 @@ fixStyle syn = case (isAHole (head syn), isAHole (last syn)) of
   (False,False) -> Nonfix
 
 
-notationNames :: NewNotation -> [Name]
-notationNames (_, _, ps) = [Name noRange [Id x] | IdPart x <- ps ]
+notationNames :: NewNotation -> [QName]
+notationNames (q, _, ps) = zipWith ($) (requal : repeat QName) [Name noRange [Id x] | IdPart x <- ps ]
+  where
+    ms       = init (qnameParts q)
+    requal x = foldr Qual (QName x) ms
 
-buildParser :: forall e. IsExpr e => Range -> UseBoundNames -> ScopeM (ReadP e e)
-buildParser r use = do
-    (names, ops) <- localNames
-    cons         <- getDefinedNames [ConName]
+buildParser :: forall e. IsExpr e => Range -> FlatScope -> UseBoundNames -> ScopeM (ReadP e e)
+buildParser r flat use = do
+    (names, ops) <- localNames flat
+    let cons = getDefinedNames [ConName, PatternSynName] flat
+    reportSLn "scope.operators" 50 $ unlines
+      [ "names = " ++ show names
+      , "ops   = " ++ show ops
+      , "cons  = " ++ show cons ]
     let conparts   = Set.fromList $ concatMap notationNames $ map oldToNewNotation cons
         opsparts   = Set.fromList $ concatMap notationNames $ ops
         allParts   = Set.union conparts opsparts
@@ -169,13 +196,9 @@ buildParser r use = do
         ++ map (nonfixP . opP p) non -- for things with no outer "holes"
         ++ [ const $ atomP isAtom ]
     where
-
-
-        level :: NewNotation -> Nat
+        level :: NewNotation -> Integer
         level (_name, fixity, _syn) = fixityLevel fixity
 
-        on f g x y = f (g x) (g y)
-
         isinfixl, isinfixr, isinfix, nonfix, isprefix, ispostfix :: NewNotation -> Bool
 
         isinfixl (_, LeftAssoc _ _, syn)  = isInfix syn
@@ -222,7 +245,7 @@ buildParser r use = do
 
 instance IsExpr Expr where
     exprView e = case e of
-        Ident (QName x) -> LocalV x
+        Ident x         -> LocalV x
         App _ e1 e2     -> AppV e1 e2
         OpApp r d es    -> OpAppV d es
         HiddenArg _ e   -> HiddenArgV e
@@ -232,7 +255,7 @@ instance IsExpr Expr where
         Underscore{}    -> WildV e
         _               -> OtherV e
     unExprView e = case e of
-        LocalV x      -> Ident (QName x)
+        LocalV x      -> Ident x
         AppV e1 e2    -> App (fuseRange e1 e2) e1 e2
         OpAppV d es   -> OpApp (fuseRange d es) d es
         HiddenArgV e  -> HiddenArg (getRange e) e
@@ -245,27 +268,69 @@ instance IsExpr Expr where
 
 instance IsExpr Pattern where
     exprView e = case e of
-        IdentP (QName x) -> LocalV x
-        AppP e1 e2       -> AppV e1 e2
-        OpAppP r d es    -> OpAppV d (map Ordinary es)
-        HiddenP _ e      -> HiddenArgV e
-        InstanceP _ e    -> InstanceArgV e
-        ParenP _ e       -> ParenV e
-        WildP{}          -> WildV e
-        _                -> OtherV e
+        IdentP x      -> LocalV x
+        AppP e1 e2    -> AppV e1 e2
+        OpAppP r d es -> OpAppV d (map Ordinary es)
+        HiddenP _ e   -> HiddenArgV e
+        InstanceP _ e -> InstanceArgV e
+        ParenP _ e    -> ParenV e
+        WildP{}       -> WildV e
+        _             -> OtherV e
     unExprView e = case e of
-        LocalV x         -> IdentP (QName x)
-        AppV e1 e2       -> AppP e1 e2
-        OpAppV d es      -> let ess :: [Pattern]
-                                ess = (map (fromOrdinary __IMPOSSIBLE__) es)
-                            in OpAppP (fuseRange d es) d ess
-        HiddenArgV e     -> HiddenP (getRange e) e
-        InstanceArgV e   -> InstanceP (getRange e) e
-        ParenV e         -> ParenP (getRange e) e
-        LamV _ _         -> __IMPOSSIBLE__
-        WildV e          -> e
-        OtherV e         -> e
+        LocalV x       -> IdentP x
+        AppV e1 e2     -> AppP e1 e2
+        OpAppV d es    -> let ess :: [Pattern]
+                              ess = (map (fromOrdinary __IMPOSSIBLE__) es)
+                          in OpAppP (fuseRange d es) d ess
+        HiddenArgV e   -> HiddenP (getRange e) e
+        InstanceArgV e -> InstanceP (getRange e) e
+        ParenV e       -> ParenP (getRange e) e
+        LamV _ _       -> __IMPOSSIBLE__
+        WildV e        -> e
+        OtherV e       -> e
 
+{- TRASH
+instance IsExpr LHSCore where
+    exprView e = case e of
+        LHSHead f ps -> foldl AppV (LocalV f) $ map exprView ps
+        LHSProj d ps1 e ps2 -> foldl AppV (LocalV d) $
+          map exprView ps1 ++ exprView e : map exprView ps2
+    unExprView e = LHSHead f ps
+      where p :: Pattern
+            p = unExprView
+            (f, ps) = lhsArgs p
+-}
+
+---------------------------------------------------------------------------
+-- * Helpers for pattern and lhs parsing
+---------------------------------------------------------------------------
+
+-- Andreas, 2011-11-24 moved here from ConcreteToAbstract
+lhsArgs :: Pattern -> (Name, [NamedArg Pattern])
+lhsArgs p = case lhsArgs' p of
+              Just (x, args) -> (x, args)
+              Nothing        -> __IMPOSSIBLE__
+
+-- | @lhsArgs' p@ splits a lhs @f ps@, given as a pattern @p@,
+--   into @(f, ps)@.
+lhsArgs' :: Pattern -> Maybe (Name, [NamedArg Pattern])
+lhsArgs' p = case patternAppView p of
+    Arg _ _ (Named _ (IdentP (QName x))) : ps -> Just (x, ps)
+    _                                         -> Nothing
+
+-- | View a pattern @p@ as a list @p0 .. pn@ where @p0@ is the identifier
+--   (in most cases a constructor).
+--
+--  Pattern needs to be parsed already (operators resolved).
+patternAppView :: Pattern -> [NamedArg Pattern]
+patternAppView p = case p of
+    AppP p arg    -> patternAppView p ++ [arg]
+    OpAppP _ x ps -> mkHead (IdentP x) : map notHidden ps
+    ParenP _ p    -> patternAppView p
+    RawAppP _ _   -> __IMPOSSIBLE__
+    _             -> [ mkHead p ]
+  where mkHead    = Arg NotHidden Relevant . unnamed
+        notHidden = Arg NotHidden Relevant . unnamed
 
 
 ---------------------------------------------------------------------------
@@ -273,66 +338,211 @@ instance IsExpr Pattern where
 ---------------------------------------------------------------------------
 
 -- | Returns the list of possible parses.
-parsePattern :: ReadP Pattern Pattern -> Pattern -> [Pattern]
-parsePattern prs p = case p of
-    AppP p (Arg h r q) -> fullParen' <$> (AppP <$> parsePattern prs p <*> (Arg h r <$> traverse (parsePattern prs) q))
-    RawAppP _ ps     -> fullParen' <$> (parsePattern prs =<< parse prs ps)
-    OpAppP r d ps    -> fullParen' . OpAppP r d <$> mapM (parsePattern prs) ps
+parsePat :: ReadP Pattern Pattern -> Pattern -> [Pattern]
+parsePat prs p = case p of
+    AppP p (Arg h r q) -> fullParen' <$> (AppP <$> parsePat prs p <*> (Arg h r <$> traverse (parsePat prs) q))
+    RawAppP _ ps     -> fullParen' <$> (parsePat prs =<< parse prs ps)
+    OpAppP r d ps    -> fullParen' . OpAppP r d <$> mapM (parsePat prs) ps
     HiddenP _ _      -> fail "bad hidden argument"
     InstanceP _ _    -> fail "bad instance argument"
-    AsP r x p        -> AsP r x <$> parsePattern prs p
+    AsP r x p        -> AsP r x <$> parsePat prs p
     DotP r e         -> return $ DotP r e
-    ParenP r p       -> fullParen' <$> parsePattern prs p
+    ParenP r p       -> fullParen' <$> parsePat prs p
     WildP _          -> return p
     AbsurdP _        -> return p
     LitP _           -> return p
     IdentP _         -> return p
 
 
+{- Implement parsing of copattern left hand sides, e.g.
+
+  record Tree (A : Set) : Set where
+    field
+      label : A
+      child : Bool -> Tree A
+
+  -- corecursive function defined by copattern matching
+  alternate : {A : Set}(a b : A) -> Tree A
+  -- shallow copatterns
+         label (alternate a b)              = a
+         child (alternate a b) True         = alternate b a
+  -- deep copatterns:
+  label (child (alternate a b) False)       = b
+  child (child (alternate a b) False) True  = alternate a b
+  child (child (alternate a b) False) False = alternate a b
+
+  Delivers an infinite tree
+
+                   a
+              b        b
+            a   a    a   a
+           b b b b  b b b b
+                 ...
+
+  Each lhs is a pattern tree with a distinct path of destructors
+  ("child", "label") from the root to the defined symbol ("alternate").
+  All branches besides this distinct path are patterns.
+
+  Syntax.Concrete.LHSCore represents a lhs
+   - the destructor path
+   - the side patterns
+   - the defined function symbol
+   - the applied patterns
+-}
+
+type ParseLHS = Either Pattern (Name, LHSCore)
+
+parseLHS' :: LHSOrPatSyn -> Maybe Name -> Pattern -> ScopeM ParseLHS
+parseLHS' lhsOrPatSyn top p = do
+    let ms = qualifierModules $ patternQNames p
+    flat <- flattenScope ms <$> getScope
+    patP <- buildParser (getRange p) flat DontUseBoundNames
+    let cons = getNames [ConName, PatternSynName] flat
+    let flds = getNames [FldName] flat
+    case [ res | p' <- parsePat patP p
+               , res <- validPattern (PatternCheckConfig top cons flds) p' ] of
+        [(p,lhs)] -> return lhs
+        []        -> typeError $ NoParseForLHS lhsOrPatSyn p
+        rs        -> typeError $ AmbiguousParseForLHS lhsOrPatSyn p $
+                       map (fullParen . fst) rs
+    where
+        getNames kinds flat = map fst $ getDefinedNames kinds flat
+
+        -- validPattern returns an empty or singleton list (morally a Maybe)
+        validPattern :: PatternCheckConfig -> Pattern -> [(Pattern, ParseLHS)]
+        validPattern conf p = case (classifyPattern conf p, top) of
+            (Just r@(Left _), Nothing) -> [(p, r)] -- expect pattern
+            (Just r@(Right _), Just{}) -> [(p, r)] -- expect lhs
+            _ -> []
+
+-- | Name sets for classifying a pattern.
+data PatternCheckConfig = PatternCheckConfig
+  { topName  :: Maybe Name  -- ^ name of defined symbol
+  , conNames :: [QName]     -- ^ valid constructor names
+  , fldNames :: [QName]     -- ^ valid field names
+  }
+
+-- | Returns zero or one classified patterns.
+classifyPattern :: PatternCheckConfig -> Pattern -> Maybe ParseLHS
+classifyPattern conf p =
+  case patternAppView p of
+
+    -- case @f ps@
+    Arg _ _ (Named _ (IdentP x@(QName f))) : ps | Just f == topName conf -> do
+      guard $ all validPat ps
+      return $ Right (f, LHSHead f ps)
+
+    -- case @d ps@
+    Arg _ _ (Named _ (IdentP x)) : ps | x `elem` fldNames conf -> do
+      -- ps0 :: [NamedArg ParseLHS]
+      ps0 <- mapM classPat ps
+      let (ps1, rest) = span (isLeft . namedArg) ps0
+      (p2, ps3) <- uncons rest -- when (null rest): no field pattern or def pattern found
+      guard $ all (isLeft . namedArg) ps3
+      let (f, lhs)      = fromR p2
+          (ps', _:ps'') = splitAt (length ps1) ps
+      return $ Right (f, LHSProj x ps' lhs ps'')
+
+    -- case: ordinary pattern
+    _ -> do
+      guard $ validConPattern (conNames conf) p
+      return $ Left p
+
+  where -- allNames = conNames conf ++ fldNames conf
+        validPat = validConPattern (conNames conf) . namedArg
+        classPat :: NamedArg Pattern -> Maybe (NamedArg ParseLHS)
+        classPat = Trav.mapM (Trav.mapM (classifyPattern conf))
+        fromR :: NamedArg (Either a (b, c)) -> (b, NamedArg c)
+        fromR (Arg h r (Named n (Right (b, c)))) = (b, Arg h r (Named n c))
+        fromR (Arg h r (Named n (Left  a     ))) = __IMPOSSIBLE__
+
+
+
 -- | Parses a left-hand side, and makes sure that it defined the expected name.
 --   TODO: check the arities of constructors. There is a possible ambiguity with
 --   postfix constructors:
 --      Assume _ * is a constructor. Then 'true *' can be parsed as either the
 --      intended _* applied to true, or as true applied to a variable *. If we
 --      check arities this problem won't appear.
-parseLHS :: Maybe Name -> Pattern -> ScopeM Pattern
+parseLHS :: Name -> Pattern -> ScopeM LHSCore
 parseLHS top p = do
-    patP <- buildParser (getRange p) DontUseBoundNames
-    cons <- getNames [ConName]
-    case filter (validPattern top cons) $ parsePattern patP p of
-        [p] -> return p
-        []  -> typeError $ NoParseForLHS p
-        ps  -> typeError $ AmbiguousParseForLHS p $ map fullParen ps
-    where
-        getNames kinds = map fst <$> getDefinedNames kinds
+  res <- parseLHS' IsLHS (Just top) p
+  case res of
+    Right (f, lhs) -> return lhs
+    _ -> typeError $ NoParseForLHS IsLHS p
 
-        validPattern :: Maybe Name -> [Name] -> Pattern -> Bool
-        validPattern (Just top) cons p = case appView p of
-            IdentP (QName x) : ps -> x == top && all (validPat cons) ps
-            _                     -> False
-        validPattern Nothing cons p = validPat cons p
+-- | Parses a pattern.
+--   TODO: check the arities of constructors. There is a possible ambiguity with
+--   postfix constructors:
+--      Assume _ * is a constructor. Then 'true *' can be parsed as either the
+--      intended _* applied to true, or as true applied to a variable *. If we
+--      check arities this problem won't appear.
+parsePattern :: Pattern -> ScopeM Pattern
+parsePattern = parsePatternOrSyn IsLHS
 
-        validPat :: [Name] -> Pattern -> Bool
-        validPat cons p = case appView p of
-            [_]                   -> True
-            IdentP (QName x) : ps -> elem x cons && all (validPat cons) ps
-            ps                    -> all (validPat cons) ps
+parsePatternSyn :: Pattern -> ScopeM Pattern
+parsePatternSyn = parsePatternOrSyn IsPatSyn
 
-        appView :: Pattern -> [Pattern]
-        appView p = case p of
-            AppP p a         -> appView p ++ [namedThing (unArg a)]
-            OpAppP _ op ps   -> IdentP (QName op) : ps
-            ParenP _ p       -> appView p
-            RawAppP _ _      -> __IMPOSSIBLE__
-            HiddenP _ _      -> __IMPOSSIBLE__
-            InstanceP _ _    -> __IMPOSSIBLE__
-            _                -> [p]
+parsePatternOrSyn :: LHSOrPatSyn -> Pattern -> ScopeM Pattern
+parsePatternOrSyn lhsOrPatSyn p = do
+  res <- parseLHS' lhsOrPatSyn Nothing p
+  case res of
+    Left p -> return p
+    _      -> typeError $ NoParseForLHS lhsOrPatSyn p
 
+-- | Helper function for 'parseLHS' and 'parsePattern'.
+validConPattern :: [QName] -> Pattern -> Bool
+validConPattern cons p = case appView p of
+    [_]           -> True
+    IdentP x : ps -> elem x cons && all (validConPattern cons) ps
+    _             -> False
+-- Andreas, 2012-06-04: I do not know why the following line was
+-- the catch-all case.  It seems that the new catch-all works also
+-- and is more logical.
+--    ps            -> all (validConPattern cons) ps
+
+-- | Helper function for 'parseLHS' and 'parsePattern'.
+appView :: Pattern -> [Pattern]
+appView p = case p of
+    AppP p a         -> appView p ++ [namedThing (unArg a)]
+    OpAppP _ op ps   -> IdentP op : ps
+    ParenP _ p       -> appView p
+    RawAppP _ _      -> __IMPOSSIBLE__
+    HiddenP _ _      -> __IMPOSSIBLE__
+    InstanceP _ _    -> __IMPOSSIBLE__
+    _                -> [p]
+
+-- | Collect all names in a pattern into a list of qualified names.
+patternQNames :: Pattern -> [QName]
+patternQNames p = case p of
+  RawAppP _ ps     -> concatMap patternQNames ps
+  IdentP q         -> [q]
+  ParenP _ p       -> patternQNames p
+  HiddenP _ p      -> patternQNames (namedThing p)
+  InstanceP _ p    -> patternQNames (namedThing p)
+  OpAppP r d ps    -> __IMPOSSIBLE__
+  AppP{}           -> __IMPOSSIBLE__
+  AsP{}            -> __IMPOSSIBLE__
+  AbsurdP{}        -> []
+  WildP{}          -> []
+  DotP{}           -> []
+  LitP{}           -> []
+
+-- | Return all qualifiers occuring in a list of 'QName's.
+--   Each qualifier is returned as a list of names, e.g.
+--   for @Data.Nat._+_@ we return the list @[Data,Nat]@.
+qualifierModules :: [QName] -> [[Name]]
+qualifierModules qs =
+  nub $ filter (not . null) $ map (init . qnameParts) qs
+
+-- | Parse a list of expressions into an application.
 parseApplication :: [Expr] -> ScopeM Expr
 parseApplication [e] = return e
 parseApplication es = do
     -- Build the parser
-    p <- buildParser (getRange es) UseBoundNames
+    let ms = qualifierModules [ q | Ident q <- es ]
+    flat <- flattenScope ms <$> getScope
+    p <- buildParser (getRange es) flat UseBoundNames
 
     -- Parse
     case parse p es of
@@ -340,14 +550,16 @@ parseApplication es = do
         []  -> do
           -- When the parser fails and a name is not in scope, it is more
           -- useful to say that to the user rather than just "failed".
-          inScope <- partsInScope
-          case [ QName x | Ident (QName x) <- es, not (Set.member x inScope) ] of
+          inScope <- partsInScope flat
+          case [ x | Ident x <- es, not (Set.member x inScope) ] of
                []  -> typeError $ NoParseForApplication es
                xs  -> typeError $ NotInScope xs
 
         es' -> typeError $ AmbiguousParseForApplication es $ map fullParen es'
 
--- Inserting parenthesis --------------------------------------------------
+---------------------------------------------------------------------------
+-- * Inserting parenthesis
+---------------------------------------------------------------------------
 
 fullParen :: IsExpr e => e -> e
 fullParen e = case exprView $ fullParen' e of
@@ -373,7 +585,7 @@ fullParen' e = case exprView e of
     where
         par = unExprView . ParenV
 
-paren :: Monad m => (Name -> m Fixity) -> Expr -> m (Precedence -> Expr)
+paren :: Monad m => (QName -> m Fixity) -> Expr -> m (Precedence -> Expr)
 paren _   e@(App _ _ _)        = return $ \p -> mparen (appBrackets p) e
 paren f   e@(OpApp _ op _)     = do fx <- f op; return $ \p -> mparen (opBrackets fx p) e
 paren _   e@(Lam _ _ _)        = return $ \p -> mparen (lamBrackets p) e
