@@ -8,6 +8,7 @@ module Agda.Syntax.Concrete.Pretty where
 import Data.Char
 
 import Agda.Syntax.Common
+import Agda.Syntax.Position
 import Agda.Syntax.Concrete
 import Agda.Syntax.Fixity
 import Agda.Syntax.Literal
@@ -46,7 +47,12 @@ pHidden Instance    = dbraces . pretty
 pHidden NotHidden   = pretty
 
 pRelevance :: Pretty a => Relevance -> a -> Doc
+pRelevance rel a =
+  let d = pretty a
+  in  if render d == "_" then d else pretty rel <> d
+{-
 pRelevance Forced     a = pretty a
+pRelevance UnusedArg  a = pretty a
 pRelevance Relevant   a = pretty a
 pRelevance Irrelevant a =
   let d = pretty a
@@ -54,11 +60,6 @@ pRelevance Irrelevant a =
 pRelevance NonStrict a =
   let d = pretty a
   in  if render d == "_" then d else text ".." <> d
-
-{- UNUSED
--- | Use for printing non-dependent function types
-prettyWithRelevance :: Pretty a => Arg a -> Doc
-prettyWithRelevance a = pRelevance (argRelevance a) a
 -}
 
 instance Pretty (ThingWithFixity Name) where
@@ -91,6 +92,7 @@ showChar' c
 
 instance Pretty Relevance where
   pretty Forced     = empty
+  pretty UnusedArg  = empty
   pretty Relevant   = empty
   pretty Irrelevant = text "."
   pretty NonStrict  = text ".."
@@ -109,7 +111,8 @@ instance Pretty Expr where
 	    Ident x	     -> pretty x
 	    Lit l	     -> pretty l
 	    QuestionMark _ n -> text "?" <> maybe empty (text . show) n
-	    Underscore _ n   -> underscore <> maybe empty (text . show) n
+	    Underscore _ n   -> maybe underscore text n
+--	    Underscore _ n   -> underscore <> maybe empty (text . show) n
 	    App _ _ _	     ->
 		case appView e of
 		    AppView e1 args	->
@@ -118,13 +121,7 @@ instance Pretty Expr where
 -- 			    , nest 2 $ fsep $ map pretty args
 -- 			    ]
 	    RawApp _ es   -> fsep $ map pretty es
-	    OpApp _ (Name _ xs) es -> fsep $ prOp xs es
-		where
-		    prOp (Hole : xs) (e : es) = pretty e : prOp xs es
-		    prOp (Hole : _)  []       = __IMPOSSIBLE__
-		    prOp (Id x : xs) es       = text x : prOp xs es
-		    prOp []	     es       = map pretty es
-	    OpApp _ (NoName _ _) _ -> __IMPOSSIBLE__
+	    OpApp _ q es -> fsep $ prettyOpApp q es
 
 	    WithApp _ e es -> fsep $
 	      pretty e : map ((text "|" <+>) . pretty) es
@@ -139,7 +136,7 @@ instance Pretty Expr where
             AbsurdLam _ Instance -> lambda <+> text "{{}}"
             AbsurdLam _ Hidden -> lambda <+> text "{}"
 	    ExtendedLam _ pes ->
-		lambda <+> braces (fsep $ punctuate (text ";") (map (\(x,y,z) -> prettyClause x y z) pes))
+		lambda <+> braces' (fsep $ punctuate (text ";") (map (\(x,y,z) -> prettyClause x y z) pes))
                    where prettyClause lhs rhs wh = sep [ pretty lhs
                                                        , nest 2 $ pretty' rhs
                                                        ] $$ nest 2 (pretty wh)
@@ -247,6 +244,13 @@ instance Pretty LHS where
       pThing thing (e : es) = fsep $ (text thing <+> pretty e)
 			           : map ((text "|" <+>) . pretty) es
 
+instance Show LHSCore where show = show . pretty
+instance Pretty LHSCore where
+  pretty (LHSHead f ps) = sep $ pretty f : map (parens . pretty) ps
+  pretty (LHSProj d ps lhscore ps') = sep $
+    pretty d : map (parens . pretty) ps ++
+    parens (pretty lhscore) : map (parens . pretty) ps'
+
 instance Pretty [Declaration] where
   pretty = vcat . map pretty
 
@@ -308,19 +312,25 @@ instance Pretty Declaration where
 			    , pretty e
 			    ]
 		    ]
-	    Record _ x con tel (Just e) cs ->
+	    Record _ x ind con tel me cs ->
 		sep [ hsep  [ text "record"
 			    , pretty x
 			    , fcat (map pretty tel)
 			    ]
-		    , nest 2 $ hsep
+		    , nest 2 $ pType me
+		    ] $$ nest 2 (vcat $ pInd ++
+                                        pCon ++
+                                        map pretty cs)
+              where pType (Just e) = hsep
 			    [ text ":"
 			    , pretty e
 			    , text "where"
 			    ]
-		    ] $$ nest 2 (vcat $ maybe [] (\c -> [text "constructor" <+> pretty c])
-                                              con ++
-                                        map pretty cs)
+                    pType Nothing  =
+                              text "where"
+                    pInd = maybe [] (\i -> [text $ show i]) ind
+                    pCon = maybe [] (\c -> [text "constructor" <+> pretty c]) con
+{- ELIMINATED CUT-AND-PASTE CODE
 	    Record _ x con tel Nothing cs ->
 		sep [ hsep  [ text "record"
 			    , pretty x
@@ -330,10 +340,12 @@ instance Pretty Declaration where
 		    ] $$ nest 2 (vcat $ maybe [] (\c -> [text "constructor" <+> pretty c])
                                               con ++
                                         map pretty cs)
-
+-}
             Infix f xs	->
 		pretty f <+> (fsep $ punctuate comma $ map pretty xs)
             Syntax n xs -> text "syntax" <+> pretty n <+> text "..."
+            PatternSyn _ n as p -> text "pattern" <+> pretty n <+> fsep (map pretty as)
+                                     <+> text "=" <+> pretty p
 	    Mutual _ ds	    -> namedBlock "mutual" ds
 	    Abstract _ ds   -> namedBlock "abstract" ds
 	    Private _ ds    -> namedBlock "private" ds
@@ -396,6 +408,7 @@ instance Pretty Pragma where
       hsep $ [text "IMPOSSIBLE"]
     pretty (EtaPragma _ x) =
       hsep $ [text "ETA", pretty x]
+    pretty (NoTerminationCheckPragma _) = text "NO_TERMINATION_CHECK"
 
 instance Pretty Fixity where
     pretty (LeftAssoc _ n)  = text "infixl" <+> text (show n)
@@ -418,25 +431,31 @@ instance Pretty [Pattern] where
 instance Pretty Pattern where
     pretty p =
 	case p of
-	    IdentP x	       -> pretty x
-	    AppP p1 p2	       -> sep [ pretty p1, nest 2 $ pretty p2 ]
-	    RawAppP _ ps       -> fsep $ map pretty ps
-	    OpAppP _ (Name _ xs) ps -> fsep $ prOp xs ps
-		where
-		    prOp (Hole : xs) (e : es) = pretty e : prOp xs es
-		    prOp (Hole : _)  []	      = __IMPOSSIBLE__
-		    prOp (Id x : xs) es       = text x : prOp xs es
-		    prOp []	     []       = []
-		    prOp []	     es       = map pretty es
-	    OpAppP _ (NoName _ _) _ -> __IMPOSSIBLE__
-	    HiddenP _ p	       -> braces' $ pretty p
-	    InstanceP _ p      -> dbraces $ pretty p
-	    ParenP _ p	       -> parens $ pretty p
-	    WildP _	       -> underscore
-	    AsP _ x p	       -> pretty x <> text "@" <> pretty p
-	    DotP _ p	       -> text "." <> pretty p
-	    AbsurdP _	       -> text "()"
-	    LitP l	       -> pretty l
+	    IdentP x      -> pretty x
+	    AppP p1 p2    -> sep [ pretty p1, nest 2 $ pretty p2 ]
+	    RawAppP _ ps  -> fsep $ map pretty ps
+	    OpAppP _ q ps -> fsep $ prettyOpApp q ps
+	    HiddenP _ p   -> braces' $ pretty p
+	    InstanceP _ p -> dbraces $ pretty p
+	    ParenP _ p    -> parens $ pretty p
+	    WildP _       -> underscore
+	    AsP _ x p     -> pretty x <> text "@" <> pretty p
+	    DotP _ p      -> text "." <> pretty p
+	    AbsurdP _     -> text "()"
+	    LitP l        -> pretty l
+
+prettyOpApp :: Pretty a => QName -> [a] -> [Doc]
+prettyOpApp q es = prOp ms xs es
+  where
+    ms = init (qnameParts q)
+    xs = case unqualify q of
+           Name _ xs -> xs
+           NoName{}  -> __IMPOSSIBLE__
+    prOp ms (Hole : xs) (e : es) = pretty e : prOp ms xs es
+    prOp _  (Hole : _)  []       = __IMPOSSIBLE__
+    prOp ms (Id x : xs) es       = pretty (foldr Qual (QName (Name noRange $ [Id x])) ms) : prOp [] xs es
+    prOp _  []	     []          = []
+    prOp _  []	     es          = map pretty es
 
 instance Pretty ImportDirective where
     pretty i =
