@@ -4,7 +4,7 @@
 -}
 module Agda.Syntax.Fixity where
 
-import Data.Generics (Typeable, Data)
+import Data.Typeable (Typeable)
 import Data.Foldable
 import Data.Traversable
 
@@ -18,15 +18,15 @@ import Agda.Syntax.Notation
 data Fixity' = Fixity'
     {theFixity :: Fixity,
      theNotation :: Notation}
-  deriving (Typeable, Data, Show, Eq)
+  deriving (Typeable, Show, Eq)
 
-data ThingWithFixity x = ThingWithFixity x Fixity' deriving (Functor,Foldable,Traversable,Typeable,Data,Show)
+data ThingWithFixity x = ThingWithFixity x Fixity' deriving (Functor,Foldable,Traversable,Typeable,Show)
 -- | All the notation information related to a name.
-type NewNotation = (Name, Fixity, Notation)
+type NewNotation = (QName, Fixity, Notation)
 
 -- | If an operator has no specific notation, recover it from its name.
-oldToNewNotation :: (Name, Fixity') -> NewNotation
-oldToNewNotation (name, Fixity' f []) = (name, f, syntaxOf name)
+oldToNewNotation :: (QName, Fixity') -> NewNotation
+oldToNewNotation (name, Fixity' f []) = (name, f, syntaxOf $ unqualify name)
 oldToNewNotation (name, Fixity' f syn) = (name, f, syn)
 
 syntaxOf :: Name -> Notation
@@ -41,13 +41,13 @@ syntaxOf (Name _ xs)  = mkSyn 0 xs
 
 defaultFixity' = Fixity' defaultFixity defaultNotation
 
-noFixity = NonAssoc noRange (negate 666)
+noFixity = NonAssoc noRange (negate 666) -- ts,ts,ts, why the number of the beast?  Revelation 13, 18
 
 -- | Fixity of operators.
-data Fixity = LeftAssoc  Range Nat
-	    | RightAssoc Range Nat
-	    | NonAssoc   Range Nat
-    deriving (Typeable, Data, Show)
+data Fixity = LeftAssoc  Range Integer
+	    | RightAssoc Range Integer
+	    | NonAssoc   Range Integer
+    deriving (Typeable, Show)
 
 instance Eq Fixity where
     LeftAssoc _ n   == LeftAssoc _ m	= n == m
@@ -55,7 +55,7 @@ instance Eq Fixity where
     NonAssoc _ n    == NonAssoc _ m	= n == m
     _		    == _		= False
 
-fixityLevel :: Fixity -> Nat
+fixityLevel :: Fixity -> Integer
 fixityLevel (LeftAssoc	_ n) = n
 fixityLevel (RightAssoc _ n) = n
 fixityLevel (NonAssoc	_ n) = n
@@ -69,7 +69,7 @@ data Precedence = TopCtx | FunctionSpaceDomainCtx
 		| LeftOperandCtx Fixity | RightOperandCtx Fixity
 		| FunctionCtx | ArgumentCtx | InsideOperandCtx
                 | WithFunCtx | WithArgCtx | DotPatternCtx
-    deriving (Show,Typeable,Data)
+    deriving (Show,Typeable)
 
 
 -- | The precedence corresponding to a possibly hidden argument.
@@ -135,3 +135,11 @@ instance HasRange Fixity where
     getRange (LeftAssoc  r _)	= r
     getRange (RightAssoc r _)	= r
     getRange (NonAssoc   r _)	= r
+
+instance KillRange Fixity where
+  killRange (LeftAssoc  _ n) = LeftAssoc  noRange n
+  killRange (RightAssoc _ n) = RightAssoc noRange n
+  killRange (NonAssoc   _ n) = NonAssoc   noRange n
+
+instance KillRange Fixity' where
+  killRange (Fixity' f n) = killRange1 (flip Fixity' n) f
