@@ -7,7 +7,7 @@
 
 module Agda.Syntax.Info where
 
-import Data.Generics (Typeable, Data)
+import Data.Typeable (Typeable)
 import Text.Show.Functions
 
 import qualified Agda.Syntax.Concrete.Name as C
@@ -15,7 +15,7 @@ import Agda.Syntax.Common
 import Agda.Syntax.Position
 import Agda.Syntax.Concrete
 import Agda.Syntax.Fixity
-import Agda.Syntax.Scope.Base (ScopeInfo)
+import Agda.Syntax.Scope.Base (ScopeInfo, emptyScopeInfo)
 
 {--------------------------------------------------------------------------
     No information
@@ -31,8 +31,17 @@ data MetaInfo =
 	MetaInfo { metaRange	:: Range
 		 , metaScope	:: ScopeInfo
 		 , metaNumber	:: Maybe Nat
+                 , metaNameSuggestion :: String
 		 }
-  deriving (Typeable, Data, Show)
+  deriving (Typeable, Show)
+
+emptyMetaInfo :: MetaInfo
+emptyMetaInfo = MetaInfo
+  { metaRange          = noRange
+  , metaScope          = emptyScopeInfo
+  , metaNumber         = Nothing
+  , metaNameSuggestion = ""
+  }
 
 instance HasRange MetaInfo where
   getRange = metaRange
@@ -51,7 +60,7 @@ data ExprInfo
 	| ExprSource Range (Precedence -> Expr)
 	    -- ^ Even if we store the original expression we have to know
 	    --	 whether to put parenthesis around it.
-  deriving (Typeable, Data, Show)
+  deriving (Typeable, Show)
 
 instance HasRange ExprInfo where
   getRange (ExprRange  r  ) = r
@@ -77,13 +86,16 @@ data ModuleInfo =
                    , minfoDirective :: Maybe ImportDirective
                      -- Retained for abstractToConcrete of ModuleMacro
 		   }
-  deriving (Typeable, Data)
+  deriving (Typeable)
 
 deriving instance (Show OpenShortHand, Show ImportDirective) => Show ModuleInfo
 
 instance HasRange ModuleInfo where
   getRange = minfoRange
 
+instance SetRange ModuleInfo where
+  setRange r i = i { minfoRange = r }
+
 instance KillRange ModuleInfo where
   killRange m = m { minfoRange = killRange $ minfoRange m }
 
@@ -92,7 +104,7 @@ instance KillRange ModuleInfo where
 ---------------------------------------------------------------------------
 
 newtype LetInfo = LetRange Range
-  deriving (Typeable, Data, Show)
+  deriving (Typeable, Show)
 
 instance HasRange LetInfo where
   getRange (LetRange r)   = r
@@ -110,7 +122,7 @@ data DefInfo =
 		, defAbstract :: IsAbstract
 		, defInfo     :: DeclInfo
 		}
-  deriving (Typeable, Data, Show)
+  deriving (Typeable, Show)
 
 mkDefInfo :: Name -> Fixity' -> Access -> IsAbstract -> Range -> DefInfo
 mkDefInfo x f a ab r = DefInfo f a ab (DeclInfo x r)
@@ -118,6 +130,9 @@ mkDefInfo x f a ab r = DefInfo f a ab (DeclInfo x r)
 instance HasRange DefInfo where
   getRange = getRange . defInfo
 
+instance SetRange DefInfo where
+  setRange r i = i { defInfo = setRange r (defInfo i) }
+
 instance KillRange DefInfo where
   killRange i = i { defInfo = killRange $ defInfo i }
 
@@ -129,20 +144,39 @@ data DeclInfo =
 	DeclInfo { declName  :: Name
 		 , declRange :: Range
 		 }
-  deriving (Typeable, Data, Show)
+  deriving (Typeable, Show)
 
 instance HasRange DeclInfo where
   getRange = declRange
 
+instance SetRange DeclInfo where
+  setRange r i = i { declRange = r }
+
 instance KillRange DeclInfo where
   killRange i = i { declRange = killRange $ declRange i }
 
+{--------------------------------------------------------------------------
+    Mutual block information
+ --------------------------------------------------------------------------}
+
+data MutualInfo =
+     MutualInfo { mutualTermCheck :: Bool  -- ^ termination check (default=True)
+		, mutualRange     :: Range
+		}
+  deriving (Typeable, Show)
+
+instance HasRange MutualInfo where
+  getRange = mutualRange
+
+instance KillRange MutualInfo where
+  killRange i = i { mutualRange = killRange $ mutualRange i }
+
 {--------------------------------------------------------------------------
     Left hand side information
  --------------------------------------------------------------------------}
 
 newtype LHSInfo = LHSRange Range
-  deriving (Typeable, Data, Show)
+  deriving (Typeable, Show)
 
 instance HasRange LHSInfo where
   getRange (LHSRange r) = r
@@ -159,7 +193,7 @@ instance KillRange LHSInfo where
 
 data PatInfo = PatRange Range
 	     | PatSource Range (Precedence -> Pattern)
-  deriving (Typeable, Data)
+  deriving (Typeable)
 
 instance Show PatInfo where
   show (PatRange r) = "PatRange " ++ show r
