@@ -16,11 +16,12 @@ data Keyword
 	| KwOpen | KwImport | KwModule | KwPrimitive
 	| KwInfix | KwInfixL | KwInfixR | KwWith | KwRewrite
 	| KwSet | KwProp | KwForall | KwRecord | KwConstructor | KwField
+        | KwInductive | KwCoInductive
 	| KwHiding | KwUsing | KwRenaming | KwTo | KwPublic
 	| KwOPTIONS | KwBUILTIN | KwLINE
 	| KwCOMPILED_DATA | KwCOMPILED_TYPE | KwCOMPILED | KwCOMPILED_EPIC | KwCOMPILED_JS
-	| KwIMPORT | KwIMPOSSIBLE | KwETA | KwSTATIC
-	| KwQuoteGoal | KwQuote | KwQuoteTerm | KwUnquote | KwSyntax
+	| KwIMPORT | KwIMPOSSIBLE | KwETA | KwNO_TERMINATION_CHECK | KwSTATIC
+	| KwQuoteGoal | KwQuote | KwQuoteTerm | KwUnquote | KwSyntax | KwPatternSyn
     deriving (Eq, Show)
 
 layoutKeywords :: [Keyword]
@@ -43,7 +44,9 @@ data Token
 	= TokKeyword Keyword Interval
 	  -- Identifiers and operators
 	| TokId		(Interval, String)
-	| TokQId	[(Interval, String)] -- non empty namespace
+	| TokQId	[(Interval, String)]
+                        -- Non-empty namespace. The intervals for
+                        -- "A.B.x" correspond to "A.", "B." and "x".
 	  -- Literals
 	| TokLiteral	Literal
 	  -- Special symbols
