@@ -1,4 +1,4 @@
-{-# LANGUAGE CPP, DeriveDataTypeable, GADTs, ScopedTypeVariables, TupleSections #-}
+{-# LANGUAGE CPP, DeriveDataTypeable, GADTs, ScopedTypeVariables, TupleSections, RankNTypes #-}
 
 {-| This module defines the notion of a scope and operations on scopes.
 -}
@@ -6,7 +6,7 @@ module Agda.Syntax.Scope.Base where
 
 import Control.Arrow ((***), (&&&))
 import Control.Applicative
-import Data.Generics (Typeable, Data)
+import Data.Typeable (Typeable)
 import Data.List
 import Data.Map (Map)
 import qualified Data.Map as Map
@@ -31,15 +31,16 @@ import Agda.Utils.Impossible
 -- | A scope is a named collection of names partitioned into public and private
 --   names.
 data Scope = Scope
-      { scopeName          :: A.ModuleName
-      , scopeParents       :: [A.ModuleName]
-      , scopeNameSpaces    :: [(NameSpaceId, NameSpace)]
-      , scopeImports       :: Map C.QName A.ModuleName
+      { scopeName           :: A.ModuleName
+      , scopeParents        :: [A.ModuleName]
+      , scopeNameSpaces     :: [(NameSpaceId, NameSpace)]
+      , scopeImports        :: Map C.QName A.ModuleName
+      , scopeDatatypeModule :: Bool
       }
-  deriving (Typeable, Data)
+  deriving (Typeable)
 
 data NameSpaceId = PrivateNS | PublicNS | ImportedNS | OnlyQualifiedNS
-  deriving (Typeable, Data, Eq, Bounded, Enum)
+  deriving (Typeable, Eq, Bounded, Enum)
 
 localNameSpace :: Access -> NameSpaceId
 localNameSpace PublicAccess  = PublicNS
@@ -61,7 +62,7 @@ data ScopeInfo = ScopeInfo
       , scopeLocals	:: LocalVars
       , scopePrecedence :: Precedence
       }
-  deriving (Typeable, Data)
+  deriving (Typeable)
 
 -- | Local variables
 type LocalVars = [(C.Name, A.Name)]
@@ -73,7 +74,7 @@ data NameSpace = NameSpace
       { nsNames	  :: NamesInScope
       , nsModules :: ModulesInScope
       }
-  deriving (Typeable, Data)
+  deriving (Typeable)
 
 type ThingsInScope a = Map C.Name [a]
 type NamesInScope    = ThingsInScope AbstractName
@@ -101,8 +102,11 @@ instance InScope AbstractModule where
   inScopeTag = ModuleTag
 
 -- | We distinguish constructor names from other names.
-data KindOfName = ConName | DefName
-  deriving (Eq, Show, Typeable, Data)
+data KindOfName = ConName | FldName | DefName | PatternSynName
+  deriving (Eq, Show, Typeable, Enum, Bounded)
+
+allKindsOfNames :: [KindOfName]
+allKindsOfNames = [minBound..maxBound]
 
 -- | Apart from the name, we also record whether it's a constructor or not and
 --   what the fixity is.
@@ -110,13 +114,13 @@ data AbstractName = AbsName
       { anameName   :: A.QName
       , anameKind   :: KindOfName
       }
-  deriving (Typeable, Data)
+  deriving (Typeable)
 
 -- | For modules we record the arity. I'm not sure that it's every used anywhere.
 data AbstractModule = AbsModule
       { amodName    :: A.ModuleName
       }
-  deriving (Typeable, Data)
+  deriving (Typeable)
 
 instance Eq AbstractName where
   (==) = (==) `on` anameName
@@ -234,11 +238,12 @@ mapNameSpaceM fd fm ns = do
 
 -- | The empty scope.
 emptyScope :: Scope
-emptyScope = Scope { scopeName	     = noModuleName
-                   , scopeParents    = []
-		   , scopeNameSpaces = [ (nsid, emptyNameSpace) | nsid <- [minBound..maxBound] ]
-                   , scopeImports    = Map.empty
-		   }
+emptyScope = Scope { scopeName	         = noModuleName
+                   , scopeParents        = []
+		   , scopeNameSpaces     = [ (nsid, emptyNameSpace) | nsid <- [minBound..maxBound] ]
+                   , scopeImports        = Map.empty
+		   , scopeDatatypeModule = False
+                   }
 
 -- | The empty scope info.
 emptyScopeInfo :: ScopeInfo
@@ -474,6 +479,45 @@ everythingInScope scope =
       Nothing -> __IMPOSSIBLE__
     current = this : parents
 
+-- | Compute a flattened scope. Only include unqualified names or names
+-- qualified by modules in the first argument.
+flattenScope :: [[C.Name]] -> ScopeInfo -> Map C.QName [AbstractName]
+flattenScope ms scope =
+  -- Map.filterKeys (\q -> elem (init $ C.qnameParts q) ([]:ms)) $
+  Map.unionWith (++)
+    (build ms allNamesInScope root)
+    imported
+  where
+    current = moduleScope $ scopeCurrent scope
+    root    = mergeScopes $ current : map moduleScope (scopeParents current)
+
+    imported = Map.unionsWith (++)
+               [ qual c (build ms' exportedNamesInScope $ moduleScope a)
+               | (c, a) <- Map.toList $ scopeImports root
+               , let m   = C.qnameParts c
+                     ms' = map (drop (length m)) $ filter (m `isPrefixOf`) ms
+               , not $ null ms' ]
+    qual c = Map.mapKeys (q c)
+      where
+        q (C.QName x)  = C.Qual x
+        q (C.Qual m x) = C.Qual m . q x
+
+    build :: [[C.Name]] -> (forall a. InScope a => Scope -> ThingsInScope a) -> Scope -> Map C.QName [AbstractName]
+    build ms getNames s =
+      Map.unionWith (++)
+        (Map.mapKeys (\x -> C.QName x) (getNames s))
+        $ Map.unionsWith (++) $
+          [ Map.mapKeys (\y -> C.Qual x y) $ build ms' exportedNamesInScope $ moduleScope m
+          | (x, mods) <- Map.toList (getNames s)
+          , let ms' = [ ms' | m':ms' <- ms, m' == x ]
+          , not $ null ms'
+          , AbsModule m <- mods ]
+
+    moduleScope :: A.ModuleName -> Scope
+    moduleScope name = case Map.lookup name (scopeModules scope) of
+      Nothing -> __IMPOSSIBLE__
+      Just s  -> s
+
 -- | Look up a name in the scope
 scopeLookup :: InScope a => C.QName -> ScopeInfo -> [a]
 scopeLookup q scope = map fst $ scopeLookup' q scope
