@@ -46,6 +46,11 @@ type ScopeM = TCM
 notInScope :: C.QName -> ScopeM a
 notInScope x = typeError $ NotInScope [x]
 
+isDatatypeModule :: A.ModuleName -> ScopeM Bool
+isDatatypeModule m = do
+   sc <- getScope
+   return $ maybe __IMPOSSIBLE__ scopeDatatypeModule (Map.lookup m (scopeModules sc))
+
 -- * General operations
 
 getCurrentModule :: ScopeM A.ModuleName
@@ -82,12 +87,14 @@ getNamedScope m = do
 getCurrentScope :: ScopeM Scope
 getCurrentScope = getNamedScope =<< getCurrentModule
 
--- | Create a new module with an empty scope
-createModule :: A.ModuleName -> ScopeM ()
-createModule m = do
+-- | Create a new module with an empty scope (Bool is True if it is a datatype module)
+createModule :: Bool -> A.ModuleName -> ScopeM ()
+createModule b m = do
   s <- getCurrentScope
   let parents = scopeName s : scopeParents s
-  modifyScopes $ Map.insert m emptyScope { scopeName = m, scopeParents = parents }
+  modifyScopes $ Map.insert m emptyScope { scopeName           = m
+                                         , scopeParents        = parents
+                                         , scopeDatatypeModule = b }
 
 -- | Apply a function to the scope info.
 modifyScopeInfo :: (ScopeInfo -> ScopeInfo) -> ScopeM ()
@@ -184,7 +191,9 @@ freshAbstractQName fx x = do
 
 data ResolvedName = VarName A.Name
                   | DefinedName Access AbstractName
+                  | FieldName AbstractName           -- ^ record fields names need to be distinguished to parse copatterns
                   | ConstructorName [AbstractName]
+                  | PatternSynResName AbstractName
                   | UnknownName
   deriving (Show)
 
@@ -200,6 +209,9 @@ resolveName x = do
       ds | all ((==ConName) . anameKind . fst) ds ->
         return $ ConstructorName
                $ map (\ (d, _) -> updateConcreteName d $ unqualify x) ds
+      [(d, a)] | anameKind d == FldName -> return $ FieldName $ updateConcreteName d (unqualify x)
+      [(d, a)] | anameKind d == PatternSynName ->
+                  return $ PatternSynResName (updateConcreteName d $ unqualify x)
       [(d, a)] -> return $ DefinedName a $ updateConcreteName d (unqualify x)
       ds  -> typeError $ AmbiguousName x (map (anameName . fst) ds)
   where
@@ -223,13 +235,15 @@ getFixity x = do
   case r of
     VarName y          -> return $ nameFixity y
     DefinedName _ d    -> return $ nameFixity $ qnameName $ anameName d
+    FieldName d        -> return $ nameFixity $ qnameName $ anameName d
     ConstructorName ds
       | null fs        -> __IMPOSSIBLE__
       | allEqual fs    -> return $ head fs
       | otherwise      -> return defaultFixity'
       where
         fs = map (nameFixity . qnameName . anameName) ds
-    UnknownName        -> __IMPOSSIBLE__
+    PatternSynResName n -> return $ nameFixity $ qnameName $ anameName n
+    UnknownName         -> __IMPOSSIBLE__
 
 -- * Binding names
 
@@ -244,13 +258,15 @@ bindName :: Access -> KindOfName -> C.Name -> A.QName -> ScopeM ()
 bindName acc kind x y = do
   r  <- resolveName (C.QName x)
   ys <- case r of
+    FieldName d        -> typeError $ ClashingDefinition (C.QName x) $ anameName d
     DefinedName _ d    -> typeError $ ClashingDefinition (C.QName x) $ anameName d
     VarName z          -> typeError $ ClashingDefinition (C.QName x) $ A.qualify (mnameFromList []) z
     ConstructorName [] -> __IMPOSSIBLE__
     ConstructorName ds
       | kind == ConName && all ((==ConName) . anameKind) ds -> return [ AbsName y kind ]
       | otherwise -> typeError $ ClashingDefinition (C.QName x) $ anameName (head' ds)
-    UnknownName        -> return [AbsName y kind]
+    PatternSynResName n -> typeError $ ClashingDefinition (C.QName x) $ anameName n
+    UnknownName         -> return [AbsName y kind]
   modifyCurrentScope $ addNamesToScope (localNameSpace acc) x ys
   where
     head' []    = {- ' -} __IMPOSSIBLE__
@@ -355,7 +371,7 @@ copyScope new s = runStateT (copy new s) (Map.empty, Map.empty)
           addMod x y
 
           -- We need to copy the contents of included modules recursively
-          s0 <- lift $ createModule y >> getNamedScope x
+          s0 <- lift $ createModule False y >> getNamedScope x
           s  <- withCurrentModule' y $ copy y s0
           lift $ modifyNamedScope y (const s)
           return y
