@@ -35,6 +35,7 @@ import qualified Data.Traversable as Trav
 import Agda.Syntax.Common
 import Agda.Syntax.Position
 import Agda.Syntax.Info
+import Agda.Syntax.Internal (MetaId(..))
 import Agda.Syntax.Fixity
 import Agda.Syntax.Concrete as C
 import Agda.Syntax.Concrete.Pretty
@@ -43,7 +44,7 @@ import Agda.Syntax.Abstract.Views as AV
 import Agda.Syntax.Scope.Base
 
 import Agda.TypeChecking.Monad.State (getScope)
-import Agda.TypeChecking.Monad.Base  (TCM)
+import Agda.TypeChecking.Monad.Base  (TCM, NamedMeta(..))
 
 import Agda.Utils.Maybe
 import Agda.Utils.Monad hiding (bracket)
@@ -311,9 +312,9 @@ instance ToConcrete A.Expr C.Expr where
     toConcrete (A.QuestionMark i)   = return $ C.QuestionMark
                                                 (getRange i)
                                                 (metaNumber i)
-    toConcrete (A.Underscore i)     = return $ C.Underscore
-                                                (getRange i)
-                                                (metaNumber i)
+    toConcrete (A.Underscore i)     = return $
+      C.Underscore (getRange i) $
+       fmap (show . NamedMeta (metaNameSuggestion i) . MetaId) (metaNumber i)
 
     toConcrete e@(A.App i e1 e2)    =
         tryToRecoverOpApp e
@@ -350,10 +351,14 @@ instance ToConcrete A.Expr C.Expr where
                     (bs@(A.DomainFull _ : _), e)   -> (b:bs, e)
                     _                              -> ([b], e)
             lamView e = ([], e)
-    toConcrete (A.ExtendedLam i di qname cs) = do
+    toConcrete (A.ExtendedLam i di qname cs) =
+        bracket lamBrackets $ do
           decls <- toConcrete cs
-          let removeApp (C.RawAppP r (lam:es)) = C.RawAppP r es
-              removeApp (C.AppP p np) = namedThing $ unArg np
+              -- we know all lhs are of the form `.extlam p1 p2 ... pn`,
+              -- with the name .extlam leftmost. It is our mission to remove it.
+          let removeApp (C.RawAppP r (_:es)) = C.RawAppP r es
+              removeApp (C.AppP (C.IdentP _) np) = namedArg np
+              removeApp (C.AppP p np) = (C.AppP (removeApp p) np)
               removeApp _ = __IMPOSSIBLE__
           let decl2clause (C.FunClause lhs rhs wh) = (lhs {lhsOriginalPattern = removeApp $ lhsOriginalPattern lhs},rhs,wh)
               decl2clause _ = __IMPOSSIBLE__
@@ -373,10 +378,14 @@ instance ToConcrete A.Expr C.Expr where
         bracket piBrackets
         $ do a' <- toConcreteCtx (if irr then DotPatternCtx else FunctionSpaceDomainCtx) a
              b' <- toConcreteCtx TopCtx b
-             return $ C.Fun (getRange i) (addDot a' $ mkArg a') b'
+             return $ C.Fun (getRange i) (addRel a' $ mkArg a') b'
         where
-            irr        = argRelevance a == Irrelevant
-            addDot a e = if irr then Dot (getRange a) e else e
+            irr        = argRelevance a `elem` [Irrelevant, NonStrict]
+            addRel a e = case argRelevance a of
+                           Irrelevant -> addDot a e
+                           NonStrict  -> addDot a (addDot a e)
+                           _          -> e
+            addDot a e = Dot (getRange a) e
             mkArg (Arg Hidden    r e) = HiddenArg (getRange e) (unnamed e)
             mkArg (Arg Instance  r e) = InstanceArg (getRange e) (unnamed e)
             mkArg (Arg NotHidden r e) = e
@@ -419,13 +428,16 @@ instance ToConcrete A.Expr C.Expr where
     toConcrete (A.QuoteTerm i) = return $ C.QuoteTerm (getRange i)
     toConcrete (A.Unquote i) = return $ C.Unquote (getRange i)
 
+    -- Andreas, 2012-04-02: TODO!  print DontCare as irrAxiom
     -- Andreas, 2010-10-05 print irrelevant things as ordinary things
-    toConcrete (A.DontCare e) = toConcrete e
+    toConcrete (A.DontCare e) = C.Dot r . C.Paren r  <$> toConcrete e
+       where r = getRange e
 --    toConcrete (A.DontCare e) = C.DontCare <$> toConcreteCtx TopCtx e
 {-
     -- Andreas, 2010-09-21 abuse C.Underscore to print irrelevant things
     toConcrete (A.DontCare) = return $ C.Underscore noRange Nothing
 -}
+    toConcrete (A.PatternSyn n) = C.Ident <$> toConcrete n
 
 makeDomainFree :: A.LamBinding -> A.LamBinding
 makeDomainFree b@(A.DomainFull (A.TypedBindings r (Arg h rel (A.TBind _ [x] t)))) =
@@ -465,12 +477,17 @@ instance ToConcrete LetBinding [C.Declaration] where
                 , C.FunClause (C.LHS (C.IdentP $ C.QName x) [] [] [])
                               e C.NoWhere
                 ]
+    -- TODO: bind variables
+    bindToConcrete (LetPatBind i p e) ret = do
+        p <- toConcrete p
+        e <- toConcrete e
+        ret [ C.FunClause (C.LHS p [] [] []) (C.RHS e) NoWhere ]
     bindToConcrete (LetApply i x modapp _ _) ret = do
       x' <- unqualify <$> toConcrete x
       modapp <- toConcrete modapp
       let r = getRange modapp
           open = maybe DontOpen id $ minfoOpenShort i
-          dir  = maybe (ImportDirective r (Hiding []) [] False) id $ minfoDirective i
+          dir  = maybe defaultImportDir{ importDirRange = r } id $ minfoDirective i
       -- This is no use since toAbstract LetDefs is in localToAbstract.
       local (openModule' x dir id) $
         ret [ C.ModuleMacro (getRange i) x' modapp open dir ]
@@ -493,8 +510,8 @@ instance ToConcrete AsWhereDecls WhereClause where
     ret . AnyWhere =<< declsToConcrete ds
 
 mergeSigAndDef :: [C.Declaration] -> [C.Declaration]
-mergeSigAndDef (C.RecordSig _ x bs e : C.Record r y c _ Nothing fs : ds)
-  | x == y = C.Record r y c bs (Just e) fs : mergeSigAndDef ds
+mergeSigAndDef (C.RecordSig _ x bs e : C.Record r y ind c _ Nothing fs : ds)
+  | x == y = C.Record r y ind c bs (Just e) fs : mergeSigAndDef ds
 mergeSigAndDef (C.DataSig _ _ x bs e : C.Data r i y _ Nothing cs : ds)
   | x == y = C.Data r i y bs (Just e) cs : mergeSigAndDef ds
 mergeSigAndDef (d : ds) = d : mergeSigAndDef ds
@@ -610,7 +627,7 @@ instance ToConcrete A.Declaration [C.Declaration] where
       return [C.Primitive (getRange i) [C.TypeSig Relevant x' t']]
         -- Primitives are always relevant.
 
-  toConcrete (A.FunDef i _ cs) =
+  toConcrete (A.FunDef i _ _ cs) =
     withAbstractPrivate i $ concat <$> toConcrete cs
 
   toConcrete (A.DataSig i x bs t) =
@@ -633,11 +650,11 @@ instance ToConcrete A.Declaration [C.Declaration] where
       t' <- toConcreteCtx TopCtx t
       return [ C.RecordSig (getRange i) x' (map C.DomainFull tel') t' ]
 
-  toConcrete (A.RecDef  i x c bs t cs) =
+  toConcrete (A.RecDef  i x ind c bs t cs) =
     withAbstractPrivate i $
     bindToConcrete (map makeDomainFree bs) $ \tel' -> do
       (x',cs') <- (unsafeQNameToName -*- id) <$> toConcrete (x, map Constr cs)
-      return [ C.Record (getRange i) x' Nothing tel' Nothing cs' ]
+      return [ C.Record (getRange i) x' ind Nothing tel' Nothing cs' ]
 
   toConcrete (A.Mutual i ds) = declsToConcrete ds
 
@@ -652,7 +669,7 @@ instance ToConcrete A.Declaration [C.Declaration] where
     modapp <- toConcrete modapp
     let r = getRange modapp
         open = maybe DontOpen id $ minfoOpenShort i
-        dir  = maybe (ImportDirective r (Hiding []) [] False) id $ minfoDirective i
+        dir  = maybe defaultImportDir{ importDirRange = r } id $ minfoDirective i
     return [ C.ModuleMacro (getRange i) x modapp open dir ]
 
   toConcrete (A.Import i x) = do
@@ -669,6 +686,7 @@ instance ToConcrete A.Declaration [C.Declaration] where
     x <- toConcrete x
     return [C.Open (getRange i) x defaultImportDir]
 
+
 data RangeAndPragma = RangeAndPragma Range A.Pragma
 
 instance ToConcrete RangeAndPragma C.Pragma where
@@ -699,18 +717,39 @@ instance ToConcrete RangeAndPragma C.Pragma where
 
 -- Left hand sides --------------------------------------------------------
 
-noImplicitArgs = filter (noImplicit . namedThing . unArg)
+noImplicitArgs = filter (noImplicit . namedArg)
 noImplicitPats = filter noImplicit
 
 noImplicit (A.ImplicitP _) = False
 noImplicit _               = True
 
 instance ToConcrete A.LHS C.LHS where
-    bindToConcrete (A.LHS i x args wps) ret = do
+    bindToConcrete (A.LHS i lhscore wps) ret = do
+      bindToConcreteCtx TopCtx lhscore $ \lhs ->
+        bindToConcreteCtx TopCtx (noImplicitPats wps) $ \wps ->
+          ret $ C.LHS lhs wps [] []
+{-
+    bindToConcrete (A.LHS i (A.LHSHead x args) wps) ret = do
       bindToConcreteCtx TopCtx (A.DefP info x args) $ \lhs ->
         bindToConcreteCtx TopCtx (noImplicitPats wps) $ \wps ->
           ret $ C.LHS lhs wps [] []
       where info = PatRange (getRange i)
+-}
+
+instance ToConcrete A.LHSCore C.Pattern where
+    bindToConcrete = bindToConcrete . lhsCoreToPattern
+{-
+    bindToConcrete (A.LHSHead x args) ret = do
+      bindToConcreteCtx TopCtx (A.DefP info x args) $ \ lhs ->
+        ret $ lhs
+      where info = PatRange noRange -- seems to be unused anyway
+    bindToConcrete (A.LHSProj d ps1 lhscore ps2) ret = do
+      d <- toConcrete d
+      bindToConcrete ps1 $ \ ps1 ->
+        bindToConcrete lhscore $ \ p ->
+          bindToConcrete ps2 $ \ ps2 ->
+            ret $ makePattern d ps1 p ps2
+  -}
 
 appBrackets' :: [arg] -> Precedence -> Bool
 appBrackets' []    _   = False
@@ -744,12 +783,14 @@ instance ToConcrete A.Pattern C.Pattern where
         return $ C.DotP (getRange i) e
     -- just for debugging purposes (shouldn't show up in practise)
     toConcrete (A.ImplicitP i) = return $ C.IdentP (C.QName $ C.Name noRange [C.Id "(implicit)"])
+    toConcrete (A.PatternSynP i n _) = IdentP <$> toConcrete n
+
 
 -- Helpers for recovering C.OpApp ------------------------------------------
 
 data Hd = HdVar A.Name | HdCon A.QName | HdDef A.QName
 
-cOpApp :: Range -> C.Name -> [C.Expr] -> C.Expr
+cOpApp :: Range -> C.QName -> [C.Expr] -> C.Expr
 cOpApp r n es = C.OpApp r n (map Ordinary es)
 
 tryToRecoverOpApp :: A.Expr -> AbsToCon C.Expr -> AbsToCon C.Expr
@@ -771,83 +812,96 @@ tryToRecoverOpAppP p def = recoverOpApp bracketP_ C.OpAppP view p def
       DefP _ f            ps -> Just (HdDef f, ps)
       _                      -> Nothing
 
-recoverOpApp :: (ToConcrete a c, HasRange c) =>
-                ((Precedence -> Bool) -> AbsToCon c -> AbsToCon c) ->
-                (Range -> C.Name -> [c] -> c) -> (a -> Maybe (Hd, [NamedArg a])) -> a ->
-                AbsToCon c -> AbsToCon c
-recoverOpApp bracket opApp view e mdefault = case view e of
-  Nothing -> mdefault
+recoverOpApp :: (ToConcrete a c, HasRange c)
+  => ((Precedence -> Bool) -> AbsToCon c -> AbsToCon c)
+  -> (Range -> C.QName -> [c] -> c)
+  -> (a -> Maybe (Hd, [NamedArg a]))
+  ->  a
+  -> AbsToCon c
+  -> AbsToCon c
+recoverOpApp bracket opApp view e mDefault = case view e of
+  Nothing -> mDefault
   Just (hd, args)
     | all notHidden args  -> do
-      let  args' = map (namedThing . unArg) args
+      let  args' = map namedArg args
       case hd of
-        HdVar n | isNoName n -> mdefault
-                | otherwise  -> do
-          x <- toConcrete n
-          doCName (theFixity $ nameFixity n) x args'
-        HdDef qn -> doQName qn args'
-        HdCon qn -> doQName qn args'
-    | otherwise -> mdefault
+        HdVar  n
+          | isNoName n    -> mDefault
+          | otherwise     -> doQNameHelper id        C.QName  n args'
+        HdDef qn          -> doQNameHelper qnameName id      qn args'
+        HdCon qn          -> doQNameHelper qnameName id      qn args'
+    | otherwise           -> mDefault
   where
 
-  isNoName x = C.isNoName $ A.nameConcrete x
+  isNoName :: A.Name -> Bool
+  isNoName  x = C.isNoName $ A.nameConcrete x
 
+  notHidden :: NamedArg a -> Bool
   notHidden a = argHiding a == NotHidden
 
-  -- qualified names can't use mixfix syntax
-  doQName qn as = do
-    x <- toConcrete qn
-    case x of
-      C.QName x -> doCName (theFixity $ nameFixity $ qnameName qn) x as
-      _         -> mdefault
+  doQNameHelper fixityHelper conHelper n as = do
+    x <- toConcrete n
+    doQName (theFixity $ nameFixity $ fixityHelper n) (conHelper x) as
 
   -- fall-back (wrong number of arguments or no holes)
-  doCName _ cn@(C.Name _ xs) es
-    | length xs == 1        = mdefault
-    | length es /= numHoles = mdefault
-    | List.null es          = mdefault
-    where numHoles = length [ () | Hole <- xs ]
-          msg = "doCName " ++ showList xs "" ++ " on " ++ show (length es) ++ " args"
+  doQName _ n es
+    | length xs == 1        = mDefault
+    | length es /= numHoles = mDefault
+    | List.null es          = mDefault
+    where
+      xs       = C.nameParts $ C.unqualify n
+      numHoles = length (filter (== Hole) xs)
+      msg      = concat [ "doQName "
+                        , showList xs ""
+                        , " on "
+                        , show (length es)
+                        , " args" ]
 
   -- binary case
-  doCName fixity cn@(C.Name _ xs) as
+  doQName fixity n as
     | Hole <- head xs
     , Hole <- last xs = do
-        let a1     = head as
-            an     = last as
-            as'    = case as of
-                       as@(_:_:_) -> init $ tail as
-                       _          -> __IMPOSSIBLE__
+        let a1  = head as
+            an  = last as
+            as' = case as of
+                    as@(_ : _ : _) -> init $ tail as
+                    _              -> __IMPOSSIBLE__
         e1 <- toConcreteCtx (LeftOperandCtx fixity) a1
         es <- mapM (toConcreteCtx InsideOperandCtx) as'
         en <- toConcreteCtx (RightOperandCtx fixity) an
         bracket (opBrackets fixity)
-            $ return $ opApp (getRange (e1,en)) cn ([e1] ++ es ++ [en])
+          $ return $ opApp (getRange (e1, en)) n ([e1] ++ es ++ [en])
+    where
+      xs = C.nameParts $ C.unqualify n
 
   -- prefix
-  doCName fixity cn@(C.Name _ xs) as
+  doQName fixity n as
     | Hole <- last xs = do
         let an  = last as
             as' = case as of
-                    as@(_:_) -> init as
-                    _        -> __IMPOSSIBLE__
+                    as@(_ : _) -> init as
+                    _          -> __IMPOSSIBLE__
         es <- mapM (toConcreteCtx InsideOperandCtx) as'
         en <- toConcreteCtx (RightOperandCtx fixity) an
         bracket (opBrackets fixity)
-            $ return $ opApp (getRange (cn,en)) cn (es ++ [en])
+          $ return $ opApp (getRange (n, en)) n (es ++ [en])
+    where
+      xs = C.nameParts $ C.unqualify n
 
   -- postfix
-  doCName fixity cn@(C.Name _ xs) as
+  doQName fixity n as
     | Hole <- head xs = do
-        let a1     = head as
-            as'    = tail as
+        let a1  = head as
+            as' = tail as
         e1 <- toConcreteCtx (LeftOperandCtx fixity) a1
         es <- mapM (toConcreteCtx InsideOperandCtx) as'
         bracket (opBrackets fixity)
-            $ return $ opApp (getRange (e1,cn)) cn ([e1] ++ es)
+          $ return $ opApp (getRange (e1, n)) n ([e1] ++ es)
+    where
+      xs = C.nameParts $ C.unqualify n
 
   -- roundfix
-  doCName _ cn as = do
+  doQName _ n as = do
     es <- mapM (toConcreteCtx InsideOperandCtx) as
     bracket roundFixBrackets
-      $ return $ opApp (getRange cn) cn es
+      $ return $ opApp (getRange n) n es
