@@ -1,5 +1,6 @@
 {-# LANGUAGE CPP, MultiParamTypeClasses, FunctionalDependencies,
-             FlexibleInstances, UndecidableInstances, OverlappingInstances
+             FlexibleInstances, UndecidableInstances, OverlappingInstances,
+             ScopedTypeVariables
   #-}
 
 {-| Translation from "Agda.Syntax.Concrete" to "Agda.Syntax.Abstract". Involves scope analysis,
@@ -26,28 +27,28 @@ import Control.Applicative
 import Control.Monad.Reader hiding (mapM)
 import Control.Monad.Error hiding (mapM)
 import Data.Typeable
-import Data.Traversable (mapM)
-import Data.List ((\\), nub)
+import Data.Traversable (mapM, traverse)
+import Data.List ((\\), nub, foldl')
 import qualified Data.Map as Map
 
 import Agda.Syntax.Concrete as C hiding (topLevelModuleName)
 import Agda.Syntax.Concrete.Operators
+-- import qualified Agda.Syntax.Concrete.Copatterns as Cop -- merged into Operators
 import Agda.Syntax.Abstract as A
+import Agda.Syntax.Abstract.Copatterns
 import Agda.Syntax.Position
 import Agda.Syntax.Common
 import Agda.Syntax.Info
 import Agda.Syntax.Concrete.Definitions as C
-import Agda.Syntax.Concrete.Operators
 import Agda.Syntax.Concrete.Pretty
 import Agda.Syntax.Abstract.Pretty
 import Agda.Syntax.Fixity
 import Agda.Syntax.Notation
 import Agda.Syntax.Scope.Base
 import Agda.Syntax.Scope.Monad
-import Agda.Syntax.Strict
 
 import Agda.TypeChecking.Monad.Base (TypeError(..), Call(..), typeError,
-                                     TCErr(..), TCErr'(..), extendlambdaname)
+                                     TCErr(..), extendlambdaname)
 import Agda.TypeChecking.Monad.Trace (traceCall, traceCallCPS, setCurrentRange)
 import Agda.TypeChecking.Monad.State
 import Agda.TypeChecking.Monad.Options
@@ -85,26 +86,12 @@ printLocals v s = verboseS "scope.top" v $ do
 printScope :: String -> Int -> String -> ScopeM ()
 printScope tag v s = verboseS ("scope." ++ tag) v $ do
   scope <- getScope
-  reportSLn "" 0 $ s ++ " " ++ show scope
+  reportSDoc "" 0 $ return $ vcat [ text s, text $ show scope ]
 
 {--------------------------------------------------------------------------
     Helpers
  --------------------------------------------------------------------------}
 
-lhsArgs :: C.Pattern -> (C.Name, [NamedArg C.Pattern])
-lhsArgs p = case appView p of
-    Arg _ _ (Named _ (IdentP (C.QName x))) : ps -> (x, ps)
-    _                                         -> __IMPOSSIBLE__
-    where
-        mkHead    = Arg NotHidden Relevant . unnamed
-        notHidden = Arg NotHidden Relevant . unnamed
-        appView p = case p of
-            AppP p arg    -> appView p ++ [arg]
-            OpAppP _ x ps -> mkHead (IdentP $ C.QName x) : map notHidden ps
-            ParenP _ p    -> appView p
-            RawAppP _ _   -> __IMPOSSIBLE__
-            _             -> [ mkHead p ]
-
 annotateDecl :: ScopeM A.Declaration -> ScopeM A.Declaration
 annotateDecl m = annotateDecls $ (:[]) <$> m
 
@@ -134,15 +121,18 @@ checkPatternLinearity ps = case xs \\ nub xs of
     xs = concatMap vars ps
     vars :: A.Pattern' e -> [C.Name]
     vars p = case p of
-      A.VarP x        -> [nameConcrete x]
-      A.ConP _ _ args -> concatMap (vars . namedThing . unArg) args
-      A.WildP _       -> []
-      A.AsP _ x p     -> nameConcrete x : vars p
-      A.DotP _ _      -> []
-      A.AbsurdP _     -> []
-      A.LitP _        -> []
-      A.DefP _ _ args -> __IMPOSSIBLE__
-      A.ImplicitP _   -> __IMPOSSIBLE__
+      A.VarP x               -> [nameConcrete x]
+      A.ConP _ _ args        -> concatMap (vars . namedArg) args
+      A.WildP _              -> []
+      A.AsP _ x p            -> nameConcrete x : vars p
+      A.DotP _ _             -> []
+      A.AbsurdP _            -> []
+      A.LitP _               -> []
+      A.DefP _ _ args        -> concatMap (vars . namedArg) args
+        -- Projection pattern, @args@ should be empty unless we have
+        -- indexed records.
+      A.ImplicitP _          -> __IMPOSSIBLE__
+      A.PatternSynP _ _ args -> concatMap (vars . namedArg) args
 
 -- | Compute the type of the record constructor (with bogus target type)
 recordConstructorType :: [NiceDeclaration] -> C.Expr
@@ -156,7 +146,8 @@ recordConstructorType fields = build fs
     build (NiceField r f _ _ x (Arg h rel e) : fs) =
         C.Pi [C.TypedBindings r $ Arg h rel (C.TBind r [BName x f] e)] $ build fs
       where r = getRange x
-    build (d : fs)                     = C.Let noRange (notSoNiceDeclarations [d]) $ build fs
+    build (d : fs)                     = C.Let noRange [killRange $ notSoNiceDeclaration d] $
+                                           build fs
     build []                           = C.SetN noRange 0 -- todo: nicer
 
 checkModuleApplication (C.SectionApp _ tel e) m0 x dir' =
@@ -180,12 +171,12 @@ checkModuleApplication (C.SectionApp _ tel e) m0 x dir' =
     printScope "mod.inst" 20 "copied source module"
     reportSLn "scope.mod.inst" 30 $ "renamings:\n  " ++ show renD ++ "\n  " ++ show renM
     return ((A.SectionApp tel' m1 args'), renD, renM)
-checkModuleApplication (C.RecordModuleIFS _ rec) m0 x dir' =
+checkModuleApplication (C.RecordModuleIFS _ recN) m0 x dir' =
   withCurrentModule m0 $ do
-    m1 <- toAbstract $ OldModuleName rec
+    m1 <- toAbstract $ OldModuleName recN
     s <- getNamedScope m1
     (s', (renM, renD)) <- copyScope m0 s
-    s' <- applyImportDirectiveM rec dir' s'
+    s' <- applyImportDirectiveM recN dir' s'
     modifyCurrentScope $ const s'
 
     printScope "mod.inst" 20 "copied record module"
@@ -318,21 +309,26 @@ instance ToAbstract (NewName C.BoundName) A.Name where
 nameExpr :: AbstractName -> A.Expr
 nameExpr d = mk (anameKind d) $ anameName d
   where
-    mk DefName = Def
-    mk ConName = Con . AmbQ . (:[])
+    mk DefName        = Def
+    mk FldName        = Def
+    mk ConName        = Con . AmbQ . (:[])
+    mk PatternSynName = A.PatternSyn
 
 instance ToAbstract OldQName A.Expr where
   toAbstract (OldQName x) = do
     qx <- resolveName x
     reportSLn "scope.name" 10 $ "resolved " ++ show x ++ ": " ++ show qx
     case qx of
-      VarName x'         -> return $ A.Var x'
-      DefinedName _ d    -> return $ nameExpr d
-      ConstructorName ds -> return $ A.Con $ AmbQ (map anameName ds)
-      UnknownName        -> notInScope x
+      VarName x'          -> return $ A.Var x'
+      DefinedName _ d     -> return $ nameExpr d
+      FieldName     d     -> return $ nameExpr d
+      ConstructorName ds  -> return $ A.Con $ AmbQ (map anameName ds)
+      UnknownName         -> notInScope x
+      PatternSynResName d -> return $ nameExpr d
 
 data APatName = VarPatName A.Name
               | ConPatName [AbstractName]
+              | PatternSynPatName AbstractName
 
 instance ToAbstract PatName APatName where
   toAbstract (PatName x) = do
@@ -341,9 +337,11 @@ instance ToAbstract PatName APatName where
     z  <- case (rx, x) of
       -- TODO: warn about shadowing
       (VarName y,       C.QName x)                          -> return $ Left x -- typeError $ RepeatedVariableInPattern y x
+      (FieldName d,     C.QName x)                          -> return $ Left x
       (DefinedName _ d, C.QName x) | DefName == anameKind d -> return $ Left x
       (UnknownName,     C.QName x)                          -> return $ Left x
-      (ConstructorName ds, _)                               -> return $ Right ds
+      (ConstructorName ds, _)                               -> return $ Right (Left ds)
+      (PatternSynResName d, _)                              -> return $ Right (Right d)
       _                                                     ->
         typeError $ GenericError $
           "Cannot pattern match on " ++ show x ++ ", because it is not a constructor"
@@ -353,9 +351,13 @@ instance ToAbstract PatName APatName where
         p <- VarPatName <$> toAbstract (NewName x)
         printLocals 10 "bound it:"
         return p
-      Right cs -> do
-        reportSLn "scope.pat" 10 $ "it was a con: " ++ show (map anameName cs)
-        return $ ConPatName cs
+      Right (Left ds) -> do
+        reportSLn "scope.pat" 10 $ "it was a con: " ++ show (map anameName ds)
+        return $ ConPatName ds
+      Right (Right d) -> do
+        reportSLn "scope.pat" 10 $ "it was a pat syn: " ++ show (anameName d)
+        return $ PatternSynPatName d
+
 
 -- Should be a defined name.
 instance ToAbstract OldName A.QName where
@@ -375,9 +377,10 @@ freshQModule m x = A.qualifyM m . mnameFromList . (:[]) <$> freshAbstractName_ x
 checkForModuleClash :: C.Name -> ScopeM ()
 checkForModuleClash x = do
   ms <- scopeLookup (C.QName x) <$> getScope
-  unless (null ms) $
+  unless (null ms) $ do
+    reportSLn "scope.clash" 20 $ "clashing modules ms = " ++ show ms
     setCurrentRange (getRange x) $
-    typeError $ ShadowedModule $
+      typeError $ ShadowedModule x $
                 map ((`withRangeOf` x) . amodName) ms
 
 instance ToAbstract NewModuleName A.ModuleName where
@@ -385,7 +388,7 @@ instance ToAbstract NewModuleName A.ModuleName where
     checkForModuleClash x
     m <- getCurrentModule
     y <- freshQModule m x
-    createModule y
+    createModule False y
     return y
 
 instance ToAbstract NewModuleQName A.ModuleName where
@@ -393,7 +396,7 @@ instance ToAbstract NewModuleQName A.ModuleName where
     where
       toAbs m (C.QName x)  = do
         y <- freshQModule m x
-        createModule y
+        createModule False y
         return y
       toAbs m (C.Qual x q) = do
         m' <- freshQModule m x
@@ -477,13 +480,15 @@ instance ToAbstract C.Expr A.Expr where
                     { metaRange  = r
                     , metaScope  = scope
                     , metaNumber = n
+                    , metaNameSuggestion = ""
                     }
       C.Underscore r n -> do
         scope <- getScope
         return $ A.Underscore $ MetaInfo
                     { metaRange  = r
                     , metaScope  = scope
-                    , metaNumber = n
+                    , metaNumber = maybe Nothing __IMPOSSIBLE__ n
+                    , metaNameSuggestion = maybe "" id n
                     }
 
   -- Raw application
@@ -537,11 +542,11 @@ instance ToAbstract C.Expr A.Expr where
             insertApp _ = __IMPOSSIBLE__
             insertHead (C.LHS p wps eqs with) = C.LHS (insertApp p) wps eqs with
             insertHead (C.Ellipsis r wps eqs with) = C.Ellipsis r wps eqs with
-        scdef <- toAbstract (C.FunDef r [] defaultFixity' ConcreteDef cname
+        scdef <- toAbstract (C.FunDef r [] defaultFixity' ConcreteDef True cname
                                (map (\(lhs,rhs,wh) -> -- wh = NoWhere, see parser for more info
                                       C.Clause cname (insertHead lhs) rhs wh []) cs))
         case scdef of
-          (A.ScopedDecl si [A.FunDef di qname' cs]) -> do
+          (A.ScopedDecl si [A.FunDef di qname' NotDelayed cs]) -> do
             setScope si
             return $ A.ExtendedLam (ExprRange r) di qname' cs
           _ -> __IMPOSSIBLE__
@@ -557,7 +562,7 @@ instance ToAbstract C.Expr A.Expr where
 -- Irrelevant non-dependent function type
 
       C.Fun r e1 e2 -> do
-        Arg h rel (e0, dotted) <- fmapM (toAbstractDot FunctionSpaceDomainCtx) $ mkArg e1
+        Arg h rel (e0, dotted) <- traverse (toAbstractDot FunctionSpaceDomainCtx) $ mkArg e1
         let e1 = Arg h (if dotted then Irrelevant else rel) e0
         e2 <- toAbstractCtx TopCtx e2
         let info = ExprRange r
@@ -698,11 +703,18 @@ instance ToAbstract (TopLevel [C.Declaration]) TopLevelInfo where
 -- | runs Syntax.Concrete.Definitions.niceDeclarations on main module
 niceDecls :: [C.Declaration] -> ScopeM [NiceDeclaration]
 niceDecls ds = case runNice $ niceDeclarations ds of
-  Left e   -> throwError $ TCErr Nothing $ Exception (getRange e) (show e)
+  Left e   -> throwError $ Exception (getRange e) (show e)
   Right ds -> return ds
 
 instance ToAbstract [C.Declaration] [A.Declaration] where
-  toAbstract ds = toAbstract =<< niceDecls ds
+  toAbstract ds = do
+    -- don't allow to switch off termination checker in --safe mode
+    ds <- ifM (optSafe <$> commandLineOptions) (mapM noNoTermCheck ds) (return ds)
+    toAbstract =<< niceDecls ds
+   where
+    noNoTermCheck (C.Pragma (NoTerminationCheckPragma r)) =
+      typeError $ SafeFlagNoTerminationCheck
+    noNoTermCheck d = return d
 
 newtype LetDefs = LetDefs [C.Declaration]
 newtype LetDef = LetDef NiceDeclaration
@@ -714,7 +726,7 @@ instance ToAbstract LetDefs [A.LetBinding] where
 instance ToAbstract LetDef [A.LetBinding] where
     toAbstract (LetDef d) =
         case d of
-            NiceMutual _ d@[C.FunSig _ fx _ rel x t, C.FunDef _ _ _ abstract _ [cl]] ->
+            NiceMutual _ _ d@[C.FunSig _ fx _ rel _ x t, C.FunDef _ _ _ abstract _ _ [cl]] ->
                 do  when (abstract == AbstractDef) $ do
                       typeError $ GenericError $ "abstract not allowed in let expressions"
                     e <- letToAbstract cl
@@ -722,6 +734,15 @@ instance ToAbstract LetDef [A.LetBinding] where
                     x <- toAbstract (NewName $ C.BName x fx)
                     return [ A.LetBind (LetRange $ getRange d) rel x t e ]
 
+            -- irrefutable let binding, like  (x , y) = rhs
+            NiceFunClause r PublicAccess ConcreteDef termCheck d@(C.FunClause (C.LHS p [] [] []) (C.RHS rhs) NoWhere) -> do
+              rhs <- toAbstract rhs
+              p   <- parsePattern p
+              p   <- toAbstract p
+              checkPatternLinearity [p]
+              p   <- toAbstract p
+              return [ A.LetPatBind (LetRange r) p rhs ]
+
             -- You can't open public in a let
             NiceOpen r x dirs | not (C.publicOpen dirs) -> do
               m       <- toAbstract (OldModuleName x)
@@ -743,8 +764,17 @@ instance ToAbstract LetDef [A.LetBinding] where
             _   -> notAValidLetBinding d
         where
             letToAbstract (C.Clause top clhs@(C.LHS p [] [] []) (C.RHS rhs) NoWhere []) = do
-                p    <- parseLHS (Just top) p
+{-
+                p    <- parseLHS top p
                 localToAbstract (snd $ lhsArgs p) $ \args ->
+-}
+                (x, args) <- do
+                  res <- parseLHS top p
+                  case res of
+                    C.LHSHead x args -> return (x, args)
+                    C.LHSProj{} -> typeError $ GenericError $ "copatterns not allowed in let bindings"
+
+                localToAbstract args $ \args ->
                     do  rhs <- toAbstract rhs
                         foldM lambda rhs (reverse args)  -- just reverse because these DomainFree
             letToAbstract _ = notAValidLetBinding d
@@ -770,15 +800,15 @@ instance ToAbstract NiceDeclaration A.Declaration where
 
   -- Axiom
     C.Axiom r f p rel x t -> do
+      -- check that we do not postulate in --safe mode
       clo <- commandLineOptions
       when (optSafe clo) (typeError (SafeFlagPostulate x))
-      t' <- toAbstractCtx TopCtx t
-      y  <- freshAbstractQName f x
-      bindName p DefName x y
-      return [ A.Axiom (mkDefInfo x f p ConcreteDef r) rel y t' ]
+      -- check the postulate
+      toAbstractNiceAxiom d
 
   -- Fields
     C.NiceField r f p a x t -> do
+      unless (p == PublicAccess) $ typeError $ GenericError "Record fields can not be private"
       t' <- toAbstractCtx TopCtx t
       y  <- freshAbstractQName f x
       irrProj <- optIrrelevantProjections <$> pragmaOptions
@@ -786,7 +816,7 @@ instance ToAbstract NiceDeclaration A.Declaration where
         -- Andreas, 2010-09-24: irrelevant fields are not in scope
         -- this ensures that projections out of irrelevant fields cannot occur
         -- Ulf: unless you turn on --irrelevant-projections
-        bindName p DefName x y
+        bindName p FldName x y
       return [ A.Field (mkDefInfo x f p a r) y t' ]
 
   -- Primitive function
@@ -797,10 +827,9 @@ instance ToAbstract NiceDeclaration A.Declaration where
       return [ A.Primitive (mkDefInfo x f p a r) y t' ]
 
   -- Definitions (possibly mutual)
-    NiceMutual r ds -> do
+    NiceMutual r termCheck ds -> do
       ds' <- toAbstract ds
-      return [ A.Mutual (DeclInfo C.noName_ r) ds' ]
-                          -- TODO: what does the info mean here?
+      return [ A.Mutual (MutualInfo termCheck r) ds' ]
 
     C.NiceRecSig r f a x ls t -> withLocalVars $ do
         let toTypeBinding :: C.LamBinding -> C.TypedBindings
@@ -820,16 +849,25 @@ instance ToAbstract NiceDeclaration A.Declaration where
                _            -> __IMPOSSIBLE__
         ls' <- toAbstract (map toTypeBinding ls)
         x'  <- freshAbstractQName f x
+        {- -- Andreas, 2012-01-16: remember number of parameters
+        bindName a (DataName (length ls)) x x' -}
         bindName a DefName x x'
         t' <- toAbstract t
         return [ A.DataSig (mkDefInfo x f a ConcreteDef r) x' ls' t' ]
   -- Type signatures
-    C.FunSig r f p rel x t -> (:[]) <$> toAbstract (C.Axiom r f p rel x t)
+    C.FunSig r f p rel tc x t -> toAbstractNiceAxiom (C.Axiom r f p rel x t)
   -- Function definitions
-    C.FunDef r ds f a x cs -> do
+    C.FunDef r ds f a tc x cs -> do
         printLocals 10 $ "checking def " ++ show x
-        (x',cs') <- toAbstract (OldName x,cs)
-        return [ A.FunDef (mkDefInfo x f PublicAccess a r) x' cs' ]
+        (x',cs) <- toAbstract (OldName x,cs)
+        (delayed, cs) <- translateCopatternClauses cs
+        return [ A.FunDef (mkDefInfo x f PublicAccess a r) x' delayed cs ]
+
+  -- Uncategorized function clauses
+    C.NiceFunClause r acc abs termCheck (C.FunClause lhs rhs wcls) ->
+      typeError $ GenericError $
+        "Missing type signature for left hand side " ++ show lhs
+    C.NiceFunClause{} -> __IMPOSSIBLE__
 
   -- Data definitions
     C.DataDef r f a x pars cons -> withLocalVars $ do
@@ -848,7 +886,7 @@ instance ToAbstract NiceDeclaration A.Declaration where
         -- Create the module for the qualified constructors
         checkForModuleClash x -- disallow shadowing previously defined modules
         let m = mnameFromList $ qnameToList x'
-        createModule m
+        createModule True m
         bindModule p x m  -- make it a proper module
         cons <- toAbstract (map (ConstrDecl NoRec m a p) cons)
         -- Open the module
@@ -860,7 +898,7 @@ instance ToAbstract NiceDeclaration A.Declaration where
         conName _ = __IMPOSSIBLE__
 
   -- Record definitions (mucho interesting)
-    C.RecDef r f a x cm pars fields ->
+    C.RecDef r f a x ind cm pars fields ->
       withLocalVars $ do
         -- Check that the generated module doesn't clash with a previously
         -- defined module
@@ -872,7 +910,7 @@ instance ToAbstract NiceDeclaration A.Declaration where
         m0     <- getCurrentModule
         let m = A.qualifyM m0 $ mnameFromList $ (:[]) $ last $ qnameToList x'
         printScope "rec" 15 "before record"
-        createModule m
+        createModule False m
         afields <- withCurrentModule m $ do
           afields <- toAbstract fields
           printScope "rec" 15 "checked fields"
@@ -880,14 +918,24 @@ instance ToAbstract NiceDeclaration A.Declaration where
         bindModule p x m
         cm' <- mapM (\(ThingWithFixity c f) -> bindConstructorName m c f a p YesRec) cm
         printScope "rec" 15 "record complete"
-        return [ A.RecDef (mkDefInfo x f PublicAccess a r) x' cm' pars contel afields ]
+        return [ A.RecDef (mkDefInfo x f PublicAccess a r) x' ind cm' pars contel afields ]
 
+    -- Andreas, 2012-10-30 anonymous modules are like Coq sections
     NiceModule r p a (C.QName name) tel ds ->
       traceCall (ScopeCheckDeclaration $ NiceModule r p a (C.QName name) tel []) $ do
+      (name, p, isSection) <- if not (C.isNoName name)
+        then return (name, p, False)
+        else do
+          (i :: NameId) <- fresh
+          return (C.NoName (getRange name) i, PrivateAccess, True)
       aname <- toAbstract (NewModuleName name)
-      x <- snd <$> scopeCheckModule r (C.QName name) aname tel ds
+      ds <- snd <$> scopeCheckModule r (C.QName name) aname tel ds
       bindModule p name aname
-      return x
+      -- if the module was anonymous open it public
+      when isSection $
+        openModule_ (C.QName name) $
+          defaultImportDir { publicOpen = True }
+      return ds
 
     NiceModule _ _ _ C.Qual{} _ _ -> __IMPOSSIBLE__
 
@@ -962,11 +1010,38 @@ instance ToAbstract NiceDeclaration A.Declaration where
                            })
                         m ]
 
+    NicePatternSyn r fx n as p -> do
+      reportSLn "scope.pat" 10 $ "found nice pattern syn: " ++ show r
+
+      isparameterised <- not . null <$> getLocalVars
+      when isparameterised $ typeError $ NotSupported
+          "pattern synonym in parameterised module"
+
+      y <- freshAbstractQName fx n
+      bindName PublicAccess PatternSynName n y
+      defn <- withLocalVars $ do
+               p'   <- killRange <$> (toAbstract =<< toAbstract =<< parsePatternSyn p)
+               as'  <- mapM (\a -> unVarName =<< resolveName (C.QName a)) as
+               return (as', p')
+      modifyPatternSyns (Map.insert y defn)
+      return []
+      where unVarName (VarName a) = return a
+            unVarName _           = typeError $ UnusedVariableInPatternSynonym
+
+    where
+      -- checking postulate or type sig. without checking safe flag
+      toAbstractNiceAxiom (C.Axiom r f p rel x t) = do
+        t' <- toAbstractCtx TopCtx t
+        y  <- freshAbstractQName f x
+        bindName p DefName x y
+        return [ A.Axiom (mkDefInfo x f p ConcreteDef r) rel y t' ]
+      toAbstractNiceAxiom _ = __IMPOSSIBLE__
+
 
 data IsRecordCon = YesRec | NoRec
 data ConstrDecl = ConstrDecl IsRecordCon A.ModuleName IsAbstract Access C.NiceDeclaration
 
-bindConstructorName m x f a p rec = do
+bindConstructorName m x f a p record = do
   -- The abstract name is the qualified one
   y <- withCurrentModule m $ freshAbstractQName f x
   -- Bind it twice, once unqualified and once qualified
@@ -979,17 +1054,17 @@ bindConstructorName m x f a p rec = do
     p' = case a of
            AbstractDef -> PrivateAccess
            _           -> p
-    p'' = case (a, rec) of
+    p'' = case (a, record) of
             (AbstractDef, _) -> PrivateAccess
             (_, YesRec)      -> OnlyQualified   -- record constructors aren't really in the record module
             _                -> PublicAccess
 
 instance ToAbstract ConstrDecl A.Declaration where
-  toAbstract (ConstrDecl rec m a p (C.Axiom r f _ rel x t)) = do -- rel==Relevant
+  toAbstract (ConstrDecl record m a p (C.Axiom r f _ rel x t)) = do -- rel==Relevant
     t' <- toAbstractCtx TopCtx t
     -- The abstract name is the qualified one
     -- Bind it twice, once unqualified and once qualified
-    y <- bindConstructorName m x f a p rec
+    y <- bindConstructorName m x f a p record
     printScope "con" 15 "bound constructor"
     return $ A.Axiom (mkDefInfo x f p ConcreteDef r) rel y t'
 
@@ -1046,11 +1121,16 @@ instance ToAbstract C.Pragma [A.Pragma] where
       case e of
         A.Def x -> return [ A.EtaPragma x ]
         _       -> fail "Bad ETA pragma"
+    -- NO_TERMINATION_CHECK is handled by the nicifier
+    toAbstract (C.NoTerminationCheckPragma _) = __IMPOSSIBLE__
 
 instance ToAbstract C.Clause A.Clause where
     toAbstract (C.Clause top C.Ellipsis{} _ _ _) = fail "bad '...'" -- TODO: errors message
     toAbstract (C.Clause top lhs@(C.LHS p wps eqs with) rhs wh wcs) = withLocalVars $ do
-      let wcs' = map (expandEllipsis p wps) wcs
+-- WAS:     let wcs' = map (expandEllipsis p wps) wcs
+      -- Andreas, 2012-02-14: need to reset local vars before checking subclauses
+      vars <- getLocalVars
+      let wcs' = map (\ c -> setLocalVars vars >> do return $ expandEllipsis p wps c) wcs
       lhs' <- toAbstract (LeftHandSide top p wps)
       printLocals 10 "after lhs:"
       let (whname, whds) = case wh of
@@ -1085,9 +1165,16 @@ whereToAbstract r whname whds inner = do
   bindModule acc m am
   return (x, ds)
 
-data RightHandSide = RightHandSide [C.Expr] [C.Expr] [C.Clause] C.RHS [C.Declaration]
+data RightHandSide = RightHandSide
+  { rhsRewriteEqn :: [C.RewriteEqn]  -- ^ @rewrite e@ (many)
+  , rhsWithExpr   :: [C.WithExpr]    -- ^ @with e@ (many)
+  , rhsSubclauses :: [ScopeM C.Clause] -- ^ the subclauses spawned by a with (monadic because we need to reset the local vars before checking these clauses)
+  , rhs           :: C.RHS
+  , rhsWhereDecls :: [C.Declaration]
+  }
+
 data AbstractRHS = AbsurdRHS'
-                 | WithRHS' [A.Expr] [C.Clause]  -- ^ The with clauses haven't been translated yet
+                 | WithRHS' [A.Expr] [ScopeM C.Clause]  -- ^ The with clauses haven't been translated yet
                  | RHS' A.Expr
                  | RewriteRHS' [A.Expr] AbstractRHS [A.Declaration]
 
@@ -1110,7 +1197,7 @@ instance ToAbstract AbstractRHS A.RHS where
     return $ RewriteRHS auxs eqs rhs wh
   toAbstract (WithRHS' es cs) = do
     aux <- withFunctionName "with-"
-    A.WithRHS aux es <$> toAbstract cs
+    A.WithRHS aux es <$> do toAbstract =<< sequence cs
 
 instance ToAbstract RightHandSide AbstractRHS where
   toAbstract (RightHandSide eqs@(_:_) es cs rhs wh) = do
@@ -1139,18 +1226,46 @@ data LeftHandSide = LeftHandSide C.Name C.Pattern [C.Pattern]
 instance ToAbstract LeftHandSide A.LHS where
     toAbstract (LeftHandSide top lhs wps) =
       traceCall (ScopeCheckLHS top lhs) $ do
-        p <- parseLHS (Just top) lhs
+        lhscore <- parseLHS top lhs
+        reportSLn "scope.lhs" 5 $ "parsed lhs: " ++ show lhscore
         printLocals 10 "before lhs:"
-        let (x, ps) = lhsArgs p
+        -- error if copattern parsed but no --copatterns option
+        haveCoPats <- optCopatterns <$> pragmaOptions
+        unless haveCoPats $
+          case lhscore of
+            C.LHSHead x ps -> return ()
+            C.LHSProj{} -> typeError $ NeedOptionCopatterns
+        -- scope check patterns except for dot patterns
+        lhscore <- toAbstract lhscore
+        reportSLn "scope.lhs" 5 $ "parsed lhs patterns: " ++ show lhscore
+        wps  <- toAbstract =<< mapM parsePattern wps
+        checkPatternLinearity $ lhsCoreAllPatterns lhscore ++ wps
+        printLocals 10 "checked pattern:"
+        -- scope check dot patterns
+        lhscore <- toAbstract lhscore
+        reportSLn "scope.lhs" 5 $ "parsed lhs dot patterns: " ++ show lhscore
+        wps     <- toAbstract wps
+        printLocals 10 "checked dots:"
+        return $ A.LHS (LHSRange $ getRange (lhs, wps)) lhscore wps
+
+-- does not check pattern linearity
+instance ToAbstract C.LHSCore (A.LHSCore' C.Expr) where
+    toAbstract (C.LHSHead x ps) = do
         x    <- withLocalVars $ setLocalVars [] >> toAbstract (OldName x)
         args <- toAbstract ps
-        wps  <- toAbstract =<< mapM (parseLHS Nothing) wps
-        checkPatternLinearity (map (namedThing . unArg) args ++ wps)
-        printLocals 10 "checked pattern:"
-        args <- toAbstract args -- take care of dot patterns
-        wps  <- toAbstract wps
-        printLocals 10 "checked dots:"
-        return $ A.LHS (LHSRange $ getRange (lhs, wps)) x args wps
+        return $ A.LHSHead x args
+    toAbstract (C.LHSProj d ps1 l ps2) = do
+        qx <- resolveName d
+        d  <- case qx of
+                FieldName d -> return $ anameName d
+                UnknownName -> notInScope d
+                _           -> typeError $ GenericError $
+                  "head of copattern needs to be a field identifier, but "
+                  ++ show d ++ " isn't one"
+        args1 <- toAbstract ps1
+        l     <- toAbstract l
+        args2 <- toAbstract ps2
+        return $ A.LHSProj d args1 l args2
 
 instance ToAbstract c a => ToAbstract (Arg c) (Arg a) where
     toAbstract (Arg h r e) = Arg h r <$> toAbstractCtx (hiddenArgumentCtx h) e
@@ -1158,48 +1273,85 @@ instance ToAbstract c a => ToAbstract (Arg c) (Arg a) where
 instance ToAbstract c a => ToAbstract (Named name c) (Named name a) where
     toAbstract (Named n e) = Named n <$> toAbstract e
 
+{- DOES NOT WORK ANYMORE with pattern synonyms
+instance ToAbstract c a => ToAbstract (A.LHSCore' c) (A.LHSCore' a) where
+    toAbstract = mapM toAbstract
+-}
+
+instance ToAbstract (A.LHSCore' C.Expr) (A.LHSCore' A.Expr) where
+    toAbstract (A.LHSHead f ps)             = A.LHSHead f <$> mapM toAbstract ps
+    toAbstract (A.LHSProj d ps lhscore ps') = A.LHSProj d <$> mapM toAbstract ps
+      <*> mapM toAbstract lhscore <*> mapM toAbstract ps'
+
 -- Patterns are done in two phases. First everything but the dot patterns, and
 -- then the dot patterns. This is because dot patterns can refer to variables
 -- bound anywhere in the pattern.
 
-instance ToAbstract c a => ToAbstract (A.Pattern' c) (A.Pattern' a) where
-    toAbstract = mapM toAbstract
+instance ToAbstract (A.Pattern' C.Expr) (A.Pattern' A.Expr) where
+    toAbstract (A.VarP x)             = return $ A.VarP x
+    toAbstract (A.ConP i ds as)       = A.ConP i ds <$> mapM toAbstract as
+    toAbstract (A.DefP i x as)        = A.DefP i x <$> mapM toAbstract as
+    toAbstract (A.WildP i)            = return $ A.WildP i
+    toAbstract (A.AsP i x p)          = A.AsP i x <$> toAbstract p
+    toAbstract (A.DotP i e)           = A.DotP i <$> toAbstract e
+    toAbstract (A.AbsurdP i)          = return $ A.AbsurdP i
+    toAbstract (A.LitP l)             = return $ A.LitP l
+    toAbstract (A.ImplicitP i)        = return $ A.ImplicitP i
+    toAbstract (A.PatternSynP i x as) = do
+        p   <- lookupPatternSyn x
+        as' <- mapM toAbstract as
+        instPatternSyn p as'
+      where
+        instPatternSyn :: A.PatternSynDefn -> [NamedArg A.Pattern] -> ScopeM A.Pattern
+        instPatternSyn (ns, p) as
+            | length ns == length as = return $ substPattern s $ setRange (getRange i) p
+            | otherwise              = typeError $ PatternSynonymArityMismatch x
+          where
+          s = zipWith' (\n a -> (n, namedThing (unArg a))) ns as
+
 
 instance ToAbstract C.Pattern (A.Pattern' C.Expr) where
 
     toAbstract p@(C.IdentP x) = do
         px <- toAbstract (PatName x)
         case px of
-            VarPatName y  -> return $ VarP y
-            ConPatName ds -> return $ ConP (PatRange (getRange p))
-                                           (AmbQ $ map anameName ds)
-                                           []
+            VarPatName y        -> return $ VarP y
+            ConPatName ds       -> return $ ConP (PatRange (getRange p))
+                                                 (AmbQ $ map anameName ds)
+                                                 []
+            PatternSynPatName d -> return $ PatternSynP (PatRange (getRange p))
+                                                        (anameName d) []
 
     toAbstract p0@(AppP p q) = do
         (p', q') <- toAbstract (p,q)
         case p' of
-            ConP _ x as -> return $ ConP info x (as ++ [q'])
-            DefP _ x as -> return $ DefP info x (as ++ [q'])
-            _           -> typeError $ InvalidPattern p0
+            ConP _ x as        -> return $ ConP info x (as ++ [q'])
+            DefP _ x as        -> return $ DefP info x (as ++ [q'])
+            PatternSynP _ x as -> return $ PatternSynP info x (as ++ [q'])
+            _                  -> typeError $ InvalidPattern p0
         where
             r = getRange p0
             info = PatSource r $ \pr -> if appBrackets pr then ParenP r p0 else p0
 
     toAbstract p0@(OpAppP r op ps) = do
-        p <- toAbstract (IdentP $ C.QName op)
+        p <- toAbstract (IdentP op)
         ps <- toAbstract ps
         case p of
-          ConP _ x as -> return $ ConP info x (as ++ map (Arg NotHidden Relevant . unnamed) ps)
-          DefP _ x as -> return $ DefP info x (as ++ map (Arg NotHidden Relevant. unnamed) ps)
-          _           -> __IMPOSSIBLE__
+          ConP        _ x as -> return $ ConP info x
+                                    (as ++ map (Arg NotHidden Relevant . unnamed) ps)
+          DefP        _ x as -> return $ DefP info x
+                                    (as ++ map (Arg NotHidden Relevant . unnamed) ps)
+          PatternSynP _ x as -> return $ PatternSynP info x
+                                    (as ++ map (Arg NotHidden Relevant . unnamed) ps)
+          _                  -> __IMPOSSIBLE__
         where
-            r = getRange p0
+            r    = getRange p0
             info = PatSource r $ \pr -> if appBrackets pr then ParenP r p0 else p0
 
     -- Removed when parsing
-    toAbstract (HiddenP _ _) = __IMPOSSIBLE__
+    toAbstract (HiddenP _ _)   = __IMPOSSIBLE__
     toAbstract (InstanceP _ _) = __IMPOSSIBLE__
-    toAbstract (RawAppP _ _) = __IMPOSSIBLE__
+    toAbstract (RawAppP _ _)   = __IMPOSSIBLE__
 
     toAbstract p@(C.WildP r)    = return $ A.WildP (PatSource r $ const p)
     toAbstract (C.ParenP _ p)   = toAbstract p
@@ -1216,18 +1368,16 @@ instance ToAbstract C.Pattern (A.Pattern' C.Expr) where
     toAbstract p0@(C.DotP r e) = return $ A.DotP info e
         where info = PatSource r $ \_ -> p0
     toAbstract p0@(C.AbsurdP r) = return $ A.AbsurdP info
-        where
-            info = PatSource r $ \_ -> p0
+        where info = PatSource r $ \_ -> p0
 
 -- | Turn an operator application into abstract syntax. Make sure to record the
 -- right precedences for the various arguments.
-toAbstractOpApp :: C.Name -> [OpApp C.Expr] -> ScopeM A.Expr
-toAbstractOpApp op@(C.NoName _ _) es = __IMPOSSIBLE__
-toAbstractOpApp op@(C.Name _ _) es = do
-    f  <- getFixity (C.QName op)
+toAbstractOpApp :: C.QName -> [OpApp C.Expr] -> ScopeM A.Expr
+toAbstractOpApp op es = do
+    f  <- getFixity op
     let (_,_,parts) = oldToNewNotation $ (op, f)
-    op <- toAbstract (OldQName $ C.QName op)
-    foldl app op <$> left (theFixity f) [p | p <- parts, not (isBindingHole p)] es
+    op <- toAbstract (OldQName op)
+    foldl' app op <$> left (theFixity f) [p | p <- parts, not (isBindingHole p)] es
     where
         app e arg = A.App (ExprRange (fuseRange e arg)) e
                   $ Arg NotHidden Relevant $ unnamed arg
