@@ -213,7 +213,8 @@ This makes order multiplication associative.
 
 -}
 collapse :: (?cutoff :: Int) => Matrix Integer Order -> Order
-collapse m = case (toLists (Matrix.transpose m)) of
+collapse m = -- if not $ Matrix.matrixInvariant m then __IMPOSSIBLE__ else
+  case toLists $ Matrix.transpose m of
    [] -> __IMPOSSIBLE__   -- This can never happen if order matrices are generated by the smart constructor
    m' -> foldl1 (.*.) $ map (foldl1 maxO) m'
 
