@@ -1,4 +1,4 @@
-{-# LANGUAGE CPP, PatternGuards, ImplicitParams #-}
+{-# LANGUAGE CPP, PatternGuards, ImplicitParams, TupleSections #-}
 
 {- Checking for Structural recursion
    Authors: Andreas Abel, Nils Anders Danielsson, Ulf Norell,
@@ -8,25 +8,30 @@
  -}
 
 module Agda.Termination.TermCheck
-    ( termDecls
+    ( termDecl
     , Result, DeBruijnPat
     ) where
 
 import Control.Applicative
 import Control.Monad.Error
+import Control.Monad.State
+
 import Data.List as List
 import qualified Data.Map as Map
 import Data.Map (Map)
 import qualified Data.Maybe as Maybe
+import Data.Monoid
 import qualified Data.Set as Set
 import Data.Set (Set)
 
 import qualified Agda.Syntax.Abstract as A
+-- import Agda.Syntax.Abstract.Pretty (prettyA)
 import Agda.Syntax.Internal
 import qualified Agda.Syntax.Info as Info
 import Agda.Syntax.Position
 import Agda.Syntax.Common
 import Agda.Syntax.Literal (Literal(LitString))
+import Agda.Syntax.Translation.InternalToAbstract
 
 import Agda.Termination.CallGraph   as Term
 import qualified Agda.Termination.SparseMatrix as Term
@@ -35,23 +40,28 @@ import qualified Agda.Termination.Termination  as Term
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Reduce (reduce, normalise, instantiate, instantiateFull)
-import Agda.TypeChecking.Records (isRecordConstructor)
+import Agda.TypeChecking.Records (isRecordConstructor, isRecord, isInductiveRecord)
 import Agda.TypeChecking.Rules.Builtin.Coinduction
 import Agda.TypeChecking.Rules.Term (isType_)
-import Agda.TypeChecking.Substitute (abstract,raise,substs)
+import Agda.TypeChecking.Substitute (abstract,raise)
 import Agda.TypeChecking.Telescope
 import Agda.TypeChecking.EtaContract
 import Agda.TypeChecking.Monad.Builtin
-import Agda.TypeChecking.Monad.Signature (isProjection)
+import Agda.TypeChecking.Monad.Signature (isProjection, mutuallyRecursive)
 import Agda.TypeChecking.Primitive (constructorForm)
 import Agda.TypeChecking.Level (reallyUnLevelView)
 import Agda.TypeChecking.Substitute
+import Agda.TypeChecking.SizedTypes
 
 import qualified Agda.Interaction.Highlighting.Range as R
 import Agda.Interaction.Options
 
+import Agda.Utils.List
 import Agda.Utils.Size
-import Agda.Utils.Monad (thread, (<$>), ifM)
+import Agda.Utils.Monad ((<$>), mapM', forM', ifM)
+-- import Agda.Utils.NubList
+import Agda.Utils.Pointed
+import Agda.Utils.Permutation
 
 #include "../undefined.h"
 import Agda.Utils.Impossible
@@ -60,12 +70,15 @@ type Calls = Term.CallGraph (Set CallInfo)
 type MutualNames = [QName]
 
 -- | The result of termination checking a module.
-
+--   Must be 'Pointed' and a 'Monoid'.
 type Result = [TerminationError]
 
+-- use of a NubList did not achieve the desired effect, now unnecessary
+-- type Result = NubList TerminationError
+
 -- | Termination check a sequence of declarations.
 termDecls :: [A.Declaration] -> TCM Result
-termDecls ds = fmap concat $ mapM termDecl ds
+termDecls ds = concat <$> mapM termDecl ds
 
 -- | Termination check a single declaration.
 termDecl :: A.Declaration -> TCM Result
@@ -73,30 +86,27 @@ termDecl (A.ScopedDecl scope ds) = do
   setScope scope
   termDecls ds
 termDecl d = case d of
-    A.Axiom {}            -> return []
-    A.Field {}            -> return []
-    A.Primitive {}        -> return []
+    A.Axiom {}            -> return mempty
+    A.Field {}            -> return mempty
+    A.Primitive {}        -> return mempty
     A.Mutual _ ds
-      | [A.RecSig{}, A.RecDef _ r _ _ _ rds] <- unscopeDefs ds
-                          -> do
-        let m = mnameFromList $ qnameToList r
-        setScopeFromDefs ds
-        termSection m rds
-    A.Mutual i ds  -> termMutual i ds
+      | [A.RecSig{}, A.RecDef _ r _ _ _ _ rds] <- unscopeDefs ds
+                          -> checkRecDef ds r rds
+    A.Mutual i ds         -> termMutual i ds
     A.Section _ x _ ds    -> termSection x ds
-    A.Apply {}            -> return []
-    A.Import {}           -> return []
-    A.Pragma {}           -> return []
-    A.Open {}             -> return []
+    A.Apply {}            -> return mempty
+    A.Import {}           -> return mempty
+    A.Pragma {}           -> return mempty
+    A.Open {}             -> return mempty
         -- open is just an artifact from the concrete syntax
     A.ScopedDecl{}        -> __IMPOSSIBLE__
         -- taken care of above
+    A.RecSig{}            -> return mempty
+    A.RecDef _ r _ _ _ _ ds -> checkRecDef [] r ds
     -- These should all be wrapped in mutual blocks
     A.FunDef{}  -> __IMPOSSIBLE__
     A.DataSig{} -> __IMPOSSIBLE__
     A.DataDef{} -> __IMPOSSIBLE__
-    A.RecSig{}  -> __IMPOSSIBLE__
-    A.RecDef{}  -> __IMPOSSIBLE__
   where
     setScopeFromDefs = mapM_ setScopeFromDef
     setScopeFromDef (A.ScopedDecl scope d) = setScope scope
@@ -107,32 +117,36 @@ termDecl d = case d of
     unscopeDef (A.ScopedDecl _ ds) = unscopeDefs ds
     unscopeDef d = [d]
 
-collectCalls :: (a -> TCM Calls) -> [a] -> TCM Calls
-collectCalls f [] = return Term.empty
-collectCalls f (a : as) = do c1 <- f a
-                             c2 <- collectCalls f as
-                             return (c1 `Term.union` c2)
+    checkRecDef ds r rds = do
+      setScopeFromDefs ds
+      termSection (mnameFromList $ qnameToList r) rds
 
 -- | Termination check a bunch of mutually inductive recursive definitions.
-termMutual :: Info.DeclInfo -> [A.Declaration] -> TCM Result
-termMutual i ds = if names == [] then return [] else
-  do -- get list of sets of mutually defined names from the TCM
+termMutual :: Info.MutualInfo -> [A.Declaration] -> TCM Result
+termMutual i ds = if names == [] then return mempty else
+  -- we set the range to avoid panics when printing error messages
+  traceCall (SetRange (Info.mutualRange i)) $ do
+
+  mutualBlock <- findMutualBlock (head names)
+  let allNames = Set.elems mutualBlock
+
+  if not (Info.mutualTermCheck i) then do
+      reportSLn "term.warn.yes" 2 $ "Skipping termination check for " ++ show names
+      forM_ allNames $ \ q -> setTerminates q True -- considered terminating!
+      return mempty
+   else do
+     -- get list of sets of mutually defined names from the TCM
      -- this includes local and auxiliary functions introduced
      -- during type-checking
 
      cutoff <- optTerminationDepth <$> pragmaOptions
-     let ?cutoff = cutoff
+     let ?cutoff = cutoff -- needed for Term.terminates
 
      reportSLn "term.top" 10 $ "Termination checking " ++ show names ++
        " with cutoff=" ++ show cutoff ++ "..."
-     mutualBlock <- findMutualBlock (head names)
-     let allNames = Set.elems mutualBlock
-
-     -- collect all recursive calls in the block
-     let collect use = collectCalls (termDef use allNames) allNames
 
      -- Get the name of size suc (if sized types are enabled)
-     suc <- sizeSuc
+     suc <- sizeSucName
 
      -- The name of sharp (if available).
      sharp <- fmap nameOfSharp <$> coinductionKit
@@ -140,59 +154,28 @@ termMutual i ds = if names == [] then return [] else
      guardingTypeConstructors <-
        optGuardingTypeConstructors <$> pragmaOptions
 
-     -- first try to termination check ignoring the dot patterns
      let conf = DBPConf
            { useDotPatterns           = False
            , guardingTypeConstructors = guardingTypeConstructors
            , withSizeSuc              = suc
            , sharp                    = sharp
+           , currentTarget            = Nothing
            }
-     calls1 <- collect conf{ useDotPatterns = False }
-     reportS "term.lex" 20 $ unlines
-       [ "Calls (no dot patterns): " ++ show calls1
-       ]
-     reportSDoc "term.behaviours" 20 $ vcat
-       [ text "Recursion behaviours (no dot patterns):"
-       , nest 2 $ return $ Term.prettyBehaviour (Term.complete calls1)
-       ]
-     reportSDoc "term.matrices" 30 $ vcat
-       [ text "Call matrices (no dot patterns):"
-       , nest 2 $ pretty $ Term.complete calls1
-       ]
-     r <- do let r = Term.terminates calls1
-             case r of
-               Right _ -> return r
-               Left _  -> do
-                 -- Try again, but include the dot patterns this time.
-                 calls2 <- collect conf{ useDotPatterns = True }
-                 reportS "term.lex" 20 $ unlines
-                   [ "Calls (dot patterns): " ++ show calls2
-                   ]
-                 reportSDoc "term.behaviours" 20 $ vcat
-                   [ text "Recursion behaviours (dot patterns):"
-                   , nest 2 $ return $
-                       Term.prettyBehaviour (Term.complete calls2)
-                   ]
-                 reportSDoc "term.matrices" 30 $ vcat
-                   [ text "Call matrices (dot patterns):"
-                   , nest 2 $ pretty $ Term.complete calls2
-                   ]
-                 return $ Term.terminates calls2
-     case r of
-       Left calls -> do
-         return [TerminationError
-                   { termErrFunctions = names
-                     -- TODO: This could be changed to allNames.
-                   , termErrCalls     = Set.toList calls
-                   }
-                ]
-       Right _ -> do
-         reportSLn "term.warn.yes" 2
-                     (show (names) ++ " does termination check")
-         return []
+
+     -- new check currently only makes a difference for copatterns
+     -- since it is slow, only invoke it if --copatterns
+     res <- ifM (optCopatterns <$> pragmaOptions)
+       (forM' allNames $ termFunction conf names allNames) -- new check one after another
+       (termMutual' conf names allNames) -- old check, all at once
+
+     -- record result of termination check in signature
+     let terminates = null res
+     forM_ allNames $ \ q -> setTerminates q terminates
+     return res
+
   where
-  getName (A.FunDef i x cs)       = [x]
-  getName (A.RecDef _ _ _ _ _ ds) = concatMap getName ds
+  getName (A.FunDef i x delayed cs) = [x]
+  getName (A.RecDef _ _ _ _ _ _ ds) = concatMap getName ds
   getName (A.Mutual _ ds)       = concatMap getName ds
   getName (A.Section _ _ _ ds)  = concatMap getName ds
   getName (A.ScopedDecl _ ds)   = concatMap getName ds
@@ -201,8 +184,112 @@ termMutual i ds = if names == [] then return [] else
   -- the mutual names mentioned in the abstract syntax
   names = concatMap getName ds
 
-  concat' :: Ord a => [Set a] -> [a]
-  concat' = Set.toList . Set.unions
+
+-- | @termMutual' conf names allNames@ checks @allNames@ for termination.
+--
+--   @names@ is taken from the 'Abstract' syntax, so it contains only
+--   the names the user has declared.  This is for error reporting.
+--
+--   @allNames@ is taken from 'Internal' syntax, it contains also
+--   the definitions created by the type checker (e.g., with-functions).
+--
+termMutual' :: (?cutoff :: Int) => DBPConf -> [QName] -> MutualNames -> TCM Result
+termMutual' conf names allNames = do
+
+     -- collect all recursive calls in the block
+     let collect conf = mapM' (termDef conf allNames) allNames
+
+     -- first try to termination check ignoring the dot patterns
+     calls1 <- collect conf{ useDotPatterns = False }
+     reportCalls "no " calls1
+
+     r <- case Term.terminates calls1 of
+            r@Right{} -> return r
+            Left{}    -> do
+              -- Try again, but include the dot patterns this time.
+              calls2 <- collect conf{ useDotPatterns = True }
+              reportCalls "" calls2
+              return $ Term.terminates calls2
+     case r of
+       Left calls -> do
+         return $ point $ TerminationError
+                   { termErrFunctions = names
+                   , termErrCalls     = Set.toList calls
+                   }
+       Right _ -> do
+         reportSLn "term.warn.yes" 2
+                     (show (names) ++ " does termination check")
+         return mempty
+  where
+
+reportCalls no calls = do
+   reportS "term.lex" 20 $ unlines
+     [ "Calls (" ++ no ++ "dot patterns): " ++ show calls
+     ]
+   reportSDoc "term.behaviours" 20 $ vcat
+     [ text $ "Recursion behaviours (" ++ no ++ "dot patterns):"
+     , nest 2 $ return $ Term.prettyBehaviour (Term.complete calls)
+     ]
+   reportSDoc "term.matrices" 30 $ vcat
+     [ text $ "Call matrices (" ++ no ++ "dot patterns):"
+     , nest 2 $ pretty $ Term.complete calls
+     ]
+
+-- | @termFunction conf names allNames name@ checks @name@ for termination.
+--
+--   @names@ is taken from the 'Abstract' syntax, so it contains only
+--   the names the user has declared.  This is for error reporting.
+--
+--   @allNames@ is taken from 'Internal' syntax, it contains also
+--   the definitions created by the type checker (e.g., with-functions).
+--
+termFunction :: (?cutoff :: Int) => DBPConf -> [QName] -> MutualNames -> QName -> TCM Result
+termFunction conf0 names allNames name = do
+
+     let index = toInteger $ maybe __IMPOSSIBLE__ id $
+           List.elemIndex name allNames
+
+     conf <- do
+       r <- typeEndsInDef =<< typeOfConst name
+       reportTarget r
+       return $ conf0 { currentTarget = r }
+
+     -- collect all recursive calls in the block
+     let collect conf = mapM' (termDef conf allNames) allNames
+
+     -- first try to termination check ignoring the dot patterns
+     calls1 <- collect conf{ useDotPatterns = False }
+     reportCalls "no " calls1
+
+     r <- case Term.terminatesFilter (== index) calls1 of
+            r@Right{} -> return r
+            Left{}    -> do
+              -- Try again, but include the dot patterns this time.
+              calls2 <- collect conf{ useDotPatterns = True }
+              reportCalls "" calls2
+              return $ Term.terminatesFilter (== index) calls2
+     case r of
+       Left calls -> do
+         return $ point $ TerminationError
+                   { termErrFunctions = if name `elem` names then [name] else []
+                   , termErrCalls     = Set.toList calls
+                   }
+       Right _ -> do
+         reportSLn "term.warn.yes" 2
+                     (show (name) ++ " does termination check")
+         return mempty
+  where
+    reportTarget r = reportSLn "term.target" 20 $ maybe
+      ("  target type not recognized")
+      (\ q -> "  target type ends in " ++ show q)
+      r
+
+typeEndsInDef :: Type -> TCM (Maybe QName)
+typeEndsInDef t = do
+  TelV _ core <- telView t
+  case ignoreSharing $ unEl core of
+    Def d vs -> return $ Just d
+    _        -> return Nothing
 
 -- | Termination check a module.
 termSection :: ModuleName -> [A.Declaration] -> TCM Result
@@ -228,8 +315,8 @@ termDef use names name = do
 	      , nest 2 $ text ":" <+> (prettyTCM $ defType def)
 	      ]
         case (theDef def) of
-          Function{ funClauses = cls } ->
-            collectCalls (termClause use names name) cls
+          Function{ funClauses = cls, funDelayed = delayed } ->
+            mapM' (termClause use names name delayed) cls
           _ -> return Term.empty
 
 
@@ -286,15 +373,44 @@ liftDBP = adjIndexDBP (1+)
 
 {- | Configuration parameters to termination checker.
 -}
-data DBPConf = DBPConf { useDotPatterns           :: Bool
-                       , guardingTypeConstructors :: Bool
-                         -- ^ Do we assume that record and data type
-                         --   constructors preserve guardedness?
-                       , withSizeSuc              :: Maybe QName
-                       , sharp                    :: Maybe QName
-                         -- ^ The name of the sharp constructor, if
-                         --   any.
-                       }
+data DBPConf = DBPConf
+  { useDotPatterns           :: Bool
+  , guardingTypeConstructors :: Bool
+    -- ^ Do we assume that record and data type constructors preserve guardedness?
+  , withSizeSuc              :: Maybe QName
+  , sharp                    :: Maybe QName
+    -- ^ The name of the sharp constructor, if any.
+  , currentTarget            :: Maybe Target
+    -- ^ Target type of the function we are currently termination checking.
+    --   Only the constructors of 'Target' are considered guarding.
+  }
+
+type Target = QName
+
+targetElem :: DBPConf -> [Target] -> Bool
+targetElem conf ds = maybe False (`elem` ds) (currentTarget conf)
+
+{-
+-- | Check wether a 'Target" corresponds to the current one.
+matchingTarget :: DBPConf -> Target -> TCM Bool
+matchingTarget conf d = maybe (return True) (mutuallyRecursive d) (currentTarget conf)
+-}
+
+{-
+-- | The target type of the considered recursive definition.
+data Target
+  = Set        -- ^ Constructing a Set (only meaningful with 'guardingTypeConstructors').
+  | Data QName -- ^ Constructing a coinductive or mixed type (could be data or record).
+  deriving (Eq, Show)
+
+-- | Check wether a 'Target" corresponds to the current one.
+matchingTarget :: DBPConf -> Target -> TCM Bool
+matchingTarget conf t = maybe (return True) (match t) (currentTarget conf)
+  where
+    match Set      Set       = return True
+    match (Data d) (Data d') = mutuallyRecursive d d'
+    match _ _                = return False
+-}
 
 {- | Convert a term (from a dot pattern) to a DeBruijn pattern.
 -}
@@ -304,7 +420,7 @@ termToDBP conf t
   | not $ useDotPatterns conf = return $ unusedVar
   | otherwise                 = do
     t <- stripProjections =<< constructorForm t
-    case t of
+    case ignoreSharing t of
       Var i []    -> return $ VarDBP i
       Con c args  -> ConDBP c <$> mapM (termToDBP conf . unArg) args
       Def s [arg]
@@ -326,6 +442,7 @@ stripCoConstructors conf p = case p of
       Inductive   -> ConDBP c <$> mapM (stripCoConstructors conf) args
       CoInductive -> return unusedVar
 
+{- Andreas, 2012-09-19 BAD CODE, RETIRED
 {- | stripBind i p b = Just (i', dbp, b')
 
   converts a pattern into a de Bruijn pattern
@@ -367,23 +484,67 @@ stripBinds use i (p:ps) b = do
         Just (i2, dbps, b2) -> return $ Just (i2, dbp:dbps, b2)
         Nothing -> return Nothing
     Nothing -> return Nothing
+-}
+
+-- | cf. 'TypeChecking.Coverage.Match.buildMPatterns'
+openClause :: DBPConf -> Permutation -> [Pattern] -> ClauseBody -> TCM ([DeBruijnPat], Maybe Term)
+openClause conf perm ps body = do
+  -- invariant: xs has enough variables for the body
+  unless (permRange perm == genericLength xs) __IMPOSSIBLE__
+  dbps <- evalStateT (mapM build ps) xs
+  return . (dbps,) $ case body `apply` map (defaultArg . var) xs of
+    NoBody -> Nothing
+    Body v -> Just v
+    _      -> __IMPOSSIBLE__
+  where
+    -- length of the telescope
+    n    = size perm
+    -- the variables as a map from the body variables to the clause telescope
+    xs   = permute (invertP perm) $ downFrom (size perm)
+
+    tick = do x : xs <- get; put xs; return x
+
+    build :: Pattern -> StateT [Nat] TCM DeBruijnPat
+    build (VarP _)        = VarDBP <$> tick
+    build (ConP con _ ps) = ConDBP con <$> mapM (build . unArg) ps
+    build (DotP t)        = tick *> do lift $ termToDBP conf t
+    build (LitP l)        = return $ LitDBP l
+
 
 -- | Extract recursive calls from one clause.
-termClause :: DBPConf -> MutualNames -> QName -> Clause -> TCM Calls
-termClause use names name (Clause { clauseTel  = tel
-                                  , clausePerm = perm
-                                  , clausePats = argPats'
-                                  , clauseBody = body }) =
+termClause :: DBPConf -> MutualNames -> QName -> Delayed -> Clause -> TCM Calls
+termClause conf names name delayed
+    (Clause { clauseTel  = tel
+            , clausePerm = perm
+            , clausePats = argPats'
+            , clauseBody = body }) = do
+  reportSDoc "term.check.clause" 25 $ vcat
+    [ text "termClause"
+    , nest 2 $ text "tel      =" <+> prettyTCM tel
+    , nest 2 $ text ("perm     = " ++ show perm)
+    -- how to get the following right?
+    -- , nest 2 $ text "argPats' =" <+> do prettyA =<< reifyPatterns tel perm argPats'
+    ]
+  addCtxTel tel $ do
+    ps <- normalise $ map unArg argPats'
+    (dbpats, res) <- openClause conf perm ps body
+    case res of
+       Nothing -> return Term.empty
+       Just t -> do
+          dbpats <- mapM (stripCoConstructors conf) dbpats
+          termTerm conf names name delayed dbpats t
+
+{-
   addCtxTel tel $ do
     argPats' <- normalise argPats'
     -- The termination checker doesn't know about reordered telescopes
     let argPats = substs (renamingR perm) argPats'
-    dbs <- stripBinds use (nVars - 1) (map unArg argPats) body
+    dbs <- stripBinds conf (nVars - 1) (map unArg argPats) body
     case dbs of
        Nothing -> return Term.empty
        Just (-1, dbpats, Body t) -> do
-          dbpats <- mapM (stripCoConstructors use) dbpats
-          termTerm use names name dbpats t
+          dbpats <- mapM (stripCoConstructors conf) dbpats
+          termTerm conf names name delayed dbpats t
           -- note: convert dB levels into dB indices
        Just (n, dbpats, Body t) -> internalError $ "termClause: misscalculated number of vars: guess=" ++ show nVars ++ ", real=" ++ show (nVars - 1 - n)
        Just (n, dbpats, b)  -> internalError $ "termClause: not a Body" -- ++ show b
@@ -392,20 +553,35 @@ termClause use names name (Clause { clauseTel  = tel
     boundVars (Bind b)   = 1 + boundVars (absBody b)
     boundVars NoBody     = 0
     boundVars (Body _)   = 0
+-}
 
 -- | Extract recursive calls from a term.
-termTerm :: DBPConf -> MutualNames -> QName -> [DeBruijnPat] -> Term -> TCM Calls
-termTerm conf names f pats0 t0 = do
+termTerm :: DBPConf -> MutualNames -> QName -> Delayed -> [DeBruijnPat] -> Term -> TCM Calls
+termTerm conf names f delayed pats0 t0 = do
  cutoff <- optTerminationDepth <$> pragmaOptions
  let ?cutoff = cutoff
  do
   reportSDoc "term.check.clause" 6
-    (sep [ text "termination checking clause of" <+> prettyTCM f
+    (sep [ text ("termination checking " ++
+             (if delayed == Delayed then "delayed " else "") ++ "clause of")
+           <+> prettyTCM f
          , nest 2 $ text "lhs:" <+> hsep (map prettyTCM pats0)
          , nest 2 $ text "rhs:" <+> prettyTCM t0
          ])
-  loop pats0 Term.le t0
+  {-
+  -- if we are checking a delayed definition, we treat it as if there were
+  -- a guarding coconstructor (sharp)
+  let guarded = case delayed of
+        Delayed    -> Term.lt
+        NotDelayed -> Term.le
+  -}
+  let guarded = Term.le -- not initially guarded
+  loop pats0 guarded t0
   where
+       -- only a delayed definition can be guarded
+       ifDelayed o | Term.decreasing o && delayed == NotDelayed = Term.le
+                   | otherwise                                  = o
+
        Just fInd = toInteger <$> List.elemIndex f names
 
        -- sorts can contain arb. terms of type Nat,
@@ -414,10 +590,18 @@ termTerm conf names f pats0 t0 = do
        -- a subgrammar of Term, then we would not need this boilerplate.
        loopSort :: (?cutoff :: Int) => [DeBruijnPat] -> Sort -> TCM Calls
        loopSort pats s = do
+         reportSDoc "term.sort" 20 $ text "extracting calls from sort" <+> prettyTCM s
+         reportSDoc "term.sort" 50 $ text ("s = " ++ show s)
+         -- s <- instantiateFull s -- Andreas, 2012-09-05 NOT NECESSARY
+         -- instantiateFull resolves problems with reallyUnLevelView
+         -- in the absense of level built-ins.
+         -- However, the termination checker should only receive terms
+         -- that are already fully instantiated.
+
          case s of
            Type (Max [])              -> return Term.empty
            Type (Max [ClosedLevel _]) -> return Term.empty
-           Type t -> loop pats Term.unknown (Level t)
+           Type t -> loop pats Term.unknown (Level t) -- no guarded levels
            Prop   -> return Term.empty
            Inf    -> return Term.empty
            DLub s1 (NoAbs x s2) -> Term.union <$> loopSort pats s1 <*> loopSort pats s2
@@ -438,7 +622,6 @@ termTerm conf names f pats0 t0 = do
          -> TCM Calls
        loop pats guarded t = do
          t <- instantiate t          -- instantiate top-level MetaVar
-         suc <- sizeSuc
 
              -- Handles constructor applications.
          let constructor
@@ -453,7 +636,7 @@ termTerm conf names f pats0 t0 = do
                   --   argument should be viewed as preserving
                   --   guardedness.
                -> TCM Calls
-             constructor c ind args = collectCalls loopArg args
+             constructor c ind args = mapM' loopArg args
                where
                loopArg (arg , preserves) = do
                  loop pats g' (unArg arg)
@@ -469,22 +652,22 @@ termTerm conf names f pats0 t0 = do
                args2 <- mapM instantiateFull args1
 
                -- We have to reduce constructors in case they're reexported.
-               let reduceCon (Con c vs) = (`apply` vs) <$> reduce (Con c [])  -- make sure we don't reduce the arguments
-                   reduceCon t          = return t
+               let reduceCon t = case ignoreSharing t of
+                      Con c vs -> (`apply` vs) <$> reduce (Con c [])  -- make sure we don't reduce the arguments
+                      _        -> return t
                args2 <- mapM reduceCon args2
                args  <- mapM etaContract args2
 
                -- If the function is a projection, then preserve guardedness
                -- for its principal argument.
-               isProj <- isProjection g
+               isProj <- isProjectionButNotFlat g
                let unguards = repeat Term.unknown
-               let guards = maybe unguards -- not a proj. ==> unguarded
-                              (\ _ -> guarded : unguards)
-                                -- proj. => preserve g. of princ. arg. (counting starts with 1)
-                              isProj
+               let guards = if isProj then guarded : unguards
+                                           -- proj => preserve guardedness of principal argument
+                                      else unguards -- not a proj ==> unguarded
                -- collect calls in the arguments of this call
-               calls <- collectCalls (uncurry (loop pats)) (zip guards args)
-               -- calls <- collectCalls (loop pats Term.unknown) args
+               calls <- mapM' (uncurry (loop pats)) (zip guards args)
+               -- calls <- mapM' (loop pats Term.unknown) args
 
 
                reportSDoc "term.found.call" 20
@@ -501,8 +684,9 @@ termTerm conf names f pats0 t0 = do
                   -- call is to one of the mutally recursive functions
                   Just gInd' -> do
 
-                     matrix <- compareArgs suc pats args
-                     let (nrows, ncols, matrix') = addGuardedness guarded
+                     matrix <- compareArgs (withSizeSuc conf) pats args
+                     let (nrows, ncols, matrix') = addGuardedness
+                            (ifDelayed guarded)  -- only delayed defs can be guarded
                             (genericLength args) -- number of rows
                             (genericLength pats) -- number of cols
                             matrix
@@ -530,58 +714,76 @@ termTerm conf names f pats0 t0 = do
                          calls)
 
 
-         case t of
+         case ignoreSharing t of
 
             -- Constructed value.
             Con c args
               | Just c == sharp conf ->
                 constructor c CoInductive $ zip args (repeat True)
-              | otherwise ->
-                constructor c Inductive $ zip args (repeat True)
+              | otherwise -> do
+                -- If we encounter a coinductive record constructor
+                -- in a type mutual with the current target
+                -- then we count it as guarding.
+                ind <- do
+                  r <- isRecordConstructor c
+                  case r of
+                    Nothing       -> return Inductive
+                    Just (q, def) -> return . (\ b -> if b then CoInductive else Inductive) $
+                      and [ recRecursive def
+                          , recInduction def == CoInductive
+                          , targetElem conf (q : recMutual def)
+                          ]
+                constructor c ind $ zip args (repeat True)
 
             Def g args0
               | guardingTypeConstructors conf -> do
-                gDef <- theDef <$> getConstInfo g
-                case gDef of
+                def <- getConstInfo g
+                let occs = defArgOccurrences def
+                case theDef def of
+{-
                   Datatype {dataArgOccurrences = occs} -> con occs
                   Record   {recArgOccurrences  = occs} -> con occs
-                  _                                    -> fun
+-}
+                  Datatype{} -> con occs
+                  Record{}   -> con occs
+                  _          -> fun
               | otherwise -> fun
               where
               -- Data or record type constructor.
               con occs =
                 constructor g Inductive $   -- guardedness preserving
-                -- constructor g CoInductive $ -- guarding! (Andreas, 2011-04-10) -- does not work, might lead to infinite unfolding in eq. checking (Ripley!)
                   zip args0 (map preserves occs ++ repeat False)
                 where
-                preserves Positive = True
-                preserves Negative = False
+                preserves = (StrictPos <=)   -- everything which is at least strictly positive
+{- SPELLED OUT, this means:
                 preserves Unused   = True
+                preserves GuardPos = True
+                preserves StrictPos = True
+                preserves Mixed = False
+-}
 
               -- Call to defined function.
               fun = function g args0
 
             -- Abstraction. Preserves guardedness.
-            Lam h (Abs x t) -> addCtxString x (Arg { argHiding    = h
-                                                   , argRelevance = __IMPOSSIBLE__
-                                                   , unArg        = __IMPOSSIBLE__
-                                                   }) $
+            Lam h (Abs x t) -> addCtxString_ x $
               loop (map liftDBP pats) guarded t
             Lam h (NoAbs _ t) -> loop pats guarded t
 
             -- Neutral term. Destroys guardedness.
-            Var i args -> collectCalls (loop pats Term.unknown) (map unArg args)
+            Var i args -> mapM' (loop pats Term.unknown) (map unArg args)
 
             -- Dependent function space.
             Pi a (Abs x b) ->
-               do g1 <- loopType pats Term.unknown (unArg a)
+               do g1 <- loopType pats Term.unknown (unDom a)
+                  a  <- maskSizeLt a
                   g2 <- addCtxString x a $
                         loopType (map liftDBP pats) piArgumentGuarded b
                   return $ g1 `Term.union` g2
 
             -- Non-dependent function space.
             Pi a (NoAbs _ b) ->
-               do g1 <- loopType pats Term.unknown (unArg a)
+               do g1 <- loopType pats Term.unknown (unDom a)
                   g2 <- loopType pats piArgumentGuarded b
                   return $ g1 `Term.union` g2
 
@@ -599,7 +801,13 @@ termTerm conf names f pats0 t0 = do
             DontCare t -> loop pats guarded t
 
             -- Level.
-            Level l -> loop pats guarded =<< reallyUnLevelView l
+            Level l -> do
+              l <- catchError (reallyUnLevelView l) $ const $ internalError $
+                "Termination checker: cannot view level expression, " ++
+                "probably due to missing level built-ins."
+              loop pats guarded l
+
+            Shared{} -> __IMPOSSIBLE__
 
          where
          -- Should function and Π type constructors be treated as
@@ -607,10 +815,23 @@ termTerm conf names f pats0 t0 = do
          piArgumentGuarded =
            if guardingTypeConstructors conf then
              guarded   -- preserving guardedness
-             -- Term.lt      -- guarding! (Andreas, 2011-04-10)  -- SEE ABOVE
             else
              Term.unknown
 
+-- | Rewrite type @tel -> Size< u@ to @tel -> Size@.
+maskSizeLt :: Dom Type -> TCM (Dom Type)
+maskSizeLt dom@(Dom h r a) = do
+  (msize, msizelt) <- getBuiltinSize
+  case (msize, msizelt) of
+    (_ , Nothing) -> return dom
+    (Nothing, _)  -> __IMPOSSIBLE__
+    (Just size, Just sizelt) -> do
+      TelV tel c <- telView a
+      case ignoreSharingType a of
+        El s (Def d [v]) | d == sizelt -> return $ Dom h r $
+          abstract tel $ El s $ Def size []
+        _ -> return dom
+
 {- | compareArgs suc pats ts
 
      compare a list of de Bruijn patterns (=parameters) @pats@
@@ -644,21 +865,9 @@ addGuardedness g nrows ncols m =
 
 -- | Stripping off a record constructor is not counted as decrease, in
 --   contrast to a data constructor.
-decreaseFromConstructor :: QName -> TCM Order
-decreaseFromConstructor c = do
-  isRC <- isRecordConstructor c
-  return $ if isRC then Term.le else Term.lt
-
-increaseFromConstructor :: (?cutoff :: Int) => QName -> TCM Order
-increaseFromConstructor c = do
-  isRC <- isRecordConstructor c
-  return $ if isRC then Term.le else Term.decr (-1)
-
-{-
-increaseFromConstructor c = negateOrder <$> decreaseFromConstructor c
-  where negateOrder (Decr k) = Decr (- k)
-        negateOrder _ = __IMPOSSIBLE__
--}
+--   A record constructor increases/decreases by 0, a data constructor by 1.
+offsetFromConstructor :: QName -> TCM Int
+offsetFromConstructor c = maybe 1 (const 0) <$> isRecordConstructor c
 
 -- | Compute the sub patterns of a 'DeBruijnPat'.
 subPatterns :: DeBruijnPat -> [DeBruijnPat]
@@ -678,16 +887,33 @@ compareTerm t p = Term.supremum $ compareTerm' t p : map cmp (subPatterns p)
     cmp p' = (Term..*.) Term.lt (compareTerm' t p')
 -}
 
+-- | For termination checking purposes flat should not be considered a
+--   projection. That is, it flat doesn't preserve either structural order
+--   or guardedness like other projections do.
+--   Andreas, 2012-06-09: the same applies to projections of recursive records.
+isProjectionButNotFlat :: QName -> TCM Bool
+isProjectionButNotFlat qn = do
+  flat <- fmap nameOfFlat <$> coinductionKit
+  if Just qn == flat
+    then return False
+    else do
+      mp <- isProjection qn
+      case mp of
+        Nothing     -> return False
+        Just (r, _) -> isInductiveRecord r
+
+
 -- | Remove projections until a term is no longer a projection.
 --   Also, remove 'DontCare's.
 stripProjections :: Term -> TCM Term
-stripProjections (DontCare t) = stripProjections t
-stripProjections t@(Def qn ts@(~(r : _))) = do
-  isProj <- isProjection qn
-  case isProj of
-    Just{} | not (null ts) -> stripProjections $ unArg r
-    _ -> return t
-stripProjections t = return t
+stripProjections t = case ignoreSharing t of
+  DontCare t -> stripProjections t
+  Def qn ts@(~(r : _)) -> do
+    isProj <- isProjectionButNotFlat qn
+    case isProj of
+      True | not (null ts) -> stripProjections $ unArg r
+      _ -> return t
+  _ -> return t
 
 -- | Remove all projections from an algebraic term (not going under binders).
 class StripAllProjections a where
@@ -702,62 +928,41 @@ instance StripAllProjections a => StripAllProjections [a] where
 instance StripAllProjections Term where
   stripAllProjections t = do
     t <- stripProjections t
-    case t of
+    case ignoreSharing t of
       Con c ts -> Con c <$> stripAllProjections ts
       Def d ts -> Def d <$> stripAllProjections ts
       _ -> return t
 
-{-
--- | Remove all projections from an algebraic term (not going under binders).
-stripAllProjections :: Term -> TCM Term
-stripAllProjections t = do
-  t <- stripProjections t
-  case t of
-    Con c ts -> Con c <$> mapM stripAllProjections ts
-    Def d ts -> Def d <$> mapM stripAllProjections ts
-    _ -> return t
--}
-
 -- | compareTerm t dbpat
 --   Precondition: top meta variable resolved
 compareTerm' :: (?cutoff :: Int) => Maybe QName -> Term -> DeBruijnPat -> TCM Term.Order
-compareTerm' _ (Var i _)      p = compareVar i p
+compareTerm' suc (Shared x)   p = compareTerm' suc (derefPtr x) p
+compareTerm' suc (Var i _)    p = compareVar suc i p
 compareTerm' suc (DontCare t) p = compareTerm' suc t p
 compareTerm' _ (Lit l)    (LitDBP l')
   | l == l'   = return Term.le
   | otherwise = return Term.unknown
 compareTerm' suc (Lit l) p = do
   t <- constructorForm (Lit l)
-  case t of
+  case ignoreSharing t of
     Lit _ -> return Term.unknown
     _     -> compareTerm' suc t p
 -- Andreas, 2011-04-19 give subterm priority over matrix order
 compareTerm' _ t@Con{} (ConDBP c ps)
-  | any (isSubTerm t) ps = decreaseFromConstructor c
+  | any (isSubTerm t) ps = decrease <$> offsetFromConstructor c <*> return Term.le
 compareTerm' suc (Con c ts) (ConDBP c' ps)
   | c == c' = compareConArgs suc ts ps
 compareTerm' suc (Def s ts) (ConDBP s' ps)
   | s == s' && Just s == suc = compareConArgs suc ts ps
 -- new cases for counting constructors / projections
 -- register also increase
-compareTerm' suc (Def s ts) p | Just s == suc = do
-    os <- mapM (\ t -> compareTerm' suc (unArg t) p) ts
-    return $ decr (-1) .*. infimum os
-{- Andreas 2011-07-07 Projections are now being removed in a preprocess
--- projections are size preserving
-compareTerm' suc (Def qn ts) p = do
-    isProj <- isProjection qn
-    case isProj of
-      -- strip off projection (n is the number of the record argument, counting starts with 1)
-      Just n | length ts >= n && n >= 1 ->
-        compareTerm' suc (unArg (ts !! (n - 1))) p
-      -- not a projection or underapplied:
-      _ -> return Term.unknown
--}
+compareTerm' suc (Def s [t]) p | Just s == suc = do
+    -- Andreas, 2012-10-19 do not cut off here
+    increase 1 <$> compareTerm' suc (unArg t) p
+compareTerm' suc (Con c []) p = return Term.le
 compareTerm' suc (Con c ts) p = do
-    os <- mapM (\ t -> compareTerm' suc (unArg t) p) ts
-    oc <- increaseFromConstructor c
-    return $ if (null os) then Term.unknown else oc .*. infimum os
+    increase <$> offsetFromConstructor c
+             <*> (infimum <$> mapM (\ t -> compareTerm' suc (unArg t) p) ts)
 compareTerm' suc t p | isSubTerm t p = return Term.le
 compareTerm' _ _ _ = return Term.unknown
 
@@ -765,6 +970,7 @@ compareTerm' _ _ _ = return Term.unknown
 isSubTerm :: Term -> DeBruijnPat -> Bool
 isSubTerm t p = equal t p || properSubTerm t p
   where
+    equal (Shared p) dbp = equal (derefPtr p) dbp
     equal (Con c ts) (ConDBP c' ps) =
       and $ (c == c')
           : (length ts == length ps)
@@ -797,11 +1003,19 @@ compareConArgs suc ts ps =
        -- allows examples like (x, y) < (Succ x, y)
 -}
 
-compareVar :: (?cutoff :: Int) => Nat -> DeBruijnPat -> TCM Term.Order
-compareVar i (VarDBP j)    = return $ if i == j then Term.le else Term.unknown
-compareVar i (LitDBP _)    = return $ Term.unknown
-compareVar i (ConDBP c ps) = do
-  os <- mapM (compareVar i) ps
-  let o = Term.supremum os
-  oc <- decreaseFromConstructor c
-  return $ (Term..*.) oc o
+compareVar :: (?cutoff :: Int) => Maybe QName -> Nat -> DeBruijnPat -> TCM Term.Order
+compareVar suc i (VarDBP j)    = compareVarVar suc i j
+compareVar suc i (LitDBP _)    = return $ Term.unknown
+compareVar suc i (ConDBP c ps) = do
+  decrease <$> offsetFromConstructor c
+           <*> (Term.supremum <$> mapM (compareVar suc i) ps)
+
+-- | Compare two variables
+compareVarVar :: (?cutoff :: Int) => Maybe QName -> Nat -> Nat -> TCM Term.Order
+compareVarVar suc i j
+  | i == j    = return Term.le
+  | otherwise = do
+      res <- isBounded i
+      case res of
+        BoundedNo  -> return Term.unknown
+        BoundedLt v -> decrease 1 <$> compareTerm' suc v (VarDBP j)
