@@ -14,7 +14,7 @@ import Agda.Utils.Impossible
 piAbstractTerm :: Term -> Type -> Type -> Type
 piAbstractTerm v a b = fun a (abstractTerm v b)
   where
-    fun a b = El s $ Pi (Arg NotHidden Relevant a) $ mkAbs "w" b
+    fun a b = El s $ Pi (Dom NotHidden Relevant a) $ mkAbs "w" b
       where s = (sLub `on` getSort) a b
 
 class AbstractTerm a where
@@ -34,9 +34,13 @@ instance AbstractTerm Term where
     Sort s      -> Sort $ absT s
     MetaV m vs  -> MetaV m $ absT vs
     DontCare mv -> DontCare $ absT mv
+    Shared p    -> Shared $ absT p
     where
       absT x = abstractTerm u x
 
+instance AbstractTerm a => AbstractTerm (Ptr a) where
+  abstractTerm u = fmap (abstractTerm u)
+
 instance AbstractTerm Type where
   abstractTerm u (El s v) = El (abstractTerm u s) (abstractTerm u v)
 
@@ -65,19 +69,21 @@ instance AbstractTerm LevelAtom where
 instance AbstractTerm a => AbstractTerm (Arg a) where
   abstractTerm = fmap . abstractTerm
 
+instance AbstractTerm a => AbstractTerm (Dom a) where
+  abstractTerm = fmap . abstractTerm
+
 instance AbstractTerm a => AbstractTerm [a] where
   abstractTerm = fmap . abstractTerm
 
 instance AbstractTerm a => AbstractTerm (Maybe a) where
   abstractTerm = fmap . abstractTerm
 
-instance (Raise a, AbstractTerm a) => AbstractTerm (Abs a) where
+instance (Subst a, AbstractTerm a) => AbstractTerm (Abs a) where
   abstractTerm u (NoAbs x v) = NoAbs x $ abstractTerm u v
-  abstractTerm u (Abs   x v) = Abs x $ rename swap $ abstractTerm (raise 1 u) v
+  abstractTerm u (Abs   x v) = Abs x $ applySubst swap $ abstractTerm (raise 1 u) v
     where
-      swap 0 = 1
-      swap 1 = 0
-      swap i = i
+      -- This swaps var 0 and var 1 (we hope)
+      swap = var 1 :# liftS 1 (raiseS 1)
 
 instance (AbstractTerm a, AbstractTerm b) => AbstractTerm (a, b) where
   abstractTerm u (x, y) = (abstractTerm u x, abstractTerm u y)
