@@ -4,20 +4,24 @@ module Agda.TypeChecking.Conversion where
 
 import Control.Applicative
 import Control.Monad
+import Control.Monad.Reader
 import Control.Monad.State
 import Control.Monad.Error
 import Data.Traversable hiding (mapM, sequence)
 import Data.List hiding (sort)
 import qualified Data.List as List
 
+import Agda.Syntax.Abstract.Views (isSet)
 import Agda.Syntax.Literal
 import Agda.Syntax.Common
 import Agda.Syntax.Internal
+import Agda.Syntax.Translation.InternalToAbstract (reify)
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.MetaVars
 import Agda.TypeChecking.MetaVars.Occurs (killArgs,PruneResult(..))
-import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Reduce
+import Agda.TypeChecking.Substitute
+import Agda.TypeChecking.Telescope
 import Agda.TypeChecking.Constraints
 import Agda.TypeChecking.Errors
 import Agda.TypeChecking.Primitive (constructorForm)
@@ -25,15 +29,19 @@ import Agda.TypeChecking.Free
 import Agda.TypeChecking.Records
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Injectivity
+import Agda.TypeChecking.Polarity
 import Agda.TypeChecking.SizedTypes
 import Agda.TypeChecking.Monad.Builtin
 import Agda.TypeChecking.Level
+import Agda.TypeChecking.Implicit (implicitArgs)
 import Agda.TypeChecking.Irrelevance
 import Agda.TypeChecking.EtaContract
 import Agda.TypeChecking.Eliminators
 -- import Agda.TypeChecking.UniversePolymorphism
 
+import Agda.Utils.Size
 import Agda.Utils.Monad
+import Agda.Utils.Maybe
 
 import Agda.TypeChecking.Monad.Debug
 
@@ -43,9 +51,6 @@ import Agda.Utils.Impossible
 mlevel :: TCM (Maybe Term)
 mlevel = liftTCM $ (Just <$> primLevel) `catchError` \_ -> return Nothing
 
-nextPolarity []       = (Invariant, [])
-nextPolarity (p : ps) = (p, ps)
-
 -- | Check if to lists of arguments are the same (and all variables).
 --   Precondition: the lists have the same length.
 sameVars :: Args -> Args -> Bool
@@ -73,37 +78,80 @@ equalAtom = compareAtom CmpEq
 equalType :: Type -> Type -> TCM ()
 equalType = compareType CmpEq
 
+{- Comparing in irrelevant context always succeeds.
+
+   However, we might want to dig for solutions of irrelevant metas.
+
+   To this end, we can just ignore errors during conversion checking.
+ -}
+
+-- convError ::  MonadTCM tcm => TypeError -> tcm a
+-- | Ignore errors in irrelevant context.
+convError :: TypeError -> TCM ()
+convError err = ifM ((==) Irrelevant <$> asks envRelevance) (return ()) $ typeError err
+
 -- | Type directed equality on values.
 --
 compareTerm :: Comparison -> Type -> Term -> Term -> TCM ()
   -- If one term is a meta, try to instantiate right away. This avoids unnecessary unfolding.
-compareTerm cmp a u v = liftTCM $ do
-  (u, v) <- instantiate (u, v)
+  -- Andreas, 2012-02-14: This is UNSOUND for subtyping!
+compareTerm cmp a u v = do
+  -- Check syntactic equality first. This actually saves us quite a bit of work.
+  (u, v) <- instantiateFull (u, v)
+  if u == v then unifyPointers cmp u v $ verboseS "profile.sharing" 20 $ tick "equal terms" else do
+  verboseS "profile.sharing" 20 $ tick "unequal terms"
+  let checkPointerEquality def | not $ null $ List.intersect (pointerChain u) (pointerChain v) = do
+        verboseS "profile.sharing" 10 $ tick "pointer equality"
+        return ()
+      checkPointerEquality def = def
+  checkPointerEquality $ do
   reportSDoc "tc.conv.term" 10 $ sep [ text "compareTerm"
                                      , nest 2 $ prettyTCM u <+> prettyTCM cmp <+> prettyTCM v
                                      , nest 2 $ text ":" <+> prettyTCM a ]
   let fallback = compareTerm' cmp a u v
-  case (u, v) of
-    (u@(MetaV x us), v@(MetaV y vs))
-      | x /= y    -> solve1 `orelse` solve2 `orelse` compareTerm' cmp a u v
+      unlessSubtyping cont =
+          if cmp == CmpEq then cont else do
+            -- do not short circuit size comparison!
+            isSize <- isJust <$> do isSizeTypeTest <*> reduce a
+            if isSize then fallback else cont
+  case (ignoreSharing u, ignoreSharing v) of
+    (MetaV x us, MetaV y vs)
+      | x /= y    -> unlessSubtyping $ solve1 `orelse` solve2 `orelse` compareTerm' cmp a u v
       | otherwise -> fallback
       where
         (solve1, solve2) | x > y     = (assign x us v, assign y vs u)
                          | otherwise = (assign y vs u, assign x us v)
-    (u@(MetaV x us), v) -> assign x us v `orelse` fallback
-    (u, v@(MetaV y vs)) -> assign y vs u `orelse` fallback
-    _                   -> fallback
+    (MetaV x us, _) -> unlessSubtyping $ assign x us v `orelse` fallback
+    (_, MetaV y vs) -> unlessSubtyping $ assign y vs u `orelse` fallback
+    _               -> fallback
   where
     assign x us v = do
-      reportSDoc "tc.conv.term" 20 $ sep [ text "attempting shortcut"
-                                         , nest 2 $ prettyTCM (MetaV x us) <+> text ":=" <+> prettyTCM v ]
+      reportSDoc "tc.conv.term.shortcut" 20 $ sep
+        [ text "attempting shortcut"
+        , nest 2 $ prettyTCM (MetaV x us) <+> text ":=" <+> prettyTCM v
+        ]
       ifM (isInstantiatedMeta x) patternViolation (assignV x us v)
+      instantiate u
+      -- () <- seq u' $ return ()
+      reportSLn "tc.conv.term.shortcut" 50 $
+        "shortcut successful\n  result: " ++ show u
     -- Should be ok with catchError_ but catchError is much safer since we don't
     -- rethrow errors.
-    m `orelse` h = m `catchError` \err -> case errError err of
+    m `orelse` h = m `catchError` \err -> case err of
                     PatternErr s -> put s >> h
                     _            -> h
 
+unifyPointers _ _ _ action = action
+-- unifyPointers cmp _ _ action | cmp /= CmpEq = action
+-- unifyPointers _ u v action = do
+--   old <- gets stDirty
+--   modify $ \s -> s { stDirty = False }
+--   action
+--   (u, v) <- instantiate (u, v)
+--   dirty <- gets stDirty
+--   modify $ \s -> s { stDirty = old }
+--   when (not dirty) $ forceEqualTerms u v
+
 compareTerm' :: Comparison -> Type -> Term -> Term -> TCM ()
 compareTerm' cmp a m n =
   verboseBracket "tc.conv.term" 20 "compareTerm" $ do
@@ -112,18 +160,19 @@ compareTerm' cmp a m n =
     reportSDoc "tc.conv.term" 30 $ fsep
       [ text "compareTerm", prettyTCM m, prettyTCM cmp, prettyTCM n, text ":", prettyTCM a' ]
     proofIrr <- proofIrrelevance
-    isSize   <- isSizeType a'
+    isSize   <- isJust <$> isSizeType a'
     s        <- reduce $ getSort a'
     mlvl     <- mlevel
     case s of
       Prop | proofIrr -> return ()
       _    | isSize   -> compareSizes cmp m n
-      _               -> case unEl a' of
+      _               -> case ignoreSharing $ unEl a' of
         a | Just a == mlvl -> do
           a <- levelView m
           b <- levelView n
           equalLevel a b
-        Pi a _    -> equalFun (a,a') m n
+-- OLD:        Pi dom _  -> equalFun (dom, a') m n
+        a@Pi{}    -> equalFun a m n
         Lam _ _   -> __IMPOSSIBLE__
         Def r ps  -> do
           isrec <- isEtaRecord r
@@ -157,23 +206,40 @@ compareTerm' cmp a m n =
 -- Andreas, 2010-10-11: allowing neutrals to be blocked things does not seem
 -- to change Agda's behavior
 --    isNeutral Blocked{}          = False
-    isNeutral (NotBlocked Con{}) = False
-    isNeutral _                  = True
-    isMeta (NotBlocked MetaV{})  = True
-    isMeta _                     = False
+    isNeutral = isNeutral' . fmap ignoreSharing
+    isMeta    = isMeta'    . fmap ignoreSharing
+    isNeutral' (NotBlocked Con{}) = False
+    isNeutral' _                  = True
+    isMeta' (NotBlocked MetaV{})  = True
+    isMeta' _                     = False
 
-    equalFun (a,t) m n = do
+    -- equality at function type (accounts for eta)
+    equalFun :: Term -> Term -> Term -> TCM ()
+    equalFun (Shared p) m n = equalFun (derefPtr p) m n
+    equalFun (Pi dom@(Dom h r _) b) m n = do
+        name <- freshName_ $ properName $ absName b
+        addCtx name dom $ compareTerm cmp (absBody b) m' n'
+      where
+        (m',n') = raise 1 (m,n) `apply` [Arg h r $ var 0]
+        properName "_" = "x"
+        properName  x  =  x
+    equalFun _ _ _ = __IMPOSSIBLE__
+{- OLD CODE
+    equalFun :: (Dom Type, Type) -> Term -> Term -> TCM ()
+    equalFun (a@(Dom h r _), t) m n = do
         name <- freshName_ (suggest $ unEl t)
         addCtx name a $ compareTerm cmp t' m' n'
       where
-        p	= fmap (const $ Var 0 []) a
+        p	= Arg h r $ var 0
         (m',n') = raise 1 (m,n) `apply` [p]
         t'	= raise 1 t `piApply` [p]
         suggest (Pi _ b) = absName b
         suggest _	 = __IMPOSSIBLE__
+-}
 
--- | @compareTel t1 t2 cmp tel1 tel1@ checks whether pointwise @tel1 `cmp` tel2@
---   and complains that @t2 `cmp` t1@ failed if not.
+-- | @compareTel t1 t2 cmp tel1 tel1@ checks whether pointwise
+--   @tel1 \`cmp\` tel2@ and complains that @t2 \`cmp\` t1@ failed if
+--   not.
 compareTel :: Type -> Type ->
   Comparison -> Telescope -> Telescope -> TCM ()
 compareTel t1 t2 cmp tel1 tel2 =
@@ -182,17 +248,20 @@ compareTel t1 t2 cmp tel1 tel2 =
     (EmptyTel, EmptyTel) -> return ()
     (EmptyTel, _)        -> bad
     (_, EmptyTel)        -> bad
-    (ExtendTel arg1@(Arg h1 r1 a1) tel1, ExtendTel arg2@(Arg h2 r2 a2) tel2)
+    (ExtendTel arg1@(Dom h1 r1 a1) tel1, ExtendTel arg2@(Dom h2 r2 a2) tel2)
       | h1 /= h2 -> bad
         -- Andreas, 2011-09-11 do not test r1 == r2 because they could differ
         -- e.g. one could be Forced and the other Relevant (see fail/UncurryMeta)
       | otherwise -> do
-          let (tel1', tel2') = raise 1 (tel1, tel2)
-              arg            = Var 0 []
           name <- freshName_ (suggest (absName tel1) (absName tel2))
           let checkArg = escapeContext 1 $ compareType cmp a1 a2
-          let c = TelCmp t1 t2 cmp (absApp tel1' arg) (absApp tel2' arg)
-          let r = max r1 r2  -- take "most irrelevant"
+{- OLD
+          let (tel1', tel2') = raise 1 (tel1, tel2)
+              arg            = var 0
+              c = TelCmp t1 t2 cmp (absApp tel1' arg) (absApp tel2' arg)
+-}
+              c = TelCmp t1 t2 cmp (absBody tel1) (absBody tel2)
+              r = max r1 r2  -- take "most irrelevant"
               dependent = (r /= Irrelevant) && isBinderUsed tel2
           addCtx name arg1 $
             if dependent
@@ -213,13 +282,24 @@ compareAtom cmp t m n =
     verboseBracket "tc.conv.atom" 20 "compareAtom" $
     -- if a PatternErr is thrown, rebuild constraint!
     catchConstraint (ValueCmp cmp t m n) $ do
-      let unLevel (Level l) = reallyUnLevelView l
-          unLevel v = return v
-      -- constructorForm changes literal to constructors
+      reportSDoc "tc.conv.atom" 50 $
+	text "compareAtom" <+> fsep [ prettyTCM m <+> prettyTCM cmp
+                                    , prettyTCM n
+                                    , text ":" <+> prettyTCM t ]
       -- Andreas: what happens if I cut out the eta expansion here?
       -- Answer: Triggers issue 245, does not resolve 348
-      mb <- traverse unLevel =<< traverse constructorForm =<< etaExpandBlocked =<< reduceB m
-      nb <- traverse unLevel =<< traverse constructorForm =<< etaExpandBlocked =<< reduceB n
+      mb' <- etaExpandBlocked =<< reduceB m
+      nb' <- etaExpandBlocked =<< reduceB n
+
+      -- constructorForm changes literal to constructors
+      -- only needed if the other side is not a literal
+      (mb'', nb'') <- case (ignoreSharing $ ignoreBlocking mb', ignoreSharing $ ignoreBlocking nb') of
+	(Lit _, Lit _) -> return (mb', nb')
+        _ -> (,) <$> traverse constructorForm mb'
+                 <*> traverse constructorForm nb'
+
+      mb <- traverse unLevel mb''
+      nb <- traverse unLevel nb''
 
       let m = ignoreBlocking mb
           n = ignoreBlocking nb
@@ -233,11 +313,13 @@ compareAtom cmp t m n =
                 then return ()	-- Check syntactic equality for blocked terms
                 else postpone
 
+      unifyPointers cmp (ignoreBlocking mb') (ignoreBlocking nb') $ do    -- this needs to go after eta expansion to avoid creating infinite terms
+
       reportSDoc "tc.conv.atom" 30 $
 	text "compareAtom" <+> fsep [ prettyTCM mb <+> prettyTCM cmp
                                     , prettyTCM nb
                                     , text ":" <+> prettyTCM t ]
-      case (mb, nb) of
+      case (ignoreSharing <$> mb, ignoreSharing <$> nb) of
         -- equate two metas x and y.  if y is the younger meta,
         -- try first y := x and then x := y
         (NotBlocked (MetaV x xArgs), NotBlocked (MetaV y yArgs))
@@ -265,7 +347,7 @@ compareAtom cmp t m n =
                       where l = assignV x xArgs n
                             r = assignV y yArgs m
 
-                    try m h = m `catchError_` \err -> case errError err of
+                    try m h = m `catchError_` \err -> case err of
                       PatternErr s -> put s >> h
                       _            -> throwError err
 
@@ -280,7 +362,7 @@ compareAtom cmp t m n =
         (Blocked{}, Blocked{})	-> checkSyntacticEquality
         (Blocked{}, _)    -> useInjectivity cmp t m n
         (_,Blocked{})     -> useInjectivity cmp t m n
-        _ -> case (m, n) of
+        _ -> case (ignoreSharing m, ignoreSharing n) of
 	    (Pi{}, Pi{}) -> equalFun m n
 
 	    (Sort s1, Sort s2) -> compareSort CmpEq s1 s2
@@ -293,11 +375,24 @@ compareAtom cmp t m n =
             (Def{}, Def{}) -> do
               ev1 <- elimView m
               ev2 <- elimView n
+              reportSDoc "tc.conv.atom" 50 $
+                sep [ text $ "ev1 = " ++ show ev1
+                    , text $ "ev2 = " ++ show ev2 ]
               case (ev1, ev2) of
                 (VarElim x els1, VarElim y els2) | x == y -> cmpElim (typeOfBV x) (Var x []) els1 els2
-                (ConElim x els1, ConElim y els2) | x == y -> cmpElim (conType x t) (Con x []) els1 els2
+                (ConElim x els1, ConElim y els2) | x == y ->
+                  cmpElim (conType x t) (Con x []) els1 els2
+                  -- Andreas 2012-01-17 careful!  In the presence of
+                  -- projection eliminations, t is NOT the datatype x belongs to
+                  -- Ulf 2012-07-12: actually projection likeness is carefully
+                  -- set up so that there can't be any projections from
+                  -- constructor applications at this point, so t really is the
+                  -- datatype of x. See issue 676 for an example where it
+                  -- failed.
                 (DefElim x els1, DefElim y els2) | x == y ->
                   cmpElim (defType <$> getConstInfo x) (Def x []) els1 els2
+                (DefElim x els1, DefElim y els2) ->
+                  trySizeUniv cmp t m n x els1 y els2
                 (MetaElim{}, _) -> __IMPOSSIBLE__   -- projections from metas should have been eta expanded
                 (_, MetaElim{}) -> __IMPOSSIBLE__
                 _ -> typeError $ UnequalTerms cmp m n t
@@ -324,49 +419,43 @@ compareAtom cmp t m n =
                     compareArgs [] a' (Con x []) xArgs yArgs
             _ -> typeError $ UnequalTerms cmp m n t
     where
-        conType c (El _ (Def d args)) = do
-          npars <- do
-            def <- theDef <$> getConstInfo d
-            return $ case def of Datatype{dataPars = n} -> n
-                                 Record{recPars = n}    -> n
-                                 _                      -> __IMPOSSIBLE__
-          a <- defType <$> getConstInfo c
-          return $ piApply a (genericTake npars args)
-        conType _ _ = __IMPOSSIBLE__
+        conType c (El _ v) = case ignoreSharing v of
+          Def d args -> do
+            npars <- do
+              def <- theDef <$> getConstInfo d
+              return $ case def of Datatype{dataPars = n} -> n
+                                   Record{recPars = n}    -> n
+                                   _                      -> __IMPOSSIBLE__
+            a <- defType <$> getConstInfo c
+            return $ piApply a (genericTake npars args)
+          _ -> __IMPOSSIBLE__
 
-	equalFun t1@(Pi arg1@(Arg h1 r1 a1) _) t2@(Pi (Arg h2 r2 a2) _)
-	    | h1 /= h2	= typeError $ UnequalHiding ty1 ty2
+        equalFun t1 t2 = case (ignoreSharing t1, ignoreSharing t2) of
+	  (Pi dom1@(Dom h1 r1 a1) b1, Pi (Dom h2 r2 a2) b2)
+	    | h1 /= h2	-> typeError $ UnequalHiding t1 t2
             -- Andreas 2010-09-21 compare r1 and r2, but ignore forcing annotations!
-	    | ignoreForced r1 /= ignoreForced r2 = typeError $ UnequalRelevance ty1 ty2
-	    | otherwise = verboseBracket "tc.conv.fun" 15 "compare function types" $ do
+	    | not (compareRelevance cmp (ignoreForced r2) (ignoreForced r1)) -> typeError $ UnequalRelevance cmp t1 t2
+	    | otherwise -> verboseBracket "tc.conv.fun" 15 "compare function types" $ do
                 reportSDoc "tc.conv.fun" 20 $ nest 2 $ vcat
                   [ text "t1 =" <+> prettyTCM t1
                   , text "t2 =" <+> prettyTCM t2 ]
-                let (ty1',ty2') = raise 1 (ty1,ty2)
-                    arg	    = Arg h1 r1 (Var 0 [])
-                name <- freshName_ (suggest t1 t2)
-                let checkArg = escapeContext 1 $ compareType cmp a2 a1
-                    c        = TypeCmp cmp (piApply ty1' [arg]) (piApply ty2' [arg])
-
+                let checkDom = escapeContext 1 $ compareType cmp a2 a1
+                    conCoDom = TypeCmp cmp (absBody b1) (absBody b2)
                 -- We only need to require a1 == a2 if t2 is a dependent function type.
                 -- If it's non-dependent it doesn't matter what we add to the context.
-                let dependent = case t2 of
-                                    Pi _ b  -> isBinderUsed b
-                                    _       -> __IMPOSSIBLE__
-                addCtx name arg1 $
-                  if dependent
-                  then guardConstraint c checkArg
-                  else checkArg >> solveConstraint_ c
+                name <- freshName_ (suggest b1 b2)
+                addCtx name dom1 $
+                  if isBinderUsed b2 -- dependent function type?
+                  then guardConstraint conCoDom checkDom
+                  else checkDom >> solveConstraint_ conCoDom
 	    where
-		ty1 = El (getSort a1) t1    -- TODO: wrong (but it doesn't matter)
-		ty2 = El (getSort a2) t2
-		suggest t1 t2 = case concatMap name [t1,t2] of
-				    []	-> "_"
-				    x:_	-> x
-		    where
-			name (Pi _ b) = filter (/= "_") [absName b]
-			name _        = __IMPOSSIBLE__
-	equalFun _ _ = __IMPOSSIBLE__
+		suggest b1 b2 = head $
+                  [ x | x <- map absName [b1,b2], x /= "_"] ++ ["_"]
+	  _ -> __IMPOSSIBLE__
+
+compareRelevance :: Comparison -> Relevance -> Relevance -> Bool
+compareRelevance CmpEq  = (==)
+compareRelevance CmpLeq = (<=)
 
 -- | Type-directed equality on eliminator spines
 compareElims :: [Polarity] -> Type -> Term -> [Elim] -> [Elim] -> TCM ()
@@ -380,29 +469,31 @@ compareElims pols0 a v els01@(Apply arg1 : els1) els02@(Apply arg2 : els2) =
   reportSDoc "tc.conv.elim" 25 $ nest 2 $ vcat
     [ text "a    =" <+> prettyTCM a
     , text "v    =" <+> prettyTCM v
-    , text "els1 =" <+> prettyTCM els01
-    , text "els2 =" <+> prettyTCM els02
+    , text "arg1 =" <+> prettyTCM arg1
+    , text "arg2 =" <+> prettyTCM arg2
+    ]
+  reportSDoc "tc.conv.elim" 50 $ nest 2 $ vcat
+    [ text "v    =" <+> text (show v)
+    , text "arg1 =" <+> text (show arg1)
+    , text "arg2 =" <+> text (show arg2)
     ]
   let (pol, pols) = nextPolarity pols0
   ab <- reduceB a
   let a = ignoreBlocking ab
   catchConstraint (ElimCmp pols0 a v els01 els02) $ do
-  case unEl <$> ab of
+  case ignoreSharing . unEl <$> ab of
     Blocked{}                     -> patternViolation
     NotBlocked MetaV{}            -> patternViolation
-    NotBlocked (Pi (Arg _ r b) _) -> do
-      let cmp x y = case pol of
-                      Invariant     -> compareTerm CmpEq  b x y
-                      Covariant     -> compareTerm CmpLeq b x y
-                      Contravariant -> compareTerm CmpLeq b y x
+    NotBlocked (Pi (Dom _ r b) _) -> do
       mlvl <- mlevel
-      let checkArg = case r of
+      let checkArg = applyRelevanceToContext r $ case r of
             Forced     -> return ()
-            Irrelevant -> return () -- Andreas: ignore irr. func. args.
-            _          -> applyRelevanceToContext r $
-                            cmp (unArg arg1) (unArg arg2)
-          dependent = case unEl a of
-            Pi (Arg _ _ (El _ lvl')) c -> 0 `freeInIgnoringSorts` absBody c
+            r | irrelevantOrUnused r ->
+                          compareIrrelevant b (unArg arg1) (unArg arg2)
+            _          -> compareWithPol pol (flip compareTerm b)
+                            (unArg arg1) (unArg arg2)
+          dependent = case ignoreSharing $ unEl a of
+            Pi (Dom _ _ (El _ lvl')) c -> 0 `freeInIgnoringSorts` absBody c
                                           && Just lvl' /= mlvl
             _ -> False
 
@@ -417,26 +508,74 @@ compareElims pols a v els01@(Proj f : els1) els02@(Proj f' : els2)
   | f /= f'   = typeError . GenericError . show =<< prettyTCM f <+> text "/=" <+> prettyTCM f'
   | otherwise = do
     a <- reduce a
-    case unEl a of
+    case ignoreSharing $ unEl a of
       Def r us -> do
         let (pol, _) = nextPolarity pols
-        ft <- defType <$> getConstInfo f
+        ft <- defType <$> getConstInfo f  -- get type of projection function
         let arg = Arg NotHidden Relevant v  -- TODO: not necessarily relevant?
         let c = piApply ft (us ++ [arg])
         (cmp, els1, els2) <- return $ case pol of
               Invariant     -> (CmpEq, els1, els2)
               Covariant     -> (CmpLeq, els1, els2)
-              Contravariant -> (CmpLeq, els2, els2)
+              Contravariant -> (CmpLeq, els2, els1)
+              Nonvariant    -> __IMPOSSIBLE__ -- the polarity should be Invariant
         pols' <- getPolarity' cmp f
         compareElims pols' c (Def f [arg]) els1 els2
       _ -> __IMPOSSIBLE__
 
+-- | "Compare" two terms in irrelevant position.  This always succeeds.
+--   However, we can dig for solutions of irrelevant metas in the
+--   terms we compare.
+--   (Certainly not the systematic solution, that'd be proof search...)
+compareIrrelevant :: Type -> Term -> Term -> TCM ()
+{- 2012-04-02 DontCare no longer present
+compareIrrelevant a (DontCare v) w = compareIrrelevant a v w
+compareIrrelevant a v (DontCare w) = compareIrrelevant a v w
+-}
+compareIrrelevant a v w = do
+  reportSDoc "tc.conv.irr" 20 $ vcat
+    [ text "compareIrrelevant"
+    , nest 2 $ text "v =" <+> prettyTCM v
+    , nest 2 $ text "w =" <+> prettyTCM w
+    ]
+  reportSDoc "tc.conv.irr" 50 $ vcat
+    [ nest 2 $ text $ "v = " ++ show v
+    , nest 2 $ text $ "w = " ++ show w
+    ]
+  try v w $ try w v $ return ()
+  where
+    try (Shared p) w fallback = try (derefPtr p) w fallback
+    try (MetaV x vs) w fallback = do
+      mv <- lookupMeta x
+      let rel  = getMetaRelevance mv
+          inst = case mvInstantiation mv of
+                   InstV{} -> True
+                   InstS{} -> True
+                   _       -> False
+      reportSDoc "tc.conv.irr" 20 $ vcat
+        [ nest 2 $ text $ "rel  = " ++ show rel
+        , nest 2 $ text $ "inst = " ++ show inst
+        ]
+      if not (irrelevantOrUnused rel) || inst then fallback else assignV x vs w
+        -- the value of irrelevant or unused meta does not matter
+    try v w fallback = fallback
+
+compareWithPol :: Polarity -> (Comparison -> a -> a -> TCM ()) -> a -> a -> TCM ()
+compareWithPol Invariant     cmp x y = cmp CmpEq x y
+compareWithPol Covariant     cmp x y = cmp CmpLeq x y
+compareWithPol Contravariant cmp x y = cmp CmpLeq y x
+compareWithPol Nonvariant    cmp x y = return ()
+
 -- | Type-directed equality on argument lists
 --
 compareArgs :: [Polarity] -> Type -> Term -> Args -> Args -> TCM ()
 compareArgs pol a v args1 args2 =
   compareElims pol a v (map Apply args1) (map Apply args2)
 
+---------------------------------------------------------------------------
+-- * Types
+---------------------------------------------------------------------------
+
 -- | Equality on Types
 compareType :: Comparison -> Type -> Type -> TCM ()
 compareType cmp ty1@(El s1 a1) ty2@(El s2 a2) =
@@ -449,35 +588,71 @@ compareType cmp ty1@(El s1 a1) ty2@(El s2 a2) =
           ]
 -- Andreas, 2011-4-27 should not compare sorts, but currently this is needed
 -- for solving sort and level metas
---        let cs1 = []
-	compareSort CmpEq s1 s2 `catchError` \err -> case errError err of
-                  TypeError _ _ -> do
-                    reportSDoc "tc.conv.type" 30 $ vcat
-                      [ text "sort comparison failed"
-                      , nest 2 $ vcat
-                        [ text "s1 =" <+> prettyTCM s1
-                        , text "s2 =" <+> prettyTCM s2
-                        ]
-                      ]
-                    -- This error will probably be more informative
-                    compareTerm cmp (sort s1) a1 a2
-                    -- Throw the original error if the above doesn't
-                    -- give an error (for instance, due to pending
-                    -- constraints).
-                    -- Or just ignore it... We run into this with irrelevant levels
-                    -- which may show up in sort constraints, causing them to fail.
-                    -- In any case it's not safe to ignore the error, for instance
-                    -- a1 might be Set and a2 a meta of type Set, in which case we
-                    -- really need the sort comparison to fail, instead of silently
-                    -- instantiating the meta.
-                    throwError err
-                  _             -> throwError err
+	compareSort CmpEq s1 s2 `catchError` \err -> case err of
+          TypeError _ e -> do
+            reportSDoc "tc.conv.type" 30 $ vcat
+              [ text "sort comparison failed"
+              , nest 2 $ vcat
+                [ text "s1 =" <+> prettyTCM s1
+                , text "s2 =" <+> prettyTCM s2
+                ]
+              ]
+            case clValue e of
+              -- Issue 659: Better error message
+              SetOmegaNotValidType -> typeError $ UnequalBecauseOfUniverseConflict cmp a1 a2
+              _ -> do
+                -- This error will probably be more informative
+                compareTerm cmp (sort s1) a1 a2
+                -- Throw the original error if the above doesn't
+                -- give an error (for instance, due to pending
+                -- constraints).
+                -- Or just ignore it... We run into this with irrelevant levels
+                -- which may show up in sort constraints, causing them to fail.
+                -- In any case it's not safe to ignore the error, for instance
+                -- a1 might be Set and a2 a meta of type Set, in which case we
+                -- really need the sort comparison to fail, instead of silently
+                -- instantiating the meta.
+                throwError err
+          _             -> throwError err
 	compareTerm cmp (sort s1) a1 a2
 	return ()
 
 leqType :: Type -> Type -> TCM ()
 leqType = compareType CmpLeq
 
+-- | @coerce v a b@ coerces @v : a@ to type @b@, returning a @v' : b@
+--   with maybe extra hidden applications or hidden abstractions.
+--
+--   In principle, this function can host coercive subtyping, but
+--   currently it only tries to fix problems with hidden function types.
+coerce :: Term -> Type -> Type -> TCM Term
+coerce v t1 t2 = blockTerm t2 $ do
+  verboseS "tc.conv.coerce" 10 $ do
+    (a1,a2) <- reify (t1,t2)
+    let dbglvl = if isSet a1 && isSet a2 then 50 else 10
+    reportSDoc "tc.conv.coerce" dbglvl $
+      text "coerce" <+> vcat
+        [ text "term      v  =" <+> prettyTCM v
+        , text "from type t1 =" <+> prettyTCM a1
+        , text "to type   t2 =" <+> prettyTCM a2
+        ]
+  -- v <$ do workOnTypes $ leqType t1 t2
+  -- take off hidden/instance domains from t1 and t2
+  TelV tel1 b1 <- telViewUpTo' (-1) ((NotHidden /=) . domHiding) t1
+  TelV tel2 b2 <- telViewUpTo' (-1) ((NotHidden /=) . domHiding) t2
+  let n = size tel1 - size tel2
+  -- the crude solution would be
+  --   v' = λ {tel2} → v {tel1}
+  -- however, that may introduce unneccessary many function types
+  -- If n  > 0 and b2 is not blocked, it is safe to
+  -- insert n many hidden args
+  if n <= 0 then fallback else do
+    ifBlockedType b2 (\ _ _ -> fallback) $ \ _ -> do
+      (args, t1') <- implicitArgs n (NotHidden /=) t1
+      v `apply` args <$ do workOnTypes $ leqType t1' t2
+  where
+    fallback = v <$ do workOnTypes $ leqType t1 t2
+
 ---------------------------------------------------------------------------
 -- * Sorts
 ---------------------------------------------------------------------------
@@ -497,7 +672,7 @@ leqSort s1 s2 =
               , nest 2 $ fsep [ prettyTCM s1 <+> text "=<"
                               , prettyTCM s2 ]
               ]
-	case (s1,s2) of
+	case (s1, s2) of
 
             (Type a, Type b) -> leqLevel a b
 
@@ -558,6 +733,9 @@ leqLevel a b = liftTCM $ do
                         [n] -> Just n
                         _   -> Nothing
 
+        -- Andreas, 2012-10-02 raise error on unsolvable constraint
+        ([ClosedLevel n], [ClosedLevel m]) -> if n <= m then ok else notok
+
         -- closed ≤ bs
         ([ClosedLevel n], bs)
           | n <= maximum (map constant bs) -> ok
@@ -589,7 +767,7 @@ leqLevel a b = liftTCM $ do
         postpone = patternViolation
 
         wrap m = catchError m $ \e ->
-          case errError e of
+          case e of
             TypeError{} -> notok
             _           -> throwError e
 
@@ -613,7 +791,7 @@ leqLevel a b = liftTCM $ do
 
 --     choice []     = patternViolation
 --     choice (m:ms) = noConstraints m `catchError` \_ -> choice ms
---       case errError e of
+--       case e of
 --         PatternErr{} -> choice ms
 --         _            -> throwError e
 
@@ -680,7 +858,10 @@ equalLevel a b = do
         _ | any isMeta (as ++ bs) -> postpone
 
         -- neutral/closed == neutral/closed
-        _ | all isNeutralOrClosed (as ++ bs) -> as =!= bs
+        _ | all isNeutralOrClosed (as ++ bs) ->
+          if length as == length bs
+            then zipWithM_ (\a b -> [a] =!= [b]) as bs
+            else notok
 
         -- more cases?
         _ -> postpone
@@ -704,7 +885,7 @@ equalLevel a b = do
 
         -- Make sure to give a sensible error message
         wrap m = m `catchError` \err ->
-          case errError err of
+          case err of
             TypeError{} -> notok
             _           -> throwError err
 
@@ -750,7 +931,7 @@ equalSort s1 s2 =
                                      , text (show s2) ]
                      ]
               ]
-	case (s1,s2) of
+	case (s1, s2) of
 
             (Type a  , Type b  ) -> equalLevel a b
 
@@ -778,5 +959,6 @@ equalSort s1 s2 =
         isInf notok ClosedLevel{} = notok
         isInf notok (Plus _ l) = case l of
           MetaLevel x vs          -> assignV x vs (Sort Inf)
+          NeutralLevel (Shared p) -> isInf notok (Plus 0 $ NeutralLevel $ derefPtr p)
           NeutralLevel (Sort Inf) -> return ()
           _                       -> notok
