@@ -3,12 +3,17 @@ module Agda.TypeChecking.Errors
     ( prettyError
     , PrettyTCM(..)
     , tcErrString
+    , Warnings(..)
+    , warningsToError
     ) where
 
 import Control.Applicative ( (<$>) )
 import Control.Monad.State
 import Control.Monad.Error
+
+import Data.List (nub)
 import qualified Data.Map as Map (empty)
+
 import System.FilePath
 
 import Agda.Syntax.Common
@@ -24,8 +29,10 @@ import qualified Agda.Syntax.Concrete.Pretty as P
 import Agda.Syntax.Translation.InternalToAbstract
 import Agda.Syntax.Translation.AbstractToConcrete
 import Agda.Syntax.Scope.Base (ScopeInfo(..))
-
-import Agda.TypeChecking.Monad
+import Agda.Syntax.Scope.Monad (isDatatypeModule)
+import Agda.TypeChecking.Monad.Base
+import Agda.TypeChecking.Monad.Context
+import Agda.TypeChecking.Monad.Options
 import Agda.TypeChecking.Pretty
 
 import Agda.Utils.FileName
@@ -49,6 +56,34 @@ prettyError err = liftTCM $ liftM show $
 	pwords "I give up! Approximations of errors:" )
 	$$ vcat (map (text . tcErrString) [err,err',err'',err'''])
 
+---------------------------------------------------------------------------
+-- * Warnings
+---------------------------------------------------------------------------
+
+-- | Warnings.
+--
+-- Invariant: The fields are never empty at the same time.
+
+data Warnings = Warnings
+  { terminationProblems   :: [TerminationError]
+    -- ^ Termination checking problems are not reported if
+    -- 'optTerminationCheck' is 'False'.
+  , unsolvedMetaVariables :: [Range]
+    -- ^ Meta-variable problems are reported as type errors unless
+    -- 'optAllowUnsolved' is 'True'.
+  , unsolvedConstraints   :: Constraints
+    -- ^ Same as 'unsolvedMetaVariables'.
+  }
+
+-- | Turns warnings into an error. Even if several errors are possible
+-- only one is raised.
+
+warningsToError :: Warnings -> TypeError
+warningsToError (Warnings [] [] [])    = __IMPOSSIBLE__
+warningsToError (Warnings _ w@(_:_) _) = UnsolvedMetas w
+warningsToError (Warnings _ _ w@(_:_)) = UnsolvedConstraints w
+warningsToError (Warnings w@(_:_) _ _) = TerminationCheckFailed w
+
 ---------------------------------------------------------------------------
 -- * Helpers
 ---------------------------------------------------------------------------
@@ -70,7 +105,7 @@ nameWithBinding q =
     r = nameBindingSite $ qnameName q
 
 tcErrString :: TCErr -> String
-tcErrString err = show (getRange err) ++ " " ++ case errError err of
+tcErrString err = show (getRange err) ++ " " ++ case err of
     TypeError _ cl  -> errorString $ clValue cl
     Exception r s   -> show r ++ " " ++ s
     IOException r e -> show r ++ " " ++ show e
@@ -83,6 +118,7 @@ errorString err = case err of
     AmbiguousName{}                          -> "AmbiguousName"
     AmbiguousParseForApplication{}           -> "AmbiguousParseForApplication"
     AmbiguousParseForLHS{}                   -> "AmbiguousParseForLHS"
+--    AmbiguousParseForPatternSynonym{}        -> "AmbiguousParseForPatternSynonym"
     AmbiguousTopLevelModuleName {}           -> "AmbiguousTopLevelModuleName"
     BothWithAndRHS                           -> "BothWithAndRHS"
     BuiltinInParameterisedModule{}           -> "BuiltinInParameterisedModule"
@@ -100,6 +136,7 @@ errorString err = case err of
     CoverageCantSplitType{}                  -> "CoverageCantSplitType"
     CyclicModuleDependency{}                 -> "CyclicModuleDependency"
     DataMustEndInSort{}                      -> "DataMustEndInSort"
+-- UNUSED:    DataTooManyParameters{}                  -> "DataTooManyParameters"
     DifferentArities                         -> "DifferentArities"
     DoesNotConstructAnElementOf{}            -> "DoesNotConstructAnElementOf"
     DuplicateBuiltinBinding{}                -> "DuplicateBuiltinBinding"
@@ -109,11 +146,12 @@ errorString err = case err of
     FieldOutsideRecord                       -> "FieldOutsideRecord"
     FileNotFound{}                           -> "FileNotFound"
     GenericError{}                           -> "GenericError"
+    GenericDocError{}                        -> "GenericDocError"
     IFSNoCandidateInScope{}                  -> "IFSNoCandidateInScope"
     IlltypedPattern{}                        -> "IlltypedPattern"
     IncompletePatternMatching{}              -> "IncompletePatternMatching"
-    IndexFreeInParameter{}                   -> "IndexFreeInParameter"
     IndexVariablesNotDistinct{}              -> "IndexVariablesNotDistinct"
+    IndicesFreeInParameters{}                -> "IndicesFreeInParameters"
     IndicesNotConstructorApplications{}      -> "IndicesNotConstructorApplications"
     InternalError{}                          -> "InternalError"
     InvalidPattern{}                         -> "InvalidPattern"
@@ -124,9 +162,11 @@ errorString err = case err of
     ModuleDefinedInOtherFile {}              -> "ModuleDefinedInOtherFile"
     ModuleDoesntExport{}                     -> "ModuleDoesntExport"
     ModuleNameDoesntMatchFileName {}         -> "ModuleNameDoesntMatchFileName"
+    NeedOptionCopatterns{}                   -> "NeedOptionCopatterns"
     NoBindingForBuiltin{}                    -> "NoBindingForBuiltin"
     NoParseForApplication{}                  -> "NoParseForApplication"
     NoParseForLHS{}                          -> "NoParseForLHS"
+--    NoParseForPatternSynonym{}               -> "NoParseForPatternSynonym"
     NoRHSRequiresAbsurdPattern{}             -> "NoRHSRequiresAbsurdPattern"
     NotInductive {}                          -> "NotInductive"
     AbsurdPatternRequiresNoRHS{}             -> "AbsurdPatternRequiresNoRHS"
@@ -135,6 +175,7 @@ errorString err = case err of
     NoSuchPrimitiveFunction{}                -> "NoSuchPrimitiveFunction"
     NotAModuleExpr{}                         -> "NotAModuleExpr"
     NotAProperTerm                           -> "NotAProperTerm"
+    SetOmegaNotValidType                     -> "SetOmegaNotValidType"
     NotAValidLetBinding{}                    -> "NotAValidLetBinding"
     NotAnExpression{}                        -> "NotAnExpression"
     NotImplemented{}                         -> "NotImplemented"
@@ -146,10 +187,12 @@ errorString err = case err of
     NothingAppliedToInstanceArg{}            -> "NothingAppliedToInstanceArg"
     OverlappingProjects {}                   -> "OverlappingProjects"
     PatternShadowsConstructor {}             -> "PatternShadowsConstructor"
+    PatternSynonymArityMismatch {}           -> "PatternSynonymArityMismatch"
     PropMustBeSingleton                      -> "PropMustBeSingleton"
     RepeatedVariablesInPattern{}             -> "RepeatedVariablesInPattern"
     SafeFlagPostulate{}                      -> "SafeFlagPostulate"
     SafeFlagPragma{}                         -> "SafeFlagPragma"
+    SafeFlagNoTerminationCheck{}             -> "SafeFlagNoTerminationCheck"
     SafeFlagPrimTrustMe{}                    -> "SafeFlagPrimTrustMe"
     ShadowedModule{}                         -> "ShadowedModule"
     ShouldBeASort{}                          -> "ShouldBeASort"
@@ -158,6 +201,7 @@ errorString err = case err of
     ShouldBeEmpty{}                          -> "ShouldBeEmpty"
     ShouldBePi{}                             -> "ShouldBePi"
     ShouldBeRecordType{}                     -> "ShouldBeRecordType"
+    ShouldBeRecordPattern{}                  -> "ShouldBeRecordPattern"
     ShouldEndInApplicationOfTheDatatype{}    -> "ShouldEndInApplicationOfTheDatatype"
     TerminationCheckFailed{}                 -> "TerminationCheckFailed"
     TooFewFields{}                           -> "TooFewFields"
@@ -166,6 +210,7 @@ errorString err = case err of
     SplitOnIrrelevant{}                      -> "SplitOnIrrelevant"
     DefinitionIsIrrelevant{}                 -> "DefinitionIsIrrelevant"
     VariableIsIrrelevant{}                   -> "VariableIsIrrelevant"
+    UnequalBecauseOfUniverseConflict{}       -> "UnequalBecauseOfUniverseConflict"
     UnequalRelevance{}                       -> "UnequalRelevance"
     UnequalHiding{}                          -> "UnequalHiding"
     UnequalLevel{}                           -> "UnequalLevel"
@@ -180,6 +225,7 @@ errorString err = case err of
     UnreachableClauses{}                     -> "UnreachableClauses"
     UnsolvedConstraints{}                    -> "UnsolvedConstraints"
     UnsolvedMetas{}                          -> "UnsolvedMetas"
+    UnusedVariableInPatternSynonym           -> "UnusedVariableInPatternSynonym"
     WithClausePatternMismatch{}              -> "WithClausePatternMismatch"
     WrongHidingInApplication{}               -> "WrongHidingInApplication"
     WrongHidingInLHS{}                       -> "WrongHidingInLHS"
@@ -188,13 +234,10 @@ errorString err = case err of
     WrongNumberOfConstructorArguments{}      -> "WrongNumberOfConstructorArguments"
 
 instance PrettyTCM TCErr where
-    prettyTCM err = case errError err of
-	TypeError s e -> do
-	    s0 <- get
+    prettyTCM err = case err of
+	TypeError s e -> localState $ do
 	    put s
-	    d <- sayWhen (envRange $ clEnv e) (envCall $ clEnv e) $ prettyTCM e
-	    put s0
-	    return d
+	    sayWhen (envRange $ clEnv e) (envCall $ clEnv e) $ prettyTCM e
 	Exception r s   -> sayWhere r $ fwords s
 	IOException r e -> sayWhere r $ fwords $ show e
 	PatternErr _    -> sayWhere err $ panic "uncaught pattern violation"
@@ -208,6 +251,7 @@ instance PrettyTCM TypeError where
 	    NotSupported s -> fwords $ "Not supported: " ++ s
 	    CompilationError s -> sep [fwords "Compilation error:", text s]
 	    GenericError s   -> fwords s
+	    GenericDocError d   -> return d
 	    TerminationCheckFailed because ->
               fwords "Termination checking failed for the following functions:"
               $$ (nest 2 $
@@ -219,12 +263,15 @@ instance PrettyTCM TypeError where
                                case show (callInfoRange c) of
                                  "" -> call
                                  r  -> call $$ nest 2 (text "(at" <+> text r <> text ")"))
-                        (concatMap termErrCalls because))
+                        (nub $ concatMap termErrCalls because))
 	    PropMustBeSingleton -> fwords
 		"Datatypes in Prop must have at most one constructor when proof irrelevance is enabled"
 	    DataMustEndInSort t -> fsep $
 		pwords "The type of a datatype must end in a sort."
 		++ [prettyTCM t] ++ pwords "isn't a sort."
+{- UNUSED:
+	    DataTooManyParameters -> fsep $ pwords "Too many parameters given to data type."
+-}
 	    ShouldEndInApplicationOfTheDatatype t -> fsep $
 		pwords "The target of a constructor must be the datatype applied to its parameters,"
 		++ [prettyTCM t] ++ pwords "isn't"
@@ -236,6 +283,8 @@ instance PrettyTCM TypeError where
 		pwords "which is not the right datatype"
 	    ShouldBeRecordType t -> fsep $
 		pwords "Expected record type, found " ++ [prettyTCM t]
+	    ShouldBeRecordPattern p -> fsep $
+		pwords "Expected record pattern" -- ", found " ++ [prettyTCM p]
 	    DifferentArities ->
 		fwords "The number of arguments in the defining equations differ"
 	    WrongHidingInLHS t -> do
@@ -252,7 +301,7 @@ instance PrettyTCM TypeError where
               pwords "Failed to infer the value of dotted pattern"
             IlltypedPattern p a -> fsep $
               pwords "Type mismatch"
-            TooManyArgumentsInLHS n a -> fsep $
+            TooManyArgumentsInLHS a -> fsep $
               pwords "Left hand side gives too many arguments to a function of type" ++ [prettyTCM a]
             WrongNumberOfConstructorArguments c expect given -> fsep $
               pwords "The constructor" ++ [prettyTCM c] ++ pwords "expects" ++
@@ -263,21 +312,33 @@ instance PrettyTCM TypeError where
             ConstructorPatternInWrongDatatype c d -> fsep $
               [prettyTCM c] ++ pwords "is not a constructor of the datatype" ++ [prettyTCM d]
             IndicesNotConstructorApplications is ->
-              fsep (pwords "The following indices are not constructors" ++
-                    pwords "(or literals) applied to variables:")
+              fwords "The indices"
               $$ nest 2 (vcat $ map prettyTCM is)
-            IndexVariablesNotDistinct is ->
-              fsep (pwords "The variables in the following indices are not distinct:")
+              $$ fsep (pwords "are not constructors (or literals) applied to variables" ++
+                       pwords "(note that parameters count as constructor arguments)")
+            IndexVariablesNotDistinct vs is ->
+              fwords "The variables"
+              $$ nest 2 (vcat $ map (\v -> prettyTCM (I.Var v [])) vs)
+              $$ fwords "in the indices"
               $$ nest 2 (vcat $ map prettyTCM is)
-            IndexFreeInParameter i pars ->
-              fsep (pwords "The index" ++ [prettyTCM (I.Var i [])] ++
-                    pwords "is free in the following parameters:")
+              $$ fwords "are not distinct (note that parameters count as constructor arguments)"
+            IndicesFreeInParameters vs indices pars ->
+              fwords "The variables"
+              $$ nest 2 (vcat $ map (\v -> prettyTCM (I.Var v [])) vs)
+              $$ fwords "which are used (perhaps as constructor parameters) in the index expressions"
+              $$ nest 2 (vcat $ map prettyTCM indices)
+              $$ fwords "are free in the parameters"
               $$ nest 2 (vcat $ map prettyTCM pars)
-            ShadowedModule [] -> __IMPOSSIBLE__
-            ShadowedModule ms@(m : _) -> fsep $
-              pwords "Duplicate definition of module" ++ [prettyTCM m <> text "."] ++
-              pwords "Previous definition at" ++ [text $ show r]
+            ShadowedModule x [] -> __IMPOSSIBLE__
+            ShadowedModule x ms@(m : _) -> fsep $
+              pwords "Duplicate definition of module" ++ [prettyTCM x <> text "."] ++
+              pwords "Previous definition of" ++ [help m] ++ pwords "module" ++ [prettyTCM x] ++
+              pwords "at" ++ [text $ show r]
               where
+                help m = do
+                  b <- isDatatypeModule m
+                  if b then text "datatype" else empty
+
                 r = case [ r | r <- map (defSiteOfLast . mnameToList) ms
                              , r /= noRange ] of
                       []    -> noRange
@@ -304,6 +365,8 @@ instance PrettyTCM TypeError where
 		[prettyTCM t] ++ pwords "should be a function type, but it isn't"
 	    NotAProperTerm ->
 		fwords "Found a malformed term"
+	    SetOmegaNotValidType ->
+		fwords "Setω is not a valid type"
             SplitOnIrrelevant p t -> fsep $
                 pwords "Cannot pattern match" ++ [prettyA p] ++
                 pwords "against irrelevant type" ++ [prettyTCM t]
@@ -311,40 +374,30 @@ instance PrettyTCM TypeError where
                 text "Identifier" : prettyTCM x : pwords "is declared irrelevant, so it cannot be used here"
             VariableIsIrrelevant x -> fsep $
                 text "Variable" : prettyTCM x : pwords "is declared irrelevant, so it cannot be used here"
+	    UnequalBecauseOfUniverseConflict cmp s t -> fsep $
+		[prettyTCM s, notCmp cmp, prettyTCM t, text "because this would result in an invalid use of Setω" ]
  	    UnequalTerms cmp s t a -> fsep $
-		[prettyTCM s, f cmp, prettyTCM t] ++ pwords "of type" ++ [prettyTCM a]
-                where
-                  f CmpEq  = text "!="
-                  f CmpLeq = text "!=<"
+		[prettyTCM s, notCmp cmp, prettyTCM t] ++ pwords "of type" ++ [prettyTCM a]
 	    UnequalLevel cmp s t -> fsep $
-		[prettyTCM s, f cmp, prettyTCM t]
-                where
-                  f CmpEq  = text "!="
-                  f CmpLeq = text "!=<"
+		[prettyTCM s, notCmp cmp, prettyTCM t]
 	    UnequalTelescopes cmp a b -> fsep $
-		[prettyTCM a, f cmp, prettyTCM b]
-                where
-                  f CmpEq  = text "!="
-                  f CmpLeq = text "!=<"
+		[prettyTCM a, notCmp cmp, prettyTCM b]
 	    UnequalTypes cmp a b -> fsep $
-		[prettyTCM a, f cmp, prettyTCM b]
-                where
-                  f CmpEq  = text "!="
-                  f CmpLeq = text "!=<"
+		[prettyTCM a, notCmp cmp, prettyTCM b]
 	    HeterogeneousEquality u a v b -> fsep $
                 pwords "Refuse to solve heterogeneous constraint" ++
                 [prettyTCM u] ++ pwords ":" ++ [prettyTCM a] ++ pwords "=?=" ++
                 [prettyTCM v] ++ pwords ":" ++ [prettyTCM b]
-	    UnequalRelevance a b -> fsep $
-		[prettyTCM a] ++ pwords "!=" ++ [prettyTCM b] ++
+	    UnequalRelevance cmp a b -> fsep $
+		[prettyTCM a, notCmp cmp, prettyTCM b] ++
 -- Andreas 2010-09-21 to reveal Forced annotations, print also uglily
---		[text $ show a] ++ pwords "!=" ++ [text $ show b] ++
+--		[text $ show a, notCmp cmp, text $ show b] ++
 		pwords "because one is a relevant function type and the other is an irrelevant function type"
 	    UnequalHiding a b -> fsep $
-		[prettyTCM a] ++ pwords "!=" ++ [prettyTCM b] ++
+		[prettyTCM a, text "!=", prettyTCM b] ++
 		pwords "because one is an implicit function type and the other is an explicit function type"
 	    UnequalSorts s1 s2 -> fsep $
-		[prettyTCM s1] ++ pwords "!=" ++ [prettyTCM s2]
+		[prettyTCM s1, text "!=", prettyTCM s2]
 	    NotLeqSort s1 s2 -> fsep $
 		pwords "The type of the constructor does not fit in the sort of the datatype, since"
 		++ [prettyTCM s1] ++ pwords "is not less or equal than" ++ [prettyTCM s2]
@@ -460,8 +513,13 @@ instance PrettyTCM TypeError where
 	    AmbiguousModule x ys -> vcat
 	      [ fsep $ pwords "Ambiguous module name" ++ [pretty x <> text "."] ++
 		       pwords "It could refer to any one of"
-	      , nest 2 $ vcat $ map prettyTCM ys
+	      , nest 2 $ vcat $ map help ys
 	      ]
+              where
+                help :: ModuleName -> TCM Doc
+                help m = do
+                  b <- isDatatypeModule m
+                  sep [prettyTCM m, if b then text "(datatype module)" else empty]
 	    UninstantiatedModule x -> fsep (
 		    pwords "Cannot access the contents of the parameterised module" ++ [pretty x <> text "."] ++
 		    pwords "To do this the module first has to be instantiated. For instance:"
@@ -509,12 +567,28 @@ instance PrettyTCM TypeError where
 		    pwords "Don't know how to parse" ++ [pretty (C.RawApp noRange es) <> text "."] ++
 		    pwords "Could mean any one of:"
 		) $$ nest 2 (vcat $ map pretty es')
-	    NoParseForLHS p -> fsep $
+            UnusedVariableInPatternSynonym -> fsep $
+                pwords "Unused variable in pattern synonym."
+            PatternSynonymArityMismatch x -> fsep $
+                pwords "Arity mismatch when using pattern synonym" ++ [prettyTCM x]
+	    NoParseForLHS IsLHS p -> fsep $
 		pwords "Could not parse the left-hand side" ++ [pretty p]
-	    AmbiguousParseForLHS p ps -> fsep (
+	    NoParseForLHS IsPatSyn p -> fsep $
+		pwords "Could not parse the pattern synonym" ++ [pretty p]
+{- UNUSED
+	    NoParseForPatternSynonym p -> fsep $
+		pwords "Could not parse the pattern synonym" ++ [pretty p]
+-}
+	    AmbiguousParseForLHS lhsOrPatSyn p ps -> fsep (
 		    pwords "Don't know how to parse" ++ [pretty p <> text "."] ++
 		    pwords "Could mean any one of:"
 		) $$ nest 2 (vcat $ map pretty ps)
+{- UNUSED
+	    AmbiguousParseForPatternSynonym p ps -> fsep (
+		    pwords "Don't know how to parse" ++ [pretty p <> text "."] ++
+		    pwords "Could mean any one of:"
+		) $$ nest 2 (vcat $ map pretty ps)
+-}
 	    IncompletePatternMatching v args -> fsep $
 		pwords "Incomplete pattern matching for" ++ [prettyTCM v <> text "."] ++
 		pwords "No match for" ++ map prettyTCM args
@@ -585,21 +659,27 @@ instance PrettyTCM TypeError where
                            | otherwise      = "s"
                 in fsep $ [fwords ("Cannot set OPTION pragma" ++ plural)]
                           ++ map text xs ++ [fwords "with safe flag."]
+            SafeFlagNoTerminationCheck -> fsep (pwords "Cannot use NO_TERMINATION_CHECK pragma with safe flag.")
             SafeFlagPrimTrustMe -> fsep (pwords "Cannot use primTrustMe with safe flag")
+            NeedOptionCopatterns -> fsep (pwords "Option --copatterns needed to enable destructor patterns")
           where
             mpar n args
               | n > 0 && not (null args) = parens
               | otherwise                = id
 
+            showArg :: Arg I.Pattern -> TCM Doc
             showArg (Arg Hidden r x)    = braces $ showPat 0 x
             showArg (Arg Instance r x)  = dbraces $ showPat 0 x
             showArg (Arg NotHidden r x) = showPat 1 x
 
+            showPat :: Integer -> I.Pattern -> TCM Doc
             showPat _ (I.VarP _)        = text "_"
             showPat _ (I.DotP _)        = text "._"
             showPat n (I.ConP c _ args) = mpar n args $ prettyTCM c <+> fsep (map showArg args)
             showPat _ (I.LitP l)        = text (show l)
 
+notCmp :: Comparison -> TCM Doc
+notCmp cmp = text $ "!" ++ show cmp
 
 
 instance PrettyTCM Call where
@@ -653,7 +733,7 @@ instance PrettyTCM Call where
 	    fsep $ pwords "when inferring the type of" ++ [prettyTCM x]
 	InferDef _ x _ ->
 	    fsep $ pwords "when inferring the type of" ++ [prettyTCM x]
-        CheckIsEmpty t _ ->
+        CheckIsEmpty r t _ ->
             fsep $ pwords "when checking that" ++ [prettyTCM t] ++ pwords "has no constructors"
 	ScopeCheckExpr e _ ->
 	    fsep $ pwords "when scope checking" ++ [pretty e]
@@ -663,8 +743,7 @@ instance PrettyTCM Call where
 	ScopeCheckLHS x p _ ->
 	    fsep $ pwords "when scope checking the left-hand side" ++ [pretty p] ++
 		   pwords "in the definition of" ++ [pretty x]
-	TermFunDef _ f _ _ ->
-	    fsep $ pwords "when termination checking the definition of" ++ [prettyTCM f]
+        NoHighlighting _ -> empty
 	SetRange r _ ->
 	    fsep $ pwords "when doing something at" ++ [text $ show r]
         CheckSectionApplication _ m1 modapp _ -> fsep $
@@ -676,6 +755,4 @@ instance PrettyTCM Call where
 	where
 	    hPretty a = pretty =<< abstractToConcreteCtx (hiddenArgumentCtx (argHiding a)) a
 
-	    simpleDecl d = d'
-              where [d'] = D.notSoNiceDeclarations [d]
-
+	    simpleDecl = D.notSoNiceDeclaration
