@@ -9,8 +9,8 @@ module Agda.TypeChecking.Free
     , relevantVars
     , rigidVars
     , freeIn, isBinderUsed
-    , freeInIgnoringSorts
-    , relevantIn
+    , freeInIgnoringSorts, freeInIgnoringSortAnn
+    , relevantIn, relevantInIgnoringSortAnn
     , Occurrence(..)
     , occurrence
     ) where
@@ -40,7 +40,8 @@ data FreeVars = FV
   { stronglyRigidVars :: VarSet -- ^ variables at top and under constructors
   , weaklyRigidVars   :: VarSet -- ^ ord. rigid variables, e.g., in arguments of variables
   , flexibleVars      :: VarSet -- ^ variables occuring in arguments of metas. These are potentially free, depending how the meta variable is instantiated.
-  , irrelevantVars    :: VarSet -- ^ variables under a @DontCare@, i.e., in irrelevant positions
+  , irrelevantVars    :: VarSet -- ^ variables in irrelevant arguments and under a @DontCare@, i.e., in irrelevant positions
+  , unusedVars        :: VarSet -- ^ variables in 'UnusedArg'uments
   }
 
 rigidVars :: FreeVars -> VarSet
@@ -48,7 +49,7 @@ rigidVars fv = Set.union (stronglyRigidVars fv) (weaklyRigidVars fv)
 
 -- | @allVars fv@ includes irrelevant variables.
 allVars :: FreeVars -> VarSet
-allVars fv = Set.unions [rigidVars fv, flexibleVars fv, irrelevantVars fv]
+allVars fv = Set.unions [rigidVars fv, flexibleVars fv, irrelevantVars fv, unusedVars fv]
 
 -- | All but the irrelevant variables.
 relevantVars :: FreeVars -> VarSet
@@ -56,17 +57,23 @@ relevantVars fv = Set.unions [rigidVars fv, flexibleVars fv]
 
 data Occurrence
   = NoOccurrence
+  | Irrelevantly
   | StronglyRigid
   | WeaklyRigid
   | Flexible
+  | Unused
   deriving (Eq,Show)
 
+{- NO LONGER
 -- | @occurrence x fv@ ignores irrelevant variables in @fv@
+-}
 occurrence :: Nat -> FreeVars -> Occurrence
 occurrence x fv
   | x `Set.member` stronglyRigidVars fv = StronglyRigid
   | x `Set.member` weaklyRigidVars   fv = WeaklyRigid
   | x `Set.member` flexibleVars      fv = Flexible
+  | x `Set.member` irrelevantVars    fv = Irrelevantly
+  | x `Set.member` unusedVars        fv = Unused
   | otherwise                           = NoOccurrence
 
 -- | Mark variables as flexible.  Useful when traversing arguments of metas.
@@ -88,47 +95,56 @@ weakly fv = fv
 irrelevantly :: FreeVars -> FreeVars
 irrelevantly fv = empty { irrelevantVars = allVars fv }
 
+-- | Mark all free variables as unused, except for irrelevant vars.
+unused :: FreeVars -> FreeVars
+unused fv = empty
+  { irrelevantVars = irrelevantVars fv
+  , unusedVars     = Set.unions [ rigidVars fv, flexibleVars fv, unusedVars fv ]
+  }
+
 -- | Pointwise union.
 union :: FreeVars -> FreeVars -> FreeVars
-union (FV sv1 rv1 fv1 iv1) (FV sv2 rv2 fv2 iv2) =
-  FV (Set.union sv1 sv2) (Set.union rv1 rv2) (Set.union fv1 fv2) (Set.union iv1 iv2)
+union (FV sv1 rv1 fv1 iv1 uv1) (FV sv2 rv2 fv2 iv2 uv2) =
+  FV (Set.union sv1 sv2) (Set.union rv1 rv2) (Set.union fv1 fv2) (Set.union iv1 iv2) (Set.union uv1 uv2)
 
 unions :: [FreeVars] -> FreeVars
 unions = foldr union empty
 
 empty :: FreeVars
-empty = FV Set.empty Set.empty Set.empty Set.empty
+empty = FV Set.empty Set.empty Set.empty Set.empty Set.empty
 
 -- | @delete x fv@ deletes variable @x@ from variable set @fv@.
 delete :: Nat -> FreeVars -> FreeVars
-delete n (FV sv rv fv iv) = FV (Set.delete n sv) (Set.delete n rv) (Set.delete n fv) (Set.delete n iv)
+delete n (FV sv rv fv iv uv) = FV (Set.delete n sv) (Set.delete n rv) (Set.delete n fv) (Set.delete n iv) (Set.delete n uv)
 
 -- | @subtractFV n fv@ subtracts $n$ from each free variable in @fv@.
 subtractFV :: Nat -> FreeVars -> FreeVars
-subtractFV n (FV sv rv fv iv) = FV (Set.subtract n sv) (Set.subtract n rv) (Set.subtract n fv) (Set.subtract n iv)
+subtractFV n (FV sv rv fv iv uv) = FV (Set.subtract n sv) (Set.subtract n rv) (Set.subtract n fv) (Set.subtract n iv) (Set.subtract n uv)
 
 -- | A single (strongly) rigid variable.
 singleton :: Nat -> FreeVars
-singleton x = FV { stronglyRigidVars = Set.singleton x
-		 , weaklyRigidVars   = Set.empty -- WAS: Set.singleton x
-		 , flexibleVars      = Set.empty
-                 , irrelevantVars    = Set.empty
-		 }
+singleton x = empty { stronglyRigidVars = Set.singleton x }
 
 -- * Collecting free variables.
 
 class Free a where
-  freeVars' :: FreeConf -> a -> FreeVars
+  freeVars'   :: FreeConf -> a -> FreeVars
+
+-- | Where should we skip sorts in free variable analysis?
+data IgnoreSorts
+  = IgnoreNot            -- ^ Do not skip.
+  | IgnoreInAnnotations  -- ^ Skip when annotation to a type.
+  | IgnoreAll            -- ^ Skip unconditionally.
+  deriving (Eq, Show)
 
 data FreeConf = FreeConf
-  { fcIgnoreSorts :: Bool
-    -- ^ Ignore free variables in sorts.
+  { fcIgnoreSorts   :: IgnoreSorts -- ^ Ignore free variables in sorts.
   }
 
 -- | Doesn't go inside solved metas, but collects the variables from a
 -- metavariable application @X ts@ as @flexibleVars@.
 freeVars :: Free a => a -> FreeVars
-freeVars = freeVars' FreeConf{ fcIgnoreSorts = False }
+freeVars = freeVars' FreeConf{ fcIgnoreSorts = IgnoreNot }
 
 instance Free Term where
   freeVars' conf t = case t of
@@ -145,14 +161,17 @@ instance Free Term where
     Level l    -> freeVars' conf l
     MetaV _ ts -> flexible $ freeVars' conf ts
     DontCare mt -> irrelevantly $ freeVars' conf mt
+    Shared p    -> freeVars' conf (derefPtr p)
 
 instance Free Type where
-  freeVars' conf (El s t) = freeVars' conf (s, t)
+  freeVars' conf (El s t)
+    | fcIgnoreSorts conf == IgnoreNot = freeVars' conf (s, t)
+    | otherwise                       = freeVars' conf t
 
 instance Free Sort where
   freeVars' conf s
-    | fcIgnoreSorts conf = empty
-    | otherwise          = case s of
+    | fcIgnoreSorts conf == IgnoreAll = empty
+    | otherwise                       = case s of
       Type a     -> freeVars' conf a
       Prop       -> empty
       Inf        -> empty
@@ -182,7 +201,12 @@ instance (Free a, Free b) => Free (a,b) where
   freeVars' conf (x,y) = freeVars' conf x `union` freeVars' conf y
 
 instance Free a => Free (Arg a) where
-  freeVars' conf = freeVars' conf . unArg
+  freeVars' conf (Arg h Irrelevant a) = irrelevantly $ freeVars' conf a
+  freeVars' conf (Arg h UnusedArg  a) = unused $ freeVars' conf a
+  freeVars' conf (Arg h r          a) = freeVars' conf a
+
+instance Free a => Free (Dom a) where
+  freeVars' conf (Dom h r a) = freeVars' conf a
 
 instance Free a => Free (Abs a) where
   freeVars' conf (Abs   _ b) = subtractFV 1 $ delete 0 $ freeVars' conf b
@@ -202,10 +226,18 @@ freeIn v t = v `Set.member` allVars (freeVars t)
 
 freeInIgnoringSorts :: Free a => Nat -> a -> Bool
 freeInIgnoringSorts v t =
-  v `Set.member` allVars (freeVars' FreeConf{ fcIgnoreSorts = True } t)
+  v `Set.member` allVars (freeVars' FreeConf{ fcIgnoreSorts = IgnoreAll } t)
+
+freeInIgnoringSortAnn :: Free a => Nat -> a -> Bool
+freeInIgnoringSortAnn v t =
+  v `Set.member` allVars (freeVars' FreeConf{ fcIgnoreSorts = IgnoreInAnnotations } t)
+
+relevantInIgnoringSortAnn :: Free a => Nat -> a -> Bool
+relevantInIgnoringSortAnn v t =
+  v `Set.member` relevantVars (freeVars' FreeConf{ fcIgnoreSorts = IgnoreInAnnotations } t)
 
 relevantIn :: Free a => Nat -> a -> Bool
-relevantIn v t = v `Set.member` relevantVars (freeVars' FreeConf{ fcIgnoreSorts = True } t)
+relevantIn v t = v `Set.member` relevantVars (freeVars' FreeConf{ fcIgnoreSorts = IgnoreAll } t)
 
 -- | Is the variable bound by the abstraction actually used?
 isBinderUsed :: Free a => Abs a -> Bool
