@@ -9,7 +9,7 @@ import Control.Monad.Reader
 
 import qualified Data.Map as Map
 
-import Agda.Interaction.Options
+import Agda.Interaction.Options hiding (tests)
 
 import Agda.Syntax.Common
 
@@ -17,14 +17,25 @@ import Agda.TypeChecking.Monad
 
 #include "../undefined.h"
 import Agda.Utils.Impossible
+import Agda.Utils.QuickCheck
+import Agda.Utils.TestHelpers
 
 -- | data 'Relevance'
 --   see 'Agda.Syntax.Common'
 
+irrelevantOrUnused :: Relevance -> Bool
+irrelevantOrUnused Irrelevant = True
+irrelevantOrUnused UnusedArg  = True
+irrelevantOrUnused NonStrict  = False
+irrelevantOrUnused Relevant   = False
+irrelevantOrUnused Forced     = False
+
 -- | @unusableRelevance rel == True@ iff we cannot use a variable of @rel@.
 unusableRelevance :: Relevance -> Bool
 unusableRelevance rel = NonStrict `moreRelevant` rel
 
+-- | 'Relevance' composition.
+--   'Irrelevant' is dominant, 'Relevant' is neutral.
 composeRelevance :: Relevance -> Relevance -> Relevance
 composeRelevance r r' =
   case (r, r') of
@@ -34,27 +45,31 @@ composeRelevance r r' =
     (_, NonStrict)  -> NonStrict
     (Forced, _)     -> Forced
     (_, Forced)     -> Forced
+    (UnusedArg, _)  -> UnusedArg
+    (_, UnusedArg)  -> UnusedArg
     (Relevant, Relevant) -> Relevant
 
 -- | @inverseComposeRelevance r x@ returns the most irrelevant @y@
 --   such that forall @x@, @y@ we have
---   @x `moreRelevant` (r `composeRelevance` y)@
+--   @x \`moreRelevant\` (r \`composeRelevance\` y)@
 --   iff
---   @(r `inverseComposeRelevance` x) `moreRelevant` y@ (Galois connection).
+--   @(r \`inverseComposeRelevance\` x) \`moreRelevant\` y@ (Galois connection).
 inverseComposeRelevance :: Relevance -> Relevance -> Relevance
 inverseComposeRelevance r x =
   case (r, x) of
-    (_, Forced)          -> Forced     -- preserve @Forced@
     (Relevant, x)        -> x          -- going to relevant arg.: nothing changes
-    (Forced, Relevant)   -> Forced     -- going forced: relevants become forced
+    _ | r == x           -> Relevant   -- because Relevant is comp.-neutral
+    (UnusedArg, x)       -> x
+    (Forced, UnusedArg)  -> Relevant
     (Forced, x)          -> x
-    (Irrelevant, _)      -> Relevant   -- going irrelevant: every thing usable
+    (Irrelevant, x)      -> Relevant   -- going irrelevant: every thing usable
     (_, Irrelevant)      -> Irrelevant -- otherwise: irrelevant things remain unusable
     (NonStrict, _)       -> Relevant   -- but @NonStrict@s become usable
 
--- | For comparing @Relevance@ ignoring @Forced@.
+-- | For comparing @Relevance@ ignoring @Forced@ and @UnusedArg@.
 ignoreForced :: Relevance -> Relevance
 ignoreForced Forced     = Relevant
+ignoreForced UnusedArg  = Relevant
 ignoreForced Relevant   = Relevant
 ignoreForced NonStrict  = NonStrict
 ignoreForced Irrelevant = Irrelevant
@@ -69,40 +84,31 @@ nonStrictToIrr :: Relevance -> Relevance
 nonStrictToIrr NonStrict = Irrelevant
 nonStrictToIrr rel       = rel
 
--- * Operations on 'Arg'.
+-- * Operations on 'Dom'.
 
 -- | Prepare parts of a parameter telescope for abstraction in constructors
 --   and projections.
-hideAndRelParams :: Arg a -> Arg a
+hideAndRelParams :: Dom a -> Dom a
 hideAndRelParams a = a
-  { argRelevance = nonStrictToIrr (argRelevance a)
-  , argHiding    = Hidden
+  { domRelevance = nonStrictToIrr (domRelevance a)
+  , domHiding    = Hidden
   }
 
+{- UNUSED
 -- | @modifyArgRelevance f arg@ applies @f@ to the 'argRelevance' component of @arg@.
 modifyArgRelevance :: (Relevance -> Relevance) -> Arg a -> Arg a
 modifyArgRelevance f a = a { argRelevance = f (argRelevance a) }
+-}
 
 -- | Used to modify context when going into a @rel@ argument.
-inverseApplyRelevance :: Relevance -> Arg a -> Arg a
-inverseApplyRelevance rel = modifyArgRelevance (rel `inverseComposeRelevance`)
--- inverseApplyRelevance rel a = a { argRelevance = rel `inverseComposeRelevance` argRelevance a }
+inverseApplyRelevance :: Relevance -> Dom a -> Dom a
+inverseApplyRelevance rel = mapDomRelevance (rel `inverseComposeRelevance`)
 
 -- | Compose two relevance flags.
 --   This function is used to update the relevance information
 --   on pattern variables @a@ after a match against something @rel@.
-applyRelevance :: Relevance -> Arg a -> Arg a
-applyRelevance rel = modifyArgRelevance (rel `composeRelevance`)
--- applyRelevance rel a = a { argRelevance = rel `composeRelevance` argRelevance a }
-
-{- Andreas, 2011-04-26: the combination Irrelevant Forced does not arise
-applyRelevance Irrelevant a | argRelevance a == Relevant =
-  a { argRelevance = Irrelevant }
-applyRelevance Forced a | argRelevance a == Relevant =
-  a { argRelevance = Forced }
-applyRelevance rel a = a -- ^ do nothing if rel == Relevant or a is
-                         -- already Forced or Irrelevant
--}
+applyRelevance :: Relevance -> Dom a -> Dom a
+applyRelevance rel = mapDomRelevance (rel `composeRelevance`)
 
 -- * Operations on 'Context'.
 
@@ -122,7 +128,7 @@ doWorkOnTypes = verboseBracket "tc.irr" 20 "workOnTypes" . workOnTypes' True
 workOnTypes' :: Bool -> TCM a -> TCM a
 workOnTypes' allowed cont =
   if allowed then
-    liftTCM $ modifyContext (modifyContextEntries $ modifyArgRelevance $ irrToNonStrict) cont
+    liftTCM $ modifyContext (modifyContextEntries $ mapDomRelevance $ irrToNonStrict) cont
    else cont
 
 -- | (Conditionally) wake up irrelevant variables and make them relevant.
@@ -140,7 +146,8 @@ applyRelevanceToContext rel =
           (fmap $ \ (t, a) -> (t, inverseApplyRelevance rel a))
           (envLetBindings e)
                                                   -- enable local  irr. defs
-      , envRelevance = rel                        -- enable global irr. defs
+      , envRelevance = composeRelevance rel (envRelevance e)
+                                                  -- enable global irr. defs
       }
 
 -- | Wake up irrelevant variables and make them relevant.  For instance,
@@ -148,14 +155,18 @@ applyRelevanceToContext rel =
 --   may be used, so they are awoken before type checking the argument.
 wakeIrrelevantVars :: TCM a -> TCM a
 wakeIrrelevantVars = applyRelevanceToContext Irrelevant
-{-
-wakeIrrelevantVars = local $ \ e -> e
-   { envContext = map wakeVar (envContext e) -- enable local  irr. defs
-   , envIrrelevant = True                    -- enable global irr. defs
-   }
-  where wakeVar ce = ce { ctxEntry = makeRelevant (ctxEntry ce) }
 
-applyRelevanceToContext :: Relevance -> TCM a -> TCM a
-applyRelevanceToContext Irrelevant cont = wakeIrrelevantVars cont
-applyRelevanceToContext _          cont = cont
--}
+
+------------------------------------------------------------------------
+-- * Tests
+------------------------------------------------------------------------
+
+prop_galois :: Relevance -> Relevance -> Relevance -> Bool
+prop_galois r x y =
+  x `moreRelevant` (r `composeRelevance` y) ==
+  (r `inverseComposeRelevance` x) `moreRelevant` y
+
+tests :: IO Bool
+tests = runTests "Agda.TypeChecking.Irrelevance"
+  [ quickCheck' prop_galois
+  ]
