@@ -35,10 +35,10 @@ levelSucFunction = do
 
 builtinLevelKit :: TCM (Maybe LevelKit)
 builtinLevelKit = liftTCM $ do
-    level@(Def l []) <- primLevel
-    zero@(Def z []) <- primLevelZero
-    suc@(Def s []) <- primLevelSuc
-    max@(Def m []) <- primLevelMax
+    level@(Def l []) <- ignoreSharing <$> primLevel
+    zero@(Def z [])  <- ignoreSharing <$> primLevelZero
+    suc@(Def s [])   <- ignoreSharing <$> primLevelSuc
+    max@(Def m [])   <- ignoreSharing <$> primLevelMax
     let a @@ b = a `apply` [defaultArg b]
     return $ Just $ LevelKit
       { lvlType  = level
@@ -61,6 +61,11 @@ requireLevels = do
     Nothing -> sequence_ [primLevel, primLevelZero, primLevelSuc, primLevelMax] >> __IMPOSSIBLE__
     Just k  -> return k
 
+unLevel :: Term -> TCM Term
+unLevel (Level l)  = reallyUnLevelView l
+unLevel (Shared p) = unLevel (derefPtr p)
+unLevel v = return v
+
 reallyUnLevelView :: Level -> TCM Term
 reallyUnLevelView nv =
   case nv of
@@ -100,7 +105,7 @@ levelView a = do
   mmax <- maybePrimDef primLevelMax
   let view a = do
         a <- reduce a
-        case a of
+        case ignoreSharing a of
           Level l -> return l
           Con s [arg]
             | Just s == msuc -> inc <$> view (unArg arg)
@@ -115,10 +120,10 @@ levelView a = do
   where
     mkAtom a = do
       b <- reduceB a
-      return $ case b of
+      return $ case ignoreSharing <$> b of
         NotBlocked (MetaV m as) -> atom $ MetaLevel m as
-        NotBlocked a            -> atom $ NeutralLevel a
-        Blocked m a             -> atom $ BlockedLevel m a
+        NotBlocked _            -> atom $ NeutralLevel (ignoreBlocking b)
+        Blocked m _             -> atom $ BlockedLevel m (ignoreBlocking b)
 
     atom a = Max [Plus 0 a]
 
