@@ -1,14 +1,32 @@
 
 module Agda.TypeChecking.Monad.Builtin where
 
-import Data.Functor
+import Control.Monad.Error
 import Control.Monad.State
+
+import Data.Functor
 import qualified Data.Map as Map
 
 import Agda.Syntax.Position
+import Agda.Syntax.Literal
 import Agda.Syntax.Internal
 import Agda.TypeChecking.Monad.Base
 
+import Agda.Utils.Monad (when_)
+
+litType :: Literal -> TCM Type
+litType l = case l of
+    LitInt _ n	  -> do
+      primZero
+      when_ (n > 0) $ primSuc
+      el <$> primNat
+    LitFloat _ _  -> el <$> primFloat
+    LitChar _ _   -> el <$> primChar
+    LitString _ _ -> el <$> primString
+    LitQName _ _  -> el <$> primQName
+  where
+    el t = El (mkType 0) t
+
 getBuiltinThing :: String -> TCM (Maybe (Builtin PrimFun))
 getBuiltinThing b = liftM2 mplus (Map.lookup b <$> gets stLocalBuiltins)
                     (Map.lookup b <$> gets stImportedBuiltins)
@@ -61,17 +79,19 @@ getPrimitive x = do
 primInteger, primFloat, primChar, primString, primBool, primTrue, primFalse,
     primList, primNil, primCons, primIO, primNat, primSuc, primZero,
     primNatPlus, primNatMinus, primNatTimes, primNatDivSucAux, primNatModSucAux,
-    primNatEquality, primNatLess, primSize, primSizeSuc, primSizeInf,
+    primNatEquality, primNatLess,
+    primSize, primSizeLt, primSizeSuc, primSizeInf,
     primInf, primSharp, primFlat,
     primEquality, primRefl,
     primLevel, primLevelZero, primLevelSuc, primLevelMax,
+    primIrrAxiom,
     -- builtins for reflection:
     primQName, primArg, primArgArg, primAgdaTerm, primAgdaTermVar,
     primAgdaTermLam, primAgdaTermDef, primAgdaTermCon, primAgdaTermPi,
     primAgdaTermSort, primAgdaTermUnsupported,
     primAgdaType, primAgdaTypeEl,
     primHiding, primHidden, primInstance, primVisible,
-    primRelvance, primRelevant, primIrrelevant,
+    primRelevance, primRelevant, primIrrelevant,
     primAgdaSort, primAgdaSortSet, primAgdaSortLit, primAgdaSortUnsupported,
     primAgdaDefinition, primAgdaDefinitionFunDef, primAgdaDefinitionDataDef, primAgdaDefinitionRecordDef,
     primAgdaDefinitionPostulate, primAgdaDefinitionPrimitive, primAgdaDefinitionDataConstructor,
@@ -99,8 +119,10 @@ primNatModSucAux = getBuiltin builtinNatModSucAux
 primNatEquality  = getBuiltin builtinNatEquals
 primNatLess      = getBuiltin builtinNatLess
 primSize         = getBuiltin builtinSize
+primSizeLt       = getBuiltin builtinSizeLt
 primSizeSuc      = getBuiltin builtinSizeSuc
 primSizeInf      = getBuiltin builtinSizeInf
+primSizeMax      = getBuiltin builtinSizeMax
 primInf          = getBuiltin builtinInf
 primSharp        = getBuiltin builtinSharp
 primFlat         = getBuiltin builtinFlat
@@ -110,6 +132,7 @@ primLevel        = getBuiltin builtinLevel
 primLevelZero    = getBuiltin builtinLevelZero
 primLevelSuc     = getBuiltin builtinLevelSuc
 primLevelMax     = getBuiltin builtinLevelMax
+primIrrAxiom     = getBuiltin builtinIrrAxiom
 primQName        = getBuiltin builtinQName
 primArg          = getBuiltin builtinArg
 primArgArg       = getBuiltin builtinArgArg
@@ -120,7 +143,7 @@ primHiding       = getBuiltin builtinHiding
 primHidden       = getBuiltin builtinHidden
 primInstance     = getBuiltin builtinInstance
 primVisible      = getBuiltin builtinVisible
-primRelvance     = getBuiltin builtinRelevance
+primRelevance    = getBuiltin builtinRelevance
 primRelevant     = getBuiltin builtinRelevant
 primIrrelevant   = getBuiltin builtinIrrelevant
 primAgdaSortSet  = getBuiltin builtinAgdaSortSet
@@ -166,8 +189,10 @@ builtinNil          = "NIL"
 builtinCons         = "CONS"
 builtinIO           = "IO"
 builtinSize         = "SIZE"
+builtinSizeLt       = "SIZELT"
 builtinSizeSuc      = "SIZESUC"
 builtinSizeInf      = "SIZEINF"
+builtinSizeMax      = "SIZEMAX"
 builtinInf          = "INFINITY"
 builtinSharp        = "SHARP"
 builtinFlat         = "FLAT"
@@ -177,6 +202,7 @@ builtinLevelMax     = "LEVELMAX"
 builtinLevel        = "LEVEL"
 builtinLevelZero    = "LEVELZERO"
 builtinLevelSuc     = "LEVELSUC"
+builtinIrrAxiom     = "IRRAXIOM"
 builtinQName        = "QNAME"
 builtinAgdaSort     = "AGDASORT"
 builtinAgdaSortSet  = "AGDASORTSET"
@@ -211,3 +237,25 @@ builtinAgdaDefinitionDataConstructor = "AGDADEFINITIONDATACONSTRUCTOR"
 builtinAgdaDefinitionPostulate       = "AGDADEFINITIONPOSTULATE"
 builtinAgdaDefinitionPrimitive       = "AGDADEFINITIONPRIMITIVE"
 builtinAgdaDefinition                = "AGDADEFINITION"
+
+-- | The coinductive primitives.
+
+data CoinductionKit = CoinductionKit
+  { nameOfInf   :: QName
+  , nameOfSharp :: QName
+  , nameOfFlat  :: QName
+  }
+
+-- | Tries to build a 'CoinductionKit'.
+
+coinductionKit :: TCM (Maybe CoinductionKit)
+coinductionKit = (do
+  Def inf   _ <- ignoreSharing <$> primInf
+  Def sharp _ <- ignoreSharing <$> primSharp
+  Def flat  _ <- ignoreSharing <$> primFlat
+  return $ Just $ CoinductionKit
+    { nameOfInf   = inf
+    , nameOfSharp = sharp
+    , nameOfFlat  = flat
+    })
+    `catchError` \_ -> return Nothing
