@@ -52,8 +52,8 @@ solvingProblem pid m = verboseBracket "tc.constr.solve" 50 ("working on problem
 
 isProblemSolved :: ProblemId -> TCM Bool
 isProblemSolved pid =
-  (&&) <$> (notElem pid <$> asks envActiveProblems)
-       <*> (all ((/= pid) . constraintProblem) <$> getAllConstraints)
+  and2M (notElem pid <$> asks envActiveProblems)
+        (all ((/= pid) . constraintProblem) <$> getAllConstraints)
 
 getConstraintsForProblem :: ProblemId -> TCM Constraints
 getConstraintsForProblem pid = List.filter ((== pid) . constraintProblem) <$> getAllConstraints
@@ -73,6 +73,23 @@ wakeConstraints wake = do
       , stAwakeConstraints    = stAwakeConstraints s ++ wakeup
       }
 
+-- danger...
+dropConstraints :: (ProblemConstraint -> Bool) -> TCM ()
+dropConstraints crit = do
+  sleepers <- gets stSleepingConstraints
+  wakers <- gets stAwakeConstraints
+  let filt = List.filter (not . crit)
+  modify $ \s -> s { stSleepingConstraints = filt sleepers
+                   , stAwakeConstraints = filt wakers
+                   }
+
+putAllConstraintsToSleep :: TCM ()
+putAllConstraintsToSleep = do
+  awakeOnes <- gets stAwakeConstraints
+  sleepers <- gets stSleepingConstraints
+  modify $ \s -> s { stSleepingConstraints = sleepers ++ awakeOnes
+                   , stAwakeConstraints = [] }
+
 takeAwakeConstraint :: TCM (Maybe ProblemConstraint)
 takeAwakeConstraint = do
   cs <- getAwakeConstraints
@@ -103,7 +120,7 @@ buildConstraint c = flip buildProblemConstraint c =<< currentProblem
 addConstraint' :: Constraint -> TCM ()
 addConstraint' c = do
     pc <- build
-    modify $ \s -> s { stSleepingConstraints = pc : stSleepingConstraints s }
+    modify $ \s -> s { stDirty = True, stSleepingConstraints = pc : stSleepingConstraints s }
   where
     build | isBlocking c = buildConstraint c
           | otherwise    = buildProblemConstraint 0 c
