@@ -32,15 +32,15 @@ withAnonymousModule m n =
 
 -- | Set the current environment to the given
 withEnv :: TCEnv -> TCM a -> TCM a
-withEnv env m = local (const env) m
+withEnv env m = local (\env0 -> env { envAllowDestructiveUpdate = envAllowDestructiveUpdate env0 }) m
 
--- | Get the current environmnet
+-- | Get the current environment
 getEnv :: TCM TCEnv
 getEnv = ask
 
--- | Leave the top level to type check local things.
-leaveTopLevel :: TCM a -> TCM a
-leaveTopLevel = local $ \ env -> env { envTopLevel = False }
-
-onTopLevel :: TCM Bool
-onTopLevel = asks envTopLevel
+-- | Increases the module nesting level by one in the given
+-- computation.
+withIncreasedModuleNestingLevel :: TCM a -> TCM a
+withIncreasedModuleNestingLevel =
+  local (\e -> e { envModuleNestingLevel =
+                     envModuleNestingLevel e + 1 })
