@@ -7,6 +7,7 @@ import Control.Monad.Reader
 import qualified Data.Map as Map
 import qualified Data.Set as Set
 
+import Agda.Syntax.Common
 import Agda.Syntax.Internal
 import Agda.Syntax.Position
 import Agda.Syntax.Scope.Base
@@ -18,6 +19,8 @@ import Agda.TypeChecking.Monad.State
 import Agda.TypeChecking.Monad.Trace
 import Agda.TypeChecking.Monad.Closure
 import Agda.TypeChecking.Monad.Open
+import Agda.TypeChecking.Monad.Options (reportSLn)
+import Agda.TypeChecking.Monad.Sharing
 import Agda.TypeChecking.Substitute
 -- import Agda.TypeChecking.Pretty -- LEADS TO import cycle
 
@@ -28,6 +31,10 @@ import Agda.Utils.Permutation
 #include "../../undefined.h"
 import Agda.Utils.Impossible
 
+-- | Switch off assignment of metas.
+dontAssignMetas :: TCM a -> TCM a
+dontAssignMetas = local $ \ env -> env { envAssignMetas = False }
+
 -- | Get the meta store.
 getMetaStore :: TCM MetaStore
 getMetaStore = gets stMetaStore
@@ -50,6 +57,11 @@ updateMetaVar m f =
 getMetaPriority :: MetaId -> TCM MetaPriority
 getMetaPriority i = mvPriority <$> lookupMeta i
 
+{- UNUSED
+getMetaRelevance :: MetaId -> TCM Relevance
+getMetaRelevance x = miRelevance . mvInfo <$> lookupMeta x
+-}
+
 isSortMeta :: MetaId -> TCM Bool
 isSortMeta m = do
   mv <- lookupMeta m
@@ -57,6 +69,13 @@ isSortMeta m = do
     HasType{} -> False
     IsSort{}  -> True
 
+getMetaType :: MetaId -> TCM Type
+getMetaType m = do
+  mv <- lookupMeta m
+  return $ case mvJudgement mv of
+    HasType{ jMetaType = t } -> t
+    IsSort{}  -> __IMPOSSIBLE__
+
 isInstantiatedMeta :: MetaId -> TCM Bool
 isInstantiatedMeta m = do
   mv <- lookupMeta m
@@ -65,10 +84,38 @@ isInstantiatedMeta m = do
     InstS{} -> True
     _       -> False
 
+-- | Create 'MetaInfo' in the current environment.
 createMetaInfo :: TCM MetaInfo
-createMetaInfo =
-    do  r <- getCurrentRange
-	buildClosure r
+createMetaInfo = createMetaInfo' RunMetaOccursCheck
+
+createMetaInfo' :: RunMetaOccursCheck -> TCM MetaInfo
+createMetaInfo' b = do
+  r   <- getCurrentRange
+  cl  <- buildClosure r
+  return MetaInfo
+    { miClosRange       = cl
+    , miMetaOccursCheck = b
+    , miNameSuggestion  = ""
+    }
+
+setValueMetaName :: Term -> MetaNameSuggestion -> TCM ()
+setValueMetaName v s = do
+  case ignoreSharing v of
+    MetaV mi _ -> setMetaNameSuggestion mi s
+    u          -> do
+      reportSLn "tc.meta.name" 70 $
+        "cannot set meta name; newMeta returns " ++ show u
+      return ()
+
+getMetaNameSuggestion :: MetaId -> TCM MetaNameSuggestion
+getMetaNameSuggestion mi = miNameSuggestion . mvInfo <$> lookupMeta mi
+
+setMetaNameSuggestion :: MetaId -> MetaNameSuggestion -> TCM ()
+setMetaNameSuggestion mi s = do
+  reportSLn "tc.meta.name" 20 $
+    "setting name of meta " ++ show mi ++ " to " ++ s
+  updateMetaVar mi $ \ mvar ->
+    mvar { mvInfo = (mvInfo mvar) { miNameSuggestion = s }}
 
 updateMetaVarRange :: MetaId -> Range -> TCM ()
 updateMetaVarRange mi r = updateMetaVar mi (setRange r)
@@ -142,8 +189,12 @@ getInteractionScope ii =
        mv <- lookupMeta mi
        return $ getMetaScope mv
 
-withMetaInfo :: MetaInfo -> TCM a -> TCM a
-withMetaInfo mI m = enterClosure mI $ \r -> setCurrentRange r m
+withMetaInfo' :: MetaVariable -> TCM a -> TCM a
+withMetaInfo' mv = withMetaInfo (miClosRange $ mvInfo mv)
+
+withMetaInfo :: Closure Range -> TCM a -> TCM a
+withMetaInfo mI cont = enterClosure mI $ \ r ->
+  setCurrentRange r cont
 
 getInstantiatedMetas :: TCM [MetaId]
 getInstantiatedMetas = do
