@@ -10,7 +10,6 @@ import Data.Map (Map)
 import qualified Data.Map as Map
 import Data.List
 import Data.Function
-import qualified Agda.Utils.IO.Locale as LocIO
 
 import Agda.Syntax.Abstract.Name
 import Agda.Syntax.Common
@@ -30,12 +29,14 @@ import Agda.TypeChecking.Substitute
 -- import Agda.TypeChecking.Pretty -- leads to cyclicity
 import {-# SOURCE #-} Agda.TypeChecking.CompiledClause.Compile
 import {-# SOURCE #-} Agda.TypeChecking.Polarity
+import {-# SOURCE #-} Agda.TypeChecking.ProjectionLike
 
 import Agda.Utils.Monad
 import Agda.Utils.Map as Map
 import Agda.Utils.Size
 import Agda.Utils.Permutation
 import Agda.Utils.Pretty
+import qualified Agda.Utils.HashMap as HMap
 
 #include "../../undefined.h"
 import Agda.Utils.Impossible
@@ -66,158 +67,80 @@ withSignature sig m =
 	setSignature sig0
         return r
 
+-- * modifiers for parts of the signature
+
+updateDefinition :: QName -> (Definition -> Definition) -> Signature -> Signature
+updateDefinition q f sig = sig { sigDefinitions = HMap.adjust f q (sigDefinitions sig) }
+
+updateTheDef :: (Defn -> Defn) -> (Definition -> Definition)
+updateTheDef f def = def { theDef = f (theDef def) }
+
+updateDefType :: (Type -> Type) -> (Definition -> Definition)
+updateDefType f def = def { defType = f (defType def) }
+
+updateDefArgOccurrences :: ([Occurrence] -> [Occurrence]) -> (Definition -> Definition)
+updateDefArgOccurrences f def = def { defArgOccurrences = f (defArgOccurrences def) }
+
+updateDefPolarity :: ([Polarity] -> [Polarity]) -> (Definition -> Definition)
+updateDefPolarity f def = def { defPolarity = f (defPolarity def) }
+
+updateDefCompiledRep :: (CompiledRepresentation -> CompiledRepresentation) -> (Definition -> Definition)
+updateDefCompiledRep f def = def { defCompiledRep = f (defCompiledRep def) }
+
 -- | Add a constant to the signature. Lifts the definition to top level.
 addConstant :: QName -> Definition -> TCM ()
 addConstant q d = do
   reportSLn "tc.signature" 20 $ "adding constant " ++ show q ++ " to signature"
   tel <- getContextTelescope
   let tel' = killRange $ case theDef d of
-	      Constructor{} -> hideTel tel
+	      Constructor{} -> fmap (mapDomHiding (const Hidden)) tel
 	      _		    -> tel
   let d' = abstract tel' $ d { defName = q }
   reportSLn "tc.signature" 30 $ "lambda-lifted definition = " ++ show d'
   modifySignature $ \sig -> sig
-    { sigDefinitions = Map.insertWith (+++) q d' $ sigDefinitions sig }
-  i <- currentMutualBlock
+    { sigDefinitions = HMap.insertWith (+++) q d' $ sigDefinitions sig }
+  i <- currentOrFreshMutualBlock
   setMutualBlock i q
   where
     new +++ old = new { defDisplay = defDisplay new ++ defDisplay old }
 
-    hideTel  EmptyTel		      = EmptyTel
-    hideTel (ExtendTel (Arg _ r t) tel) = ExtendTel (Arg Hidden r t) $ hideTel <$> tel
-
--- | Turn a definition into a projection if it looks like a projection.
-makeProjection :: QName -> TCM ()
-makeProjection x = inContext [] $ do
-  reportSLn "tc.proj.like" 30 $ "Considering " ++ show x ++ " for projection likeness"
-  defn <- getConstInfo x
-  case theDef defn of
-    -- Constructor-headed functions can't be projection-like (at the moment). The reason
-    -- for this is that invoking constructor-headedness will circumvent the inference of
-    -- the dropped arguments.
-    -- Nor can abstract definitions be projection-like since they won't reduce
-    -- outside the abstract block.
-    def@Function{funProjection = Nothing, funClauses = cls, funInv = NotInjective,
-                 funAbstr = ConcreteDef} -> do
-      ps0 <- filterM validProj (candidateArgs [] (unEl $ defType defn))
-      reportSLn "tc.proj.like" 30 $ if null ps0 then "  no candidates found"
-                                                else "  candidates: " ++ show ps0
-      ps <- return $ filter (checkOccurs cls . snd) ps0
-      when (not (null ps0) && null ps) $ reportSLn "tc.proj.like" 50 $ "  occurs check failed\n    clauses = " ++ show cls
-      case reverse ps of
-        []         -> return ()
-        (d, n) : _ -> do
-          reportSLn "tc.proj.like" 10 $ show (defName defn) ++ " is projection like in argument " ++
-                                        show n ++ " for type " ++ show d
-          let cls' = map (rewriteClause n) cls
-          cc <- compileClauses True cls'
-          reportSLn "tc.proj.like" 20 $ "  rewrote clauses to\n    " ++ show cc
-          let mapInv f NotInjective  = NotInjective
-              mapInv f (Inverse inv) = Inverse (f inv)
-              newDef = def
-                       { funProjection     = Just (d, n + 1)
-                       , funClauses        = cls'
-                       , funCompiled       = cc
-                       , funInv            = mapInv (Map.map $ rewriteClause n) $ funInv def
-                       , funArgOccurrences = drop n $ funArgOccurrences def
-                       , funPolarity       = drop n $ funPolarity def
-                       }
-          addConstant x $ defn{ theDef     = newDef
-                              , defDisplay = [] }
-    _ -> return ()
+-- | Set termination info of a defined function symbol.
+setTerminates :: QName -> Bool -> TCM ()
+setTerminates q b = modifySignature $ updateDefinition q $ updateTheDef $ setT
   where
-    validProj (_, 0) = return False
-    validProj (d, _) = do
-      defn <- theDef <$> getConstInfo d
-      return $ case defn of
-        Datatype{} -> True
-        Record{}   -> True
-        Axiom{}    -> True
-        _          -> False
-
-    rewriteClause n cl@Clause{clausePerm = Perm m p} =
-      cl{ clausePerm = Perm (m - fromIntegral n) $ map (subtract $ fromIntegral n) $ drop n p
-        , clauseTel  = telFromList $ drop n $ telToList $ clauseTel cl
-        , clausePats = drop n $ clausePats cl
-        , clauseBody = dropB n $ clauseBody cl
-        }
-      where
-        dropB 0 b          = b
-        dropB _ NoBody     = NoBody
-        dropB n (Bind b)   = dropB (n - 1) (absBody b)
-        dropB n Body{}     = __IMPOSSIBLE__
-
-    checkOccurs cls n = all (nonOccur n) cls
-
-    nonOccur n Clause{clausePerm = Perm _ p, clausePats = ps, clauseBody = b} =
-      and [ take n p == [0..fromIntegral n - 1]
-          , onlyMatch n ps  -- projection-like functions are only allowed to match on the eliminatee
-                            -- otherwise we may end up projecting from constructor applications, in
-                            -- which case we can't reconstruct the dropped parameters
-          , checkBody n b ]
-
-    onlyMatch n ps = all (noMatch . unArg) $ ps0 ++ drop 1 ps1
-      where
-        (ps0, ps1) = splitAt n ps
-        noMatch ConP{} = False
-        noMatch LitP{} = False
-        noMatch VarP{} = True
-        noMatch DotP{} = True
-
-    checkBody 0 _          = True
-    checkBody _ NoBody     = True
-    checkBody n (Bind b)   = not (isBinderUsed b) && checkBody (n - 1) (unAbs b)
-    checkBody _ Body{}     = __IMPOSSIBLE__
-
-    candidateArgs vs (Pi (Arg r h (El _ (Def d us))) b)
-      | vs == map unArg us = (d, length vs) : candidateRec vs b
-    candidateArgs vs (Pi _ b) = candidateRec vs b
-    candidateArgs _ _ = []
-
-    candidateRec vs NoAbs{} = []
-    candidateRec vs b       = candidateArgs (Var (size vs) [] : vs) (unEl $ absBody b)
+    setT def@Function{} = def { funTerminates = Just b }
+    setT def            = def
 
 addHaskellCode :: QName -> HaskellType -> HaskellCode -> TCM ()
-addHaskellCode q hsTy hsDef =
+addHaskellCode q hsTy hsDef = modifySignature $ updateDefinition q $ updateDefCompiledRep $ addHs
   -- TODO: sanity checking
-  modifySignature $ \sig -> sig
-  { sigDefinitions = Map.adjust addHs q $ sigDefinitions sig }
   where
-    addHs def = def { defCompiledRep = (defCompiledRep def) { compiledHaskell = Just $ HsDefn hsTy hsDef } }
+    addHs crep = crep { compiledHaskell = Just $ HsDefn hsTy hsDef }
 
 addHaskellType :: QName -> HaskellType -> TCM ()
-addHaskellType q hsTy =
+addHaskellType q hsTy = modifySignature $ updateDefinition q $ updateDefCompiledRep $ addHs
   -- TODO: sanity checking
-  modifySignature $ \sig -> sig
-  { sigDefinitions = Map.adjust addHs q $ sigDefinitions sig }
   where
-    addHs def = def { defCompiledRep = (defCompiledRep def) { compiledHaskell = Just $ HsType hsTy } }
+    addHs crep = crep { compiledHaskell = Just $ HsType hsTy }
 
 addEpicCode :: QName -> EpicCode -> TCM ()
-addEpicCode q epDef =
+addEpicCode q epDef = modifySignature $ updateDefinition q $ updateDefCompiledRep $ addEp
   -- TODO: sanity checking
-  modifySignature $ \sig -> sig
-  { sigDefinitions = Map.adjust addEp q $ sigDefinitions sig }
   where
-    --addEp def@Defn{theDef = con@Constructor{}} =
-      --def{theDef = con{conHsCode = Just (hsTy, hsDef)}}
-    addEp def = def { defCompiledRep = (defCompiledRep def) { compiledEpic = Just epDef } }
+    addEp crep = crep { compiledEpic = Just epDef }
 
 addJSCode :: QName -> String -> TCM ()
 addJSCode q jsDef =
   case JS.parse jsDef of
     Left e ->
-      modifySignature $ \sig -> sig
-      { sigDefinitions = Map.adjust (addJS (Just e)) q $ sigDefinitions sig }
+      modifySignature $ updateDefinition q $ updateDefCompiledRep $ addJS (Just e)
     Right s ->
       typeError (CompilationError ("Failed to parse ECMAScript (..." ++ s ++ ") for " ++ show q))
   where
-    addJS e def = def{defCompiledRep = (defCompiledRep def){compiledJS = e}}
+    addJS e crep = crep { compiledJS = e }
 
 markStatic :: QName -> TCM ()
-markStatic q =
-  modifySignature $ \sig -> sig
-  { sigDefinitions = Map.adjust mark q $ sigDefinitions sig }
+markStatic q = modifySignature $ updateDefinition q $ mark
   where
     mark def@Defn{theDef = fun@Function{}} =
       def{theDef = fun{funStatic = True}}
@@ -226,7 +149,7 @@ markStatic q =
 unionSignatures :: [Signature] -> Signature
 unionSignatures ss = foldr unionSignature emptySignature ss
   where
-    unionSignature (Sig a b) (Sig c d) = Sig (Map.union a c) (Map.union b d)
+    unionSignature (Sig a b) (Sig c d) = Sig (Map.union a c) (HMap.union b d)
 
 -- | Add a section to the signature.
 addSection :: ModuleName -> Nat -> TCM ()
@@ -247,23 +170,26 @@ lookupSection m = do
 addDisplayForms :: QName -> TCM ()
 addDisplayForms x = do
   args <- getContextArgs
-  add args x x []
+  n    <- do
+    proj <- isProjection x
+    return $ case proj of
+      Just (_, n) -> n
+      Nothing     -> 0
+  add (drop (n - 1) args) x x []
   where
-    add args top x ps = do
+    add args top x vs0 = do
       def <- getConstInfo x
       let cs = defClauses def
-          n  = case theDef def of
-                 Function{ funProjection = Just (_, n) } -> n
-                 _ -> 0
       case cs of
-	[ Clause{ clauseBody = b } ]
-	  | Just (m, Def y vs) <- strip b -> do
-	      let ps' = raise 1 (map unArg vs) ++ ps
-                  df  = Display 0 ps' $ DTerm $ Def top (drop (n - 1) args)
+	[ Clause{ clausePats = pats, clauseBody = b } ]
+	  | all (isVar . unArg) pats
+          , Just (m, Def y vs) <- strip (b `apply` vs0) -> do
+	      let ps = raise 1 (map unArg vs)
+                  df = Display 0 ps $ DTerm $ Def top args
 	      reportSLn "tc.display.section" 20 $ "adding display form " ++ show y ++ " --> " ++ show top
                                                 ++ "\n  " ++ show df
 	      addDisplayForm y df
-	      add args top y ps'
+	      add args top y vs
 	_ -> do
 	      let reason = case cs of
 		    []    -> "no clauses"
@@ -273,15 +199,18 @@ addDisplayForms x = do
 		      Just (m, Def y vs)
 			| m < length args -> "too few args"
 			| m > length args -> "too many args"
-			| otherwise	      -> "args=" ++ show args ++ " vs=" ++ show vs
+			| otherwise	  -> "args=" ++ show args ++ " vs=" ++ show vs
 		      Just (m, v) -> "not a def body"
-	      reportSLn "tc.display.section" 30 $ "no display form from" ++ show x ++ " because " ++ reason
+	      reportSLn "tc.display.section" 30 $ "no display form from " ++ show x ++ " because " ++ reason
 	      return ()
     strip (Body v)   = return (0, v)
     strip  NoBody    = Nothing
     strip (Bind b)   = do
       (n, v) <- strip $ absBody b
-      return (n + 1, v)
+      return (n + 1, ignoreSharing v)
+
+    isVar VarP{} = True
+    isVar _      = False
 
 applySection ::
   ModuleName -> Telescope -> ModuleName -> Args ->
@@ -290,7 +219,7 @@ applySection new ptel old ts rd rm = do
   sig  <- getSignature
   isig <- getImportedSignature
   let ss = getOld partOfOldM sigSections    [sig, isig]
-      ds = getOld partOfOldD sigDefinitions [sig, isig]
+      ds = getOldH partOfOldD sigDefinitions [sig, isig]
   reportSLn "tc.mod.apply" 10 $ render $ vcat
     [ text "applySection"
     , text "new  =" <+> text (show new)
@@ -309,6 +238,8 @@ applySection new ptel old ts rd rm = do
   where
     getOld partOfOld fromSig sigs =
       Map.toList $ Map.filterKeys partOfOld $ Map.unions $ map fromSig sigs
+    getOldH partOfOld fromSig sigs =
+      HMap.toList $ HMap.filterWithKey (const . partOfOld) $ HMap.unions $ map fromSig sigs
 
     partOfOldM x = x `isSubModuleOf` old
     partOfOldD x = x `isInModule`    old
@@ -323,49 +254,65 @@ applySection new ptel old ts rd rm = do
 	  addConstant y =<< nd y
           makeProjection y
 	  -- Set display form for the old name if it's not a constructor.
+{- BREAKS fail/Issue478
+          -- Andreas, 2012-10-20 and if we are not an anonymous module
+	  -- unless (isAnonymousModuleName new || isCon || size ptel > 0) $ do
+-}
 	  unless (isCon || size ptel > 0) $ do
 	    addDisplayForms y
       where
-	t  = defType d `apply` ts
+	t   = defType d `apply` ts
+        pol = defPolarity d `apply` ts
+        occ = defArgOccurrences d `apply` ts
 	-- the name is set by the addConstant function
-	nd y = Defn (defRelevance d) y t [] (-1) noCompiledRep <$> def  -- TODO: mutual block?
+	nd y = Defn (defRelevance d) y t pol occ [] (-1) noCompiledRep <$> def  -- TODO: mutual block?
         oldDef = theDef d
 	isCon = case oldDef of
 	  Constructor{} -> True
 	  _		-> False
+{- OLD
         getOcc d = case d of
           Function { funArgOccurrences  = os } -> os
           Datatype { dataArgOccurrences = os } -> os
           Record   { recArgOccurrences  = os } -> os
           _ -> []
         oldOcc = getOcc oldDef
+-}
 	def  = case oldDef of
                 Constructor{ conPars = np, conData = d } -> return $
                   oldDef { conPars = np - size ts, conData = copyName d }
                 Datatype{ dataPars = np, dataCons = cs } -> return $
                   oldDef { dataPars = np - size ts, dataClause = Just cl, dataCons = map copyName cs
-                         , dataArgOccurrences = drop (length ts) oldOcc }
+                         {- , dataArgOccurrences = drop (length ts) oldOcc -} }
                 Record{ recPars = np, recConType = t, recTel = tel } -> return $
                   oldDef { recPars = np - size ts, recClause = Just cl
                          , recConType = apply t ts, recTel = apply tel ts
-                         , recArgOccurrences = drop (length ts) oldOcc
+                         {- , recArgOccurrences = drop (length ts) oldOcc -}
                          }
 		_ -> do
-                  cc <- compileClauses True [cl]
+                  cc <- compileClauses Nothing [cl] -- Andreas, 2012-10-07 non need for record pattern translation
                   let newDef = Function
                         { funClauses        = [cl]
                         , funCompiled       = cc
                         , funDelayed        = NotDelayed
                         , funInv            = NotInjective
+{-
                         , funPolarity       = []
                         , funArgOccurrences = drop (length ts') oldOcc
+-}
+                        , funMutual         = mutual
                         , funAbstr          = ConcreteDef
                         , funProjection     = proj
                         , funStatic         = False
+                        , funCopy           = True
+                        , funTerminates     = Just True
                         }
                   reportSLn "tc.mod.apply" 80 $ "new def for " ++ show x ++ "\n  " ++ show newDef
                   return newDef
                   where
+                    mutual = case oldDef of
+                      Function{funMutual = m} -> m
+                      _ -> []
                     proj = case oldDef of
                       Function{funProjection = Just (r, n)}
                         | size ts < n -> Just (r, n - size ts)
@@ -398,7 +345,7 @@ addDisplayForm x df = do
   modifyImportedSignature (add d)
   modifySignature (add d)
   where
-    add df sig = sig { sigDefinitions = Map.adjust addDf x defs }
+    add df sig = sig { sigDefinitions = HMap.adjust addDf x defs }
       where
 	addDf def = def { defDisplay = df : defDisplay def }
 	defs	  = sigDefinitions sig
@@ -414,6 +361,7 @@ canonicalName x = do
   where
     extract NoBody           = __IMPOSSIBLE__
     extract (Body (Def x _)) = x
+    extract (Body (Shared p)) = extract (Body $ derefPtr p)
     extract (Body _)         = __IMPOSSIBLE__
     extract (Bind b)         = extract (unAbs b)
 
@@ -424,7 +372,8 @@ whatInduction c = do
   def <- theDef <$> getConstInfo c
   case def of
     Datatype{ dataInduction = i } -> return i
-    Record{}                      -> return Inductive
+    Record{ recRecursive = False} -> return Inductive
+    Record{ recInduction = i    } -> return i
     Constructor{ conInd = i }     -> return i
     _                             -> __IMPOSSIBLE__
 
@@ -452,7 +401,7 @@ getConstInfo q = liftTCM $ join $ pureTCM $ \st env ->
   let defs  = sigDefinitions $ stSignature st
       idefs = sigDefinitions $ stImports st
       smash = (++) `on` maybe [] (:[])
-  in case smash (Map.lookup q defs) (Map.lookup q idefs) of
+  in case smash (HMap.lookup q defs) (HMap.lookup q idefs) of
       []  -> fail $ "Unbound name: " ++ show q ++ " " ++ showQNameId q
       [d] -> mkAbs env d
       ds  -> fail $ "Ambiguous name: " ++ show q
@@ -478,6 +427,11 @@ getConstInfo q = liftTCM $ join $ pureTCM $ \st env ->
         init' [] = {-'-} __IMPOSSIBLE__
         init' xs = init xs
 
+-- | Look up the polarity of a definition.
+getPolarity :: QName -> TCM [Polarity]
+getPolarity q = defPolarity <$> getConstInfo q
+
+{- OLD
 -- | Look up the polarity of a definition.
 getPolarity :: QName -> TCM [Polarity]
 getPolarity q = do
@@ -487,17 +441,25 @@ getPolarity q = do
     Datatype{ dataPolarity = p } -> return p
     Record{ recPolarity    = p } -> return p
     _                            -> return []
+-}
 
+-- | Look up polarity of a definition and compose with polarity
+--   represented by 'Comparison'.
 getPolarity' :: Comparison -> QName -> TCM [Polarity]
-getPolarity' CmpEq _  = return []
-getPolarity' CmpLeq q = getPolarity q
+getPolarity' CmpEq  q = map (composePol Invariant) <$> getPolarity q -- return []
+getPolarity' CmpLeq q = getPolarity q -- composition with Covariant is identity
 
+-- | Set the polarity of a definition.
+setPolarity :: QName -> [Polarity] -> TCM ()
+setPolarity q pol = modifySignature $ updateDefinition q $ updateDefPolarity $ const pol
+
+{- OLD
 -- | Set the polarity of a definition.
 setPolarity :: QName -> [Polarity] -> TCM ()
 setPolarity q pol = do
   modifySignature setP
   where
-    setP sig = sig { sigDefinitions = Map.adjust setPx q defs }
+    setP sig = sig { sigDefinitions = HMap.adjust setPx q defs }
       where
 	setPx def = def { theDef = setPd $ theDef def }
         setPd d   = case d of
@@ -506,7 +468,40 @@ setPolarity q pol = do
           Record{}   -> d { recPolarity  = pol }
           _          -> d
 	defs	  = sigDefinitions sig
+-}
 
+-- | Return a finite list of argument occurrences.
+getArgOccurrences :: QName -> TCM [Occurrence]
+getArgOccurrences d = defArgOccurrences <$> getConstInfo d
+
+{- OLD
+-- | Return a finite list of argument occurrences.
+getArgOccurrences :: QName -> TCM [Occurrence]
+getArgOccurrences d = do
+  def <- theDef <$> getConstInfo d
+  return $ getArgOccurrences_ def
+
+getArgOccurrences_ :: Defn -> [Occurrence]
+getArgOccurrences_ def = case def of
+    Function { funArgOccurrences  = os } -> os
+    Datatype { dataArgOccurrences = os } -> os
+    Record   { recArgOccurrences  = os } -> os
+    Constructor{}                        -> [] -- repeat StrictPos
+    _                                    -> [] -- repeat Mixed
+-}
+
+getArgOccurrence :: QName -> Nat -> TCM Occurrence
+getArgOccurrence d i = do
+  def <- getConstInfo d
+  return $ case theDef def of
+    Constructor{} -> StrictPos
+    _             -> (defArgOccurrences def ++ repeat Mixed) !! i
+
+setArgOccurrences :: QName -> [Occurrence] -> TCM ()
+setArgOccurrences d os =
+  modifySignature $ updateDefinition d $ updateDefArgOccurrences $ const os
+
+{- OLD
 getArgOccurrence :: QName -> Nat -> TCM Occurrence
 getArgOccurrence d i = do
   def <- theDef <$> getConstInfo d
@@ -514,16 +509,16 @@ getArgOccurrence d i = do
     Function { funArgOccurrences  = os } -> look i os
     Datatype { dataArgOccurrences = os } -> look i os
     Record   { recArgOccurrences  = os } -> look i os
-    Constructor{}                        -> Positive
-    _                                    -> Negative
+    Constructor{}                        -> StrictPos
+    _                                    -> Mixed
   where
-    look i os = (os ++ repeat Negative) !! fromIntegral i
+    look i os = (os ++ repeat Mixed) !! fromIntegral i
 
 setArgOccurrences :: QName -> [Occurrence] -> TCM ()
 setArgOccurrences d os =
   modifySignature setO
   where
-    setO sig = sig { sigDefinitions = Map.adjust setOx d defs }
+    setO sig = sig { sigDefinitions = HMap.adjust setOx d defs }
       where
 	setOx def = def { theDef = setOd $ theDef def }
         setOd d   = case d of
@@ -532,7 +527,30 @@ setArgOccurrences d os =
           Record{}   -> d { recArgOccurrences  = os }
           _          -> d
 	defs	  = sigDefinitions sig
+-}
 
+-- | Get the mutually recursive identifiers.
+getMutual :: QName -> TCM [QName]
+getMutual d = do
+  def <- theDef <$> getConstInfo d
+  return $ case def of
+    Function {  funMutual = m } -> m
+    Datatype { dataMutual = m } -> m
+    Record   {  recMutual = m } -> m
+    _ -> []
+
+-- | Set the mutually recursive identifiers.
+setMutual :: QName -> [QName] -> TCM ()
+setMutual d m = modifySignature $ updateDefinition d $ updateTheDef $ \ def ->
+  case def of
+    Function{} -> def { funMutual = m }
+    Datatype{} -> def {dataMutual = m }
+    Record{}   -> def { recMutual = m }
+    _          -> __IMPOSSIBLE__
+
+-- | Check whether two definitions are mutually recursive.
+mutuallyRecursive :: QName -> QName -> TCM Bool
+mutuallyRecursive d d' = (d `elem`) <$> getMutual d'
 
 -- | Look up the number of free variables of a section. This is equal to the
 --   number of parameters if we're currently inside the section and 0 otherwise.
@@ -569,8 +587,8 @@ instantiateDef d = do
   verboseS "tc.sig.inst" 30 $ do
     ctx <- getContext
     m   <- currentModule
-    liftIO $ LocIO.putStrLn $ "instDef in " ++ show m ++ ": " ++ show (defName d) ++ " " ++
-			unwords (map show . take (size vs) . reverse . map (fst . unArg) $ ctx)
+    reportSLn "" 0 $ "instDef in " ++ show m ++ ": " ++ show (defName d) ++ " " ++
+			unwords (map show . take (size vs) . reverse . map (fst . unDom) $ ctx)
   return $ d `apply` vs
 
 -- | Give the abstract view of a definition.
@@ -585,7 +603,10 @@ makeAbstract d = do def <- makeAbs $ theDef d
 
 -- | Enter abstract mode. Abstract definition in the current module are transparent.
 inAbstractMode :: TCM a -> TCM a
-inAbstractMode = local $ \e -> e { envAbstractMode = AbstractMode }
+inAbstractMode = local $ \e -> e { envAbstractMode = AbstractMode,
+                                   envAllowDestructiveUpdate = False }
+                                    -- Allowing destructive updates when seeing through
+                                    -- abstract may break the abstraction.
 
 -- | Not in abstract mode. All abstract definitions are opaque.
 inConcreteMode :: TCM a -> TCM a
@@ -593,7 +614,10 @@ inConcreteMode = local $ \e -> e { envAbstractMode = ConcreteMode }
 
 -- | Ignore abstract mode. All abstract definitions are transparent.
 ignoreAbstractMode :: TCM a -> TCM a
-ignoreAbstractMode = local $ \e -> e { envAbstractMode = IgnoreAbstractMode }
+ignoreAbstractMode = local $ \e -> e { envAbstractMode = IgnoreAbstractMode,
+                                       envAllowDestructiveUpdate = False }
+                                       -- Allowing destructive updates when ignoring
+                                       -- abstract may break the abstraction.
 
 -- | Check whether a name might have to be treated abstractly (either if we're
 --   'inAbstractMode' or it's not a local name). Returns true for things not
