@@ -58,7 +58,7 @@ matchPattern (Arg _  _  (DotP _)) arg@(Arg _ _ v) = return (Yes [v], arg)
 matchPattern (Arg h' r' (LitP l)) arg@(Arg h r v) = do
     w <- reduceB v
     let v = ignoreBlocking w
-    case w of
+    case ignoreSharing <$> w of
 	NotBlocked (Lit l')
 	    | l == l'          -> return (Yes [], Arg h r v)
 	    | otherwise        -> return (No, Arg h r v)
@@ -66,6 +66,7 @@ matchPattern (Arg h' r' (LitP l)) arg@(Arg h r v) = do
 	Blocked x _            -> return (DontKnow $ Just x, Arg h r v)
 	_                      -> return (DontKnow Nothing, Arg h r v)
 
+{- Andreas, 2012-04-02 NO LONGER UP-TO-DATE
 matchPattern (Arg h' r' (ConP c _ ps))     (Arg h Irrelevant v) = do
           -- Andreas, 2010-09-07 matching a record constructor against
           -- something irrelevant will just continue matching against
@@ -73,6 +74,7 @@ matchPattern (Arg h' r' (ConP c _ ps))     (Arg h Irrelevant v) = do
 		(m, vs) <- matchPatterns ps $
                   repeat $ Arg NotHidden Irrelevant $ DontCare __IMPOSSIBLE__
 		return (m, Arg h Irrelevant $ Con c vs)
+-}
 
 matchPattern (Arg h' r' (ConP c _ ps))     (Arg h r v) =
     do	w <- traverse constructorForm =<< reduceB v
@@ -82,7 +84,7 @@ matchPattern (Arg h' r' (ConP c _ ps))     (Arg h r v) =
         -- 2) whatInduction c sometimes crashes because c may point to
         --    an axiom at this stage (if we are checking the
         --    projection functions for a record type).
-        w <- case w of
+        w <- case ignoreSharing <$> w of
                NotBlocked (Def f args) ->
                  unfoldDefinition True reduceB (Def f []) f args
                    -- reduceB is used here because some constructors
@@ -90,7 +92,7 @@ matchPattern (Arg h' r' (ConP c _ ps))     (Arg h r v) =
                    -- unfolded (due to open public).
                _ -> return w
         let v = ignoreBlocking w
-	case w of
+	case ignoreSharing <$> w of
           -- Andreas, 2010-09-07 matching a record constructor against
           -- something irrelevant will just continue matching against
           -- irrelevant stuff
