@@ -10,6 +10,7 @@ import Agda.Syntax.Literal
 import Agda.Syntax.Translation.InternalToAbstract
 import Agda.Syntax.Translation.AbstractToConcrete
 import qualified Agda.Syntax.Abstract as A
+import qualified Agda.Syntax.Concrete as C
 import qualified Agda.Syntax.Abstract.Pretty as P
 import qualified Agda.Syntax.Concrete.Pretty as P
 
@@ -27,10 +28,12 @@ import Agda.Utils.Impossible
 
 type Doc = P.Doc
 
-empty, comma :: TCM Doc
+empty, comma, colon, equals :: TCM Doc
 
 empty	   = return P.empty
 comma	   = return P.comma
+colon      = return P.colon
+equals     = return P.equals
 pretty x   = return $ P.pretty x
 prettyA x  = P.prettyA x
 text :: String -> TCM Doc
@@ -42,8 +45,9 @@ sep ds	   = P.sep <$> sequence ds
 fsep ds    = P.fsep <$> sequence ds
 hsep ds    = P.hsep <$> sequence ds
 vcat ds    = P.vcat <$> sequence ds
-($$), (<>), (<+>) :: TCM Doc -> TCM Doc -> TCM Doc
+($$), ($+$), (<>), (<+>) :: TCM Doc -> TCM Doc -> TCM Doc
 d1 $$ d2   = (P.$$) <$> d1 <*> d2
+d1 $+$ d2  = (P.$+$) <$> d1 <*> d2
 d1 <> d2   = (P.<>) <$> d1 <*> d2
 d1 <+> d2  = (P.<+>) <$> d1 <*> d2
 nest n d   = P.nest n <$> d
@@ -72,6 +76,7 @@ instance PrettyTCM a => PrettyTCM (Closure a) where
 instance PrettyTCM a => PrettyTCM [a] where
   prettyTCM = prettyList . map prettyTCM
 
+instance PrettyTCM Nat where prettyTCM = text . show
 instance PrettyTCM Term where prettyTCM x = prettyA =<< reify x
 instance PrettyTCM Type where prettyTCM x = prettyA =<< reify x
 instance PrettyTCM Sort where prettyTCM x = prettyA =<< reify x
@@ -95,13 +100,21 @@ instance (PrettyTCM a, PrettyTCM b) => PrettyTCM (Judgement a b) where
   prettyTCM (IsSort  a t) = text "Sort" <+> prettyTCM a <+> text ":" <+> prettyTCM t
 
 instance PrettyTCM MetaId where
-  prettyTCM = text . show
+  prettyTCM x = do
+   mn <- getMetaNameSuggestion x
+   text $ show (NamedMeta mn x)
 
 instance PrettyTCM a => PrettyTCM (Blocked a) where
   prettyTCM (Blocked x a) = text "[" <+> prettyTCM a <+> text "]" <> text (show x)
   prettyTCM (NotBlocked x) = prettyTCM x
 
-instance (Reify a e, ToConcrete e c, P.Pretty c) => PrettyTCM (Arg a) where
+instance (Reify a e, ToConcrete e c, P.Pretty c) => PrettyTCM (Named String a) where
+    prettyTCM x = prettyA =<< reify x
+
+instance (ReifyWhen a e, ToConcrete e c, P.Pretty c) => PrettyTCM (Arg a) where
+    prettyTCM x = prettyA =<< reify x
+
+instance (Reify a e, ToConcrete e c, P.Pretty c) => PrettyTCM (Dom a) where
     prettyTCM x = prettyA =<< reify x
 
 instance PrettyTCM Elim where
@@ -111,11 +124,15 @@ instance PrettyTCM Elim where
 instance PrettyTCM A.Expr where
     prettyTCM = prettyA
 
+instance PrettyTCM C.Name where
+  prettyTCM = text . show
+
 instance PrettyTCM Relevance where
   prettyTCM Irrelevant = text "."
   prettyTCM NonStrict  = text ".."
   prettyTCM Relevant   = empty
   prettyTCM Forced     = empty
+  prettyTCM UnusedArg  = empty
 
 instance PrettyTCM Comparison where
   prettyTCM CmpEq  = text "=="
@@ -175,8 +192,14 @@ instance PrettyTCM Constraint where
               OpenIFS{}  -> __IMPOSSIBLE__
               InstS{} -> __IMPOSSIBLE__
               InstV{} -> __IMPOSSIBLE__
-	FindInScope m   -> sep [ text $ "Find in scope " ++ (show m) ]
-        IsEmpty t ->
+	FindInScope m cands -> do
+            t <- getMetaType m
+            sep [ text $ "Find in scope " ++ (show m) ++ " :"
+                , nest 2 $ prettyTCM t
+                , sep $ flip map cands $ \(t,ty) ->
+                           prettyTCM t <+> text ": " <+> prettyTCM ty
+                ]
+        IsEmpty r t ->
             sep [ text "Is empty:", nest 2 $ prettyTCM t ]
 
 instance PrettyTCM Literal where
@@ -203,7 +226,7 @@ instance PrettyTCM PrettyContext where
   prettyTCM (PrettyContext ctx) = P.fsep . reverse <$> pr (map ctxEntry ctx)
       where
           pr []		   = return []
-          pr (Arg h r (x,t) : ctx) = escapeContext 1 $ do
+          pr (Dom h r (x,t) : ctx) = escapeContext 1 $ do
               d    <- prettyTCM t
               x    <- prettyTCM x
               dctx <- pr ctx
