@@ -41,9 +41,11 @@ import Agda.Utils.Impossible
 
 -- | Rewrite a literal to constructor form if possible.
 constructorForm :: Term -> TCM Term
-constructorForm v = case v of
+constructorForm v = case ignoreSharing v of
+{- 2012-04-02 changed semantics of DontCare
 -- Andreas, 2011-10-03, the following line restores IrrelevantLevel
     DontCare v                  -> constructorForm v
+-}
     Lit (LitInt r n)            -> cons primZero primSuc (Lit . LitInt r) n
 --     Level (Max [])              -> primLevelZero
 --     Level (Max [ClosedLevel n]) -> cons primLevelZero primLevelSuc (Level . Max . (:[]) . ClosedLevel) n
@@ -209,9 +211,10 @@ instance (ToTerm a, FromTerm a) => FromTerm [a] where
     fromA <- toTerm
     return $ mkList nil cons toA fromA
     where
-      isCon (Lam _ b) = isCon $ absBody b
-      isCon (Con c _) = return c
-      isCon v	    = do
+      isCon (Lam _ b)  = isCon $ absBody b
+      isCon (Con c _)  = return c
+      isCon (Shared p) = isCon (derefPtr p)
+      isCon v          = do
         d <- prettyTCM v
         typeError $ GenericError $ "expected constructor in built-in binding to " ++ show d
                         -- TODO: check this when binding the things
@@ -248,7 +251,7 @@ redReturn = return . YesReduction
 fromReducedTerm :: (Term -> Maybe a) -> TCM (FromTermFunction a)
 fromReducedTerm f = return $ \t -> do
     b <- reduceB t
-    case f $ unArg (ignoreBlocking b) of
+    case f $ ignoreSharing $ unArg (ignoreBlocking b) of
 	Just x	-> return $ YesReduction x
 	Nothing	-> return $ NoReduction (reduced b)
 
@@ -264,11 +267,11 @@ primTrustMe = do
   when (optSafe clo) (typeError SafeFlagPrimTrustMe)
   t    <- hPi "a" (el primLevel) $
           hPi "A" (return $ sort $ varSort 0) $
-          hPi "x" (El (varSort 1) <$> var 0) $
-          hPi "y" (El (varSort 2) <$> var 1) $
+          hPi "x" (El (varSort 1) <$> varM 0) $
+          hPi "y" (El (varSort 2) <$> varM 1) $
           El (varSort 3) <$>
-            primEquality <#> var 3 <#> var 2 <@> var 1 <@> var 0
-  Con rf [] <- primRefl
+            primEquality <#> varM 3 <#> varM 2 <@> varM 1 <@> varM 0
+  Con rf [] <- ignoreSharing <$> primRefl
   n         <- conPars . theDef <$> getConstInfo rf
   let refl x | n == 2    = Con rf [Arg Hidden Forced x]
              | n == 3    = Con rf []
@@ -422,22 +425,30 @@ infixr 4 -->
 a --> b = do
     a' <- a
     b' <- b
-    return $ El (getSort a' `sLub` getSort b') $ Pi (defaultArg a') (NoAbs "_" b')
+    return $ El (getSort a' `sLub` getSort b') $ Pi (Dom NotHidden Relevant a') (NoAbs "_" b')
+
+infixr 4 .-->
+
+(.-->) :: TCM Type -> TCM Type -> TCM Type
+a .--> b = do
+    a' <- a
+    b' <- b
+    return $ El (getSort a' `sLub` getSort b') $ Pi (Dom NotHidden Irrelevant a') (NoAbs "_" b')
 
 gpi :: Hiding -> Relevance -> String -> TCM Type -> TCM Type -> TCM Type
 gpi h r name a b = do
   a <- a
   x <- freshName_ name
-  b <- addCtx x (Arg h r a) b
+  b <- addCtx x (Dom h r a) b
   return $ El (getSort a `dLub` Abs name (getSort b))
-              (Pi (Arg h r a) (Abs name b))
+              (Pi (Dom h r a) (Abs name b))
 
 hPi, nPi :: String -> TCM Type -> TCM Type -> TCM Type
 hPi = gpi Hidden Relevant
 nPi = gpi NotHidden Relevant
 
-var :: Integer -> TCM Term
-var n = return $ Var n []
+varM :: Int -> TCM Term
+varM = return . var
 
 infixl 9 <@>, <#>
 
@@ -466,10 +477,12 @@ tset = return $ sort (mkType 0)
 -- | Abbreviation: @argN = 'Arg' 'NotHidden' 'Relevant'@.
 
 argN = Arg NotHidden Relevant
+domN = Dom NotHidden Relevant
 
 -- | Abbreviation: @argH = 'Arg' 'Hidden' 'Relevant'@.
 
 argH = Arg Hidden Relevant
+domH = Dom Hidden Relevant
 
 ---------------------------------------------------------------------------
 -- * The actual primitive functions
@@ -499,7 +512,7 @@ primitiveFunctions = Map.fromList
     , "primNatPlus"	    |-> mkPrimFun2 ((+)			    :: Op Nat)
     , "primNatMinus"	    |-> mkPrimFun2 ((\x y -> max 0 (x - y)) :: Op Nat)
     , "primNatTimes"	    |-> mkPrimFun2 ((*)			    :: Op Nat)
-    , "primNatDivSucAux"    |-> mkPrimFun4 ((\k m n j -> k + div (n + m - j) (m + 1)) :: Nat -> Nat -> Nat -> Nat -> Nat)
+    , "primNatDivSucAux"    |-> mkPrimFun4 ((\k m n j -> k + div (max 0 $ n + m - j) (m + 1)) :: Nat -> Nat -> Nat -> Nat -> Nat)
     , "primNatModSucAux"    |->
         let aux :: Nat -> Nat -> Nat -> Nat -> Nat
             aux k m n j | n > j     = mod (n - j - 1) (m + 1)
@@ -567,10 +580,10 @@ lookupPrimitiveFunction x =
 	Just p	-> liftTCM p
 	Nothing	-> typeError $ NoSuchPrimitiveFunction x
 
--- | Rebind a primitive. Assumes everything is type correct. Used when
---   importing a module with primitives.
-rebindPrimitive :: String -> TCM PrimFun
-rebindPrimitive x = do
-    PrimImpl _ pf <- lookupPrimitiveFunction x
-    bindPrimitive x pf
-    return pf
+lookupPrimitiveFunctionQ :: QName -> TCM (String, PrimitiveImpl)
+lookupPrimitiveFunctionQ q = do
+  let s = case qnameName q of
+            Name _ x _ _ -> show x
+  PrimImpl t pf <- lookupPrimitiveFunction s
+  return (s, PrimImpl t $ pf { primFunName = q })
+
