@@ -40,8 +40,8 @@ quotingKit = do
   sucLevel <- primLevelSuc
   lub <- primLevelMax
   el <- primAgdaTypeEl
-  Con z _ <- primZero
-  Con s _ <- primSuc
+  Con z _ <- ignoreSharing <$> primZero
+  Con s _ <- ignoreSharing <$> primSuc
   unsupported <- primAgdaTermUnsupported
   let t @@ u = apply t [defaultArg u]
       quoteHiding Hidden    = hidden
@@ -51,6 +51,7 @@ quotingKit = do
       quoteRelevance Irrelevant = irrelevant
       quoteRelevance NonStrict  = relevant
       quoteRelevance Forced     = relevant
+      quoteRelevance UnusedArg  = relevant
       quoteLit (LitInt   _ n)   = iterate suc zero !! fromIntegral n
       quoteLit _                = unsupported
       -- We keep no ranges in the quoted term, so the equality on terms
@@ -68,17 +69,19 @@ quotingKit = do
       list (a : as) = cons @@ a @@ list as
       zero = con @@ quoteName z @@ nil
       suc n = con @@ quoteName s @@ list [arg @@ visible @@ relevant @@ n]
+      quoteDom q (Dom h r t) = arg @@ quoteHiding h @@ quoteRelevance r @@ q t
       quoteArg q (Arg h r t) = arg @@ quoteHiding h @@ quoteRelevance r @@ q t
       quoteArgs ts = list (map (quoteArg quote) ts)
-      quote (Var n ts) = var @@ Lit (LitInt noRange n) @@ quoteArgs ts
+      quote (Var n ts) = var @@ Lit (LitInt noRange $ fromIntegral n) @@ quoteArgs ts
       quote (Lam h t) = lam @@ quoteHiding h @@ quote (absBody t)
       quote (Def x ts) = def @@ quoteName x @@ quoteArgs ts
       quote (Con x ts) = con @@ quoteName x @@ quoteArgs ts
-      quote (Pi t u) = pi @@ quoteArg quoteType t
+      quote (Pi t u) = pi @@ quoteDom quoteType t
                           @@ quoteType (absBody u)
       quote (Level _) = unsupported
       quote (Lit lit) = quoteLit lit
       quote (Sort s)  = sort @@ quoteSort s
+      quote (Shared p) = quote $ derefPtr p
       quote MetaV{}   = unsupported
       quote DontCare{} = unsupported -- could be exposed at some point but we have to take care
   return (quote, quoteType)
@@ -100,7 +103,7 @@ qNameType = El (mkType 0) <$> primQName
 
 isCon :: QName -> TCM Term -> TCM Bool
 isCon con tm = do t <- tm
-                  case t of
+                  case ignoreSharing t of
                     Con con' _ -> return (con == con')
                     _ -> return False
 
@@ -130,7 +133,7 @@ choice ((mb, mx) : mxs) dflt = do b <- mb
 instance Unquote a => Unquote (Arg a) where
   unquote t = do
     t <- reduce t
-    case t of
+    case ignoreSharing t of
       Con c [hid,rel,x] -> do
         choice
           [(c `isCon` primArgArg, Arg <$> unquoteN hid <*> unquoteN rel <*> unquoteN x)]
@@ -140,14 +143,14 @@ instance Unquote a => Unquote (Arg a) where
 instance Unquote Integer where
   unquote t = do
     t <- reduce t
-    case t of
+    case ignoreSharing t of
       Lit (LitInt _ n) -> return n
       _ -> unquoteFailed "Integer" "not a literal integer" t
 
 instance Unquote a => Unquote [a] where
   unquote t = do
     t <- reduce t
-    case t of
+    case ignoreSharing t of
       Con c [x,xs] -> do
         choice
           [(c `isCon` primCons, (:) <$> unquoteN x <*> unquoteN xs)]
@@ -161,30 +164,34 @@ instance Unquote a => Unquote [a] where
 instance Unquote Hiding where
   unquote t = do
     t <- reduce t
-    case t of
+    let err = unquoteFailed "Hiding" "neither `hidden' nor `visible'" t
+    case ignoreSharing t of
       Con c [] -> do
         choice
           [(c `isCon` primHidden,  return Hidden)
           ,(c `isCon` primInstance, return Instance)
           ,(c `isCon` primVisible, return NotHidden)]
-          (unquoteFailed "Hiding" "neither `hidden' nor `visible'" t)
-      _ -> unquoteFailed "Hiding" "arity is not 0" t
+          err
+      Con c vs -> unquoteFailed "Hiding" "the value is a constructor, but its arity is not 0" t
+      _        -> err
 
 instance Unquote Relevance where
   unquote t = do
     t <- reduce t
-    case t of
+    let err = unquoteFailed "Relevance" "neither `relevant' or `irrelevant'" t
+    case ignoreSharing t of
       Con c [] -> do
         choice
           [(c `isCon` primRelevant,   return Relevant)
           ,(c `isCon` primIrrelevant, return Irrelevant)]
-          (unquoteFailed "Relevance" "neither `relevant' or `irrelevant'" t)
-      _ -> unquoteFailed "Relevance" "arity is not 0" t
+          err
+      Con c vs -> unquoteFailed "Relevance" "the value is a constructor, but its arity is not 0" t
+      _        -> err
 
 instance Unquote QName where
   unquote t = do
     t <- reduce t
-    case t of
+    case ignoreSharing t of
       Lit (LitQName _ x) -> return x
       _                  -> unquoteFailed "QName" "not a literal qname value" t
 
@@ -195,7 +202,7 @@ instance Unquote a => Unquote (Abs a) where
 instance Unquote Sort where
   unquote t = do
     t <- reduce t
-    case t of
+    case ignoreSharing t of
       Con c [] -> do
         choice
           [(c `isCon` primAgdaSortUnsupported, unquoteFailed "Sort" "unsupported sort" t)]
@@ -213,7 +220,7 @@ instance Unquote Level where
 instance Unquote Type where
   unquote t = do
     t <- reduce t
-    case t of
+    case ignoreSharing t of
       Con c [s, u] -> do
         choice
           [(c `isCon` primAgdaTypeEl, El <$> unquoteN s <*> unquoteN u)]
@@ -223,7 +230,7 @@ instance Unquote Type where
 instance Unquote Term where
   unquote t = do
     t <- reduce t
-    case t of
+    case ignoreSharing t of
       Con c [] ->
         choice
           [(c `isCon` primAgdaTermUnsupported, unquoteFailed "Term" "unsupported term" t)]
@@ -236,11 +243,11 @@ instance Unquote Term where
 
       Con c [x,y] ->
         choice
-          [(c `isCon` primAgdaTermVar, Var <$> unquoteN x <*> unquoteN y)
+          [(c `isCon` primAgdaTermVar, Var <$> (fromInteger <$> unquoteN x) <*> unquoteN y)
           ,(c `isCon` primAgdaTermCon, Con <$> unquoteN x <*> unquoteN y)
           ,(c `isCon` primAgdaTermDef, Def <$> unquoteN x <*> unquoteN y)
           ,(c `isCon` primAgdaTermLam, Lam <$> unquoteN x <*> unquoteN y)
-          ,(c `isCon` primAgdaTermPi,  Pi  <$> unquoteN x <*> unquoteN y)]
+          ,(c `isCon` primAgdaTermPi,  Pi  <$> (domFromArg <$> unquoteN x) <*> unquoteN y)]
           (unquoteFailed "Term" "arity 2 and none of Var, Con, Def, Lam, Pi" t)
 
       Con{} -> unquoteFailed "Term" "neither arity 0 nor 1 nor 2" t
