@@ -1,4 +1,4 @@
-{-# LANGUAGE CPP #-}
+{-# LANGUAGE CPP, TupleSections #-}
 
 module Agda.TypeChecking.Records where
 
@@ -21,8 +21,11 @@ import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Telescope
 import Agda.TypeChecking.Datatypes
+import Agda.Utils.Either
 import Agda.Utils.List
+import Agda.Utils.Maybe
 import Agda.Utils.Monad
+import qualified Agda.Utils.HashMap as HMap
 
 #include "../undefined.h"
 import Agda.Utils.Impossible
@@ -71,12 +74,12 @@ getRecordFieldNames r =
 -- | Find all records with at least the given fields.
 findPossibleRecords :: [C.Name] -> TCM [QName]
 findPossibleRecords fields = do
-  defs <- (Map.union `on` sigDefinitions) <$> getSignature <*> getImportedSignature
+  defs <- (HMap.union `on` sigDefinitions) <$> getSignature <*> getImportedSignature
   let possible def = case theDef def of
         Record{ recFields = fs } -> Set.isSubsetOf given inrecord
           where inrecord = Set.fromList $ map (nameConcrete . qnameName . unArg) fs
         _ -> False
-  return [ defName d | d <- Map.elems defs, possible d ]
+  return [ defName d | d <- HMap.elems defs, possible d ]
   where
     given = Set.fromList fields
 
@@ -84,6 +87,18 @@ findPossibleRecords fields = do
 getRecordFieldTypes :: QName -> TCM Telescope
 getRecordFieldTypes r = recTel <$> getRecordDef r
 
+-- | Get the field names belonging to a record type.
+getRecordTypeFields :: Type -> TCM [Arg QName]
+getRecordTypeFields t =
+  case ignoreSharing $ unEl t of
+    Def r _ -> do
+      rDef <- theDef <$> getConstInfo r
+      case rDef of
+        Record { recFields = fields } -> return fields
+        _ -> __IMPOSSIBLE__
+    _ -> __IMPOSSIBLE__
+
+
 -- | Get the type of the record constructor.
 getRecordConstructorType :: QName -> TCM Type
 getRecordConstructorType r = recConType <$> getRecordDef r
@@ -94,33 +109,36 @@ getRecordConstructor :: QName -> TCM QName
 getRecordConstructor r = killRange <$> recCon <$> getRecordDef r
 
 -- | Check if a name refers to a record.
-isRecord :: QName -> TCM Bool
+--   If yes, return record definition.
+isRecord :: QName -> TCM (Maybe Defn)
 isRecord r = do
   def <- theDef <$> getConstInfo r
   return $ case def of
-    Record{} -> True
-    _        -> False
+    Record{} -> Just def
+    _        -> Nothing
 
 -- | Check if a name refers to an eta expandable record.
 isEtaRecord :: QName -> TCM Bool
-isEtaRecord r = do
-  def <- theDef <$> getConstInfo r
-  return $ case def of
-    Record{recEtaEquality = eta} -> eta
-    _                            -> False
+isEtaRecord r = maybe False recEtaEquality <$> isRecord r
+
+-- | Check if a name refers to a record which is not coinductive.  (Projections are then size-preserving)
+isInductiveRecord :: QName -> TCM Bool
+isInductiveRecord r = maybe False (\ d -> recInduction d == Inductive || not (recRecursive d)) <$> isRecord r
 
 -- | Check if a type is an eta expandable record and return the record identifier and the parameters.
 isEtaRecordType :: Type -> TCM (Maybe (QName, Args))
-isEtaRecordType (El _ (Def d ps)) = ifM (isEtaRecord d) (return $ Just (d, ps)) (return Nothing)
-isEtaRecordType _ = return Nothing
+isEtaRecordType a = case ignoreSharing $ unEl a of
+  Def d ps -> ifM (isEtaRecord d) (return $ Just (d, ps)) (return Nothing)
+  _        -> return Nothing
 
 -- | Check if a name refers to a record constructor.
-isRecordConstructor :: QName -> TCM Bool
+--   If yes, return record definition.
+isRecordConstructor :: QName -> TCM (Maybe (QName, Defn))
 isRecordConstructor c = do
   def <- theDef <$> getConstInfo c
   case def of
-    Constructor{ conData = r } -> isRecord r
-    _                          -> return False
+    Constructor{ conData = r } -> fmap (r,) <$> isRecord r
+    _                          -> return Nothing
 
 -- | Check if a constructor name is the internally generated record constructor.
 isGeneratedRecordConstructor :: QName -> TCM Bool
@@ -134,6 +152,20 @@ isGeneratedRecordConstructor c = do
         _                             -> return False
     _ -> return False
 
+-- | Mark record type as unguarded.
+--   No eta-expansion.  Projections do not preserve guardedness.
+unguardedRecord :: QName -> TCM ()
+unguardedRecord q = modifySignature $ updateDefinition q $ updateTheDef $ updateRecord
+  where updateRecord r@Record{} = r { recEtaEquality = False, recRecursive = True }
+        updateRecord _          = __IMPOSSIBLE__
+
+-- | Mark record type as recursive.
+--   Projections do not preserve guardedness.
+recursiveRecord :: QName -> TCM ()
+recursiveRecord q = modifySignature $ updateDefinition q $ updateTheDef $ updateRecord
+  where updateRecord r@Record{} = r { recRecursive = True }
+        updateRecord _          = __IMPOSSIBLE__
+
 {-| Compute the eta expansion of a record. The first argument should be
     the name of a record type. Given
 
@@ -143,50 +175,48 @@ isGeneratedRecordConstructor c = do
 -}
 etaExpandRecord :: QName -> Args -> Term -> TCM (Telescope, Args)
 etaExpandRecord r pars u = do
-  Record{ recFields = xs, recTel = tel } <- getRecordDef r
+  Record{ recFields = xs, recTel = tel, recEtaEquality = eta } <- getRecordDef r
+  unless eta __IMPOSSIBLE__ -- make sure we do not expand non-eta records
   let tel' = apply tel pars
-  case u of
+  case ignoreSharing u of
     Con _ args -> return (tel', args)  -- Already expanded.
     _          -> do
-      -- irrelevant fields are expanded to DontCare
-      -- this is sound because etaExpandRecord is only called during conversion
-      -- WARNING: do not use etaExpandRecord to expand MetaVars!!
-      let proj (Arg h Irrelevant x) = Arg h Irrelevant $ DontCare (Def x [defaultArg u])
-          proj (Arg h rel x)        = Arg h rel $
-            Def x [defaultArg u]
+      let xs' = map (fmap (\ x -> Def x [defaultArg u])) xs
       reportSDoc "tc.record.eta" 20 $ vcat
         [ text "eta expanding" <+> prettyTCM u <+> text ":" <+> prettyTCM r
         , nest 2 $ vcat
           [ text "tel' =" <+> prettyTCM tel'
-          , text "args =" <+> prettyTCM (map proj xs)
+          , text "args =" <+> prettyTCM xs'
           ]
         ]
-      return (tel', map proj xs)
-  where
-    hide a = a { argHiding = Hidden }
+      return (tel', xs')
 
 -- | The fields should be eta contracted already.
+--
+--   We can eta constract if all fields @f = ...@ are irrelevant
+--   or the corresponding projection @f = f v@ of the same value @v@,
+--   but we need at least one relevant field to find the value @v@.
 etaContractRecord :: QName -> QName -> Args -> TCM Term
 etaContractRecord r c args = do
   Record{ recPars = npars, recFields = xs } <- getRecordDef r
-  let check a ax = do
+  let check :: Arg Term -> Arg QName -> Maybe (Maybe Term)
+      check a ax = do
       -- @a@ is the constructor argument, @ax@ the corr. record field name
         -- skip irrelevant record fields by returning DontCare
-        case (argRelevance a, unArg a) of
-          (Irrelevant, v) -> return $ Just $ DontCare __IMPOSSIBLE__ -- thrown out later
+        case (argRelevance a, ignoreSharing $ unArg a) of
+          (Irrelevant, _) -> Just Nothing
           -- if @a@ is the record field name applied to a single argument
           -- then it passes the check
-          (_, Def y [arg]) | unArg ax == y
-                         -> return (Just $ unArg arg)
-          _              -> return Nothing
+          (_, Def f [arg]) | unArg ax == f
+                         -> Just $ Just $ unArg arg
+          _              -> Nothing
       fallBack = return (Con c args)
   case compare (length args) (length xs) of
     LT -> fallBack       -- Not fully applied
     GT -> __IMPOSSIBLE__ -- Too many arguments. Impossible.
     EQ -> do
-      as <- zipWithM check args xs
-      case sequence as of
-        Just as -> case filter notDontCare as of
+      case zipWithM check args xs of
+        Just as -> case [ a | Just a <- as ] of
           (a:as) ->
             if all (a ==) as
               then do
@@ -196,41 +226,66 @@ etaContractRecord r c args = do
                   ]
                 return a
               else fallBack
-          _ -> fallBack -- just DontCares
+          _ -> fallBack -- just irrelevant terms
         _ -> fallBack  -- a Nothing
-  where notDontCare (DontCare _) = False
-        notDontCare _            = True
 
 -- | Is the type a hereditarily singleton record type? May return a
 -- blocking metavariable.
 --
 -- Precondition: The name should refer to a record type, and the
 -- arguments should be the parameters to the type.
-
 isSingletonRecord :: QName -> Args -> TCM (Either MetaId Bool)
-isSingletonRecord = isSingletonRecord' False
+isSingletonRecord r ps = mapRight isJust <$> isSingletonRecord' False r ps
 
 isSingletonRecordModuloRelevance :: QName -> Args -> TCM (Either MetaId Bool)
-isSingletonRecordModuloRelevance = isSingletonRecord' True
+isSingletonRecordModuloRelevance r ps = mapRight isJust <$> isSingletonRecord' True r ps
 
-isSingletonRecord' :: Bool -> QName -> Args -> TCM (Either MetaId Bool)
-isSingletonRecord' regardIrrelevance r ps =
-  check =<< ((`apply` ps) <$> getRecordFieldTypes r)
+-- | Return the unique (closed) inhabitant if exists.
+--   In case of counting irrelevance in, the returned inhabitant
+--   contains garbage.
+isSingletonRecord' :: Bool -> QName -> Args -> TCM (Either MetaId (Maybe Term))
+isSingletonRecord' regardIrrelevance r ps = do
+  def <- getRecordDef r
+  emap (Con $ recCon def) <$> check (recTel def `apply` ps)
   where
-  check EmptyTel            = return (Right True)
-  check (ExtendTel arg tel) | regardIrrelevance && argRelevance arg == Irrelevant =
-    underAbstraction arg tel check
-  check (ExtendTel arg tel) = do
-    TelV _ t <- telView $ unArg arg
+  check :: Telescope -> TCM (Either MetaId (Maybe [Arg Term]))
+  check EmptyTel            = return $ Right $ Just []
+  check (ExtendTel arg@(Dom h Irrelevant _) tel) | regardIrrelevance =
+    underAbstraction arg tel $ \ tel ->
+      emap (Arg h Irrelevant garbage :) <$> check tel
+  check (ExtendTel arg@(Dom h r t) tel) = do
+    isSing <- isSingletonType' regardIrrelevance t
+    case isSing of
+      Left mid       -> return $ Left mid
+      Right Nothing  -> return $ Right Nothing
+      Right (Just v) -> underAbstraction arg tel $ \ tel ->
+        emap (Arg h r v :) <$> check tel
+  garbage :: Term
+  garbage = Sort Prop
+
+-- | Check whether a type has a unique inhabitant and return it.
+--   Can be blocked by a metavar.
+isSingletonType :: Type -> TCM (Either MetaId (Maybe Term))
+isSingletonType = isSingletonType' False
+
+-- | Check whether a type has a unique inhabitant (irrelevant parts ignored).
+--   Can be blocked by a metavar.
+isSingletonTypeModuloRelevance :: (MonadTCM tcm) => Type -> tcm (Either MetaId Bool)
+isSingletonTypeModuloRelevance t = liftTCM $ do
+  mapRight isJust <$> isSingletonType' True t
+
+isSingletonType' :: Bool -> Type -> TCM (Either MetaId (Maybe Term))
+isSingletonType' regardIrrelevance t = do
+    TelV tel t <- telView t
     t <- reduceB $ unEl t
-    case t of
+    case ignoreSharing <$> t of
       Blocked m _            -> return (Left m)
       NotBlocked (MetaV m _) -> return (Left m)
       NotBlocked (Def r ps)  ->
-        ifM (not <$> isRecord r) (return $ Right False) $ do
-          isRec <- isSingletonRecord' regardIrrelevance r ps
-          case isRec of
-            Left _      -> return isRec
-            Right False -> return isRec
-            Right True  -> underAbstraction arg tel $ check
-      _ -> return (Right False)
+        ifM (isNothing <$> isRecord r) (return $ Right Nothing) $ do
+          emap (abstract tel) <$> isSingletonRecord' regardIrrelevance r ps
+      _ -> return (Right Nothing)
+
+-- | Auxiliary function.
+emap :: (a -> b) -> Either c (Maybe a) -> Either c (Maybe b)
+emap = mapRight . fmap
