@@ -13,6 +13,7 @@ import qualified Data.Map as Map
 import Data.Map (Map)
 import qualified Data.Set as Set
 import Data.Traversable
+import Data.Hashable
 
 import Agda.Syntax.Position
 import Agda.Syntax.Common
@@ -34,6 +35,8 @@ import {-# SOURCE #-} Agda.TypeChecking.Patterns.Match
 import {-# SOURCE #-} Agda.TypeChecking.CompiledClause.Match
 
 import Agda.Utils.Monad
+import Agda.Utils.HashMap (HashMap)
+import qualified Agda.Utils.HashMap as HMap
 
 #include "../undefined.h"
 import Agda.Utils.Impossible
@@ -71,6 +74,7 @@ instance Instantiate Term where
       InstS _                        -> __IMPOSSIBLE__
   instantiate (Level l) = levelTm <$> instantiate l
   instantiate (Sort s) = sortTm <$> instantiate s
+  instantiate v@Shared{} = updateSharedTerm instantiate v
   instantiate t = return t
 
 instance Instantiate Level where
@@ -84,7 +88,7 @@ instance Instantiate LevelAtom where
   instantiate l = case l of
     MetaLevel m vs -> do
       v <- instantiate (MetaV m vs)
-      case v of
+      case ignoreSharing v of
         MetaV m vs -> return $ MetaLevel m vs
         _          -> return $ UnreducedLevel v
     UnreducedLevel l -> UnreducedLevel <$> instantiate l
@@ -120,6 +124,9 @@ instance Instantiate t => Instantiate (Abs t) where
 instance Instantiate t => Instantiate (Arg t) where
     instantiate = traverse instantiate
 
+instance Instantiate t => Instantiate (Dom t) where
+    instantiate = traverse instantiate
+
 instance Instantiate t => Instantiate [t] where
     instantiate = traverse instantiate
 
@@ -150,16 +157,32 @@ instance Instantiate Constraint where
   instantiate (SortCmp cmp a b)    = uncurry (SortCmp cmp) <$> instantiate (a,b)
   instantiate (Guarded c pid)      = Guarded <$> instantiate c <*> pure pid
   instantiate (UnBlock m)          = return $ UnBlock m
-  instantiate (FindInScope m)      = return $ FindInScope m
-  instantiate (IsEmpty t)          = IsEmpty <$> instantiate t
+  instantiate (FindInScope m args) = FindInScope m <$> mapM instantiate args
+  instantiate (IsEmpty r t)        = IsEmpty r <$> instantiate t
 
 instance (Ord k, Instantiate e) => Instantiate (Map k e) where
     instantiate = traverse instantiate
 
 
---
--- Reduction to weak head normal form.
---
+---------------------------------------------------------------------------
+-- * Reduction to weak head normal form.
+---------------------------------------------------------------------------
+
+ifBlocked :: MonadTCM tcm =>  Term -> (MetaId -> Term -> tcm a) -> (Term -> tcm a) -> tcm a
+ifBlocked t blocked unblocked = do
+  t <- liftTCM $ reduceB t
+  case ignoreSharing <$> t of
+    Blocked m _            -> blocked m (ignoreBlocking t)
+    NotBlocked (MetaV m _) -> blocked m (ignoreBlocking t)
+    NotBlocked _           -> unblocked (ignoreBlocking t)
+
+ifBlockedType :: MonadTCM tcm => Type -> (MetaId -> Type -> tcm a) -> (Type -> tcm a) -> tcm a
+ifBlockedType t blocked unblocked = do
+  t <- liftTCM $ reduceB t
+  case ignoreSharing . unEl <$> t of
+    Blocked m _            -> blocked m (ignoreBlocking t)
+    NotBlocked (MetaV m _) -> blocked m (ignoreBlocking t)
+    NotBlocked _           -> unblocked (ignoreBlocking t)
 
 class Reduce t where
     reduce  :: t -> TCM t
@@ -211,13 +234,14 @@ instance Reduce LevelAtom where
     where
       fromTm v = do
         bv <- reduceB v
-        case bv of
-          Blocked m v             -> return $ Blocked m  $ BlockedLevel m v
+        let v = ignoreBlocking bv
+        case ignoreSharing <$> bv of
           NotBlocked (MetaV m vs) -> return $ NotBlocked $ MetaLevel m vs
-          NotBlocked v            -> return $ NotBlocked $ NeutralLevel v
+          Blocked m _             -> return $ Blocked m  $ BlockedLevel m v
+          NotBlocked _            -> return $ NotBlocked $ NeutralLevel v
 
 
-instance (Raise t, Reduce t) => Reduce (Abs t) where
+instance (Subst t, Reduce t) => Reduce (Abs t) where
   reduce b@(Abs x _) = Abs x <$> underAbstraction_ b reduce
   reduce (NoAbs x v) = NoAbs x <$> reduce v
 
@@ -226,7 +250,13 @@ instance Reduce t => Reduce [t] where
     reduce = traverse reduce
 
 instance Reduce t => Reduce (Arg t) where
-    reduce  = traverse reduce
+    reduce a@(Arg h Irrelevant t) = return a  -- Don't reduce irr. args!?
+    reduce a                      = traverse reduce a
+
+    reduceB t = traverse id <$> traverse reduceB t
+
+instance Reduce t => Reduce (Dom t) where
+    reduce = traverse reduce
     reduceB t = traverse id <$> traverse reduceB t
 
 -- Tuples are never blocked
@@ -240,7 +270,10 @@ instance Reduce Term where
   reduceB v = {-# SCC "reduce<Term>" #-} do
     v <- instantiate v
     case v of
-      MetaV x args -> notBlocked . MetaV x <$> reduce args
+--    Andreas, 2012-11-05 not reducing meta args does not destroy anything
+--    and seems to save 2% sec on the standard library
+--      MetaV x args -> notBlocked . MetaV x <$> reduce args
+      MetaV x args -> return $ notBlocked v
       Def f args   -> unfoldDefinition False reduceB (Def f []) f args
       Con c args   -> do
           -- Constructors can reduce when they come from an
@@ -254,20 +287,22 @@ instance Reduce Term where
       Var _ _  -> return $ notBlocked v
       Lam _ _  -> return $ notBlocked v
       DontCare _ -> return $ notBlocked v
+      Shared{}   -> updateSharedTermF reduceB v
     where
       -- NOTE: reduceNat can traverse the entire term.
+      reduceNat v@Shared{} = updateSharedTerm reduceNat v
       reduceNat v@(Con c []) = do
         mz  <- getBuiltin' builtinZero
         case v of
           _ | Just v == mz  -> return $ Lit $ LitInt (getRange c) 0
           _		    -> return v
       reduceNat v@(Con c [Arg NotHidden Relevant w]) = do
-        ms  <- getBuiltin' builtinSuc
+        ms  <- fmap ignoreSharing <$> getBuiltin' builtinSuc
         case v of
           _ | Just (Con c []) == ms -> inc <$> reduce w
           _	                    -> return v
           where
-            inc w = case w of
+            inc w = case ignoreSharing w of
               Lit (LitInt r n) -> Lit (LitInt (fuseRange c r) $ n + 1)
               _                -> Con c [Arg NotHidden Relevant w]
       reduceNat v = return v
@@ -313,7 +348,7 @@ unfoldDefinition unfoldDelayed keepGoing v0 f args =
             mredToBlocked (MaybeRed NotReduced  x) = notBlocked x
             mredToBlocked (MaybeRed (Reduced b) x) = x <$ b
 
-    -- reduceNormal :: Term -> QName -> [MaybeReduced (Arg Term)] -> Delayed -> [Clause] -> Maybe CompiledClauses -> TCM (Blocked Term)
+    reduceNormal :: Term -> QName -> [MaybeReduced (Arg Term)] -> Delayed -> [Clause] -> Maybe CompiledClauses -> TCM (Blocked Term)
     reduceNormal v0 f args delayed def mcc = {-# SCC "reduceNormal" #-} do
         case def of
           _ | Delayed <- delayed,
@@ -333,59 +368,88 @@ unfoldDefinition unfoldDelayed keepGoing v0 f args =
                 reportSDoc "tc.reduce" 90 $ vcat
                   [ text "*** reduced definition: " <+> prettyTCM f
                   ]
-                reportSDoc "tc.reduce" 100 $ text "    result" <+> prettyTCM v
+                reportSDoc "tc.reduce" 100 $ text "    result" <+> prettyTCM v $$
+                                             text "    raw   " <+> text (show v)
                 keepGoing v
       where defaultResult = return $ notBlocked $ v0 `apply` (map ignoreReduced args)
 
-    -- Apply a defined function to it's arguments.
-    --   The original term is the first argument applied to the third.
-    appDef :: Term -> CompiledClauses -> MaybeReducedArgs -> TCM (Reduced (Blocked Term) Term)
-    appDef v cc args = liftTCM $ do
-      r <- matchCompiled cc args
-      case r of
+-- | Reduce a non-primitive definition if it is a copy linking to another def.
+reduceDefCopy :: QName -> Args -> TCM (Reduced () Term)
+reduceDefCopy f vs = do
+  info <- getConstInfo f
+  if (defCopy info) then reduceDef_ info f vs else return $ NoReduction ()
+
+-- | Reduce a non-primitive definition once unless it is delayed.
+reduceDef :: QName -> Args -> TCM (Reduced () Term)
+reduceDef f vs = do
+  info <- getConstInfo f
+  reduceDef_ info f vs
+
+reduceDef_ :: Definition -> QName -> Args -> TCM (Reduced () Term)
+reduceDef_ info f vs = do
+  let v0   = Def f []
+      args = map notReduced vs
+      cls  = (defClauses info)
+      mcc  = (defCompiled info)
+      delayed = (defDelayed info)
+  case delayed of
+    Delayed -> return $ NoReduction ()
+    NotDelayed -> do
+      ev <- maybe (appDef' v0 cls args)
+                  (\cc -> appDef v0 cc args) mcc
+      case ev of
         YesReduction t    -> return $ YesReduction t
-        NoReduction args' -> return $ NoReduction $ fmap (apply v) args'
+        NoReduction args' -> return $ NoReduction ()
 
-    appDef' :: Term -> [Clause] -> MaybeReducedArgs -> TCM (Reduced (Blocked Term) Term)
-    appDef' _ [] _ = {- ' -} __IMPOSSIBLE__
-    appDef' v cls@(Clause {clausePats = ps} : _) args
-      | m < n     = return $ NoReduction $ notBlocked $ v `apply` map ignoreReduced args
-      | otherwise = do
-        let (args0, args1) = splitAt n args
-        r <- goCls cls (map ignoreReduced args0)
-        case r of
-          YesReduction u -> return $ YesReduction $ u `apply` map ignoreReduced args1
-          NoReduction v  -> return $ NoReduction $ (`apply` map ignoreReduced args1) <$> v
-      where
+-- Apply a defined function to it's arguments.
+--   The original term is the first argument applied to the third.
+appDef :: Term -> CompiledClauses -> MaybeReducedArgs -> TCM (Reduced (Blocked Term) Term)
+appDef v cc args = liftTCM $ do
+  r <- matchCompiled cc args
+  case r of
+    YesReduction t    -> return $ YesReduction t
+    NoReduction args' -> return $ NoReduction $ fmap (apply v) args'
 
-        n = genericLength ps
-        m = genericLength args
+appDef' :: Term -> [Clause] -> MaybeReducedArgs -> TCM (Reduced (Blocked Term) Term)
+appDef' _ [] _ = {- ' -} __IMPOSSIBLE__
+appDef' v cls@(Clause {clausePats = ps} : _) args
+  | m < n     = return $ NoReduction $ notBlocked $ v `apply` map ignoreReduced args
+  | otherwise = do
+    let (args0, args1) = splitAt n args
+    r <- goCls cls (map ignoreReduced args0)
+    case r of
+      YesReduction u -> return $ YesReduction $ u `apply` map ignoreReduced args1
+      NoReduction v  -> return $ NoReduction $ (`apply` map ignoreReduced args1) <$> v
+  where
 
-        goCls :: [Clause] -> Args -> TCM (Reduced (Blocked Term) Term)
-        goCls [] args = typeError $ IncompletePatternMatching v args
-        goCls (cl@(Clause { clausePats = pats
-                          , clauseBody = body }) : cls) args = do
-            (m, args) <- matchPatterns pats args
-            case m of
-                No		  -> goCls cls args
-                DontKnow Nothing  -> return $ NoReduction $ notBlocked $ v `apply` args
-                DontKnow (Just m) -> return $ NoReduction $ blocked m $ v `apply` args
-                Yes args'
-                  | hasBody body  -> return $ YesReduction (
-                      -- TODO: let matchPatterns also return the reduced forms
-                      -- of the original arguments!
-                      app args' body)
-                  | otherwise	  -> return $ NoReduction $ notBlocked $ v `apply` args
+    n = genericLength ps
+    m = genericLength args
 
-        hasBody (Body _) = True
-        hasBody NoBody   = False
-        hasBody (Bind b) = hasBody (unAbs b)
+    goCls :: [Clause] -> Args -> TCM (Reduced (Blocked Term) Term)
+    goCls [] args = typeError $ IncompletePatternMatching v args
+    goCls (cl@(Clause { clausePats = pats
+                      , clauseBody = body }) : cls) args = do
+        (m, args) <- matchPatterns pats args
+        case m of
+            No		  -> goCls cls args
+            DontKnow Nothing  -> return $ NoReduction $ notBlocked $ v `apply` args
+            DontKnow (Just m) -> return $ NoReduction $ blocked m $ v `apply` args
+            Yes args'
+              | hasBody body  -> return $ YesReduction (
+                  -- TODO: let matchPatterns also return the reduced forms
+                  -- of the original arguments!
+                  app args' body)
+              | otherwise	  -> return $ NoReduction $ notBlocked $ v `apply` args
 
-        app []		 (Body v') = v'
-        app (arg : args) (Bind b)  = app args $ absApp b arg -- CBN
-        app  _		  NoBody   = __IMPOSSIBLE__
-        app (_ : _)	 (Body _)  = __IMPOSSIBLE__
-        app []		 (Bind _)  = __IMPOSSIBLE__
+    hasBody (Body _) = True
+    hasBody NoBody   = False
+    hasBody (Bind b) = hasBody (unAbs b)
+
+    app []		 (Body v') = v'
+    app (arg : args) (Bind b)  = app args $ absApp b arg -- CBN
+    app  _		  NoBody   = __IMPOSSIBLE__
+    app (_ : _)	 (Body _)  = __IMPOSSIBLE__
+    app []		 (Bind _)  = __IMPOSSIBLE__
 
 
 instance Reduce a => Reduce (Closure a) where
@@ -402,14 +466,14 @@ instance Reduce Constraint where
     return $ ValueCmp cmp t u v
   reduce (ElimCmp cmp t v as bs) =
     ElimCmp cmp <$> reduce t <*> reduce v <*> reduce as <*> reduce bs
-  reduce (LevelCmp cmp u v)   = uncurry (LevelCmp cmp) <$> reduce (u,v)
-  reduce (TypeCmp cmp a b)    = uncurry (TypeCmp cmp) <$> reduce (a,b)
+  reduce (LevelCmp cmp u v)    = uncurry (LevelCmp cmp) <$> reduce (u,v)
+  reduce (TypeCmp cmp a b)     = uncurry (TypeCmp cmp) <$> reduce (a,b)
   reduce (TelCmp a b cmp tela telb) = uncurry (TelCmp a b cmp)  <$> reduce (tela,telb)
-  reduce (SortCmp cmp a b)    = uncurry (SortCmp cmp) <$> reduce (a,b)
-  reduce (Guarded c pid)      = Guarded <$> reduce c <*> pure pid
-  reduce (UnBlock m)          = return $ UnBlock m
-  reduce (FindInScope m)      = return $ FindInScope m
-  reduce (IsEmpty t)          = IsEmpty <$> reduce t
+  reduce (SortCmp cmp a b)     = uncurry (SortCmp cmp) <$> reduce (a,b)
+  reduce (Guarded c pid)       = Guarded <$> reduce c <*> pure pid
+  reduce (UnBlock m)           = return $ UnBlock m
+  reduce (FindInScope m cands) = FindInScope m <$> mapM reduce cands
+  reduce (IsEmpty r t)         = IsEmpty r <$> reduce t
 
 instance (Ord k, Reduce e) => Reduce (Map k e) where
     reduce = traverse reduce
@@ -446,6 +510,7 @@ instance Normalise Term where
 		Lam h b	    -> Lam h <$> normalise b
 		Sort s	    -> sortTm <$> normalise s
 		Pi a b	    -> uncurry Pi <$> normalise (a,b)
+                Shared{}    -> updateSharedTerm normalise v
                 DontCare _  -> return v
 
 instance Normalise Elim where
@@ -473,11 +538,15 @@ instance Normalise ClauseBody where
     normalise (Bind   b) = Bind   <$> normalise b
     normalise  NoBody	 = return NoBody
 
-instance (Raise t, Normalise t) => Normalise (Abs t) where
+instance (Subst t, Normalise t) => Normalise (Abs t) where
     normalise a@(Abs x _) = Abs x <$> underAbstraction_ a normalise
     normalise (NoAbs x v) = NoAbs x <$> normalise v
 
 instance Normalise t => Normalise (Arg t) where
+    normalise a@(Arg h Irrelevant t) = return a -- Andreas, 2012-04-02: Do not normalize irrelevant terms!?
+    normalise a = traverse normalise a
+
+instance Normalise t => Normalise (Dom t) where
     normalise = traverse normalise
 
 instance Normalise t => Normalise [t] where
@@ -496,7 +565,7 @@ instance Normalise a => Normalise (Closure a) where
 	x <- enterClosure cl normalise
 	return $ cl { clValue = x }
 
-instance (Raise a, Normalise a) => Normalise (Tele a) where
+instance (Subst a, Normalise a) => Normalise (Tele a) where
   normalise EmptyTel        = return EmptyTel
   normalise (ExtendTel a b) = uncurry ExtendTel <$> normalise (a, b)
 
@@ -509,14 +578,14 @@ instance Normalise Constraint where
     return $ ValueCmp cmp t u v
   normalise (ElimCmp cmp t v as bs) =
     ElimCmp cmp <$> normalise t <*> normalise v <*> normalise as <*> normalise bs
-  normalise (LevelCmp cmp u v)   = uncurry (LevelCmp cmp) <$> normalise (u,v)
-  normalise (TypeCmp cmp a b)    = uncurry (TypeCmp cmp) <$> normalise (a,b)
+  normalise (LevelCmp cmp u v)    = uncurry (LevelCmp cmp) <$> normalise (u,v)
+  normalise (TypeCmp cmp a b)     = uncurry (TypeCmp cmp) <$> normalise (a,b)
   normalise (TelCmp a b cmp tela telb) = uncurry (TelCmp a b cmp) <$> normalise (tela,telb)
-  normalise (SortCmp cmp a b)    = uncurry (SortCmp cmp) <$> normalise (a,b)
-  normalise (Guarded c pid)      = Guarded <$> normalise c <*> pure pid
-  normalise (UnBlock m)          = return $ UnBlock m
-  normalise (FindInScope m)      = return $ FindInScope m
-  normalise (IsEmpty t)          = IsEmpty <$> normalise t
+  normalise (SortCmp cmp a b)     = uncurry (SortCmp cmp) <$> normalise (a,b)
+  normalise (Guarded c pid)       = Guarded <$> normalise c <*> pure pid
+  normalise (UnBlock m)           = return $ UnBlock m
+  normalise (FindInScope m cands) = FindInScope m <$> mapM normalise cands
+  normalise (IsEmpty r t)         = IsEmpty r <$> normalise t
 
 instance Normalise Pattern where
   normalise p = case p of
@@ -574,6 +643,7 @@ instance InstantiateFull Term where
           Lam h b     -> Lam h <$> instantiateFull b
           Sort s      -> sortTm <$> instantiateFull s
           Pi a b      -> uncurry Pi <$> instantiateFull (a,b)
+          Shared{}    -> updateSharedTerm instantiateFull v
           DontCare v  -> DontCare <$> instantiateFull v
 
 instance InstantiateFull Level where
@@ -587,7 +657,7 @@ instance InstantiateFull LevelAtom where
   instantiateFull l = case l of
     MetaLevel m vs -> do
       v <- instantiateFull (MetaV m vs)
-      case v of
+      case ignoreSharing v of
         MetaV m vs -> return $ MetaLevel m vs
         _          -> return $ UnreducedLevel v
     NeutralLevel v -> NeutralLevel <$> instantiateFull v
@@ -608,13 +678,16 @@ instance InstantiateFull ClauseBody where
     instantiateFull (Bind   b) = Bind   <$> instantiateFull b
     instantiateFull  NoBody    = return NoBody
 
-instance (Raise t, InstantiateFull t) => InstantiateFull (Abs t) where
+instance (Subst t, InstantiateFull t) => InstantiateFull (Abs t) where
     instantiateFull a@(Abs x _) = Abs x <$> underAbstraction_ a instantiateFull
     instantiateFull (NoAbs x a) = NoAbs x <$> instantiateFull a
 
 instance InstantiateFull t => InstantiateFull (Arg t) where
     instantiateFull = traverse instantiateFull
 
+instance InstantiateFull t => InstantiateFull (Dom t) where
+    instantiateFull = traverse instantiateFull
+
 instance InstantiateFull t => InstantiateFull [t] where
     instantiateFull = traverse instantiateFull
 
@@ -641,14 +714,14 @@ instance InstantiateFull Constraint where
       return $ ValueCmp cmp t u v
     ElimCmp cmp t v as bs ->
       ElimCmp cmp <$> instantiateFull t <*> instantiateFull v <*> instantiateFull as <*> instantiateFull bs
-    LevelCmp cmp u v   -> uncurry (LevelCmp cmp) <$> instantiateFull (u,v)
-    TypeCmp cmp a b    -> uncurry (TypeCmp cmp) <$> instantiateFull (a,b)
+    LevelCmp cmp u v    -> uncurry (LevelCmp cmp) <$> instantiateFull (u,v)
+    TypeCmp cmp a b     -> uncurry (TypeCmp cmp) <$> instantiateFull (a,b)
     TelCmp a b cmp tela telb -> uncurry (TelCmp a b cmp) <$> instantiateFull (tela,telb)
-    SortCmp cmp a b    -> uncurry (SortCmp cmp) <$> instantiateFull (a,b)
-    Guarded c pid      -> Guarded <$> instantiateFull c <*> pure pid
-    UnBlock m          -> return $ UnBlock m
-    FindInScope m      -> return $ FindInScope m
-    IsEmpty t          -> IsEmpty <$> instantiateFull t
+    SortCmp cmp a b     -> uncurry (SortCmp cmp) <$> instantiateFull (a,b)
+    Guarded c pid       -> Guarded <$> instantiateFull c <*> pure pid
+    UnBlock m           -> return $ UnBlock m
+    FindInScope m cands -> FindInScope m <$> mapM instantiateFull cands
+    IsEmpty r t         -> IsEmpty r <$> instantiateFull t
 
 instance InstantiateFull Elim where
   instantiateFull (Apply v) = Apply <$> instantiateFull v
@@ -657,6 +730,9 @@ instance InstantiateFull Elim where
 instance (Ord k, InstantiateFull e) => InstantiateFull (Map k e) where
     instantiateFull = traverse instantiateFull
 
+instance (Eq k, Hashable k, InstantiateFull e) => InstantiateFull (HashMap k e) where
+    instantiateFull = traverse instantiateFull
+
 instance InstantiateFull ModuleName where
     instantiateFull = return
 
@@ -669,7 +745,7 @@ instance InstantiateFull Signature where
 instance InstantiateFull Section where
   instantiateFull (Section tel n) = flip Section n <$> instantiateFull tel
 
-instance (Raise a, InstantiateFull a) => InstantiateFull (Tele a) where
+instance (Subst a, InstantiateFull a) => InstantiateFull (Tele a) where
   instantiateFull EmptyTel = return EmptyTel
   instantiateFull (ExtendTel a b) = uncurry ExtendTel <$> instantiateFull (a, b)
 
@@ -677,9 +753,9 @@ instance InstantiateFull Char where
     instantiateFull = return
 
 instance InstantiateFull Definition where
-    instantiateFull (Defn rel x t df i c d) = do
+    instantiateFull (Defn rel x t pol occ df i c d) = do
       (t, (df, d)) <- instantiateFull (t, (df, d))
-      return $ Defn rel x t df i c d
+      return $ Defn rel x t pol occ df i c d
 
 instance InstantiateFull a => InstantiateFull (Open a) where
   instantiateFull (OpenThing n a) = OpenThing n <$> instantiateFull a
@@ -718,6 +794,9 @@ instance InstantiateFull FunctionInverse where
   instantiateFull NotInjective = return NotInjective
   instantiateFull (Inverse inv) = Inverse <$> instantiateFull inv
 
+instance InstantiateFull a => InstantiateFull (WithArity a) where
+  instantiateFull (WithArity n a) = WithArity n <$> instantiateFull a
+
 instance InstantiateFull a => InstantiateFull (Case a) where
   instantiateFull (Branches cs ls m) =
     Branches <$> instantiateFull cs
@@ -738,28 +817,36 @@ instance InstantiateFull Clause where
 
 instance InstantiateFull Interface where
     instantiateFull (Interface ms mod scope inside
-                               sig b hsImports highlighting pragmas) =
+                               sig b hsImports highlighting pragmas patsyns) =
 	Interface ms mod scope inside
 	    <$> instantiateFull sig
 	    <*> instantiateFull b
             <*> return hsImports
             <*> return highlighting
             <*> return pragmas
+            <*> return patsyns
 
 instance InstantiateFull a => InstantiateFull (Builtin a) where
     instantiateFull (Builtin t) = Builtin <$> instantiateFull t
     instantiateFull (Prim x)	= Prim <$> instantiateFull x
 
+instance InstantiateFull QName where
+  instantiateFull = return
+
 instance InstantiateFull a => InstantiateFull (Maybe a) where
   instantiateFull = mapM instantiateFull
 
+
+{- DUPLICATE of Telescope.telView
+
 -- | @telViewM t@ is like @telView' t@, but it reduces @t@ to expose
 --   function type constructors.
 telViewM :: Type -> TCM TelView
 telViewM t = do
   t <- reduce t -- also instantiates meta if in head position
-  case unEl t of
+  case ignoreSharing $ unEl t of
     Pi a b -> absV a (absName b) <$> telViewM (absBody b)
     _      -> return $ TelV EmptyTel t
   where
     absV a x (TelV tel t) = TelV (ExtendTel a (Abs x tel)) t
+-}
