@@ -20,7 +20,8 @@ import Agda.TypeChecking.Primitive
 import Agda.TypeChecking.Constraints
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Pretty
-
+import Agda.TypeChecking.Irrelevance
+import Agda.TypeChecking.SizedTypes ( builtinSizeHook )
 import Agda.TypeChecking.Rules.Term ( checkExpr , inferExpr )
 import {-# SOURCE #-} Agda.TypeChecking.Rules.Builtin.Coinduction
 
@@ -33,19 +34,22 @@ import Agda.Utils.Impossible
 -- * Checking builtin pragmas
 ---------------------------------------------------------------------------
 
+builtinPostulate :: TCM Type -> BuiltinDescriptor
+builtinPostulate = BuiltinPostulate Relevant
+
 coreBuiltins :: [BuiltinInfo]
 coreBuiltins = map (\(x,z) -> BuiltinInfo x z)
   [ (builtinList               |-> BuiltinData (tset --> tset) [builtinNil, builtinCons])
   , (builtinArg                |-> BuiltinData (tset --> tset) [builtinArgArg])
   , (builtinBool               |-> BuiltinData tset [builtinTrue, builtinFalse])
   , (builtinNat                |-> BuiltinData tset [builtinZero, builtinSuc])
-  , (builtinLevel              |-> BuiltinPostulate tset)
-  , (builtinInteger            |-> BuiltinPostulate tset)
-  , (builtinFloat              |-> BuiltinPostulate tset)
-  , (builtinChar               |-> BuiltinPostulate tset)
-  , (builtinString             |-> BuiltinPostulate tset)
-  , (builtinQName              |-> BuiltinPostulate tset)
-  , (builtinIO                 |-> BuiltinPostulate (tset --> tset))
+  , (builtinLevel              |-> builtinPostulate tset)
+  , (builtinInteger            |-> builtinPostulate tset)
+  , (builtinFloat              |-> builtinPostulate tset)
+  , (builtinChar               |-> builtinPostulate tset)
+  , (builtinString             |-> builtinPostulate tset)
+  , (builtinQName              |-> builtinPostulate tset)
+  , (builtinIO                 |-> builtinPostulate (tset --> tset))
   , (builtinAgdaSort           |-> BuiltinData tset [builtinAgdaSortSet, builtinAgdaSortLit, builtinAgdaSortUnsupported])
   , (builtinAgdaType           |-> BuiltinData tset [builtinAgdaTypeEl])
   , (builtinAgdaTerm           |-> BuiltinData tset
@@ -55,15 +59,15 @@ coreBuiltins = map (\(x,z) -> BuiltinInfo x z)
                                      ,builtinAgdaTermUnsupported])
   , (builtinEquality           |-> BuiltinData (hPi "a" (el primLevel) $
                                                 hPi "A" (return $ sort $ varSort 0) $
-                                                (El (varSort 1) <$> var 0) -->
-                                                (El (varSort 1) <$> var 0) -->
+                                                (El (varSort 1) <$> varM 0) -->
+                                                (El (varSort 1) <$> varM 0) -->
                                                 return (sort $ varSort 1))
                                                [builtinRefl])
   , (builtinHiding             |-> BuiltinData tset [builtinHidden, builtinInstance, builtinVisible])
   , (builtinRelevance          |-> BuiltinData tset [builtinRelevant, builtinIrrelevant])
   , (builtinRefl               |-> BuiltinDataCons (hPi "a" (el primLevel) $ hPi "A" (return $ sort $ varSort 0) $
-                                                    hPi "x" (El (varSort 1) <$> var 0) $
-                                                    El (varSort 2) <$> primEquality <#> var 2 <#> var 1 <@> var 0 <@> var 0))
+                                                    hPi "x" (El (varSort 1) <$> varM 0) $
+                                                    El (varSort 2) <$> primEquality <#> varM 2 <#> varM 1 <@> varM 0 <@> varM 0))
   , (builtinNil                |-> BuiltinDataCons (hPi "A" tset (el (list v0))))
   , (builtinCons               |-> BuiltinDataCons (hPi "A" tset (tv0 --> el (list v0) --> el (list v0))))
   , (builtinZero               |-> BuiltinDataCons tnat)
@@ -84,9 +88,17 @@ coreBuiltins = map (\(x,z) -> BuiltinInfo x z)
   , (builtinVisible            |-> BuiltinDataCons thiding)
   , (builtinRelevant           |-> BuiltinDataCons trelevance)
   , (builtinIrrelevant         |-> BuiltinDataCons trelevance)
-  , (builtinSize               |-> BuiltinPostulate tset)
-  , (builtinSizeSuc            |-> BuiltinPostulate (tsize --> tsize))
-  , (builtinSizeInf            |-> BuiltinPostulate tsize)
+  , (builtinSize               |-> builtinPostulate tset)
+  , (builtinSizeLt             |-> builtinPostulate (tsize --> tset))
+  , (builtinSizeSuc            |-> builtinPostulate (tsize --> tsize))
+  , (builtinSizeInf            |-> builtinPostulate tsize)
+  -- postulate max : {i : Size} -> Size< i -> Size< i -> Size< i
+  , (builtinSizeMax            |-> builtinPostulate (tsize --> tsize --> tsize))
+     -- (hPi "i" tsize $ let a = el $ primSizeLt <@> v0 in (a --> a --> a)))
+  -- postulate .irrelevant : {a : Level}{A : Set a} -> .A -> A
+  , (builtinIrrAxiom           |-> BuiltinPostulate Irrelevant
+                                     (hPi "a" (el primLevel) $ hPi "A" (return $ sort $ varSort 0) $
+                                      (El (varSort 1) <$> varM 0) .--> (El (varSort 1) <$> varM 0)))
   , (builtinAgdaSortSet        |-> BuiltinDataCons (tterm --> tsort))
   , (builtinAgdaSortLit        |-> BuiltinDataCons (tnat --> tsort))
   , (builtinAgdaSortUnsupported|-> BuiltinDataCons tsort)
@@ -100,9 +112,9 @@ coreBuiltins = map (\(x,z) -> BuiltinInfo x z)
   , (builtinLevelZero          |-> BuiltinPrim "primLevelZero" (const $ return ()))
   , (builtinLevelSuc           |-> BuiltinPrim "primLevelSuc" (const $ return ()))
   , (builtinLevelMax           |-> BuiltinPrim "primLevelMax" verifyMax)
-  , (builtinAgdaFunDef                |-> BuiltinPostulate tset) -- internally this is QName
-  , (builtinAgdaDataDef               |-> BuiltinPostulate tset) -- internally this is QName
-  , (builtinAgdaRecordDef             |-> BuiltinPostulate tset) -- internally this is QName
+  , (builtinAgdaFunDef                |-> builtinPostulate tset) -- internally this is QName
+  , (builtinAgdaDataDef               |-> builtinPostulate tset) -- internally this is QName
+  , (builtinAgdaRecordDef             |-> builtinPostulate tset) -- internally this is QName
   , (builtinAgdaDefinition            |-> BuiltinData tset [builtinAgdaDefinitionFunDef
                                                            ,builtinAgdaDefinitionDataDef
                                                            ,builtinAgdaDefinitionDataConstructor
@@ -119,8 +131,8 @@ coreBuiltins = map (\(x,z) -> BuiltinInfo x z)
   where
         (|->) = (,)
 
-        v0 = var 0
-        v1 = var 1
+        v0 = varM 0
+        v1 = varM 1
 
         tv0,tv1 :: TCM Type
         tv0 = el v0
@@ -137,7 +149,7 @@ coreBuiltins = map (\(x,z) -> BuiltinInfo x z)
         tsize      = el primSize
         tbool      = el primBool
         thiding    = el primHiding
-        trelevance = el primRelvance
+        trelevance = el primRelevance
         ttype      = el primAgdaType
         tsort      = el primAgdaSort
         tdefn      = el primAgdaDefinition
@@ -147,8 +159,8 @@ coreBuiltins = map (\(x,z) -> BuiltinInfo x z)
 
         verifyPlus plus =
             verify ["n","m"] $ \(@@) zero suc (==) choice -> do
-                let m = Var 0 []
-                    n = Var 1 []
+                let m = var 0
+                    n = var 1
                     x + y = plus @@ x @@ y
 
                 -- We allow recursion on any argument
@@ -161,8 +173,8 @@ coreBuiltins = map (\(x,z) -> BuiltinInfo x z)
 
         verifyMinus minus =
             verify ["n","m"] $ \(@@) zero suc (==) choice -> do
-                let m = Var 0 []
-                    n = Var 1 []
+                let m = var 0
+                    n = var 1
                     x - y = minus @@ x @@ y
 
                 -- We allow recursion on any argument
@@ -174,8 +186,8 @@ coreBuiltins = map (\(x,z) -> BuiltinInfo x z)
         verifyTimes times = do
             plus <- primNatPlus
             verify ["n","m"] $ \(@@) zero suc (==) choice -> do
-                let m = Var 0 []
-                    n = Var 1 []
+                let m = var 0
+                    n = var 1
                     x + y = plus  @@ x @@ y
                     x * y = times @@ x @@ y
 
@@ -193,10 +205,10 @@ coreBuiltins = map (\(x,z) -> BuiltinInfo x z)
         verifyDivSucAux dsAux =
             verify ["k","m","n","j"] $ \(@@) zero suc (==) choice -> do
                 let aux k m n j = dsAux @@ k @@ m @@ n @@ j
-                    k           = Var 0 []
-                    m           = Var 1 []
-                    n           = Var 2 []
-                    j           = Var 3 []
+                    k           = var 0
+                    m           = var 1
+                    n           = var 2
+                    j           = var 3
 
                 aux k m zero    j       == k
                 aux k m (suc n) zero    == aux (suc k) m n m
@@ -205,10 +217,10 @@ coreBuiltins = map (\(x,z) -> BuiltinInfo x z)
         verifyModSucAux dsAux =
             verify ["k","m","n","j"] $ \(@@) zero suc (==) choice -> do
                 let aux k m n j = dsAux @@ k @@ m @@ n @@ j
-                    k           = Var 0 []
-                    m           = Var 1 []
-                    n           = Var 2 []
-                    j           = Var 3 []
+                    k           = var 0
+                    m           = var 1
+                    n           = var 2
+                    j           = var 3
 
                 aux k m zero    j       == k
                 aux k m (suc n) zero    == aux zero m n m
@@ -219,8 +231,8 @@ coreBuiltins = map (\(x,z) -> BuiltinInfo x z)
             true  <- primTrue
             false <- primFalse
             let x == y = eq @@ x @@ y
-                m      = Var 0 []
-                n      = Var 1 []
+                m      = var 0
+                n      = var 1
             (zero  == zero ) === true
             (suc n == suc m) === (n == m)
             (suc n == zero ) === false
@@ -231,8 +243,8 @@ coreBuiltins = map (\(x,z) -> BuiltinInfo x z)
             true  <- primTrue
             false <- primFalse
             let x < y = leq @@ x @@ y
-                m     = Var 0 []
-                n     = Var 1 []
+                m     = var 0
+                n     = var 1
             (n     < zero)  === false
             (suc n < suc m) === (n < m)
             (zero  < suc m) === true
@@ -254,14 +266,14 @@ coreBuiltins = map (\(x,z) -> BuiltinInfo x z)
                 suc n  = s @@ n
                 choice = foldr1 (\x y -> x `catchError` \_ -> y)
             xs <- mapM freshName_ xs
-            addCtxs xs (defaultArg nat) $ f (@@) zero suc (==) choice
+            addCtxs xs (domFromArg $ defaultArg nat) $ f (@@) zero suc (==) choice
 
 
 inductiveCheck :: String -> Int -> Term -> TCM ()
 inductiveCheck b n t = do
     t <- etaContract =<< normalise t
     let err = typeError (NotInductive t)
-    case t of
+    case ignoreSharing t of
       Def t _ -> do
         t <- theDef <$> getConstInfo t
         case t of
@@ -313,8 +325,8 @@ bindBuiltinInfo (BuiltinInfo s d) e = do
                              level <- getBuiltin' builtinLevel
                              case (nat, level) of
                                (Just nat, Just level) -> do
-                                  Def nat   _ <- normalise nat
-                                  Def level _ <- normalise level
+                                  Def nat   _ <- ignoreSharing <$> normalise nat
+                                  Def level _ <- ignoreSharing <$> normalise level
                                   when (nat == level) $ typeError $ GenericError $
                                     builtinNat ++ " and " ++ builtinLevel ++
                                     " have to be different types."
@@ -323,9 +335,10 @@ bindBuiltinInfo (BuiltinInfo s d) e = do
 
       BuiltinDataCons t -> do
 
-        let name (Lam h b) = name (absBody b)
-            name (Con c _) = Con c []
-            name _         = __IMPOSSIBLE__
+        let name (Lam h b)  = name (absBody b)
+            name (Con c _)  = Con c []
+            name (Shared p) = name $ ignoreSharing (derefPtr p)
+            name _          = __IMPOSSIBLE__
 
         e' <- checkExpr e =<< t
 
@@ -356,15 +369,18 @@ bindBuiltinInfo (BuiltinInfo s d) e = do
 
 	  _ -> typeError $ GenericError $ "Builtin " ++ s ++ " must be bound to a function"
 
-      BuiltinPostulate t -> do
-        e' <- checkExpr e =<< t
+      BuiltinPostulate rel t -> do
+        t' <- t
+        e' <- applyRelevanceToContext rel $ checkExpr e t'
         let err = typeError $ GenericError $
                     "The argument to BUILTIN " ++ s ++ " must be a postulated name"
         case e of
           A.Def q -> do
             def <- ignoreAbstractMode $ getConstInfo q
             case theDef def of
-              Axiom {} -> bindBuiltinName s e'
+              Axiom {} -> do
+                builtinSizeHook s q e' t'
+                bindBuiltinName s e'
               _        -> err
           _ -> err
 
