@@ -7,7 +7,6 @@
 module Agda.TypeChecking.Rules.Builtin.Coinduction where
 
 import Control.Applicative
-import Control.Monad.Error
 
 import qualified Data.Map as Map
 
@@ -31,27 +30,27 @@ import Agda.Utils.Permutation
 
 #include "../../../undefined.h"
 
--- | The type of @&#x221e@.
+-- | The type of @∞@.
 
 typeOfInf :: TCM Type
 typeOfInf = hPi "a" (el primLevel) $
             (return . sort $ varSort 0) --> (return . sort $ varSort 0)
 
--- | The type of @&#x266f_@.
+-- | The type of @♯_@.
 
 typeOfSharp :: TCM Type
 typeOfSharp = hPi "a" (el primLevel) $
               hPi "A" (return . sort $ varSort 0) $
-              (El (varSort 1) <$> var 0) -->
-              (El (varSort 1) <$> primInf <#> var 1 <@> var 0)
+              (El (varSort 1) <$> varM 0) -->
+              (El (varSort 1) <$> primInf <#> varM 1 <@> varM 0)
 
--- | The type of @&#x266d@.
+-- | The type of @♭@.
 
 typeOfFlat :: TCM Type
 typeOfFlat = hPi "a" (el primLevel) $
              hPi "A" (return . sort $ varSort 0) $
-             (El (varSort 1) <$> primInf <#> var 1 <@> var 0) -->
-             (El (varSort 1) <$> var 0)
+             (El (varSort 1) <$> primInf <#> varM 1 <@> varM 0) -->
+             (El (varSort 1) <$> varM 0)
 
 -- | Binds the INFINITY builtin, but does not change the type's
 -- definition.
@@ -65,27 +64,30 @@ bindBuiltinInf e = bindPostulatedName builtinInf e $ \inf _ ->
 
 -- The following (no longer supported) definition is used:
 --
--- codata &#x221e {a} (A : Set a) : Set a where
---   &#x266f_ : (x : A) → &#x221e A
+-- codata ∞ {a} (A : Set a) : Set a where
+--   ♯_ : (x : A) → ∞ A
 
 bindBuiltinSharp :: A.Expr -> TCM ()
 bindBuiltinSharp e =
   bindPostulatedName builtinSharp e $ \sharp sharpDefn -> do
-    sharpE  <- instantiateFull =<< checkExpr (A.Def sharp) =<< typeOfSharp
-    inf     <- primInf
-    infDefn <- case inf of
-      Def inf _ -> getConstInfo inf
-      _         -> __IMPOSSIBLE__
+    sharpE    <- instantiateFull =<< checkExpr (A.Def sharp) =<< typeOfSharp
+    Def inf _ <- ignoreSharing <$> primInf
+    infDefn   <- getConstInfo inf
     addConstant (defName infDefn) $
-      infDefn { theDef = Datatype
+      infDefn { defPolarity       = [] -- not monotone
+              , defArgOccurrences = [Unused, StrictPos]
+              , theDef = Datatype
                   { dataPars           = 2
                   , dataIxs            = 0
                   , dataInduction      = CoInductive
                   , dataClause         = Nothing
                   , dataCons           = [sharp]
                   , dataSort           = varSort 1
+{-
                   , dataPolarity       = [Invariant, Invariant]
-                  , dataArgOccurrences = [Unused, Positive]
+                  , dataArgOccurrences = [Unused, StrictPos]
+-}
+                  , dataMutual         = []
                   , dataAbstr          = ConcreteDef
                   }
               }
@@ -104,74 +106,53 @@ bindBuiltinSharp e =
 
 -- The following (no longer supported) definition is used:
 --
---   &#x266d : ∀ {a} {A : Set a} → &#x221e A → A
---   &#x266d (&#x266f x) = x
+--   ♭ : ∀ {a} {A : Set a} → ∞ A → A
+--   ♭ (♯ x) = x
 
 bindBuiltinFlat :: A.Expr -> TCM ()
 bindBuiltinFlat e =
   bindPostulatedName builtinFlat e $ \flat flatDefn -> do
-    flatE  <- instantiateFull =<< checkExpr (A.Def flat) =<< typeOfFlat
-    sharp  <- (\t -> case t of
-                 Def q _ -> q
-                 _       -> __IMPOSSIBLE__)
-              <$> primSharp
-    kit    <- requireLevels
+    flatE       <- instantiateFull =<< checkExpr (A.Def flat) =<< typeOfFlat
+    Def sharp _ <- ignoreSharing <$> primSharp
+    kit         <- requireLevels
+    Def inf _   <- ignoreSharing <$> primInf
     let clause = Clause { clauseRange = noRange
-                        , clauseTel   = ExtendTel (argH (El (mkType 0) (Def (typeName kit) [])))
-                                                  (Abs "a" (ExtendTel (argH $ sort $ varSort 0)
-                                                                      (Abs "A" (ExtendTel (argN (El (varSort 1) (Var 0 [])))
+                        , clauseTel   = ExtendTel (domH (El (mkType 0) (Def (typeName kit) [])))
+                                                  (Abs "a" (ExtendTel (domH $ sort $ varSort 0)
+                                                                      (Abs "A" (ExtendTel (domN (El (varSort 1) (var 0)))
                                                                                           (Abs "x" EmptyTel)))))
-                        , clausePerm  = idP 3
-                        , clausePats  = [ argH (VarP "a")
-                                        , argH (VarP "A")
-                                        , argN (ConP sharp Nothing [argN (VarP "x")])
+                        , clausePerm  = idP 1
+                        , clausePats  = [ argN (ConP sharp Nothing [argN (VarP "x")])
                                         ]
-                        , clauseBody  = Bind $ Abs "h0" $
-                                        Bind $ Abs "h1" $
-                                        Bind $ Abs "h2" $ Body (Var 0 [])
+                        , clauseBody  = Bind $ Abs "x" $ Body (var 0)
                         }
     addConstant flat $
-      flatDefn { theDef = Function
+      flatDefn { defPolarity = []
+               , defArgOccurrences = [StrictPos]
+               , theDef = Function
                    { funClauses        = [clause]
                    , funCompiled       =
                       let hid   = Arg Hidden Relevant
                           nohid = defaultArg in
-                      Case 2 (Branches (Map.singleton sharp
-                                 (Done [hid "a", hid "A", nohid "x"] (Var 0 [])))
+                      Case 0 (Branches (Map.singleton sharp
+                                 (WithArity 1 (Done [nohid "x"] (var 0))))
                                Map.empty
                                Nothing)
                    , funInv            = NotInjective
-                   , funPolarity       = [Invariant, Invariant, Invariant]
-                   , funArgOccurrences = [Unused, Unused, Positive]
+{-
+                   , funPolarity       = [Invariant]
+                   , funArgOccurrences = [StrictPos] -- changing that to [Mixed] destroys monotonicity of 'Rec' in test/succeed/GuardednessPreservingTypeConstructors
+-}
+                   , funMutual         = []
                    , funAbstr          = ConcreteDef
                    , funDelayed        = NotDelayed
-                   , funProjection     = Nothing
-                     {- flat is a projection, but in the termination checker
-                        it destroys the (inductive) structural ordering.
-                        Thus, we do not register it as a projection. -}
+                   , funProjection     = Just (inf, 3)
                    , funStatic         = False
+                   , funCopy           = False
+                   , funTerminates     = Just True
                    }
                 }
     return flatE
 
--- | The coinductive primitives.
-
-data CoinductionKit = CoinductionKit
-  { nameOfInf   :: QName
-  , nameOfSharp :: QName
-  , nameOfFlat  :: QName
-  }
-
--- | Tries to build a 'CoinductionKit'.
-
-coinductionKit :: TCM (Maybe CoinductionKit)
-coinductionKit = (do
-  Def inf   _ <- primInf
-  Def sharp _ <- primSharp
-  Def flat  _ <- primFlat
-  return $ Just $ CoinductionKit
-    { nameOfInf   = inf
-    , nameOfSharp = sharp
-    , nameOfFlat  = flat
-    })
-    `catchError` \_ -> return Nothing
+-- The coinductive primitives.
+-- moved to TypeChecking.Monad.Builtin
