@@ -6,6 +6,8 @@ import Control.Applicative
 import Control.Monad
 import Control.Monad.Trans
 
+import Data.List (genericTake)
+
 import qualified Agda.Syntax.Abstract as A
 import Agda.Syntax.Internal
 import Agda.Syntax.Common
@@ -25,10 +27,11 @@ import Agda.TypeChecking.Forcing
 import Agda.TypeChecking.Irrelevance
 import Agda.TypeChecking.Telescope
 
-import Agda.TypeChecking.Rules.Term ( isType_ )
+import {-# SOURCE #-} Agda.TypeChecking.Rules.Term ( isType_ )
 
 import Agda.Interaction.Options
 
+import Agda.Utils.List
 import Agda.Utils.Monad
 import Agda.Utils.Size
 import Agda.Utils.Tuple
@@ -69,6 +72,7 @@ checkDataDef i name ps cs =
 
 	    -- The type we get from bindParameters is Θ -> s where Θ is the type of
 	    -- the indices. We count the number of indices and return s.
+            -- We check that s is a sort.
             (nofIxs, s) <- splitType t0
 
             when (any (`freeIn` s) [0..nofIxs - 1]) $ do
@@ -99,12 +103,16 @@ checkDataDef i name ps cs =
                                    , dataCons           = []     -- Constructors are added later
 				   , dataSort           = s
                                    , dataAbstr          = Info.defAbstract i
+{-
+                                   -- determined by the positivity checker:
                                    , dataPolarity       = []
                                    , dataArgOccurrences = []
+-}
+                                   , dataMutual         = []
                                    }
 
 	    escapeContext (size tel) $ do
-	      addConstant name ( Defn Relevant name t (defaultDisplayForm name) 0 noCompiledRep dataDef )
+	      addConstant name ( Defn Relevant name t [] [] (defaultDisplayForm name) 0 noCompiledRep dataDef )
 
 	    -- Check the types of the constructors
 	    mapM_ (checkConstructor name tel' nofIxs s) cs
@@ -128,10 +136,11 @@ checkDataDef i name ps cs =
 
 	-- Add the datatype to the signature with its constructors. It was previously
 	-- added without them.
-	addConstant name (Defn Relevant name t (defaultDisplayForm name) 0 noCompiledRep $
+	addConstant name (Defn Relevant name t [] [] (defaultDisplayForm name) 0 noCompiledRep $
                             dataDef { dataCons = map cname cs }
 			 )
-        computePolarity name
+        -- Andreas 2012-02-13: postpone polarity computation until after positivity check
+        -- computePolarity name
     where
 	cname (A.ScopedDecl _ [d]) = cname d
 	cname (A.Axiom _ _ x _)	   = x
@@ -140,12 +149,14 @@ checkDataDef i name ps cs =
 	hideTel  EmptyTel		  = EmptyTel
 	hideTel (ExtendTel a tel) = ExtendTel (hideAndRelParams a) $ hideTel <$> tel
 
-	splitType (El _ (Pi _ b))  = ((+ 1) -*- id) <$> splitType (absBody b)
-	splitType (El _ (Sort s))  = return (0, s)
-	splitType a                = do
-          s <- newSortMeta
-          equalType a (sort s)
-          return (0, s)
+        splitType :: Type -> TCM (Int, Sort)
+        splitType t = case ignoreSharing $ unEl t of
+          Pi a b -> ((+ 1) -*- id) <$> do addCtxString (absName b) a $ splitType (absBody b)
+          Sort s -> return (0, s)
+	  _      -> do
+            s <- newSortMeta
+            equalType t (sort s)
+            return (0, s)
 
 -- | Type check a constructor declaration. Checks that the constructor targets
 --   the datatype and that it fits inside the declared sort.
@@ -182,7 +193,7 @@ checkConstructor d tel nofIxs s con@(A.Axiom i _ c e) =
         -- add parameters to constructor type and put into signature
         escapeContext (size tel)
 	    $ addConstant c
-	    $ Defn Relevant c (telePi tel t') (defaultDisplayForm c) 0 noCompiledRep
+	    $ Defn Relevant c (telePi tel t') [] [] (defaultDisplayForm c) 0 noCompiledRep
 	    $ Constructor (size tel) c d (Info.defAbstract i) Inductive
   where
     debugEnter c e =
@@ -217,15 +228,19 @@ bindParameters (A.DomainFull (A.TypedBindings _ (Arg h rel (A.TBind _ xs _))) :
 bindParameters (A.DomainFull (A.TypedBindings _ (Arg h rel (A.TNoBind _))) : bs) a ret = do
   x <- freshNoName_
   bindParameters (A.DomainFree h rel x : bs) a ret
-bindParameters (A.DomainFree h rel x : ps) (El _ (Pi (Arg h' rel' a) b)) ret
+bindParameters ps0@(A.DomainFree h rel x : ps) (El _ (Pi arg@(Dom h' rel' a) b)) ret
   -- Andreas, 2011-04-07 ignore relevance information in binding?!
     | h /= h' =
 	__IMPOSSIBLE__
     | otherwise = addCtx x arg $ bindParameters ps (absBody b) $ \tel s ->
 		    ret (ExtendTel arg $ Abs (show x) tel) s
-  where
-    arg = Arg h rel' a
-bindParameters _ _ _ = __IMPOSSIBLE__
+bindParameters bs (El s (Shared p)) ret = bindParameters bs (El s $ derefPtr p) ret
+bindParameters (b : bs) t _ = __IMPOSSIBLE__
+{- Andreas, 2012-01-17 Concrete.Definitions ensures number and hiding of parameters to be correct
+-- Andreas, 2012-01-13 due to separation of data declaration/definition, the user
+-- could give more parameters than declared.
+bindParameters (b : bs) t _ = typeError $ DataTooManyParameters
+-}
 
 
 -- | Check that the arguments to a constructor fits inside the sort of the datatype.
@@ -242,12 +257,12 @@ fitsIn t s = do
   -- The line below would be simpler, but doesn't allow datatypes
   -- to be indexed by the universe level.
 --   noConstraints $ s' `leqSort` s
-  case unEl t of
-    Pi arg@(Arg h r a) _ -> do
+  case ignoreSharing $ unEl t of
+    Pi arg@(Dom h r a) _ -> do
       let s' = getSort a
       s' `leqSort` s
       x <- freshName_ (argName t)
-      let v  = Arg h r $ Var 0 []
+      let v  = Arg h r $ var 0
           t' = piApply (raise 1 t) [v]
       addCtx x arg $ fitsIn t' (raise 1 s)
     _		     -> return ()
@@ -264,7 +279,7 @@ constructs nofPars t q = constrT 0 t
         constr :: Nat -> Sort -> Term -> TCM ()
 	constr n s v = do
 	    v <- reduce v
-	    case v of
+	    case ignoreSharing v of
 		Pi _ (NoAbs _ b) -> constrT n b
 		Pi a b	-> underAbstraction a b $ \t ->
 			   constrT (n + 1) t
@@ -277,21 +292,24 @@ constructs nofPars t q = constrT 0 t
 
 	checkParams n vs = zipWithM_ sameVar vs ps
 	    where
-		ps = reverse [ i | (i,_) <- zip [n..] vs ]
+--		ps = reverse [ i | (i,_) <- zip [n..] vs ]
+                nvs = size vs
+		ps = genericTake nvs $ downFrom (n + nvs)
 
 		sameVar arg i
                   -- skip irrelevant parameters
                   | argRelevance arg == Irrelevant = return ()
 		  | otherwise = do
 		    t <- typeOfBV i
-		    equalTerm t (unArg arg) (Var i [])
+		    equalTerm t (unArg arg) (var i)
 
+{- UNUSED, Andreas 2012-09-13
 -- | Force a type to be a specific datatype.
 forceData :: QName -> Type -> TCM Type
 forceData d (El s0 t) = liftTCM $ do
     t' <- reduce t
     d  <- canonicalName d
-    case t' of
+    case ignoreSharing t' of
 	Def d' _
 	    | d == d'   -> return $ El s0 t'
 	    | otherwise	-> fail $ "wrong datatype " ++ show d ++ " != " ++ show d'
@@ -301,14 +319,15 @@ forceData d (El s0 t) = liftTCM $ do
 	    noConstraints $ leqType (El s0 t') (El s (Def d ps)) -- TODO: need equalType?
 	    reduce $ El s0 t'
 	_ -> typeError $ ShouldBeApplicationOf (El s0 t) d
+-}
 
 -- | Is the type coinductive? Returns 'Nothing' if the answer cannot
 -- be determined.
 
 isCoinductive :: Type -> TCM (Maybe Bool)
 isCoinductive t = do
-  El _ t <- normalise t
-  case t of
+  El s t <- normalise t
+  case ignoreSharing t of
     Def q _ -> do
       def <- getConstInfo q
       case theDef def of
@@ -327,4 +346,5 @@ isCoinductive t = do
     Pi    {} -> return (Just False)
     Sort  {} -> return (Just False)
     MetaV {} -> return Nothing
+    Shared{} -> __IMPOSSIBLE__
     DontCare{} -> __IMPOSSIBLE__
