@@ -1,12 +1,21 @@
-{-# LANGUAGE CPP #-}
+{-# LANGUAGE CPP, TupleSections #-}
 
 module Agda.TypeChecking.Rules.Decl where
 
 import Control.Monad
+import Control.Monad.Error
+import Control.Monad.Reader
 import Control.Monad.Trans
+import Control.Monad.State (modify)
+
+import qualified Data.Foldable as Fold
+import Data.Maybe
 import qualified Data.Map as Map
 import Data.Map (Map)
-import qualified Agda.Utils.IO.Locale as LocIO
+import qualified Data.List as List
+import qualified Data.Set as Set
+import Data.Set (Set)
+import Data.Sequence ((|>))
 
 import Agda.Syntax.Abstract (AnyAbstract(..))
 import qualified Agda.Syntax.Abstract as A
@@ -15,19 +24,24 @@ import qualified Agda.Syntax.Info as Info
 import Agda.Syntax.Position
 import Agda.Syntax.Common
 
+import Agda.TypeChecking.Errors
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Monad.Builtin
 import Agda.TypeChecking.Monad.Mutual
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Constraints
 import Agda.TypeChecking.Positivity
+import Agda.TypeChecking.Polarity
 import Agda.TypeChecking.Primitive hiding (Nat)
+import Agda.TypeChecking.ProjectionLike
 import Agda.TypeChecking.Conversion
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.SizedTypes
 import Agda.TypeChecking.Telescope
 import Agda.TypeChecking.Forcing
+import Agda.TypeChecking.Injectivity
+import Agda.TypeChecking.InstanceArguments (solveIrrelevantMetas)
 
 import Agda.TypeChecking.Rules.Term
 import Agda.TypeChecking.Rules.Data    ( checkDataDef )
@@ -39,6 +53,13 @@ import Agda.Compiler.HaskellTypes
 
 import Agda.Utils.Size
 import Agda.Utils.Monad
+import qualified Agda.Utils.HashMap as HMap
+-- import Agda.Utils.NubList -- reverted
+
+import Agda.Interaction.Highlighting.Generate
+
+import Agda.Termination.TermCheck
+import Agda.Interaction.Options
 
 #include "../../undefined.h"
 import Agda.Utils.Impossible
@@ -51,64 +72,193 @@ checkDecls ds = do
   -- whenM onTopLevel unfreezeMetas
 
 -- | Type check a single declaration.
+
 checkDecl :: A.Declaration -> TCM ()
 checkDecl d = do
     -- Issue 418 fix: freeze metas before checking an abstract things
     when isAbstract freezeMetas
-    leaveTopLevelConditionally d $ case d of
-	A.Axiom i rel x e        -> checkAxiom i rel x e
-        A.Field{}                -> typeError FieldOutsideRecord
-	A.Primitive i x e        -> checkPrimitive i x e
-	A.Mutual i ds            -> checkMutual i ds
-	A.Section i x tel ds     -> checkSection i x tel ds
-	A.Apply i x modapp rd rm -> checkSectionApplication i x modapp rd rm
-	A.Import i x             -> checkImport i x
-	A.Pragma i p             -> checkPragma i p
-	A.ScopedDecl scope ds    -> setScope scope >> checkDecls ds
-        A.FunDef i x cs          -> check x i $ checkFunDef NotDelayed i x cs
-        A.DataDef i x ps cs      -> check x i $ checkDataDef i x ps cs
-        A.RecDef i x c ps tel cs -> check x i $ checkRecDef i x c ps tel cs
-        A.DataSig i x ps t       -> checkAxiom i Relevant x (A.Pi (Info.ExprRange (fuseRange ps t)) ps t)
-        A.RecSig i x ps t        -> checkAxiom i Relevant x (A.Pi (Info.ExprRange (fuseRange ps t)) ps t)
-        A.Open _ _               -> return ()
-    top <- onTopLevel
-    when top $ solveSizeConstraints >> freezeMetas
+
+    let -- What kind of final checks/computations should be performed
+        -- if we're not inside a mutual block?
+        none       m = m >> return Nothing
+        meta       m = m >> return (Just (return []))
+        mutual     m = m >>= return . Just . mutualChecks
+        impossible m = m >> return __IMPOSSIBLE__
+                       -- We're definitely inside a mutual block.
+
+    topLevelKind <- case d of
+      A.Axiom{}                -> meta $ checkTypeSignature d
+      A.Field{}                -> typeError FieldOutsideRecord
+      A.Primitive i x e        -> meta $ checkPrimitive i x e
+      A.Mutual i ds            -> mutual $ checkMutual i ds
+      A.Section i x tel ds     -> meta $ checkSection i x tel ds
+      A.Apply i x modapp rd rm -> meta $ checkSectionApplication i x modapp rd rm
+      A.Import i x             -> none $ checkImport i x
+      A.Pragma i p             -> none $ checkPragma i p
+      A.ScopedDecl scope ds    -> none $ setScope scope >> checkDecls ds
+      A.FunDef i x delayed cs  -> impossible $ check x i $ checkFunDef delayed i x cs
+      A.DataDef i x ps cs      -> impossible $ check x i $ checkDataDef i x ps cs
+      A.RecDef i x ind c ps tel cs -> mutual $ check x i $ do
+                                    checkRecDef i x ind c ps tel cs
+                                    return (Set.singleton x)
+      A.DataSig i x ps t       -> impossible $ checkSig i x ps t
+      A.RecSig i x ps t        -> none $ checkSig i x ps t
+                                  -- A record signature is always followed by a
+                                  -- record definition. Metas should not be
+                                  -- frozen until after the definition has been
+                                  -- checked. NOTE: Metas are not frozen
+                                  -- immediately after the last field. Perhaps
+                                  -- they should be (unless we're in a mutual
+                                  -- block).
+      A.Open _ _               -> none $ return ()
+                                  -- Open is just an artifact from the concrete
+                                  -- syntax, retained for highlighting purposes.
+
+    unlessM (isJust . envMutualBlock <$> ask) $ do
+      termErrs <- case topLevelKind of
+        Nothing           -> return []
+        Just mutualChecks -> do
+
+          solveSizeConstraints
+          solveIrrelevantMetas
+          wakeupConstraints_   -- solve emptyness constraints
+          freezeMetas
+
+          mutualChecks
+
+      -- Syntax highlighting.
+      let highlight d = generateAndPrintSyntaxInfo d (Full termErrs)
+      case d of
+        A.Axiom{}                -> highlight d
+        A.Field{}                -> __IMPOSSIBLE__
+        A.Primitive{}            -> highlight d
+        A.Mutual{}               -> highlight d
+        A.Apply{}                -> highlight d
+        A.Import{}               -> highlight d
+        A.Pragma{}               -> highlight d
+        A.ScopedDecl{}           -> return ()
+        A.FunDef{}               -> __IMPOSSIBLE__
+        A.DataDef{}              -> __IMPOSSIBLE__
+        A.DataSig{}              -> __IMPOSSIBLE__
+        A.Open{}                 -> highlight d
+        A.Section i x tel _      -> highlight (A.Section i x tel [])
+          -- Each block in the section has already been highlighted,
+          -- all that remains is the module declaration.
+        A.RecSig{}               -> highlight d
+        A.RecDef i x ind c ps tel cs ->
+          highlight (A.RecDef i x ind c [] tel (fields cs))
+          -- The telescope and all record module declarations except
+          -- for the fields have already been highlighted.
+          where
+          fields (A.ScopedDecl _ ds1 : ds2) = fields ds1 ++ fields ds2
+          fields (d@A.Field{}        : ds)  = d : fields ds
+          fields (_                  : ds)  = fields ds
+          fields []                         = []
+
     where
-        unScope (A.ScopedDecl scope ds) = setScope scope >> unScope d
-        unScope d = return d
+    unScope (A.ScopedDecl scope ds) = setScope scope >> unScope d
+    unScope d = return d
 
-        check x i m = do
-          reportSDoc "tc.decl" 5 $ text "Checking" <+> prettyTCM x <> text "."
-          r <- abstract (Info.defAbstract i) m
-          reportSDoc "tc.decl" 5 $ text "Checked" <+> prettyTCM x <> text "."
-          return r
+    checkSig i x ps t = checkTypeSignature $
+      A.Axiom i Relevant x (A.Pi (Info.ExprRange (fuseRange ps t)) ps t)
 
-        isAbstract = fmap Info.defAbstract (A.getDefInfo d) == Just AbstractDef
+    check x i m = do
+      reportSDoc "tc.decl" 5 $ text "Checking" <+> prettyTCM x <> text "."
+      r <- abstract (Info.defAbstract i) m
+      reportSDoc "tc.decl" 5 $ text "Checked" <+> prettyTCM x <> text "."
+      return r
 
-        -- Concrete definitions cannot use information about abstract things.
-        abstract ConcreteDef = inConcreteMode
-        abstract AbstractDef = inAbstractMode
-	    -- open is just an artifact from the concrete syntax
-            -- retained for highlighting purposes
+    isAbstract = fmap Info.defAbstract (A.getDefInfo d) == Just AbstractDef
 
-        leaveTopLevelConditionally d =
-          case d of
-            A.Section{}    -> id
-            A.ScopedDecl{} -> id
-            _              -> leaveTopLevel
+    -- Concrete definitions cannot use information about abstract things.
+    abstract ConcreteDef = inConcreteMode
+    abstract AbstractDef = inAbstractMode
+
+    -- Some checks that should be run at the end of a mutual
+    -- block (or non-mutual record declaration). The set names
+    -- contains the names defined in the mutual block.
+    mutualChecks names = do
+
+      checkStrictlyPositive names
+
+      -- Andreas, 2012-02-13: Polarity computation uses info from
+      -- positivity check, so it needs happen after positivity
+      -- check.
+      let -- | Do we need to compute polarity information for the
+          -- definition corresponding to the given name?
+          relevant q = do
+            def <- theDef <$> getConstInfo q
+            return $ case def of
+              Function{}    -> Just q
+              Datatype{}    -> Just q
+              Record{}      -> Just q
+              Axiom{}       -> Nothing
+              Constructor{} -> Nothing
+              Primitive{}   -> Nothing
+      mapM_ computePolarity =<<
+        (catMaybes <$> mapM relevant (Set.toList names))
+
+      -- Andreas, 2012-09-11:  Injectivity check stores clauses
+      -- whose 'Relevance' is affected by polarity computation,
+      -- so do it here.
+
+      let checkInj (q, def@Defn{ theDef = d@Function{ funClauses = cs }}) = do
+            inv <- checkInjectivity q cs
+            modifySignature $ updateDefinition q $ const $
+              def { theDef = d { funInv = inv }}
+          checkInj _ = return ()
+
+      namesDefs <- mapM (\ q -> (q,) <$> getConstInfo q) $ Set.toList names
+      mapM_ checkInj namesDefs
+
+      -- Non-mutual definitions can be considered for
+      -- projection likeness
+      case Set.toList names of
+        [d] -> do
+          def <- getConstInfo d
+          case theDef def of
+            Function{} -> makeProjection (defName def)
+            _          -> return ()
+        _ -> return ()
+
+      -- Termination checking.
+      termErrs <-
+        ifM (optTerminationCheck <$> pragmaOptions)
+          (disableDestructiveUpdate $ case d of
+             A.RecDef {} -> return []
+                            -- Record module definitions should not be
+                            -- termination-checked twice.
+             _           -> do
+               termErrs <- {- nubList <$> -} termDecl d
+               modify $ \st ->
+                 st { stTermErrs = Fold.foldl' (|>) (stTermErrs st) termErrs }
+               return termErrs)
+          (return [])
+
+      return termErrs
 
 -- | Type check an axiom.
 checkAxiom :: Info.DefInfo -> Relevance -> QName -> A.Expr -> TCM ()
-checkAxiom i rel x e = do
+checkAxiom i rel0 x e = do
+  -- Andreas, 2012-04-18  if we are in irrelevant context, axioms is irrelevant
+  -- even if not declared as such (Issue 610).
+  rel <- max rel0 <$> asks envRelevance
+  -- rel <- ifM ((Irrelevant ==) <$> asks envRelevance) (return Irrelevant) (return rel0)
   t <- isType_ e
   reportSDoc "tc.decl.ax" 10 $ sep
-    [ text "checked axiom"
-    , nest 2 $ prettyTCM rel <> prettyTCM x <+> text ":" <+> (prettyTCM =<< instantiateFull t)
+    [ text $ "checked type signature"
+    , nest 2 $ prettyTCM rel <> prettyTCM x <+> text ":" <+> prettyTCM t
     ]
   -- Not safe. See Issue 330
   -- t <- addForcingAnnotations t
-  addConstant x (Defn rel x t (defaultDisplayForm x) 0 noCompiledRep Axiom)
-  solveSizeConstraints
+  addConstant x (Defn rel x t [] [] (defaultDisplayForm x) 0 noCompiledRep Axiom)
+
+  -- for top-level axioms (postulates) try to solve irrelevant metas
+  -- when postulate $
+  maybe solveIrrelevantMetas (const $ return ()) =<< asks envMutualBlock
+
+  traceCall (IsType_ e) $ solveSizeConstraints  -- need Range for error message
+
   -- Andreas, 2011-05-31, that freezing below is probably wrong:
   -- when (Info.defAbstract i == AbstractDef) $ freezeMetas
 
@@ -116,12 +266,12 @@ checkAxiom i rel x e = do
 checkPrimitive :: Info.DefInfo -> QName -> A.Expr -> TCM ()
 checkPrimitive i x e =
     traceCall (CheckPrimitive (getRange i) (qnameName x) e) $ do  -- TODO!! (qnameName)
-    PrimImpl t' pf <- lookupPrimitiveFunction (nameString $ qnameName x)
+    (_, PrimImpl t' pf) <- lookupPrimitiveFunctionQ x
     t <- isType_ e
     noConstraints $ equalType t t'
     let s  = show $ nameConcrete $ qnameName x
-    bindPrimitive s $ pf { primFunName = x }
-    addConstant x (Defn Relevant x t (defaultDisplayForm x) 0 noCompiledRep $
+    bindPrimitive s pf
+    addConstant x (Defn Relevant x t [] [] (defaultDisplayForm x) 0 noCompiledRep $
                 Primitive (Info.defAbstract i) s Nothing Nothing)
     where
 	nameString (Name _ x _ _) = show x
@@ -205,7 +355,7 @@ checkPragma r p =
 	A.OptionsPragma _   -> __IMPOSSIBLE__	-- not allowed here
         A.EtaPragma r -> modifySignature eta
           where
-            eta sig = sig { sigDefinitions = Map.adjust setEta r defs }
+            eta sig = sig { sigDefinitions = HMap.adjust setEta r defs }
               where
                 setEta def = def { theDef = setEtad $ theDef def }
                 setEtad d   = case d of
@@ -214,18 +364,21 @@ checkPragma r p =
                 defs	  = sigDefinitions sig
 
 -- | Type check a bunch of mutual inductive recursive definitions.
-checkMutual :: Info.DeclInfo -> [A.Declaration] -> TCM ()
+--
+-- All definitions which have so far been assigned to the given mutual
+-- block are returned.
+checkMutual :: Info.MutualInfo -> [A.Declaration] -> TCM (Set QName)
 checkMutual i ds = inMutualBlock $ do
+
+  verboseS "tc.decl.mutual" 20 $ do
+    blockId <- currentOrFreshMutualBlock
+    reportSDoc "" 0 $ vcat $
+      (text "Checking mutual block" <+> text (show blockId) <> text ":") :
+      map (nest 2 . prettyA) ds
+
   mapM_ checkDecl ds
 
-  checkStrictlyPositive =<< currentMutualBlock
-  let unScope (A.ScopedDecl _ ds) = concatMap unScope ds
-      unScope d = [d]
-  case concatMap unScope ds of
-    -- Non-mutual definitions can be considered for projection likeness
-    [A.Axiom _ _ x _, A.FunDef _ y _] | x == y -> makeProjection x
-    _   -> return ()
-
+  lookupMutualBlock =<< currentOrFreshMutualBlock
 
 -- | Type check the type signature of an inductive or recursive definition.
 checkTypeSignature :: A.TypeSignature -> TCM ()
@@ -248,8 +401,8 @@ checkSection i x tel ds =
       dx   <- prettyTCM x
       dtel <- mapM prettyA tel
       dtel' <- prettyTCM =<< lookupSection x
-      liftIO $ LocIO.putStrLn $ "checking section " ++ show dx ++ " " ++ show dtel
-      liftIO $ LocIO.putStrLn $ "    actual tele: " ++ show dtel'
+      reportSLn "" 0 $ "checking section " ++ show dx ++ " " ++ show dtel
+      reportSLn "" 0 $ "    actual tele: " ++ show dtel'
     withCurrentModule x $ checkDecls ds
 
 checkModuleArity :: ModuleName -> Telescope -> [NamedArg A.Expr] -> TCM Telescope
@@ -259,7 +412,7 @@ checkModuleArity m tel args = check tel args
 
     check eta []             = return eta
     check EmptyTel (_:_)     = bad
-    check (ExtendTel (Arg h _ _) btel) args0@(Arg h' _ (Named name _) : args) =
+    check (ExtendTel (Dom h _ _) btel) args0@(Arg h' _ (Named name _) : args) =
       let y   = absName btel
           tel = absBody btel in
       case (h, h', name) of
@@ -326,7 +479,7 @@ checkSectionApplication' i m1 (A.RecordModuleIFS x) rd rm = do
       telInst = instFinal tel
       instFinal :: Telescope -> Telescope
       instFinal (ExtendTel _ NoAbs{}) = __IMPOSSIBLE__
-      instFinal (ExtendTel (Arg h r t) (Abs n EmptyTel)) = ExtendTel (Arg Instance r t) (Abs n EmptyTel)
+      instFinal (ExtendTel (Dom h r t) (Abs n EmptyTel)) = ExtendTel (Dom Instance r t) (Abs n EmptyTel)
       instFinal (ExtendTel arg (Abs n tel)) = ExtendTel arg (Abs n (instFinal tel))
       instFinal EmptyTel = __IMPOSSIBLE__
       args = teleArgs tel
