@@ -20,22 +20,27 @@ import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Records -- isRecord
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Substitute hiding (Substitution)
+import qualified Agda.TypeChecking.Substitute as S (Substitution)
 import Agda.TypeChecking.Telescope
 import Agda.TypeChecking.Conversion
 import Agda.TypeChecking.Constraints
 import Agda.TypeChecking.Irrelevance
 import Agda.TypeChecking.Primitive (constructorForm)
-import {-# SOURCE #-} Agda.TypeChecking.Empty (isReallyEmptyType)
-import Agda.TypeChecking.Telescope (renamingR, teleArgs)
+import {-# SOURCE #-} Agda.TypeChecking.Empty
+-- Duplicate import??
+-- import Agda.TypeChecking.Telescope (renamingR, teleArgs)
 
-import Agda.TypeChecking.Rules.Term (checkExpr)
+import {-# SOURCE #-} Agda.TypeChecking.Rules.Term (checkExpr)
 import Agda.TypeChecking.Rules.LHS.Problem
+import Agda.TypeChecking.Rules.LHS.ProblemRest
 import Agda.TypeChecking.Rules.LHS.Unify
 import Agda.TypeChecking.Rules.LHS.Split
 import Agda.TypeChecking.Rules.LHS.Implicit
 import Agda.TypeChecking.Rules.LHS.Instantiate
 import Agda.TypeChecking.Rules.Data
 
+import Agda.Interaction.Highlighting.Generate
+
 import Agda.Utils.Permutation
 import Agda.Utils.Size
 import Agda.Utils.Monad
@@ -43,58 +48,48 @@ import Agda.Utils.Monad
 #include "../../undefined.h"
 import Agda.Utils.Impossible
 
-data DotPatternInst = DPI A.Expr Term Type
-data AsBinding      = AsB Name Term Type
-
-instance Subst DotPatternInst where
-  substs us      (DPI e v a) = uncurry (DPI e) $ substs us (v,a)
-  substUnder n u (DPI e v a) = uncurry (DPI e) $ substUnder n u (v,a)
-
-instance PrettyTCM DotPatternInst where
-  prettyTCM (DPI e v a) = sep [ prettyA e <+> text "="
-                              , nest 2 $ prettyTCM v <+> text ":"
-                              , nest 2 $ prettyTCM a
-                              ]
-
-instance Subst AsBinding where
-  substs us      (AsB x v a) = uncurry (AsB x) $ substs us (v, a)
-  substUnder n u (AsB x v a) = uncurry (AsB x) $ substUnder n u (v, a)
-
-instance Raise AsBinding where
-  raiseFrom m k (AsB x v a) = uncurry (AsB x) $ raiseFrom m k (v, a)
-  renameFrom m k (AsB x v a) = uncurry (AsB x) $ renameFrom m k (v, a)
-
-instance PrettyTCM AsBinding where
-  prettyTCM (AsB x v a) =
-    sep [ prettyTCM x <> text "@" <> parens (prettyTCM v)
-        , nest 2 $ text ":" <+> prettyTCM a
-        ]
-
 -- | Compute the set of flexible patterns in a list of patterns. The result is
 --   the deBruijn indices of the flexible patterns. A pattern is flexible if it
 --   is dotted or implicit.
-flexiblePatterns :: [NamedArg A.Pattern] -> FlexibleVars
-flexiblePatterns nps = [ i | (i, p) <- zip [0..] $ reverse ps, flexible p ]
+flexiblePatterns :: [NamedArg A.Pattern] -> TCM FlexibleVars
+flexiblePatterns nps = map fst <$> filterM (flexible . snd) (zip [0..] $ reverse ps)
   where
-    ps = map (namedThing . unArg) nps
-    flexible (A.DotP _ _)    = True
-    flexible (A.ImplicitP _) = True
-    flexible _               = False
+    ps = map namedArg nps
+    flexible (A.DotP _ _)    = return True
+    flexible (A.ImplicitP _) = return True
+    flexible (A.ConP _ (A.AmbQ [c]) qs) =
+      ifM (isJust <$> isRecordConstructor c)
+          (andM $ map (flexible . namedArg) qs)
+          (return False)
+    flexible _               = return False
 
 -- | Compute the dot pattern instantiations.
-dotPatternInsts :: [NamedArg A.Pattern] -> Substitution -> [Type] -> [DotPatternInst]
-dotPatternInsts ps s as = dpi (map (namedThing . unArg) ps) (reverse s) as
+dotPatternInsts :: [NamedArg A.Pattern] -> Substitution -> [Dom Type] -> TCM [DotPatternInst]
+dotPatternInsts ps s as = dpi (map namedArg ps) (reverse s) as
   where
     dpi (_ : _)  []            _       = __IMPOSSIBLE__
     dpi (_ : _)  (Just _ : _)  []      = __IMPOSSIBLE__
     -- the substitution also contains entries for module parameters, so it can
     -- be longer than the pattern
-    dpi []       _             _       = []
+    dpi []       _             _       = return []
     dpi (_ : ps) (Nothing : s) as      = dpi ps s as
     dpi (p : ps) (Just u : s) (a : as) =
       case p of
-        A.DotP _ e    -> DPI e u a : dpi ps s as
+        A.DotP _ e    -> (DPI e u a :) <$> dpi ps s as
         A.ImplicitP _ -> dpi ps s as
+        -- record pattern
+        A.ConP _ (A.AmbQ [c]) qs -> do
+          Def r vs   <- ignoreSharing <$> reduce (unEl $ unDom a)
+          (ftel, us) <- etaExpandRecord r vs u
+          qs <- insertImplicitPatterns ExpandLast qs ftel
+          let instTel EmptyTel _                   = []
+              instTel (ExtendTel arg tel) (u : us) = arg : instTel (absApp tel u) us
+              instTel ExtendTel{} []               = __IMPOSSIBLE__
+              bs0 = instTel ftel (map unArg us)
+              -- Andreas, 2012-09-19 propagate relevance info to dot patterns
+              bs  = map (mapDomRelevance (composeRelevance (domRelevance a))) bs0
+          dpi (map namedArg qs ++ ps) (map (Just . unArg) us ++ s) (bs ++ as)
+
         _           -> __IMPOSSIBLE__
 
 instantiatePattern :: Substitution -> Permutation -> [Arg Pattern] -> [Arg Pattern]
@@ -139,7 +134,7 @@ instantiatePattern sub perm ps
 
 -- | Check if a problem is solved. That is, if the patterns are all variables.
 isSolvedProblem :: Problem -> Bool
-isSolvedProblem = all (isVar . snd . asView . namedThing . unArg) . problemInPat
+isSolvedProblem = all (isVar . snd . asView . namedArg) . problemInPat
   where
     isVar (A.VarP _)      = True
     isVar (A.WildP _)     = True
@@ -154,26 +149,34 @@ isSolvedProblem = all (isVar . snd . asView . namedThing . unArg) . problemInPat
 --
 -- Precondition: The problem has to be solved.
 
+{-
 noShadowingOfConstructors
   :: A.Clause
      -- ^ The entire clause (used for error reporting).
   -> Problem -> TCM ()
 noShadowingOfConstructors c problem =
   traceCall (CheckPatternShadowing c) $ do
-    let pat = map (snd . asView . namedThing . unArg) $
+-}
+noShadowingOfConstructors
+  :: (Maybe r -> Call) -- ^ Trace, e.g., @CheckPatternShadowing clause@
+  -> Problem -> TCM ()
+noShadowingOfConstructors mkCall problem =
+  traceCall mkCall $ do
+    let pat = map (snd . asView . namedArg) $
                   problemInPat problem
-        tel = map (unEl . snd . unArg) $ telToList $ problemTel problem
+        tel = map (unEl . snd . unDom) $ telToList $ problemTel problem
     zipWithM' noShadowing pat tel
     return ()
   where
   noShadowing (A.WildP     {}) t = return ()
   noShadowing (A.AbsurdP   {}) t = return ()
   noShadowing (A.ImplicitP {}) t = return ()
-  noShadowing (A.ConP      {}) t = __IMPOSSIBLE__
+  noShadowing (A.ConP      {}) t = return ()  -- only happens for eta expanded record patterns
   noShadowing (A.DefP      {}) t = __IMPOSSIBLE__
   noShadowing (A.AsP       {}) t = __IMPOSSIBLE__
   noShadowing (A.DotP      {}) t = __IMPOSSIBLE__
   noShadowing (A.LitP      {}) t = __IMPOSSIBLE__
+  noShadowing (A.PatternSynP {}) t = __IMPOSSIBLE__
   noShadowing (A.VarP x)       t = do
     t <- normalise t
     case t of
@@ -197,6 +200,7 @@ noShadowingOfConstructors c problem =
       Var   {} -> return ()
       Pi    {} -> return ()
       Sort  {} -> return ()
+      Shared p -> noShadowing (A.VarP x) $ derefPtr p
       MetaV {} -> return ()
       -- TODO: If the type is a meta-variable, should the test be
       -- postponed? If there is a problem, then it will be caught when
@@ -211,7 +215,7 @@ noShadowingOfConstructors c problem =
 
 -- | Check that a dot pattern matches it's instantiation.
 checkDotPattern :: DotPatternInst -> TCM ()
-checkDotPattern (DPI e v a) =
+checkDotPattern (DPI e v (Dom h r a)) =
   traceCall (CheckDotPattern e v) $ do
   reportSDoc "tc.lhs.dot" 15 $
     sep [ text "checking dot pattern"
@@ -219,26 +223,42 @@ checkDotPattern (DPI e v a) =
         , nest 2 $ text "=" <+> prettyTCM v
         , nest 2 $ text ":" <+> prettyTCM a
         ]
-  u <- checkExpr e a
-  -- Should be ok to do noConstraints here
-  noConstraints $ equalTerm a u v
+  applyRelevanceToContext r $ do
+    u <- checkExpr e a
+    -- Should be ok to do noConstraints here
+    noConstraints $ equalTerm a u v
 
 -- | Bind the variables in a left hand side. Precondition: the patterns should
 --   all be 'A.VarP', 'A.WildP', or 'A.ImplicitP' and the telescope should have
 --   the same size as the pattern list.
+--   There could also be 'A.ConP's resulting from eta expanded implicit record
+--   patterns.
 bindLHSVars :: [NamedArg A.Pattern] -> Telescope -> TCM a -> TCM a
 bindLHSVars []       (ExtendTel _ _)   _   = __IMPOSSIBLE__
 bindLHSVars (_ : _)   EmptyTel         _   = __IMPOSSIBLE__
 bindLHSVars []        EmptyTel         ret = ret
 bindLHSVars (p : ps) (ExtendTel a tel) ret =
-  case namedThing $ unArg p of
+  case namedArg p of
     A.VarP x      -> addCtx x a $ bindLHSVars ps (absBody tel) ret
     A.WildP _     -> bindDummy (absName tel)
     A.ImplicitP _ -> bindDummy (absName tel)
-    A.AbsurdP _   -> do
-      isReallyEmptyType $ unArg a
+    A.AbsurdP pi  -> do
+      -- Andreas, 2012-03-15: allow postponement of emptyness check
+      isEmptyType (getRange pi) $ unDom a
+      -- OLD CODE: isReallyEmptyType $ unArg a
       bindDummy (absName tel)
-    _             -> __IMPOSSIBLE__
+    A.ConP _ (A.AmbQ [c]) qs -> do -- eta expanded record pattern
+      Def r vs <- reduce (unEl $ unDom a)
+      ftel     <- (`apply` vs) <$> getRecordFieldTypes r
+      let n   = size ftel
+          eta = Con c [ Var i [] <$ (namedThing <$> q) | (q, i) <- zip qs [n - 1, n - 2..0] ]
+      bindLHSVars (qs ++ ps) (ftel `abstract` absApp (raise (size ftel) tel) eta) ret
+    A.ConP{}        -> __IMPOSSIBLE__
+    A.DefP{}        -> __IMPOSSIBLE__
+    A.AsP{}         -> __IMPOSSIBLE__
+    A.DotP{}        -> __IMPOSSIBLE__
+    A.LitP{}        -> __IMPOSSIBLE__
+    A.PatternSynP{} -> __IMPOSSIBLE__
     where
       name "_" = freshNoName_
       name s   = freshName_ ("." ++ s)
@@ -256,60 +276,39 @@ bindAsPatterns (AsB x v a : asb) ret = do
         ]
   addLetBinding Relevant x v a $ bindAsPatterns asb ret
 
--- | Rename the variables in a telescope using the names from a given pattern
-useNamesFromPattern :: [NamedArg A.Pattern] -> Telescope -> Telescope
-useNamesFromPattern ps = telFromList . zipWith ren (toPats ps ++ repeat dummy) . telToList
-  where
-    dummy = A.WildP __IMPOSSIBLE__
-    ren (A.VarP x) (Arg NotHidden r (_, a)) = Arg NotHidden r (show x, a)
-    ren _ a = a
-    toPats = map (namedThing . unArg)
-
 -- | Check a LHS. Main function.
 checkLeftHandSide
-  :: A.Clause
-     -- ^ The entire clause.
+  :: (Maybe r -> Call)
+     -- ^ Trace, e.g. @CheckPatternShadowing clause@
   -> [NamedArg A.Pattern]
      -- ^ The patterns.
   -> Type
-     -- ^ The expected type.
-  -> (Telescope -> Telescope -> [Term] -> [String] -> [Arg Pattern]
-      -> Type -> Permutation -> TCM a)
+     -- ^ The expected type @a = Γ → b@.
+  -> (Maybe Telescope   -- Γ : The types of the patterns.
+                        -- 'Nothing' if more patterns than domain types in @a@.
+                        -- Used only to construct a @with@ function; see 'stripwithClausePatterns'.
+      -> Telescope      -- Δ : The types of the pattern variables.
+      -> S.Substitution -- σ : The patterns in form of a substitution Δ ⊢ σ : Γ
+      -> [String]       -- Names for the variables in Δ, for binding the body.
+      -> [Arg Pattern]  -- The patterns in internal syntax.
+      -> Type           -- The type of the body. Is @bσ@ if @Γ@ is defined.
+      -> Permutation    -- The permutation from pattern vars to @Δ@.
+      -> TCM a)
      -- ^ Continuation.
   -> TCM a
 checkLeftHandSide c ps a ret = do
-  TelV tel0' b0 <- telView a
-  ps <- insertImplicitPatterns ps tel0'
-  unless (size tel0' >= size ps) $ typeError $ TooManyArgumentsInLHS (size ps) a
-  let tel0     = useNamesFromPattern ps tel0'
-      (as, bs) = splitAt (size ps) $ telToList tel0
-      gamma    = telFromList as
-      b        = telePi (telFromList bs) b0
+  problem <- problemFromPats ps a
+  unless (noProblemRest problem) $ typeError $ TooManyArgumentsInLHS a
+  let (Problem _ _ gamma (ProblemRest _ b)) = problem
+      mgamma = if noProblemRest problem then Just gamma else Nothing
+      st     = LHSState problem idS [] []
 
-      -- internal patterns start as all variables
-      ips      = map (fmap (VarP . fst)) as
+  -- doing the splits:
+  LHSState (Problem ps (perm, qs) delta rest) sigma dpi asb <- checkLHS st
+  unless (null $ restPats rest) $ typeError $ TooManyArgumentsInLHS a
 
-      problem  = Problem ps (idP $ size ps, ips) gamma
-
-  reportSDoc "tc.lhs.top" 10 $
-    vcat [ text "checking lhs:"
-	 , nest 2 $ vcat
-	   [ text "ps    =" <+> fsep (map prettyA ps)
-	   , text "a     =" <+> (prettyTCM =<< normalise a)
-	   , text "a'    =" <+> prettyTCM (telePi tel0  b0)
-	   , text "a''   =" <+> prettyTCM (telePi tel0' b0)
-           , text "xs    =" <+> text (show $ map (fst . unArg) as)
-	   , text "tel0  =" <+> prettyTCM tel0
-	   , text "b0    =" <+> prettyTCM b0
-	   , text "gamma =" <+> prettyTCM gamma
-	   , text "b     =" <+> addCtxTel gamma (prettyTCM b)
-	   ]
-	 ]
-
-  let idsub = [ Var i [] | i <- [0..] ]
-
-  (Problem ps (perm, qs) delta, sigma, dpi, asb) <- checkLHS problem idsub [] []
-  let b' = substs sigma b
+  -- let b' = applySubst sigma b
+  let b' = restType rest
 
   noPatternMatchingOnCodata qs
 
@@ -333,18 +332,19 @@ checkLeftHandSide c ps a ret = do
     let rho = renamingR perm -- I'm not certain about this...
         Perm n _ = perm
         xs  = [ "h" ++ show n | n <- [0..n - 1] ]
-    ret gamma delta rho xs qs b' perm
+    ret mgamma delta rho xs qs b' perm
   where
-    checkLHS :: Problem -> [Term] -> [DotPatternInst] -> [AsBinding] ->
-                TCM (Problem, [Term], [DotPatternInst], [AsBinding])
-    checkLHS problem sigma dpi asb
-      | isSolvedProblem problem = do
-        problem <- insertImplicitProblem problem -- inserting implicit patterns preserves solvedness
-        noShadowingOfConstructors c problem
-        return (problem, sigma, dpi, asb)
-      | otherwise               = do
-        sp <- splitProblem =<< insertImplicitProblem problem
-        reportSDoc "tc.lhs.top" 20 $ text "splitting completed"
+    -- the loop: split at a variable in the problem until problem is solved
+    checkLHS :: LHSState -> TCM LHSState
+    checkLHS st@(LHSState problem sigma dpi asb) = do
+      problem <- insertImplicitProblem problem  -- inserting implicits no longer preserve solvedness
+      if isSolvedProblem problem                -- since we might insert eta expanded record patterns
+        then do
+          noShadowingOfConstructors c problem
+          return $ st { lhsProblem = problem }
+        else do
+        sp <- splitProblem problem
+        reportSDoc "tc.lhs.split" 20 $ text "splitting completed"
         case sp of
           Left NothingToSplit   -> nothingToSplitError problem
           Left (SplitPanic err) -> __IMPOSSIBLE__
@@ -354,27 +354,28 @@ checkLeftHandSide c ps a ret = do
 
             -- plug the hole with a lit pattern
             let ip    = plugHole (LitP lit) iph
-                iperm = expandP (fromIntegral hix) 0 $ fst (problemOutPat problem)
+                iperm = expandP hix 0 $ fst (problemOutPat problem)
 
             -- substitute the literal in p1 and sigma and dpi and asb
             let delta1 = problemTel p0
                 delta2 = absApp (fmap problemTel p1) (Lit lit)
-                rho    = [ var i | i <- [0..size delta2 - 1] ]
-                      ++ [ raise (size delta2) $ Lit lit ]
-                      ++ [ var i | i <- [size delta2 ..] ]
-                  where
-                    var i = Var i []
-                sigma'   = substs rho sigma
-                dpi'     = substs rho dpi
-                asb0     = substs rho asb
-                ip'      = substs rho ip
+                rho    = liftS (size delta2) $ singletonS (Lit lit)
+                -- rho    = [ var i | i <- [0..size delta2 - 1] ]
+                --       ++ [ raise (size delta2) $ Lit lit ]
+                --       ++ [ var i | i <- [size delta2 ..] ]
+                sigma'   = applySubst rho sigma
+                dpi'     = applySubst rho dpi
+                asb0     = applySubst rho asb
+                ip'      = applySubst rho ip
+                rest'    = applySubst rho (problemRest problem)
 
             -- Compute the new problem
             let ps'      = problemInPat p0 ++ problemInPat (absBody p1)
                 delta'   = abstract delta1 delta2
-                problem' = Problem ps' (iperm, ip') delta'
+                problem' = Problem ps' (iperm, ip') delta' rest'
                 asb'     = raise (size delta2) (map (\x -> AsB x (Lit lit) a) xs) ++ asb0
-            checkLHS problem' sigma' dpi' asb'
+            st' <- updateProblemRest (LHSState problem' sigma' dpi' asb')
+            checkLHS st'
 
           -- Split on constructor pattern
           Right (Split p0 xs (Arg h rel
@@ -396,7 +397,7 @@ checkLeftHandSide c ps a ret = do
             let delta1 = problemTel p0
             let typeOfSplitVar = Arg h rel a
 
-            reportSDoc "tc.lhs.top" 10 $ sep
+            reportSDoc "tc.lhs.split" 10 $ sep
               [ text "checking lhs"
               , nest 2 $ text "tel =" <+> prettyTCM (problemTel problem)
               , nest 2 $ text "rel =" <+> (text $ show rel)
@@ -412,11 +413,11 @@ checkLeftHandSide c ps a ret = do
                 ]
               ]
 
-            Con c' [] <- constructorForm =<< normalise (Con c [])
+            Con c' [] <- ignoreSharing <$> (constructorForm =<< normalise (Con c []))
             c  <- return $ c' `withRangeOf` c
             ca <- defType <$> getConstInfo c
 
-            reportSDoc "tc.lhs.top" 20 $ nest 2 $ vcat
+            reportSDoc "tc.lhs.split" 20 $ nest 2 $ vcat
               [ text "ca =" <+> prettyTCM ca
               , text "vs =" <+> prettyList (map prettyTCM vs)
               ]
@@ -425,7 +426,10 @@ checkLeftHandSide c ps a ret = do
             let a = ca `piApply` vs
 
             -- It will end in an application of the datatype
-            TelV gamma' ca@(El _ (Def d' us)) <- telView a
+            (gamma', ca, d', us) <- do
+              TelV gamma' ca@(El _ def) <- telView a
+              let Def d' us = ignoreSharing def
+              return (gamma', ca, d', us)
 
             -- This should be the same datatype as we split on
             unless (d == d') $ typeError $ ShouldBeApplicationOf ca d'
@@ -444,7 +448,7 @@ checkLeftHandSide c ps a ret = do
               ]
 -}
             -- Insert implicit patterns
-            qs' <- insertImplicitPatterns qs gamma'
+            qs' <- insertImplicitPatterns ExpandLast qs gamma'
 
             unless (size qs' == size gamma') $
               typeError $ WrongNumberOfConstructorArguments c (size gamma') (size qs')
@@ -455,7 +459,7 @@ checkLeftHandSide c ps a ret = do
             da <- (`piApply` vs) . defType <$> getConstInfo d
 
             -- Compute the flexible variables
-            let flex = flexiblePatterns (problemInPat p0 ++ qs')
+            flex <- flexiblePatterns (problemInPat p0 ++ qs')
 
 	    reportSDoc "tc.lhs.top" 15 $ addCtxTel delta1 $
 	      sep [ text "preparing to unify"
@@ -473,17 +477,17 @@ checkLeftHandSide c ps a ret = do
             sub0 <- addCtxTel (delta1 `abstract` gamma) $
                     unifyIndices_ flex (raise (size gamma) da) (drop (size vs) us) (raise (size gamma) ws)
 
-            -- We should subsitute c ys for x in Δ₂ and sigma
+            -- We should substitute c ys for x in Δ₂ and sigma
             let ys     = teleArgs gamma
                 delta2 = absApp (raise (size gamma) $ fmap problemTel p1) (Con c ys)
-                rho0 = [ var i | i <- [0..size delta2 - 1] ]
-                    ++ [ raise (size delta2) $ Con c ys ]
-                    ++ [ var i | i <- [size delta2 + size gamma ..] ]
-                  where
-                    var i = Var i []
-                sigma0 = substs rho0 sigma
-                dpi0   = substs rho0 dpi
-                asb0   = substs rho0 asb
+                rho0   = liftS (size delta2) $ Con c ys :# raiseS (size gamma)
+                -- rho0 = [ var i | i <- [0..size delta2 - 1] ]
+                --     ++ [ raise (size delta2) $ Con c ys ]
+                --     ++ [ var i | i <- [size delta2 + size gamma ..] ]
+                sigma0 = applySubst rho0 sigma
+                dpi0   = applySubst rho0 dpi
+                asb0   = applySubst rho0 asb
+                rest0  = applySubst rho0 (problemRest problem)
 
             reportSDoc "tc.lhs.top" 15 $ addCtxTel (delta1 `abstract` gamma) $ nest 2 $ vcat
               [ text "delta2 =" <+> prettyTCM delta2
@@ -500,14 +504,14 @@ checkLeftHandSide c ps a ret = do
             -- all variables which will be bound by patterns.
             -- Thus, it has to be raised by 1 (the "hole" variable)
             -- plus the length of delta2 (the variables coming after the hole).
-            storedPatternType <- ifM (isRecord d)
+            storedPatternType <- ifM (isJust <$> isRecord d)
               (return $ Just $ raise (1 + size delta2) $ typeOfSplitVar)
               (return $ Nothing)
 
             -- Plug the hole in the out pattern with c ys
-            let ysp = map (fmap (VarP . fst)) $ telToList gamma
+            let ysp = map (argFromDom . fmap (VarP . fst)) $ telToList gamma
                 ip  = plugHole (ConP c storedPatternType ysp) iph
-                ip0 = substs rho0 ip
+                ip0 = applySubst rho0 ip
 
             -- Δ₁Γ ⊢ sub0, we need something in Δ₁ΓΔ₂
             -- Also needs to be padded with Nothing's to have the right length.
@@ -523,8 +527,8 @@ checkLeftHandSide c ps a ret = do
               , text "ip0  =" <+> text (show ip0)
               ]
             reportSDoc "tc.lhs.top" 15 $ nest 2 $ vcat
-              [ text "rho0 =" <+> text (show $ take (size delta1 + size gamma + size delta2) rho0)
-              ]  -- Andreas, this is showing some inital segment of rho0, not necessarily the most meaningful one
+              [ text "rho0 =" <+> text (show rho0)
+              ]
 
             -- Instantiate the new telescope with the given substitution
             (delta', perm, rho, instTypes) <- instantiateTel sub newTel
@@ -547,16 +551,17 @@ checkLeftHandSide c ps a ret = do
 -}
             -- Compute the new dot pattern instantiations
             let ps0'   = problemInPat p0 ++ qs' ++ problemInPat (absBody p1)
-                newDpi = dotPatternInsts ps0' (substs rho sub) instTypes
 
             reportSDoc "tc.lhs.top" 15 $ nest 2 $ vcat
-              [ text "subst rho sub =" <+> brackets (fsep $ punctuate comma $ map (maybe (text "_") prettyTCM) (substs rho sub))
+              [ text "subst rho sub =" <+> brackets (fsep $ punctuate comma $ map (maybe (text "_") prettyTCM) (applySubst rho sub))
               , text "ps0'  =" <+> brackets (fsep $ punctuate comma $ map prettyA ps0')
               ]
 
+            newDpi <- dotPatternInsts ps0' (applySubst rho sub) instTypes
+
             -- The final dpis and asbs are the new ones plus the old ones substituted by ρ
-            let dpi' = substs rho dpi0 ++ newDpi
-                asb' = substs rho $ asb0 ++ raise (size delta2) (map (\x -> AsB x (Con c ys) ca) xs)
+            let dpi' = applySubst rho dpi0 ++ newDpi
+                asb' = applySubst rho $ asb0 ++ raise (size delta2) (map (\x -> AsB x (Con c ys) ca) xs)
 
             reportSDoc "tc.lhs.top" 15 $ nest 2 $ vcat
               [ text "dpi' = " <+> brackets (fsep $ punctuate comma $ map prettyTCM dpi')
@@ -564,7 +569,8 @@ checkLeftHandSide c ps a ret = do
               ]
 
             -- Apply the substitution to the type
-            let sigma'   = substs rho sigma0
+            let sigma'   = applySubst rho sigma0
+                rest'    = applySubst rho rest0
 
             reportSDoc "tc.lhs.inst" 15 $
               nest 2 $ text "ps0 = " <+> brackets (fsep $ punctuate comma $ map prettyA ps0')
@@ -575,15 +581,15 @@ checkLeftHandSide c ps a ret = do
            -- Compute the new permutation of the out patterns. This is the composition of
             -- the new permutation with the expansion of the old permutation to
             -- reflect the split.
-            let perm'  = expandP (fromIntegral hix) (size gamma) $ fst (problemOutPat problem)
+            let perm'  = expandP hix (size gamma) $ fst (problemOutPat problem)
                 iperm' = perm `composeP` perm'
 
             -- Instantiate the out patterns
             let ip'    = instantiatePattern sub perm' ip0
-                newip  = substs rho ip'
+                newip  = applySubst rho ip'
 
             -- Construct the new problem
-            let problem' = Problem ps' (iperm', newip) delta'
+            let problem' = Problem ps' (iperm', newip) delta' rest'
 
             reportSDoc "tc.lhs.top" 12 $ sep
               [ text "new problem"
@@ -602,8 +608,23 @@ checkLeftHandSide c ps a ret = do
               , text "newip  =" <+> text (show newip)
               ]
 
+            -- if rest type reduces,
+            -- extend the split problem by previously not considered patterns
+            st'@(LHSState problem'@(Problem ps' (iperm', ip') delta' rest')
+                          sigma' dpi' asb')
+              <- updateProblemRest $ LHSState problem' sigma' dpi' asb'
+
+            reportSDoc "tc.lhs.top" 12 $ sep
+              [ text "new problem from rest"
+              , nest 2 $ vcat
+                [ text "ps'    = " <+> fsep (map prettyA ps')
+                , text "delta' = " <+> prettyTCM delta'
+                , text "ip'    =" <+> text (show ip')
+                , text "iperm' =" <+> text (show iperm')
+                ]
+              ]
             -- Continue splitting
-            checkLHS problem' sigma' dpi' asb'
+            checkLHS st'
 
 -- Ensures that we are not performing pattern matching on codata.
 
