@@ -4,7 +4,7 @@ module Agda.TypeChecking.Rules.LHS.Split where
 
 import Control.Applicative
 import Control.Monad.Error
-import Data.Monoid
+import Data.Monoid (mempty, mappend)
 import Data.List
 import Data.Traversable hiding (mapM, sequence)
 
@@ -25,7 +25,8 @@ import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Constraints
 import Agda.TypeChecking.Conversion
 import Agda.TypeChecking.Rules.LHS.Problem
-import Agda.TypeChecking.Rules.Term
+import Agda.TypeChecking.Rules.LHS.ProblemRest
+-- import Agda.TypeChecking.Rules.Term
 import Agda.TypeChecking.Monad.Builtin
 import Agda.TypeChecking.Free
 import Agda.TypeChecking.Irrelevance
@@ -56,10 +57,10 @@ expandLitPattern p = traverse (traverse expand) p
                         ++ "the literal to the corresponding constructor pattern, so "
                         ++ "you probably don't want to do it this way."
         | otherwise -> do
-          Con z _ <- primZero
-          Con s _ <- primSuc
+          Con z _ <- ignoreSharing <$> primZero
+          Con s _ <- ignoreSharing <$> primSuc
           let zero  = A.ConP info (A.AmbQ [setRange r z]) []
-              suc p = A.ConP info (A.AmbQ [setRange r s]) [defaultArg $ unnamed p]
+              suc p = A.ConP info (A.AmbQ [setRange r s]) [defaultNamedArg p]
               info  = A.PatRange r
               p'    = foldr ($) zero $ genericReplicate n suc
           return $ foldr (A.AsP info) p' xs
@@ -68,7 +69,7 @@ expandLitPattern p = traverse (traverse expand) p
 -- | Split a problem at the first constructor of datatype type. Implicit
 --   patterns should have been inserted.
 splitProblem :: Problem -> TCM (Either SplitError SplitProblem)
-splitProblem (Problem ps (perm, qs) tel) = do
+splitProblem (Problem ps (perm, qs) tel pr) = do
     reportS "tc.lhs.split" 20 $ "initiating splitting\n"
     runErrorT $
       splitP ps (permute perm $ zip [0..] $ allHoles qs) tel
@@ -81,7 +82,7 @@ splitProblem (Problem ps (perm, qs) tel) = do
     splitP (p : ps) ((i, q) : qs) tel0@(ExtendTel a tel) = do
       let tryAgain = splitP (p : ps) ((i, q) : qs) tel0
       p <- lift $ expandLitPattern p
-      case asView $ namedThing $ unArg p of
+      case asView $ namedArg p of
 
         -- Case: literal pattern
 	(xs, p@(A.LitP lit))  -> do
@@ -90,45 +91,45 @@ splitProblem (Problem ps (perm, qs) tel) = do
           -- not indexed.
 
           -- Andreas, 2010-09-07 cannot split on irrelevant args
-          when (unusableRelevance $ argRelevance a) $
+          when (unusableRelevance $ domRelevance a) $
             typeError $ SplitOnIrrelevant p a
 	  b <- lift $ litType lit
 	  ok <- lift $ do
-	      noConstraints (equalType (unArg a) b)
+	      noConstraints (equalType (unDom a) b)
 	      return True
 	    `catchError` \_ -> return False
 	  if ok
 	    then return $
 	      Split mempty
 		    xs
-		    (fmap (LitFocus lit q i) a)
-		    (fmap (Problem ps ()) tel)
+		    (argFromDom $ fmap (LitFocus lit q i) a)
+		    (fmap (\ tel -> Problem ps () tel __IMPOSSIBLE__) tel)
 	    else keepGoing
 
         -- Case: constructor pattern
 	(xs, p@(A.ConP _ (A.AmbQ cs) args)) -> do
-	  a' <- liftTCM $ reduce $ unArg a
-	  case unEl a' of
-
-            -- Type is a meta and constructor is unambiguous,
-            -- in this case try to instantiate the meta.
-            MetaV{} | [c] <- cs -> do
-              ok <- lift $ do
-                Constructor{ conData = d } <- theDef <$> getConstInfo c
-                dt            <- defType <$> getConstInfo d
-                vs            <- newArgsMeta dt
-                El _ (Sort s) <- reduce $ apply dt vs
-                (True <$ noConstraints (equalType a' (El s $ Def d vs)))
-                  `catchError` \_ -> return False
-              if not ok then keepGoing else
-                tryAgain
+          let tryInstantiate a'
+                | [c] <- cs = do
+                    -- Type is blocked by a meta and constructor is unambiguous,
+                    -- in this case try to instantiate the meta.
+                  ok <- lift $ do
+                    Constructor{ conData = d } <- theDef <$> getConstInfo c
+                    dt     <- defType <$> getConstInfo d
+                    vs     <- newArgsMeta dt
+                    Sort s <- ignoreSharing . unEl <$> reduce (apply dt vs)
+                    (True <$ noConstraints (equalType a' (El s $ Def d vs)))
+                      `catchError` \_ -> return False
+                  if ok then tryAgain else keepGoing
+                | otherwise = keepGoing
+          ifBlockedType (unDom a) (const tryInstantiate) $ \ a' -> do
+	  case ignoreSharing $ unEl a' of
 
             -- Subcase: split type is a Def
 	    Def d vs	-> do
 	      def <- liftTCM $ theDef <$> getConstInfo d
               unless (defIsRecord def) $
                 -- cannot split on irrelevant or non-strict things
-                when (unusableRelevance $ argRelevance a) $ do
+                when (unusableRelevance $ domRelevance a) $ do
                   -- Andreas, 2011-10-04 unless allowed by option
                   allowed <- liftTCM $ optExperimentalIrrelevance <$> pragmaOptions
                   unless allowed $ typeError $ SplitOnIrrelevant p a
@@ -140,7 +141,7 @@ splitProblem (Problem ps (perm, qs) tel) = do
               case mp of
                 Nothing -> keepGoing
                 Just np ->
-		  liftTCM $ traceCall (CheckPattern p EmptyTel (unArg a)) $ do  -- TODO: wrong telescope
+		  liftTCM $ traceCall (CheckPattern p EmptyTel (unDom a)) $ do  -- TODO: wrong telescope
                   -- Check that we construct something in the right datatype
                   c <- do
                       cs' <- mapM canonicalName cs
@@ -168,50 +169,118 @@ splitProblem (Problem ps (perm, qs) tel) = do
 			 ]
 
                   whenM (optWithoutK <$> pragmaOptions) $
-                    wellFormedIndices pars ixs
+                    wellFormedIndices a'
 
 		  return $ Split mempty
 				 xs
-				 (fmap (Focus c args (getRange p) q i d pars ixs) a)
-				 (fmap (Problem ps ()) tel)
+				 (argFromDom $ fmap (Focus c args (getRange p) q i d pars ixs) a)
+				 (fmap (\ tel -> Problem ps () tel __IMPOSSIBLE__) tel)
             -- Subcase: split type is not a Def
 	    _	-> keepGoing
         -- Case: neither literal nor constructor pattern
 	p -> keepGoing
       where
 	keepGoing = do
-	  let p0 = Problem [p] () (ExtendTel a $ fmap (const EmptyTel) tel)
+	  let p0 = Problem [p] () (ExtendTel a $ fmap (const EmptyTel) tel) mempty
 	  Split p1 xs foc p2 <- underAbstraction a tel $ \tel -> splitP ps qs tel
 	  return $ Split (mappend p0 p1) xs foc p2
 
--- | Checks that the indices are constructors (or literals) applied to
--- distinct variables which do not occur free in the parameters.
+-- | Takes a type, which must be a data or record type application,
+-- and checks that the indices are constructors (or literals) applied
+-- to distinct variables which do not occur free in the parameters.
+-- For the purposes of this check parameters count as constructor
+-- arguments; parameters are reconstructed from the given type.
+--
+-- Precondition: The type must be a data or record type application.
+
+wellFormedIndices :: Type -> TCM ()
+wellFormedIndices t = do
+  t <- reduce t
+
+  reportSDoc "tc.lhs.split.well-formed" 10 $
+    fsep [ text "Checking if indices are well-formed:"
+         , nest 2 $ prettyTCM t
+         ]
+
+  (pars, ixs) <- normalise =<< case ignoreSharing $ unEl t of
+    Def d args -> do
+      def       <- getConstInfo d
+      typedArgs <- args `withTypesFrom` defType def
+
+      let noPars = case theDef def of
+            Datatype { dataPars = n } -> n
+            Record   { recPars  = n } -> n
+            _                         -> __IMPOSSIBLE__
+          (pars, ixs) = genericSplitAt noPars typedArgs
+      return (map fst pars, ixs)
+
+    _ -> __IMPOSSIBLE__
+
+  mvs <- constructorApplications ixs
+  vs  <- case mvs of
+           Nothing -> typeError $
+                        IndicesNotConstructorApplications (map fst ixs)
+           Just vs -> return vs
 
-wellFormedIndices
-  :: [Arg Term] -- ^ Parameters.
-  -> [Arg Term] -- ^ Indices.
-  -> TCM ()
-wellFormedIndices pars ixs = do
-  pars <- normalise pars
-  ixs  <- normalise ixs
-  vs   <- case constructorApplications ixs of
-            Nothing -> typeError $ IndicesNotConstructorApplications ixs
-            Just vs -> return vs
   unless (fastDistinct vs) $
-    typeError $ IndexVariablesNotDistinct ixs
-  case filter snd $ zip vs (map (`freeIn` pars) vs) of
-    []          -> return ()
-    (v , _) : _ -> typeError $ IndexFreeInParameter v pars
+    typeError $ IndexVariablesNotDistinct vs (map fst ixs)
+
+  case map fst $ filter snd $ zip vs (map (`freeIn` pars) vs) of
+    [] -> return ()
+    vs ->
+      typeError $ IndicesFreeInParameters vs (map fst ixs) pars
+
   where
   -- | If the term consists solely of constructors (or literals)
-  -- applied to variables, then the variables are returned, and
-  -- otherwise nothing.
-  constructorApplication :: Term -> Maybe [Nat]
-  constructorApplication (Var x [])   = Just [x]
-  constructorApplication (Con c args) = constructorApplications args
-  constructorApplication (Lit {})     = Just []
-  constructorApplication _            = Nothing
+  -- applied to variables (after parameter reconstruction), then the
+  -- variables are returned, and otherwise nothing.
+  constructorApplication :: Term
+                         -> Type  -- ^ The term's type.
+                         -> TCM (Maybe [Nat])
+  constructorApplication (Var x [])      _ = return (Just [x])
+  constructorApplication (Lit {})        _ = return (Just [])
+  constructorApplication (Shared p)      t  = constructorApplication (derefPtr p) t
+  constructorApplication (Con c conArgs) (El _ (Def d dataArgs)) = do
+    conDef  <- getConstInfo c
+    dataDef <- getConstInfo d
 
-  constructorApplications :: [Arg Term] -> Maybe [Nat]
-  constructorApplications args =
-    concat <$> mapM (constructorApplication . unArg) args
+    let noPars = case theDef dataDef of
+          Datatype { dataPars = n } -> n
+          Record   { recPars  = n } -> n
+          _                         -> __IMPOSSIBLE__
+        pars    = genericTake noPars dataArgs
+        allArgs = pars ++ conArgs
+
+    reportSDoc "tc.lhs.split.well-formed" 20 $
+      fsep [ text "Reconstructed parameters:"
+           , nest 2 $
+               prettyTCM (Con c []) <+>
+               text "(:" <+> prettyTCM (defType conDef) <> text ")" <+>
+               text "<<" <+> prettyTCM pars <+> text ">>" <+>
+               prettyTCM conArgs
+           ]
+
+    constructorApplications =<< allArgs `withTypesFrom` defType conDef
+
+  constructorApplication _ _ = return Nothing
+
+  constructorApplications :: [(Arg Term, Dom Type)] -> TCM (Maybe [Nat])
+  constructorApplications args = do
+    xs <- mapM (\(e, t) -> constructorApplication (unArg e) (ignoreSharingType $ unDom t))
+               args
+    return (concat <$> sequence xs)
+
+-- | @args \`withTypesFrom\` t@ returns the arguments @args@ paired up
+-- with their types, taken from @t@, which is assumed to be a @length
+-- args@-ary pi.
+--
+-- Precondition: @t@ has to start with @length args@ pis.
+
+withTypesFrom :: Args -> Type -> TCM [(Arg Term, Dom Type)]
+[]           `withTypesFrom` _ = return []
+(arg : args) `withTypesFrom` t = do
+  t <- reduce t
+  case ignoreSharing $ unEl t of
+    Pi a b -> ((arg, a) :) <$>
+              args `withTypesFrom` absApp b (unArg arg)
+    _      -> __IMPOSSIBLE__
