@@ -15,7 +15,7 @@ import Data.Map (Map)
 import qualified Data.Map as Map
 import Data.List hiding (sort)
 
-import Data.Generics (Typeable, Data)
+import Data.Typeable (Typeable)
 import Data.Foldable (Foldable)
 import Data.Traversable (Traversable,traverse)
 
@@ -32,6 +32,7 @@ import Agda.TypeChecking.Level (reallyUnLevelView)
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Substitute hiding (Substitution)
+import qualified Agda.TypeChecking.Substitute as S
 import Agda.TypeChecking.Telescope
 import Agda.TypeChecking.Free
 import Agda.TypeChecking.Records
@@ -86,12 +87,6 @@ instance Monoid Unifiable where
 reportPostponing :: Unify ()
 reportPostponing = tell Possibly
 
-{-
--- | Check whether unification proceeded without postponement.
-listenResult :: Unify () -> Unify UnifyOutput
-listenResult m = snd <$> listen m
--}
-
 -- | Check whether unification proceeded without postponement.
 ifClean :: Unify () -> Unify a -> Unify a -> Unify a
 ifClean m t e = do
@@ -100,16 +95,6 @@ ifClean m t e = do
     Definitely -> t
     Possibly ->   e
 
-{-
--- | Check whether unification proceeded without postponement.
-ifClean :: Unify a -> (a -> Unify b) -> (a -> Unify b) -> Unify b
-ifClean m t e = do
-  (ok, a) <- listen m
-  case ok of
-    Definitely -> t a
-    Possibly ->   e a
--}
-
 data Equality = Equal TypeHH Term Term
 type Sub = Map Nat Term
 
@@ -134,16 +119,6 @@ constructorMismatch a u v = throwException $ ConstructorMismatch a u v
 constructorMismatchHH :: TypeHH -> Term -> Term -> Unify a
 constructorMismatchHH aHH = constructorMismatch (leftHH aHH)
   -- do not report heterogenity
-{-
-constructorMismatchHH (Hom a) u v = constructorMismatch a u v
-constructorMismatchHH (Het a1 a2) u v = constructorMismatch a1 u v
--}
-
-{-
-instance MonadReader TCEnv Unify where
-  ask           = U . lift  $ ask
-  local k (U m) = U . lift . lift . lift $ local k m
--}
 
 instance MonadState TCState Unify where
   get = U . lift . lift . lift . lift $ get
@@ -153,13 +128,8 @@ instance MonadTCM Unify where
   liftTCM = U . lift . lift . lift . lift
 
 instance Subst Equality where
-  substs us	 (Equal a s t) = Equal (substs us a)	  (substs us s)	     (substs us t)
-  substUnder n u (Equal a s t) = Equal (substUnder n u a) (substUnder n u s) (substUnder n u t)
-
-{- UNUSED
-getSub :: Unify Sub
-getSub = U $ gets uniSub
--}
+  applySubst rho (Equal a s t) =
+    Equal (applySubst rho a) (applySubst rho s) (applySubst rho t)
 
 onSub :: (Sub -> a) -> Unify a
 onSub f = U $ gets $ f . uniSub
@@ -187,19 +157,6 @@ checkEqualityHH (Hom a) u v = do
     if ok then return () else addEquality a u v
 checkEqualityHH aHH@(Het a1 a2) u v = -- reportPostponing -- enter "dirty" mode
     addEqualityHH aHH u v -- postpone, enter "dirty" mode
-{-
-checkEqualityHH (Het a1 a2) u v = noConstraints $ do
-  equalType a1 a2
-  equalTerm a1 u v
--}
-{-
-checkEqualityHH :: MonadTCM tcm => TypeHH -> Term -> Term -> tcm ()
-checkEqualityHH aHH u v = do
-  a  <- forceHom aHH
-  cs <- equalTerm a u v
-  if null cs then return () else addEquality a u v
--}
-
 
 -- | Check whether heterogeneous situation is really homogeneous.
 --   If not, give up.
@@ -217,24 +174,6 @@ addEqualityHH aHH u v = do
   reportPostponing
   U $ modify $ \s -> s { uniConstr = Equal aHH u v : uniConstr s }
 
-{-
-addEquality :: Type -> Term -> Term -> Unify ()
-addEquality a u v = do
-  reportPostponing
-  U $ modify $ \s -> s { uniConstr = Equal a u v : uniConstr s }
--}
-
-{-
-addEquality :: Type -> Term -> Term -> Unify ()
-addEquality a u v = do
-   p <- askPostpone
-   case p of
-     MayPostpone ->  do
-       reportPostponing
-       U $ modify $ \s -> s { uniConstr = Equal a u v : uniConstr s }
-     MayNotPostpone -> checkEquality a u v
--}
-
 takeEqualities :: Unify [Equality]
 takeEqualities = U $ do
   s <- get
@@ -246,7 +185,7 @@ takeEqualities = U $ do
 occursCheck :: Nat -> Term -> Type -> Unify ()
 occursCheck i u a = do
   let fv = freeVars u
-      v  = Var i []
+      v  = var i
   case occurrence i fv of
     -- Andreas, 2011-04-14
     -- a strongly rigid recursive occurrences signals unsolvability
@@ -265,13 +204,20 @@ occursCheck i u a = do
 (|->) :: Nat -> (Term, Type) -> Unify ()
 i |-> (u, a) = do
   occursCheck i u a
-  liftTCM $ reportSDoc "tc.lhs.unify" 15 $ prettyTCM (Var i []) <+> text ":=" <+> prettyTCM u
+  liftTCM $ reportSDoc "tc.lhs.unify" 15 $ prettyTCM (var i) <+> text ":=" <+> prettyTCM u
   modSub $ Map.insert i (killRange u)
+  -- Apply substitution to itself (issue 552)
+  rho  <- onSub id
+  rho' <- traverse ureduce rho
+  modSub $ const rho'
 
-makeSubstitution :: Sub -> [Term]
-makeSubstitution sub = map val [0..]
+makeSubstitution :: Sub -> S.Substitution
+makeSubstitution sub
+  | Map.null sub = idS
+  | otherwise    = map val [0 .. highestIndex] ++# raiseS (highestIndex + 1)
   where
-    val i = maybe (Var i []) id $ Map.lookup i sub
+    highestIndex = fst $ Map.findMax sub
+    val i = maybe (var i) id $ Map.lookup i sub
 
 -- | Apply the current substitution on a term and reduce to weak head normal form.
 class UReduce t where
@@ -280,14 +226,14 @@ class UReduce t where
 instance UReduce Term where
   ureduce u = doEtaContractImplicit $ do
     rho <- onSub makeSubstitution
-    liftTCM $ etaContract =<< normalise (substs rho u)
+    liftTCM $ etaContract =<< normalise (applySubst rho u)
 -- Andreas, 2011-06-22, fix related to issue 423
 -- To make eta contraction work better, I switched reduce to normalise.
 -- I hope the performance penalty is not big (since we are dealing with
 -- l.h.s. terms only).
 -- A systematic solution would make unification type-directed and
 -- eta-insensitive...
---   liftTCM $ etaContract =<< reduce (substs rho u)
+--   liftTCM $ etaContract =<< reduce (applySubst rho u)
 
 instance UReduce Type where
   ureduce (El s t) = El s <$> ureduce t
@@ -302,7 +248,7 @@ instance UReduce t => UReduce (Maybe t) where
 
 -- | Take a substitution σ and ensure that no variables from the domain appear
 --   in the targets. The context of the targets is not changed.
---   TODO: can this be expressed using makeSubstitution and substs?
+--   TODO: can this be expressed using makeSubstitution and applySubst?
 flattenSubstitution :: Substitution -> Substitution
 flattenSubstitution s = foldr instantiate s is
   where
@@ -312,12 +258,11 @@ flattenSubstitution s = foldr instantiate s is
     instantiate :: Nat -> Substitution -> Substitution
     instantiate i s = map (fmap $ inst i u) s
       where
-	Just u = s !! fromIntegral i
+	Just u = s !! i
 
     inst :: Nat -> Term -> Term -> Term
-    inst i u v = substs us v
-      where us = [var j | j <- [0..i - 1] ] ++ [u] ++ [var j | j <- [i + 1..] ]
-	    var j = Var j []
+    inst i u v = applySubst us v
+      where us = [var j | j <- [0..i - 1] ] ++# u :# raiseS (i + 1)
 
 data UnificationResult = Unifies Substitution | NoUnify Type Term Term | DontKnow TCErr
 
@@ -325,7 +270,7 @@ data UnificationResult = Unifies Substitution | NoUnify Type Term Term | DontKno
 data HomHet a
   = Hom a    -- ^ homogeneous
   | Het a a  -- ^ heterogeneous
-  deriving (Typeable, Data, Show, Eq, Ord, Functor, Foldable, Traversable)
+  deriving (Typeable, Show, Eq, Ord, Functor, Foldable, Traversable)
 
 isHom :: HomHet a -> Bool
 isHom Hom{} = True
@@ -343,13 +288,8 @@ rightHH :: HomHet a -> a
 rightHH (Hom a) = a
 rightHH (Het a1 a2) = a2
 
-instance (Raise a) => Raise (HomHet a) where
-  raiseFrom n m  = fmap (raiseFrom n m)
-  renameFrom n f = fmap (renameFrom n f)
-
 instance (Subst a) => Subst (HomHet a) where
-  substUnder n u = fmap (substUnder n u)
-  substs us      = fmap (substs us)
+  applySubst rho u = fmap (applySubst rho) u
 
 instance (PrettyTCM a) => PrettyTCM (HomHet a) where
   prettyTCM (Hom a) = prettyTCM a
@@ -358,7 +298,7 @@ instance (PrettyTCM a) => PrettyTCM (HomHet a) where
 type TermHH    = HomHet Term
 type TypeHH    = HomHet Type
 --type FunViewHH = FunV TypeHH
-type TelHH     = Tele (Arg TypeHH)
+type TelHH     = Tele (Dom TypeHH)
 type TelViewHH = TelV TypeHH
 
 absAppHH :: SubstHH t tHH => Abs t -> TermHH -> tHH
@@ -405,6 +345,10 @@ instance SubstHH a b => SubstHH (Arg a) (Arg b) where
   substUnderHH n u = fmap $ substUnderHH n u
   trivialHH = fmap trivialHH
 
+instance SubstHH a b => SubstHH (Dom a) (Dom b) where
+  substUnderHH n u = fmap $ substUnderHH n u
+  trivialHH = fmap trivialHH
+
 instance SubstHH a b => SubstHH (Abs a) (Abs b) where
   substUnderHH n u (Abs   x v) = Abs x $ substUnderHH (n + 1) u v
   substUnderHH n u (NoAbs x v) = NoAbs x $ substUnderHH n u v
@@ -452,7 +396,7 @@ unifyIndices flex a us vs = liftTCM $ do
       Left (StronglyRigidOccurrence a u v)  -> return $ NoUnify a u v
       Left (GenericUnifyException     err)  -> fail err
       Right _                               -> do
-        checkEqualities $ substs (makeSubstitution s) eqs
+        checkEqualities $ applySubst (makeSubstitution s) eqs
         let n = maximum $ (-1) : flex
         return $ Unifies $ flattenSubstitution [ Map.lookup i s | i <- [0..n] ]
   `catchError` \err -> return $ DontKnow err
@@ -460,6 +404,7 @@ unifyIndices flex a us vs = liftTCM $ do
     flexible i = i `elem` flex
 
     flexibleTerm (Var i []) = flexible i
+    flexibleTerm (Shared p) = flexibleTerm (derefPtr p)
     flexibleTerm _          = False
 
     {- Andreas, 2011-09-12
@@ -501,7 +446,7 @@ unifyIndices flex a us vs = liftTCM $ do
     unifyConArgs _ [] (_ : _) = __IMPOSSIBLE__
     unifyConArgs _ []      [] = return ()
     unifyConArgs EmptyTel _ _ = __IMPOSSIBLE__
-    unifyConArgs tel0@(ExtendTel a@(Arg _ rel bHH) tel) us0@(arg@(Arg _ _ u) : us) vs0@(Arg _ _ v : vs) = do
+    unifyConArgs tel0@(ExtendTel a@(Dom _ rel bHH) tel) us0@(arg@(Arg _ _ u) : us) vs0@(Arg _ _ v : vs) = do
       liftTCM $ reportSDoc "tc.lhs.unify" 15 $ sep
         [ text "unifyConArgs"
 	-- , nest 2 $ parens (prettyTCM tel0)
@@ -513,7 +458,7 @@ unifyIndices flex a us vs = liftTCM $ do
         (text $ "tel0 = " ++ show tel0)
 
 
-      -- Andreas, Ulf, 2011-09-08 (AIM XVI)
+      -- Andreas, Ulf, 2011-09-08 (AIM XIV)
       -- in case of dependent function type, we cannot postpone
       -- unification of u and v, otherwise us or vs might be ill-typed
       -- skip irrelevant parts
@@ -531,11 +476,6 @@ unifyIndices flex a us vs = liftTCM $ do
       unifyConArgs (tel `absAppHH` uHH) us vs
 
 
-{-
-    unifyArgs :: Type -> [Arg Term] -> [Arg Term] -> Unify ()
-    unifyArgs a = unifyArgsHH (Hom a)
--}
-
     -- | Used for arguments of a 'Def', not 'Con'.
     unifyArgs :: Type -> [Arg Term] -> [Arg Term] -> Unify ()
     unifyArgs _ (_ : _) [] = __IMPOSSIBLE__
@@ -549,21 +489,22 @@ unifyIndices flex a us vs = liftTCM $ do
 	, nest 2 $ prettyList $ map prettyTCM vs0
         ]
       a <- ureduce a  -- Q: reduce sufficient?
-      case unEl a of
+      case ignoreSharing $ unEl a of
 	Pi b _  -> do
           -- Andreas, Ulf, 2011-09-08 (AIM XVI)
           -- in case of dependent function type, we cannot postpone
           -- unification of u and v, otherwise us or vs might be ill-typed
           let dep = dependent $ unEl a
           -- skip irrelevant parts
-	  unless (argRelevance b == Irrelevant) $
+	  unless (domRelevance b == Irrelevant) $
             (if dep then noPostponing else id) $
-              unify (unArg b) u v
+              unify (unDom b) u v
           arg <- traverse ureduce arg
 	  unifyArgs (a `piApply` [arg]) us vs
 	_	  -> __IMPOSSIBLE__
       where dependent (Pi _ NoAbs{}) = False
             dependent (Pi b c)       = 0 `relevantIn` absBody c
+            dependent (Shared p)     = dependent (derefPtr p)
             dependent _              = False
 
     -- | Check using conversion check.
@@ -604,9 +545,10 @@ unifyIndices flex a us vs = liftTCM $ do
     unifyHH aHH u v = do
       u <- liftTCM . constructorForm =<< ureduce u
       v <- liftTCM . constructorForm =<< ureduce v
+      aHH <- ureduce aHH
       liftTCM $ reportSDoc "tc.lhs.unify" 15 $
 	sep [ text "unifyHH"
-	    , nest 2 $ parens $ prettyTCM u
+	    , nest 2 $ (parens $ prettyTCM u) <+> text "=?="
 	    , nest 2 $ parens $ prettyTCM v
 	    , nest 2 $ text ":" <+> prettyTCM aHH
 	    ]
@@ -614,7 +556,7 @@ unifyIndices flex a us vs = liftTCM $ do
       isSizeName <- liftTCM isSizeNameTest
 
       -- check whether types have the same shape
-      (aHH, sh) <- shapeViewHH =<< ureduce aHH
+      (aHH, sh) <- shapeViewHH aHH
       case sh of
         ElseSh  -> checkEqualityHH aHH u v -- not a type or not same types
 
@@ -641,18 +583,18 @@ unifyIndices flex a us vs = liftTCM $ do
 	    , nest 2 $ prettyTCM v <> if flexibleTerm v then text " (flexible)" else empty
 	    , nest 2 $ text ":" <+> prettyTCM aHH
 	    ]
-      case (u, v) of
+      case (ignoreSharing u, ignoreSharing v) of
         -- Ulf, 2011-06-19
         -- We don't want to worry about levels here.
-        (Level l, v) -> do
+        (Level l, _) -> do
             u <- liftTCM $ reallyUnLevelView l
             unifyAtomHH aHH u v
-        (u, Level l) -> do
+        (_, Level l) -> do
             v <- liftTCM $ reallyUnLevelView l
             unifyAtomHH aHH u v
 	(Var i us, Var j vs) | i == j  -> checkEqualityHH aHH u v
-	(Var i [], v) | homogeneous && flexible i -> i |->> (v, a)
-	(u, Var j []) | homogeneous && flexible j -> j |->> (u, a)
+	(Var i [], _) | homogeneous && flexible i -> i |->> (v, a)
+	(_, Var j []) | homogeneous && flexible j -> j |->> (u, a)
 	(Con c us, Con c' vs)
           | c == c' -> do
               r <- ureduce =<< liftTCM (dataOrRecordTypeHH c aHH)
@@ -684,7 +626,7 @@ unifyIndices flex a us vs = liftTCM $ do
 
         -- We can instantiate metas if the other term is inert (constructor application)
         -- Andreas, 2011-09-13: test/succeed/IndexInference needs this feature.
-        (MetaV m us, v) | homogeneous -> do
+        (MetaV m us, _) | homogeneous -> do
             ok <- liftTCM $ instMeta a m us v
             liftTCM $ reportSDoc "tc.lhs.unify" 40 $
               vcat [ fsep [ text "inst meta", text $ if ok then "(ok)" else "(not ok)" ]
@@ -692,7 +634,7 @@ unifyIndices flex a us vs = liftTCM $ do
                    ]
             if ok then unify a u v
                   else addEquality a u v
-        (u, MetaV m vs) | homogeneous -> do
+        (_, MetaV m vs) | homogeneous -> do
             ok <- liftTCM $ instMeta a m vs u
             liftTCM $ reportSDoc "tc.lhs.unify" 40 $
               vcat [ fsep [ text "inst meta", text $ if ok then "(ok)" else "(not ok)" ]
@@ -728,104 +670,10 @@ unifyIndices flex a us vs = liftTCM $ do
 
     unify :: Type -> Term -> Term -> Unify ()
     unify a = unifyHH (Hom a)
-{-
-    unify :: Type -> Term -> Term -> Unify ()
-    unify a u v = do
-      u <- constructorForm =<< ureduce u
-      v <- constructorForm =<< ureduce v
-      reportSDoc "tc.lhs.unify" 15 $
-	sep [ text "unify"
-	    , nest 2 $ parens $ prettyTCM u
-	    , nest 2 $ parens $ prettyTCM v
-	    , nest 2 $ text ":" <+> prettyTCM a
-	    ]
-      isSize <- isSizeType a
-      if isSize then unifySizes u v
-                else unifyAtom a u v
--}
 
     unifyAtom :: Type -> Term -> Term -> Unify ()
     unifyAtom a = unifyAtomHH (Hom a)
 
-{-
-    unifyAtom :: Type -> Term -> Term -> Unify ()
-    unifyAtom a u v = do
-      reportSDoc "tc.lhs.unify" 15 $
-	sep [ text "unifyAtom"
-	    , nest 2 $ prettyTCM u <> if flexibleTerm u then text " (flexible)" else empty
-            , nest 2 $ text "=?="
-	    , nest 2 $ prettyTCM v <> if flexibleTerm v then text " (flexible)" else empty
-	    , nest 2 $ text ":" <+> prettyTCM a
-	    ]
-      case (u, v) of
-        -- Ulf, 2011-06-19
-        -- We don't want to worry about levels here.
-        (Level l, v) -> do
-            u <- liftTCM $ reallyUnLevelView l
-            unifyAtom a u v
-        (u, Level l) -> do
-            v <- liftTCM $ reallyUnLevelView l
-            unifyAtom a u v
-        -- Andreas, 2011-05-30
-        -- Force equality now rather than postponing it with addEquality
-	(Var i us, Var j vs) | i == j  -> checkEquality a u v
-	(Var i [], v) | flexible i -> i |->> (v, a)
-	(u, Var j []) | flexible j -> j |->> (u, a)
-	(Con c us, Con c' vs)
-          | c == c' -> do
-              a' <- dataOrRecordType c a
-              unifyArgs a' us vs
-          | otherwise -> constructorMismatch a u v
-        -- Definitions are ok as long as they can't reduce (i.e. datatypes/axioms)
-	(Def d us, Def d' vs)
-          | d == d' -> do
-              -- d must be a data, record or axiom
-              def <- getConstInfo d
-              let ok = case theDef def of
-                    Datatype{} -> True
-                    Record{}   -> True
-                    Axiom{}    -> True
-                    _          -> False
-              inj <- optInjectiveTypeConstructors <$> pragmaOptions
-              if inj && ok
-                then unifyArgs (defType def) us vs
-                else addEquality a u v
-{- Andreas: checkEquality breaks Data.Vec.Equality
-   we need an injectivity test for data types, Vec A n is injective!
-                else checkEquality a u v
-                -- Andreas: force equality now instead of postponing
-                -- We do not want to end up in a heterogeneous situation,
-                -- where u and v have different types
--}
-          -- Andreas, 2011-05-30: if heads disagree, abort
-          -- but do not raise "mismatch" because otherwise type constructors
-          -- would be distinct
-          | otherwise -> typeError $ UnequalTerms CmpEq u v a
-        (Lit l1, Lit l2)
-          | l1 == l2  -> return ()
-          | otherwise -> constructorMismatch a u v
-
-        -- We can instantiate metas if the other term is inert (constructor application)
-        (MetaV m us, v) -> do
-            ok <- liftTCM $ instMeta a m us v
-            reportSDoc "tc.lhs.unify" 40 $
-              vcat [ fsep [ text "inst meta", text $ if ok then "(ok)" else "(not ok)" ]
-                   , nest 2 $ sep [ prettyTCM u, text ":=", prettyTCM =<< normalise u ]
-                   ]
-            if ok then unify a u v
-                  else addEquality a u v
-        (u, MetaV m vs) -> do
-            ok <- liftTCM $ instMeta a m vs u
-            reportSDoc "tc.lhs.unify" 40 $
-              vcat [ fsep [ text "inst meta", text $ if ok then "(ok)" else "(not ok)" ]
-                   , nest 2 $ sep [ prettyTCM v, text ":=", prettyTCM =<< normalise v ]
-                   ]
-            if ok then unify a u v
-                  else addEquality a u v
-        -- Andreas, 2011-05-30: If I put checkEquality below, then Issue81 fails
-        -- because there are definitions blocked by flexibles that need postponement
-	_  -> addEquality a u v
--}
     -- The contexts are transient when unifying, so we should just instantiate to
     -- constructor heads and generate fresh metas for the arguments. Beware of
     -- constructors that aren't fully applied.
@@ -841,14 +689,14 @@ unifyIndices flex a us vs = liftTCM $ do
         Just (v', b, vs) -> do
             margs <- do
               -- The new metas should have the same dependencies as the original meta
-              mi <- getMetaInfo <$> lookupMeta m
+              mv <- lookupMeta m
 
               -- Only generate metas for the arguments v' is actually applied to
               -- (in case of partial application)
               TelV tel0 _ <- telView b
               let tel = telFromList $ take (length vs) $ telToList tel0
                   b'  = telePi tel (sort Prop)
-              withMetaInfo mi $ do
+              withMetaInfo' mv $ do
                 tel <- getContextTelescope
                 -- important: create the meta in the same environment as the original meta
                 newArgsMetaCtx b' tel us
@@ -858,13 +706,8 @@ unifyIndices flex a us vs = liftTCM $ do
 
     inertApplication :: Type -> Term -> TCM (Maybe (Term, Type, Args))
     inertApplication a v =
-      case v of
+      case ignoreSharing v of
         Con c vs -> fmap (\ b -> (Con c [], b, vs)) <$> dataOrRecordType c a
-{-
-        Con c vs -> do
-          b <- dataOrRecordType c a
-          return $ Just (Con c [], b, vs)
--}
         Def d vs -> do
           def <- getConstInfo d
           let ans = Just (Def d [], defType def, vs)
@@ -896,7 +739,10 @@ dataOrRecordType' ::
            --   data/record parameters
 dataOrRecordType' c a = do
   -- The telescope ends with a datatype or a record.
-  TelV _ (El _ (Def d args)) <- telView a
+  (d, args) <- do
+    TelV _ (El _ def) <- telView a
+    let Def d args = ignoreSharing def
+    return (d, args)
   def <- theDef <$> getConstInfo d
   r <- case def of
     Datatype{dataPars = n} -> Just . ((,) n) . defType <$> getConstInfo c
@@ -934,61 +780,22 @@ isEtaRecordTypeHH (Het a1 a2) = do
     _ -> return Nothing
 
 
-{-
-dataOrRecordType :: MonadTCM tcm
-  => QName -- ^ Constructor name.
-  -> Type  -- ^ Type of constructor application (must end in data/record).
-  -> tcm (Maybe Type) -- ^ Type of constructor, applied to pars.
-dataOrRecordType c a = fmap snd <$> dataOrRecordType' c a
-
-dataOrRecordType' :: MonadTCM tcm
-  => QName -- ^ Constructor name.
-  -> Type  -- ^ Type of constructor application (must end in data/record).
-  -> tcm (Maybe (QName, -- ^ Name of data/record type.
-                 Type))  -- ^ Type of constructor, applied to pars.
-dataOrRecordType' c a = do
-  -- The telescope ends with a datatype or a record.
-  TelV _ (El _ (Def d args)) <- telView a
-  def <- theDef <$> getConstInfo d
-  r <- case def of
-    Datatype{dataPars = n} -> Just . ((,) n) . defType <$> getConstInfo c
-    Record  {recPars  = n} -> Just . ((,) n) <$> getRecordConstructorType d
-    _		           -> return Nothing
-  return $ fmap (\ (n, a') -> (d, a' `apply` genericTake n args)) r
-
--- | Heterogeneous situation.
---   @a1@ and @a2@ need to end in same datatype/record.
-dataOrRecordTypeHH :: MonadTCM tcm
-  => QName      -- ^ Constructor name.
-  -> TypeHH     -- ^ Type(s) of constructor application (must end in same data/record).
-  -> tcm (Maybe TypeHH) -- ^ Type of constructor, instantiated possibly heterogeneously to parameters.
-dataOrRecordTypeHH c (Hom a) = fmap Hom <$> dataOrRecordType c a
-dataOrRecordTypeHH c (Het a1 a2) = do
-  r1 <- dataOrRecordType' c a1
-  r2 <- dataOrRecordType' c a2  -- b2 may have different parameters than b1!
-  return $ case (r1, r2) of
-    (Just (d1, b1), Just (d2, b2)) | d1 == d2 -> Just $ Het b1 b2
-    _ -> Nothing
--}
-
 -- | Views an expression (pair) as type shape.  Fails if not same shape.
 data ShapeView a
-  = PiSh (Arg a) (Abs a)
-  | FunSh (Arg a) a
+  = PiSh (Dom a) (Abs a)
+  | FunSh (Dom a) a
   | DefSh QName   -- ^ data/record
   | VarSh Nat     -- ^ neutral type
   | LitSh Literal -- ^ built-in type
   | SortSh
   | MetaSh        -- ^ some meta
   | ElseSh        -- ^ not a type or not definitely same shape
-  deriving (Typeable, Data, Show, Eq, Ord, Functor)
+  deriving (Typeable, Show, Eq, Ord, Functor)
 
 -- | Return the type and its shape.  Expects input in (u)reduced form.
 shapeView :: Type -> Unify (Type, ShapeView Type)
 shapeView t = do
---  t <- liftTCM $ reduce t  -- DO NOT REDUCE!
--- --  t <- ureduce t  -- BUG!! substitutes bound variables in telescope!
-  return . (t,) $ case unEl t of
+  return . (t,) $ case ignoreSharing $ unEl t of
     Pi a (NoAbs _ b) -> FunSh a b
     Pi a (Abs x b) -> PiSh a (Abs x b)
     Def d vs       -> DefSh d
@@ -1008,13 +815,13 @@ shapeViewHH (Het a1 a2) = do
   (a2, sh2) <- shapeView a2
   return . (Het a1 a2,) $ case (sh1, sh2) of
 
-    (PiSh (Arg h1 r1 a1) b1, PiSh (Arg h2 r2 a2) b2)
+    (PiSh (Dom h1 r1 a1) b1, PiSh (Dom h2 r2 a2) b2)
       | h1 == h2 ->
-      PiSh (Arg h1 (min r1 r2) (Het a1 a2)) (Abs (absName b1) (Het (absBody b1) (absBody b2)))
+      PiSh (Dom h1 (min r1 r2) (Het a1 a2)) (Abs (absName b1) (Het (absBody b1) (absBody b2)))
 
-    (FunSh (Arg h1 r1 a1) b1, FunSh (Arg h2 r2 a2) b2)
+    (FunSh (Dom h1 r1 a1) b1, FunSh (Dom h2 r2 a2) b2)
       | h1 == h2 ->
-      FunSh (Arg h1 (min r1 r2) (Het a1 a2)) (Het b1 b2)
+      FunSh (Dom h1 (min r1 r2) (Het a1 a2)) (Het b1 b2)
 
     (DefSh d1, DefSh d2) | d1 == d2 -> DefSh d1
     (VarSh x1, VarSh x2) | x1 == x2 -> VarSh x1
