@@ -53,10 +53,12 @@ import qualified Agda.Compiler.Epic.Interface as Epic
 
 import Agda.Syntax.Common
 import Agda.Syntax.Concrete.Name as C
+import qualified Agda.Syntax.Abstract as A
 import Agda.Syntax.Abstract.Name as A
+import Agda.Syntax.Info
 import Agda.Syntax.Internal as I
 import Agda.Syntax.Scope.Base
-import Agda.Syntax.Position (Position(..), Range)
+import Agda.Syntax.Position (Position, Range, noRange)
 import qualified Agda.Syntax.Position as P
 import Agda.Syntax.Common
 import Agda.Syntax.Fixity
@@ -69,10 +71,13 @@ import Agda.Interaction.FindFile
 
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.CompiledClause
+import Agda.TypeChecking.Pretty
 import Agda.Utils.FileName
 import Agda.Utils.Monad
 import Agda.Utils.Tuple
 import Agda.Utils.Permutation
+import Agda.Utils.HashMap (HashMap)
+import qualified Agda.Utils.HashMap as HMap
 
 #include "../undefined.h"
 import Agda.Utils.Impossible
@@ -82,7 +87,7 @@ import Agda.Utils.Impossible
 -- 32-bit machines). Word64 does not have these problems.
 
 currentInterfaceVersion :: Word64
-currentInterfaceVersion = 20111026 * 10 + 0
+currentInterfaceVersion = 20121106 * 10 + 0
 
 -- | Constructor tag (maybe omitted) and argument indices.
 
@@ -101,25 +106,17 @@ data Dict = Dict{ nodeD     :: !(HashTable Node    Int32)
                 , stringD   :: !(HashTable String  Int32)
                 , integerD  :: !(HashTable Integer Int32)
                 , doubleD   :: !(HashTable Double  Int32)
+                , termD     :: !(HashTable (Ptr Term) Int32)
                 , nodeC     :: !(IORef Int32)  -- counters for fresh indexes
                 , stringC   :: !(IORef Int32)
                 , integerC  :: !(IORef Int32)
                 , doubleC   :: !(IORef Int32)
+                , sharingStats :: !(IORef (Int32, Int32))
                 , fileMod   :: !SourceToModule
                 }
 
 data U    = forall a . Typeable a => U !a
-type Memo = HashTable (Int32, TypeRep') U    -- (node index, type rep)
-
--- | A newtype is used to avoid introducing an orphan instance of
--- Hashable.
-newtype TypeRep' = TypeRep' { unTypeRep' :: TypeRep }
-  deriving Eq
-
--- | With direct access to the internals of 'TypeRep' this instance
--- could presumably be improved.
-instance Hashable TypeRep' where
-  hash = hash . show . unTypeRep'
+type Memo = HashTable (Int32, TypeRep) U    -- (node index, type rep)
 
 data St = St
   { nodeE     :: !(Array Int32 Node)
@@ -161,12 +158,17 @@ class Typeable a => EmbPrj a where
 encode :: EmbPrj a => a -> TCM ByteString
 encode a = do
     fileMod <- sourceToModule
-    liftIO $ do
-      newD@(Dict nD sD iD dD _ _ _ _ _) <- emptyDict fileMod
+    (x, shared, total) <- liftIO $ do
+      newD@(Dict nD sD iD dD _ _ _ _ _ stats _) <- emptyDict fileMod
       root <- runReaderT (icode a) newD
       nL <- l nD; sL <- l sD; iL <- l iD; dL <- l dD
-      return $ B.encode currentInterfaceVersion `L.append`
-               G.compress (B.encode (root, nL, sL, iL, dL))
+      (shared, total) <- readIORef stats
+      return (B.encode currentInterfaceVersion `L.append`
+              G.compress (B.encode (root, nL, sL, iL, dL)), shared, total)
+    verboseS "profile.sharing" 10 $ do
+      tickN "pointers (reused)" $ fromIntegral shared
+      tickN "pointers" $ fromIntegral total
+    return x
   where
   l h = List.map fst . List.sortBy (compare `on` snd) <$> H.toList h
 
@@ -184,11 +186,11 @@ decode s = do
   -- input is malformed. The decoder is (intended to be) strict enough
   -- to ensure that all such errors can be caught by the handler here.
 
-  (mf, x) <- liftIO $ E.handle (\(E.ErrorCall {}) -> noResult) $ do
+  (mf, r) <- liftIO $ E.handle (\(E.ErrorCall s) -> noResult s) $ do
 
     (ver, s, _) <- return $ B.runGetState B.get s 0
     if ver /= currentInterfaceVersion
-     then noResult
+     then noResult "Wrong interface version."
      else do
 
       ((r, nL, sL, iL, dL), s, _) <-
@@ -196,27 +198,31 @@ decode s = do
       if s /= L.empty
          -- G.decompress seems to throw away garbage at the end, so
          -- the then branch is possibly dead code.
-       then noResult
+       then noResult "Garbage at end."
        else do
 
         st <- St (ar nL) (ar sL) (ar iL) (ar dL)
                 <$> liftIO H.new
                 <*> return mf <*> return incs
         (r, st) <- runStateT (runErrorT (value r)) st
-        return (Just (modFile st), case r of
-          Left  _ -> Nothing
-          Right x -> Just x)
+        return (Just (modFile st), r)
 
   case mf of
     Nothing -> return ()
     Just mf -> modify $ \s -> s { stModuleToSource = mf }
 
-  return x
+  case r of
+    Right x   -> return (Just x)
+    Left  err -> do
+      reportSDoc "import.iface" 5 $
+        text "Error when decoding interface file:" $+$
+        nest 2 (prettyTCM err)
+      return Nothing
 
   where
   ar l = listArray (0, List.genericLength l - 1) l
 
-  noResult = return (Nothing, Nothing)
+  noResult s = return (Nothing, Left $ GenericError s)
 
 -- | Encodes something. To ensure relocatability file paths in
 -- positions are replaced with module names.
@@ -373,8 +379,8 @@ instance EmbPrj C.QName where
                            valu _      = malformed
 
 instance EmbPrj Scope where
-  icode (Scope a b c d) = icode4' a b c d
-  value = vcase valu where valu [a, b, c, d] = valu4 Scope a b c d
+  icode (Scope a b c d e) = icode5' a b c d e
+  value = vcase valu where valu [a, b, c, d, e] = valu5 Scope a b c d e
                            valu _            = malformed
 
 instance EmbPrj NameSpaceId where
@@ -414,8 +420,11 @@ instance EmbPrj AbstractModule where
 instance EmbPrj KindOfName where
   icode DefName = icode0'
   icode ConName = icode0 1
+  icode FldName = icode0 2
+  icode PatternSynName = icode0 2
   value = vcase valu where valu []  = valu0 DefName
                            valu [1] = valu0 ConName
+                           valu [2] = valu0 FldName
                            valu _   = malformed
 
 instance EmbPrj Agda.Syntax.Fixity.Fixity where
@@ -446,6 +455,10 @@ instance EmbPrj A.QName where
   value = vcase valu where valu [a, b] = valu2 A.QName a b
                            valu _      = malformed
 
+instance EmbPrj A.AmbiguousQName where
+  icode (A.AmbQ a) = icode a
+  value n = A.AmbQ `fmap` value n
+
 instance EmbPrj A.ModuleName where
   icode (A.MName a) = icode a
   value n = A.MName `fmap` value n
@@ -455,6 +468,136 @@ instance EmbPrj A.Name where
   value = vcase valu where valu [a, b, c, d] = valu4 A.Name a b c d
                            valu _            = malformed
 
+instance (EmbPrj s, EmbPrj t) => EmbPrj (Named s t) where
+  icode (Named a b) = icode2' a b
+  value = vcase valu where valu [a, b] = valu2 Named a b
+                           valu _      = malformed
+
+instance EmbPrj A.Expr where
+  icode (A.Var n)               = icode1 0 n
+  icode (A.Def n)               = icode1 1 n
+  icode (A.Con ns)              = icode1 2 ns
+  icode (A.Lit l)               = icode1 3 l
+  icode (A.QuestionMark _)      = icode0 4
+  icode (A.Underscore _)        = icode0 5
+  icode (A.App _ a b)           = icode2 6 a b
+  icode (A.WithApp _ a b)       = icode2 7 a b
+  icode (A.Lam  _ a b)          = icode2 8 a b
+  icode (A.AbsurdLam _ a)       = icode1 9 a
+  icode (A.ExtendedLam _ _ _ _) = icode0 10
+  icode (A.Pi   _ a b)          = icode2 11 a b
+  icode (A.Fun  _ a b)          = icode2 12 a b
+  icode (A.Set  _ a)            = icode1 13 a
+  icode (A.Prop _)              = icode0 14
+  icode (A.Let  _ a b)          = icode2 15 a b
+  icode (A.ETel a)              = icode1 16 a
+  icode (A.Rec  _ a)            = icode1 17 a
+  icode (A.RecUpdate _ a b)     = icode2 18 a b
+  icode (A.ScopedExpr a b)      = icode2 19 a b
+  icode (A.QuoteGoal _ a b)     = icode2 20 a b
+  icode (A.Quote _)             = icode0 21
+  icode (A.QuoteTerm _)         = icode0 22
+  icode (A.Unquote _)           = icode0 23
+  icode (A.DontCare a)          = icode1 24 a
+  icode (A.PatternSyn a)        = icode1 25 a
+
+  value = vcase valu
+    where
+      valu [0, a]     = valu1 A.Var a
+      valu [1, a]     = valu1 A.Def a
+      valu [2, a]     = valu1 A.Con a
+      valu [3, a]     = valu1 A.Lit a
+      valu [4]        = valu0 (A.QuestionMark emptyMetaInfo)
+      valu [5]        = valu0 (A.Underscore emptyMetaInfo)
+      valu [6, a, b]  = valu2 (A.App i) a b
+      valu [7, a, b]  = valu2 (A.WithApp i) a b
+      valu [8, a, b]  = valu2 (A.Lam i) a b
+      valu [9, a]     = valu1 (A.AbsurdLam i) a
+      valu [10]       = throwError $ NotSupported
+                            "importing pattern synonym containing extended lambda"
+      valu [11, a, b] = valu2 (A.Pi i) a b
+      valu [12, a, b] = valu2 (A.Fun i) a b
+      valu [13, a]    = valu1 (A.Set i) a
+      valu [14]       = valu0 (A.Prop i)
+      valu [15, a, b] = valu2 (A.Let i) a b
+      valu [16, a]    = valu1 A.ETel a
+      valu [17, a]    = valu1 (A.Rec i) a
+      valu [18, a, b] = valu2 (A.RecUpdate i) a b
+      valu [19, a, b] = valu2 A.ScopedExpr a b
+      valu [20, a, b] = valu2 (A.QuoteGoal i) a b
+      valu [21]       = valu0 (A.Quote i)
+      valu [22]       = valu0 (A.QuoteTerm i)
+      valu [23]       = valu0 (A.Unquote i)
+      valu [24, a]    = valu1 A.DontCare a
+      valu [25, a]    = valu1 A.PatternSyn a
+      valu _          = malformed
+
+      i = ExprRange noRange
+
+instance EmbPrj A.Pattern where
+  icode (A.VarP a)            = icode1 0 a
+  icode (A.ConP _ a b)        = icode2 1 a b
+  icode (A.DefP _ a b)        = icode2 2 a b
+  icode (A.WildP _)           = icode0 3
+  icode (A.AsP _ a b)         = icode2 4 a b
+  icode (A.DotP _ a)          = icode1 5 a
+  icode (A.AbsurdP _)         = icode0 6
+  icode (A.LitP a)            = icode1 7 a
+  icode (A.ImplicitP _)       = icode0 8
+  icode (A.PatternSynP _ a b) = icode2 9 a b
+
+  value = vcase valu
+    where
+     valu [0, a]    = valu1 A.VarP a
+     valu [1, a, b] = valu2 (A.ConP i) a b
+     valu [2, a, b] = valu2 (A.DefP i) a b
+     valu [3]       = valu0 (A.WildP i)
+     valu [4, a, b] = valu2 (A.AsP i) a b
+     valu [5, a]    = valu1 (A.DotP i) a
+     valu [6]       = valu0 (A.AbsurdP i)
+     valu [7, a]    = valu1 (A.LitP) a
+     valu [8]       = valu0 (A.ImplicitP i)
+     valu [9, a, b] = valu2 (A.PatternSynP i) a b
+     valu _         = malformed
+
+     i = PatRange noRange
+
+instance EmbPrj A.LamBinding where
+  icode (A.DomainFree a b c) = icode3 0 a b c
+  icode (A.DomainFull a)     = icode1 1 a
+
+  value = vcase valu where valu [0, a, b, c] = valu3 A.DomainFree a b c
+                           valu [1, a]       = valu1 A.DomainFull a
+                           valu _            = malformed
+
+instance EmbPrj A.TypedBindings where
+  icode (A.TypedBindings a b) = icode2' a b
+
+  value = vcase valu where valu [a, b] = valu2 A.TypedBindings a b
+                           valu _      = malformed
+
+instance EmbPrj A.TypedBinding where
+  icode (A.TBind a b c) = icode3 0 a b c
+  icode (A.TNoBind a)   = icode1 1 a
+
+  value = vcase valu where valu [0, a, b, c] = valu3 A.TBind a b c
+                           valu [1, a]       = valu1 A.TNoBind a
+                           valu _            = malformed
+
+instance EmbPrj A.LetBinding where
+  icode (A.LetBind _ a b c d)  = icode4 0 a b c d
+  icode (A.LetPatBind _ a b )  = icode2 1 a b
+  icode (A.LetApply _ _ _ _ _) = icode0 2
+  icode (A.LetOpen _ _)        = icode0 2
+
+  value = vcase valu
+    where
+      valu [0, a, b, c, d] = valu4 (A.LetBind (LetRange noRange)) a b c d
+      valu [1, a, b]       = valu2 (A.LetPatBind (LetRange noRange)) a b
+      valu [2]             = throwError $ NotSupported
+                                 "importing pattern synonym containing let module"
+      valu _               = malformed
+
 instance EmbPrj NameId where
   icode (NameId a b) = icode2' a b
   value = vcase valu where valu [a, b] = valu2 NameId a b
@@ -465,6 +608,10 @@ instance EmbPrj Signature where
   value = vcase valu where valu [a, b] = valu2 Sig a b
                            valu _      = malformed
 
+instance (Eq k, Hashable k, EmbPrj k, EmbPrj v) => EmbPrj (HashMap k v) where
+  icode m = icode (HMap.toList m)
+  value m = HMap.fromList `fmap` value m
+
 instance EmbPrj Section where
   icode (Section a b) = icode2' a b
   value = vcase valu where valu [a, b] = valu2 Section a b
@@ -487,6 +634,11 @@ instance (EmbPrj a) => EmbPrj (Agda.Syntax.Common.Arg a) where
   value = vcase valu where valu [a, b, c] = valu3 Arg a b c
                            valu _         = malformed
 
+instance (EmbPrj a) => EmbPrj (Agda.Syntax.Common.Dom a) where
+  icode (Dom a b c) = icode3' a b c
+  value = vcase valu where valu [a, b, c] = valu3 Dom a b c
+                           valu _         = malformed
+
 instance EmbPrj Agda.Syntax.Common.Induction where
   icode Inductive   = icode0'
   icode CoInductive = icode0 1
@@ -508,10 +660,12 @@ instance EmbPrj Agda.Syntax.Common.Relevance where
   icode Irrelevant = icode0 1
   icode Forced     = icode0 2
   icode NonStrict  = icode0 3
+  icode UnusedArg  = icode0 4
   value = vcase valu where valu []  = valu0 Relevant
                            valu [1] = valu0 Irrelevant
                            valu [2] = valu0 Forced
                            valu [3] = valu0 NonStrict
+                           valu [4] = valu0 UnusedArg
                            valu _   = malformed
 
 instance EmbPrj I.Type where
@@ -537,16 +691,31 @@ instance EmbPrj I.Term where
   icode (MetaV    a b) = __IMPOSSIBLE__
   icode (DontCare a  ) = icode1 8 a
   icode (Level    a  ) = icode1 9 a
-  value = vcase valu where valu [0, a, b] = valu2 Var   a b
-                           valu [1, a, b] = valu2 Lam   a b
-                           valu [2, a]    = valu1 Lit   a
-                           valu [3, a, b] = valu2 Def   a b
-                           valu [4, a, b] = valu2 Con   a b
-                           valu [5, a, b] = valu2 Pi    a b
-                           valu [7, a]    = valu1 Sort  a
-                           valu [8, a]    = valu1 DontCare a
-                           valu [9, a]    = valu1 Level a
-                           valu _         = malformed
+  icode (Shared p)     = do
+    h  <- asks termD
+    mi <- liftIO $ H.lookup h p
+    st <- asks sharingStats
+    case mi of
+      Just i  -> liftIO $ modifyIORef st (\(a, b) -> ((,) $! a + 1) b) >> return i
+      Nothing -> do
+        liftIO $ modifyIORef st (\(a, b) -> (,) a $! b + 1)
+        n <- icode (derefPtr p)
+        liftIO $ H.insert h p n
+        return n
+
+  value r = vcase valu' r
+    where
+      valu' xs = shared <$> valu xs
+      valu [0, a, b] = valu2 Var   a b
+      valu [1, a, b] = valu2 Lam   a b
+      valu [2, a]    = valu1 Lit   a
+      valu [3, a, b] = valu2 Def   a b
+      valu [4, a, b] = valu2 Con   a b
+      valu [5, a, b] = valu2 Pi    a b
+      valu [7, a]    = valu1 Sort  a
+      valu [8, a]    = valu1 DontCare a
+      valu [9, a]    = valu1 Level a
+      valu _         = malformed
 
 instance EmbPrj Level where
   icode (Max a) = icode1' a
@@ -624,9 +793,9 @@ instance EmbPrj MutualId where
   value n = MutId `fmap` value n
 
 instance EmbPrj Definition where
-  icode (Defn rel a b c d e f) = icode7' rel a b c d e f
-  value = vcase valu where valu [rel, a, b, c, d, e, f] = valu7 Defn rel a b c d e f
-                           valu _                       = malformed
+  icode (Defn rel a b c d e f g h) = icode9' rel a (P.killRange b) c d e f g h
+  value = vcase valu where valu [rel, a, b, c, d, e, f, g, h] = valu9 Defn rel a b c d e f g h
+                           valu _                             = malformed
 
 instance EmbPrj HaskellRepresentation where
   icode (HsType a)   = icode1' a
@@ -688,22 +857,30 @@ instance EmbPrj Polarity where
   icode Covariant     = icode0'
   icode Contravariant = icode0 1
   icode Invariant     = icode0 2
+  icode Nonvariant    = icode0 3
 
   value = vcase valu where
     valu []  = valu0 Covariant
     valu [1] = valu0 Contravariant
     valu [2] = valu0 Invariant
+    valu [3] = valu0 Nonvariant
     valu _   = malformed
 
 instance EmbPrj Occurrence where
-  icode Positive = icode0'
-  icode Negative = icode0 1
-  icode Unused   = icode0 2
+  icode StrictPos = icode0'
+  icode Mixed     = icode0 1
+  icode Unused    = icode0 2
+  icode GuardPos  = icode0 3
+  icode JustPos   = icode0 4
+  icode JustNeg   = icode0 5
 
   value = vcase valu where
-    valu []  = valu0 Positive
-    valu [1] = valu0 Negative
+    valu []  = valu0 StrictPos
+    valu [1] = valu0 Mixed
     valu [2] = valu0 Unused
+    valu [3] = valu0 GuardPos
+    valu [4] = valu0 JustPos
+    valu [5] = valu0 JustNeg
     valu _   = malformed
 
 instance EmbPrj CompiledRepresentation where
@@ -712,20 +889,27 @@ instance EmbPrj CompiledRepresentation where
                            valu _         = malformed
 
 instance EmbPrj Defn where
-  icode Axiom                               = icode0 0
-  icode (Function    a b c d e f g h i)     = icode9 1 a b c d e f g h i
-  icode (Datatype    a b c d e f g h i)     = icode9 2 a b c d e f g h i
-  icode (Record      a b c d e f g h i j k) = icode11 3 a b c d e f g h i j k
-  icode (Constructor a b c d e)             = icode5 4 a b c d e
-  icode (Primitive   a b c d)               = icode4 5 a b c d
+  icode Axiom                                   = icode0 0
+  icode (Function    a b c d e f g h i j)       = icode10 1 a b c d e f g h i j
+  icode (Datatype    a b c d e f g h)           = icode8 2 a b c d e f g h
+  icode (Record      a b c d e f g h i j k l)   = icode12 3 a b c d e f g h i j k l
+  icode (Constructor a b c d e)                 = icode5 4 a b c d e
+  icode (Primitive   a b c d)                   = icode4 5 a b c d
   value = vcase valu where
-    valu [0]                                  = valu0 Axiom
-    valu [1, a, b, c, d, e, f, g, h, i]       = valu9 Function    a b c d e f g h i
-    valu [2, a, b, c, d, e, f, g, h, i]       = valu9 Datatype    a b c d e f g h i
-    valu [3, a, b, c, d, e, f, g, h, i, j, k] = valu11 Record     a b c d e f g h i j k
-    valu [4, a, b, c, d, e]                   = valu5 Constructor a b c d e
-    valu [5, a, b, c, d]                      = valu4 Primitive   a b c d
-    valu _                                    = malformed
+    valu [0]                                    = valu0 Axiom
+    valu [1, a, b, c, d, e, f, g, h, i, j]      = valu10 Function a b c d e f g h i j
+    valu [2, a, b, c, d, e, f, g, h]            = valu8 Datatype a b c d e f g h
+    valu [3, a, b, c, d, e, f, g, h, i, j, k, l]= valu12 Record  a b c d e f g h i j k l
+    valu [4, a, b, c, d, e]                     = valu5 Constructor a b c d e
+    valu [5, a, b, c, d]                        = valu4 Primitive   a b c d
+    valu _                                      = malformed
+
+instance EmbPrj a => EmbPrj (WithArity a) where
+  icode (WithArity a b) = icode2' a b
+
+  value = vcase valu where
+    valu [a, b] = valu2 WithArity a b
+    valu _      = malformed
 
 instance EmbPrj a => EmbPrj (Case a) where
   icode (Branches a b c) = icode3' a b c
@@ -736,7 +920,7 @@ instance EmbPrj a => EmbPrj (Case a) where
 
 instance EmbPrj CompiledClauses where
   icode Fail       = icode0'
-  icode (Done a b) = icode2' a b
+  icode (Done a b) = icode2' a (P.killRange b)
   icode (Case a b) = icode2 2 a b
 
   value = vcase valu where
@@ -855,6 +1039,8 @@ instance EmbPrj HP.OtherAspect where
   icode HP.UnsolvedMeta       = icode0 2
   icode HP.TerminationProblem = icode0 3
   icode HP.IncompletePattern  = icode0 4
+  icode HP.TypeChecks         = icode0 5
+  icode HP.UnsolvedConstraint = icode0 6
 
   value = vcase valu where
     valu [0] = valu0 HP.Error
@@ -862,6 +1048,8 @@ instance EmbPrj HP.OtherAspect where
     valu [2] = valu0 HP.UnsolvedMeta
     valu [3] = valu0 HP.TerminationProblem
     valu [4] = valu0 HP.IncompletePattern
+    valu [5] = valu0 HP.TypeChecks
+    valu [6] = valu0 HP.UnsolvedConstraint
     valu _   = malformed
 
 instance EmbPrj HP.MetaInfo where
@@ -901,10 +1089,19 @@ instance EmbPrj ScopeInfo where
   value = vcase valu where valu [a, b, c, d] = valu4 ScopeInfo a b c d
                            valu _            = malformed
 
+instance EmbPrj HP.CompressedFile where
+  icode (HP.CompressedFile f) = icode1' f
+  value = vcase valu
+    where
+    valu [f] = valu1 HP.CompressedFile f
+    valu _   = malformed
+
 instance EmbPrj Interface where
-  icode (Interface a b c d e f g h i) = icode9' a b c d e f g h i
-  value = vcase valu where valu [a, b, c, d, e, f, g, h, i] = valu9 Interface a b c d e f g h i
-                           valu _                           = malformed
+  icode (Interface a b c d e f g h i j) = icode10' a b c d e f g h i j
+  value = vcase valu
+    where
+      valu [a, b, c, d, e, f, g, h, i, j] = valu10 Interface a b c d e f g h i j
+      valu _                              = malformed
 
 -- This is used for the Epic compiler backend
 instance EmbPrj Epic.EInterface where
@@ -942,8 +1139,6 @@ instance EmbPrj Epic.Tag where
     valu [1, a] = valu1 Epic.PrimTag a
     valu _      = malformed
 
-{-# SPECIALIZE icodeX :: (Dict -> HashTable [Int32] Int32) ->
-                         (Dict -> IORef Int32) -> [Int32] -> S Int32 #-}
 icodeX :: (Eq k, Hashable k) =>
           (Dict -> HashTable k Int32) -> (Dict -> IORef Int32) ->
           k -> S Int32
@@ -968,7 +1163,7 @@ vcase :: forall a . EmbPrj a => (Node -> R a) -> Int32 -> R a
 vcase valu = \ix -> do
     memo <- gets nodeMemo
     (aTyp, maybeU) <- liftIO $ do
-      let aTyp = TypeRep' $ typeOf (undefined :: a)
+      let aTyp = typeOf (undefined :: a)
       maybeU <- H.lookup memo (ix, aTyp)
       return (aTyp, maybeU)
     case maybeU of
@@ -991,6 +1186,8 @@ icode9  tag a b c d e f g h i     = icodeN . (tag :) =<< sequence [icode a, icod
 icode10 tag a b c d e f g h i j   = icodeN . (tag :) =<< sequence [icode a, icode b, icode c, icode d, icode e, icode f, icode g, icode h, icode i, icode j]
 icode11 tag a b c d e f g h i j k = icodeN . (tag :) =<< sequence [icode a, icode b, icode c, icode d, icode e, icode f, icode g, icode h, icode i, icode j, icode k]
 icode12 tag a b c d e f g h i j k l = icodeN . (tag :) =<< sequence [icode a, icode b, icode c, icode d, icode e, icode f, icode g, icode h, icode i, icode j, icode k, icode l]
+icode13 tag a b c d e f g h i j k l m = icodeN . (tag :) =<< sequence [icode a, icode b, icode c, icode d, icode e, icode f, icode g, icode h, icode i, icode j, icode k, icode l, icode m]
+icode14 tag a b c d e f g h i j k l m n = icodeN . (tag :) =<< sequence [icode a, icode b, icode c, icode d, icode e, icode f, icode g, icode h, icode i, icode j, icode k, icode l, icode m, icode n]
 
 icode0'                        = icodeN []
 icode1'  a                     = icodeN =<< sequence [icode a]
@@ -1005,20 +1202,24 @@ icode9'  a b c d e f g h i     = icodeN =<< sequence [icode a, icode b, icode c,
 icode10' a b c d e f g h i j   = icodeN =<< sequence [icode a, icode b, icode c, icode d, icode e, icode f, icode g, icode h, icode i, icode j]
 icode11' a b c d e f g h i j k = icodeN =<< sequence [icode a, icode b, icode c, icode d, icode e, icode f, icode g, icode h, icode i, icode j, icode k]
 icode12' a b c d e f g h i j k l = icodeN =<< sequence [icode a, icode b, icode c, icode d, icode e, icode f, icode g, icode h, icode i, icode j, icode k, icode l]
+icode13' a b c d e f g h i j k l m = icodeN =<< sequence [icode a, icode b, icode c, icode d, icode e, icode f, icode g, icode h, icode i, icode j, icode k, icode l, icode m]
+icode14' a b c d e f g h i j k l m n = icodeN =<< sequence [icode a, icode b, icode c, icode d, icode e, icode f, icode g, icode h, icode i, icode j, icode k, icode l, icode m, icode n]
 
-valu0  z                         = return z
-valu1  z a                       = valu0 z                        `ap` value a
-valu2  z a b                     = valu1 z a                      `ap` value b
-valu3  z a b c                   = valu2 z a b                    `ap` value c
-valu4  z a b c d                 = valu3 z a b c                  `ap` value d
-valu5  z a b c d e               = valu4 z a b c d                `ap` value e
-valu6  z a b c d e f             = valu5 z a b c d e              `ap` value f
-valu7  z a b c d e f g           = valu6 z a b c d e f            `ap` value g
-valu8  z a b c d e f g h         = valu7 z a b c d e f g          `ap` value h
-valu9  z a b c d e f g h i       = valu8 z a b c d e f g h        `ap` value i
-valu10 z a b c d e f g h i j     = valu9 z a b c d e f g h i      `ap` value j
-valu11 z a b c d e f g h i j k   = valu10 z a b c d e f g h i j   `ap` value k
-valu12 z a b c d e f g h i j k l = valu11 z a b c d e f g h i j k `ap` value l
+valu0  z                           = return z
+valu1  z a                         = valu0 z                        `ap` value a
+valu2  z a b                       = valu1 z a                      `ap` value b
+valu3  z a b c                     = valu2 z a b                    `ap` value c
+valu4  z a b c d                   = valu3 z a b c                  `ap` value d
+valu5  z a b c d e                 = valu4 z a b c d                `ap` value e
+valu6  z a b c d e f               = valu5 z a b c d e              `ap` value f
+valu7  z a b c d e f g             = valu6 z a b c d e f            `ap` value g
+valu8  z a b c d e f g h           = valu7 z a b c d e f g          `ap` value h
+valu9  z a b c d e f g h i         = valu8 z a b c d e f g h        `ap` value i
+valu10 z a b c d e f g h i j       = valu9 z a b c d e f g h i      `ap` value j
+valu11 z a b c d e f g h i j k     = valu10 z a b c d e f g h i j   `ap` value k
+valu12 z a b c d e f g h i j k l   = valu11 z a b c d e f g h i j k `ap` value l
+valu13 z a b c d e f g h i j k l m = valu12 z a b c d e f g h i j k l `ap` value m
+valu14 z a b c d e f g h i j k l m n = valu13 z a b c d e f g h i j k l m `ap` value n
 
 -- | Creates an empty dictionary.
 
@@ -1032,8 +1233,10 @@ emptyDict fileMod = Dict
   <*> H.new
   <*> H.new
   <*> H.new
+  <*> H.new
   <*> newIORef 0
   <*> newIORef 0
   <*> newIORef 0
   <*> newIORef 0
+  <*> newIORef (0, 0)
   <*> return fileMod
