@@ -13,6 +13,7 @@ import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Free
 
+import Agda.Utils.List
 import Agda.Utils.Permutation
 import Agda.Utils.Size
 import Agda.Utils.Tuple
@@ -22,67 +23,62 @@ import qualified Agda.Utils.VarSet as Set
 #include "../undefined.h"
 import Agda.Utils.Impossible
 
--- TODO: use rename instead of substs
 -- | The permutation should permute the corresponding telescope. (left-to-right list)
 renameP :: Subst t => Permutation -> t -> t
-renameP p = substs (renaming p)
+renameP p = applySubst (renaming p)
 
--- | If @permute π : [a]Γ -> [a]Δ@, then @substs (renaming π) : Term Γ -> Term Δ@
-renaming :: Permutation -> [Term]
+-- | If @permute π : [a]Γ -> [a]Δ@, then @applySubst (renaming π) : Term Γ -> Term Δ@
+renaming :: Permutation -> Substitution
 renaming p = gamma'
   where
     n	   = size p
     gamma  = permute (reverseP $ invertP $ reverseP p) $ map var [0..]
-    gamma' = gamma ++ map var [n..]
-    var i  = Var i []
+    gamma' = gamma ++# raiseS n
 
 -- | If @permute π : [a]Γ -> [a]Δ@, then @substs (renamingR π) : Term Δ -> Term Γ@
-renamingR :: Permutation -> [Term]
-renamingR p@(Perm n _) = permute (reverseP p) (map var [0..]) ++ map var [n..]
-  where
-    var i  = Var (fromIntegral i) []
+renamingR :: Permutation -> Substitution
+renamingR p@(Perm n _) = permute (reverseP p) (map var [0..]) ++# raiseS n
 
 -- | Flatten telescope: (Γ : Tel) -> [Type Γ]
-flattenTel :: Telescope -> [Arg Type]
+flattenTel :: Telescope -> [Dom Type]
 flattenTel EmptyTel	     = []
 flattenTel (ExtendTel a tel) = raise (size tel + 1) a : flattenTel (absBody tel)
 
 -- | Order a flattened telescope in the correct dependeny order: Γ ->
 --   Permutation (Γ -> Γ~)
-reorderTel :: [Arg Type] -> Maybe Permutation
+reorderTel :: [Dom Type] -> Maybe Permutation
 reorderTel tel = topoSort comesBefore tel'
   where
-    tel' = reverse $ zip [0..] $ reverse tel
-    (i, _) `comesBefore` (_, a) = i `freeIn` unEl (unArg a) -- a tiny bit unsafe
+    tel' = zip (downFrom $ size tel) tel
+    (i, _) `comesBefore` (_, a) = i `freeIn` unEl (unDom a) -- a tiny bit unsafe
 
-reorderTel_ :: [Arg Type] -> Permutation
+reorderTel_ :: [Dom Type] -> Permutation
 reorderTel_ tel = case reorderTel tel of
   Nothing -> __IMPOSSIBLE__
   Just p  -> p
 
 -- | Unflatten: turns a flattened telescope into a proper telescope. Must be
 --   properly ordered.
-unflattenTel :: [String] -> [Arg Type] -> Telescope
+unflattenTel :: [String] -> [Dom Type] -> Telescope
 unflattenTel []	  []	        = EmptyTel
 unflattenTel (x : xs) (a : tel) = ExtendTel a' (Abs x tel')
   where
     tel' = unflattenTel xs tel
-    a'   = substs rho a
-    rho  = replicate (size tel + 1) __IMPOSSIBLE_TERM__ ++ map var [0..]
-      where var i = Var i []
+    a'   = applySubst rho a
+    rho  = parallelS (replicate (size tel + 1) __IMPOSSIBLE_TERM__)
 unflattenTel [] (_ : _) = __IMPOSSIBLE__
 unflattenTel (_ : _) [] = __IMPOSSIBLE__
 
 -- | Get the suggested names from a telescope
 teleNames :: Telescope -> [String]
-teleNames = map (fst . unArg) . telToList
+teleNames = map (fst . unDom) . telToList
 
 teleArgNames :: Telescope -> [Arg String]
-teleArgNames = map (fmap fst) . telToList
+teleArgNames = map (argFromDom . fmap fst) . telToList
 
 teleArgs :: Telescope -> Args
-teleArgs tel =
-  reverse [ Arg h r (Var i []) | (i, Arg h r _) <- zip [0..] $ reverse (telToList tel) ]
+teleArgs tel = [ Arg h r (var i) | (i, Dom h r _) <- zip (downFrom $ size l) l ]
+  where l = telToList tel
 
 -- | A telescope split in two.
 data SplitTel = SplitTel
@@ -127,14 +123,20 @@ telView :: Type -> TCM TelView
 telView = telViewUpTo (-1)
 
 -- | @telViewUpTo n t@ takes off the first @n@ function types of @t@.
--- Takes off all if $n < 0$.
+-- Takes off all if @n < 0@.
 telViewUpTo :: Int -> Type -> TCM TelView
-telViewUpTo 0 t = return $ TelV EmptyTel t
-telViewUpTo n t = do
+telViewUpTo n t = telViewUpTo' n (const True) t
+
+-- | @telViewUpTo' n p t@ takes off $t$
+--   the first @n@ (or arbitrary many if @n < 0@) function domains
+--   as long as they satify @p@.
+telViewUpTo' :: Int -> (Dom Type -> Bool) -> Type -> TCM TelView
+telViewUpTo' 0 p t = return $ TelV EmptyTel t
+telViewUpTo' n p t = do
   t <- reduce t
-  case unEl t of
-    Pi a b -> absV a (absName b) <$> telViewUpTo (n - 1) (absBody b)
-    _      -> return $ TelV EmptyTel t
+  case ignoreSharing $ unEl t of
+    Pi a b | p a -> absV a (absName b) <$> telViewUpTo' (n - 1) p (absBody b)
+    _            -> return $ TelV EmptyTel t
   where
     absV a x (TelV tel t) = TelV (ExtendTel a (Abs x tel)) t
 
@@ -144,6 +146,6 @@ piApplyM :: Type -> Args -> TCM Type
 piApplyM t []           = return t
 piApplyM t (arg : args) = do
   t <- reduce t
-  case (t, arg) of
-    (El _ (Pi  _ b), arg) -> absApp b (unArg arg) `piApplyM` args
-    _                     -> __IMPOSSIBLE__
+  case ignoreSharing $ unEl t of
+    Pi  _ b -> absApp b (unArg arg) `piApplyM` args
+    _       -> __IMPOSSIBLE__
