@@ -193,7 +193,9 @@ instance GenC Hiding where
 
 instance GenC a => GenC (Arg a) where
   genC conf = (\ (h, a) -> Arg h Relevant a) <$> genC conf
---   genC conf = uncurry Arg <$> genC conf
+
+instance GenC a => GenC (Dom a) where
+  genC conf = (\ (h, a) -> Dom h Relevant a) <$> genC conf
 
 instance GenC a => GenC (Abs a) where
   genC conf = Abs "x" <$> genC (extendConf conf)
@@ -403,7 +405,10 @@ instance ShrinkC a b => ShrinkC (Abs a) (Abs b) where
 
 instance ShrinkC a b => ShrinkC (Arg a) (Arg b) where
   shrinkC conf (Arg h r x) = (\ (h,x) -> Arg h r x) <$> shrinkC conf (h, x)
---  shrinkC conf (Arg h x) = uncurry Arg <$> shrinkC conf (h, x)
+  noShrink = fmap noShrink
+
+instance ShrinkC a b => ShrinkC (Dom a) (Dom b) where
+  shrinkC conf (Dom h r x) = (\ (h,x) -> Dom h r x) <$> shrinkC conf (h, x)
   noShrink = fmap noShrink
 
 instance ShrinkC a b => ShrinkC (Blocked a) (Blocked b) where
@@ -434,7 +439,7 @@ instance ShrinkC Type Type where
 instance ShrinkC Term Term where
   shrinkC conf (DontCare _)  = []
   shrinkC conf (Sort Prop) = []
-  shrinkC conf t	   = filter validType $ case t of
+  shrinkC conf t	   = filter validType $ case ignoreSharing t of
     Var i args   -> map unArg args ++
 		    (uncurry Var <$> shrinkC conf (VarName i, NoType args))
     Def d args   -> map unArg args ++
@@ -444,11 +449,12 @@ instance ShrinkC Term Term where
     Lit l	 -> Lit <$> shrinkC conf l
     Level l      -> [] -- TODO
     Lam h b      -> killAbs b : (uncurry Lam <$> shrinkC conf (h, b))
-    Pi a b       -> unEl (unArg a) : unEl (killAbs b) :
+    Pi a b       -> unEl (unDom a) : unEl (killAbs b) :
 		    (uncurry Pi <$> shrinkC conf (a, b))
     Sort s       -> Sort <$> shrinkC conf s
     MetaV m args -> map unArg args ++ (MetaV m <$> shrinkC conf (NoType args))
     DontCare _   -> []
+    Shared{}     -> __IMPOSSIBLE__
     where
       validType t
 	| not (tcIsType conf) = True
@@ -467,7 +473,7 @@ class KillVar a where
   killVar :: Nat -> a -> a
 
 instance KillVar Term where
-  killVar i t = case t of
+  killVar i t = case ignoreSharing t of
     Var j args | j == i	   -> DontCare (Var j [])
 	       | j >  i	   -> Var (j - 1) $ killVar i args
 	       | otherwise -> Var j	  $ killVar i args
@@ -480,6 +486,7 @@ instance KillVar Term where
     Pi a b		   -> uncurry Pi  $ killVar i (a, b)
     MetaV m args	   -> MetaV m	  $ killVar i args
     DontCare mv            -> DontCare    $ killVar i mv
+    Shared{}               -> __IMPOSSIBLE__
 
 instance KillVar Type where
   killVar i (El s t) = El s $ killVar i t
@@ -491,6 +498,9 @@ instance KillVar Telescope where
 instance KillVar a => KillVar (Arg a) where
   killVar i = fmap (killVar i)
 
+instance KillVar a => KillVar (Dom a) where
+  killVar i = fmap (killVar i)
+
 instance KillVar a => KillVar (Abs a) where
   killVar i = fmap (killVar (i + 1))
 
