@@ -24,6 +24,7 @@ import Agda.TypeChecking.Abstract
 import Agda.TypeChecking.EtaContract
 import Agda.TypeChecking.Telescope
 
+import Agda.Utils.List
 import Agda.Utils.Permutation
 import Agda.Utils.Size
 
@@ -58,12 +59,14 @@ buildWithFunction :: QName -> Telescope -> [Arg Pattern] -> Permutation ->
                      Nat -> Nat -> [A.Clause] -> TCM [A.Clause]
 buildWithFunction aux gamma qs perm n1 n cs = mapM buildWithClause cs
   where
-    buildWithClause (A.Clause (LHS i _ ps wps) rhs wh) = do
+    buildWithClause (A.Clause (LHS i (A.LHSProj{}) wps) rhs wh) =
+      typeError $ NotImplemented "with clauses for definitions by copatterns"
+    buildWithClause (A.Clause (LHS i (A.LHSHead _ ps) wps) rhs wh) = do
       let (wps0, wps1) = genericSplitAt n wps
-          ps0          = map (defaultArg . unnamed) wps0
+          ps0          = map defaultNamedArg wps0
       rhs <- buildRHS rhs
       (ps1, ps2)  <- genericSplitAt n1 <$> stripWithClausePatterns gamma qs perm ps
-      let result = A.Clause (LHS i aux (ps1 ++ ps0 ++ ps2) wps1) rhs wh
+      let result = A.Clause (LHS i (A.LHSHead aux (ps1 ++ ps0 ++ ps2)) wps1) rhs wh
       reportSDoc "tc.with" 20 $ vcat
         [ text "buildWithClause returns" <+> prettyA result
         ]
@@ -90,7 +93,7 @@ buildWithFunction aux gamma qs perm n1 n cs = mapM buildWithClause cs
 -}
 stripWithClausePatterns :: Telescope -> [Arg Pattern] -> Permutation -> [NamedArg A.Pattern] -> TCM [NamedArg A.Pattern]
 stripWithClausePatterns gamma qs perm ps = do
-  psi <- insertImplicitPatterns ps gamma
+  psi <- insertImplicitPatterns ExpandLast ps gamma
   unless (size psi == size gamma) $ fail $ "wrong number of arguments in with clause: given " ++ show (size psi) ++ ", expected " ++ show (size gamma)
   reportSDoc "tc.with.strip" 10 $ vcat
     [ text "stripping patterns"
@@ -128,36 +131,37 @@ stripWithClausePatterns gamma qs perm ps = do
           ps <- underAbstraction a tel $ \tel -> strip tel ps qs
           return $ p : ps
 
-        DotP v  -> case namedThing $ unArg p of
+        DotP v  -> case namedArg p of
           A.DotP _ _    -> ok
           A.ImplicitP _ -> ok
           _ -> do
             d <- prettyA p
             typeError $ GenericError $
                 "Inaccessible (dotted) patterns from the parent clause must " ++
-                "also be inaccesible in the with clause, when checking the " ++
+                "also be inaccessible in the with clause, when checking the " ++
                 "pattern " ++ show d ++ ","
           where
             ok = do
               ps <- strip (tel `absApp` v) ps qs
               return $ p : ps
 
-        ConP c _ qs' -> case namedThing $ unArg p of
+        ConP c _ qs' -> case namedArg p of
           A.ConP _ (A.AmbQ cs') ps' -> do
 
-            Con c' [] <- constructorForm =<< reduce (Con c [])
+            Con c' [] <- ignoreSharing <$> (constructorForm =<< reduce (Con c []))
             c <- return $ c' `withRangeOf` c
             let getCon (Con c []) = c
+                getCon (Shared p) = getCon (derefPtr p)
                 getCon _ = __IMPOSSIBLE__
             cs' <- map getCon <$> (mapM constructorForm =<< mapM (\c' -> reduce $ Con c' []) cs')
 
             unless (elem c cs') mismatch
 
             -- The type is a datatype
-            Def d us <- normalise $ unEl (unArg a)
+            Def d us <- ignoreSharing <$> normalise (unEl $ unDom a)
 
             -- Compute the argument telescope for the constructor
-            Con c []    <- constructorForm =<< normalise (Con c [])
+            Con c []    <- ignoreSharing <$> (constructorForm =<< normalise (Con c []))
             Defn {defType = ct, theDef = Constructor{conPars = np}}  <- getConstInfo c
             let ct' = ct `apply` genericTake np us
             TelV tel' _ <- telView ct'
@@ -171,7 +175,8 @@ stripWithClausePatterns gamma qs perm ps = do
                    ]
 
             -- Compute the new telescope
-            let v     = Con c $ reverse [ Arg h r (Var i []) | (i, Arg h r _) <- zip [0..] $ reverse qs' ]
+            let v     = Con c [ Arg h r (var i) | (i, Arg h r _) <- zip (downFrom $ size qs') qs' ]
+--            let v     = Con c $ reverse [ Arg h r (var i) | (i, Arg h r _) <- zip [0..] $ reverse qs' ]
                 tel'' = tel' `abstract` absApp (raise (size tel') tel) v
 
             reportSDoc "tc.with.strip" 15 $ sep
@@ -181,26 +186,28 @@ stripWithClausePatterns gamma qs perm ps = do
               ]
 
             -- Insert implicit patterns (just for the constructor arguments)
-            psi' <- insertImplicitPatterns ps' tel'
+            psi' <- insertImplicitPatterns ExpandLast ps' tel'
             unless (size psi' == size tel') $ typeError $ WrongNumberOfConstructorArguments c (size tel') (size psi')
 
             -- Do it again for everything (is this necessary?)
-            psi' <- insertImplicitPatterns (psi' ++ ps) tel''
+            psi' <- insertImplicitPatterns ExpandLast (psi' ++ ps) tel''
 
             -- Keep going
             strip tel'' psi' (qs' ++ qs)
           _ -> mismatch
 
-        LitP lit -> case namedThing $ unArg p of
+        LitP lit -> case namedArg p of
           A.LitP lit' | lit == lit' -> strip (tel `absApp` Lit lit) ps qs
           _ -> mismatch
       where
-        mismatch = typeError $ WithClausePatternMismatch (namedThing $ unArg p0) (unArg q)
+        mismatch = typeError $ WithClausePatternMismatch (namedArg p0) (unArg q)
     strip tel ps qs = error $ "huh? " ++ show (size tel) ++ " " ++ show (size ps) ++ " " ++ show (size qs)
 
 -- | Construct the display form for a with function. It will display
 --   applications of the with function as applications to the original function.
 --   For instance, @aux a b c@ as @f (suc a) (suc b) | c@
+--
+--   @n@ is the number of with arguments.
 withDisplayForm :: QName -> QName -> Telescope -> Telescope -> Nat -> [Arg Pattern] -> Permutation -> TCM DisplayForm
 withDisplayForm f aux delta1 delta2 n qs perm@(Perm m _) = do
   topArgs <- raise (n + size delta1 + size delta2) <$> getContextArgs
@@ -208,15 +215,20 @@ withDisplayForm f aux delta1 delta2 n qs perm@(Perm m _) = do
   let wild = Def (qualify (mnameFromList []) x) []
 
   let top = genericLength topArgs
-      vs = map (fmap DTerm) topArgs ++ (substs (sub ys wild) $ patsToTerms qs)
+      vs = map (fmap DTerm) topArgs ++ (applySubst (sub ys wild) $ patsToTerms qs)
       dt = DWithApp (DDef f vs : map DTerm withArgs) []
-      withArgs = reverse $ map var [size delta2..size delta2 + n - 1]
-      pats = genericReplicate (n + size delta1 + size delta2 + top) (Var 0 [])
+      withArgs = map var $ genericTake n $ downFrom $ size delta2 + n
+--      withArgs = reverse $ map var [size delta2..size delta2 + n - 1]
+      pats = genericReplicate (n + size delta1 + size delta2 + top) (var 0)
       -- Building the arguments to the with function
       (ys0, ys1) = splitAt (size delta1) (permute perm $ map Just [m - 1, m - 2..0])
       ys = reverse $ ys0 ++ genericReplicate n Nothing ++ ys1
 
   let display = Display (n + size delta1 + size delta2 + top) pats dt
+      addFullCtx = addCtxTel delta1
+                 . flip (foldr addCtxString_) (map ("w" ++) $ map show [1..n])
+                 . addCtxTel delta2
+          -- Andreas 2012-09-17: this seems to be the right order of contexts
 
   reportSDoc "tc.with.display" 20 $ vcat
     [ text "withDisplayForm"
@@ -224,40 +236,39 @@ withDisplayForm f aux delta1 delta2 n qs perm@(Perm m _) = do
       [ text "f      =" <+> text (show f)
       , text "aux    =" <+> text (show aux)
       , text "delta1 =" <+> prettyTCM delta1
-      , text "delta2 =" <+> prettyTCM delta2
+      , text "delta2 =" <+> do addCtxTel delta1 $ prettyTCM delta2
       , text "n      =" <+> text (show n)
       , text "perm   =" <+> text (show perm)
-      , text "top    =" <+> prettyTCM topArgs
+      , text "top    =" <+> do addFullCtx $ prettyTCM topArgs
       , text "qs     =" <+> text (show qs)
-      , text "dt     =" <+> prettyTCM dt
+      , text "dt     =" <+> do addFullCtx $ prettyTCM dt
       , text "ys     =" <+> text (show ys)
       , text "raw    =" <+> text (show display)
-      , text "qsToTm =" <+> prettyTCM (patsToTerms qs)
-      , text "sub qs =" <+> prettyTCM (substs (sub ys wild) $ patsToTerms qs)
+      , text "qsToTm =" <+> prettyTCM (patsToTerms qs) -- ctx would be permuted form of delta1 ++ delta2
+      , text "sub qs =" <+> prettyTCM (applySubst (sub ys wild) $ patsToTerms qs)
       ]
     ]
 
   return display
   where
-    var i = Var i []
-    sub rho wild = map term [0..] -- m - 1]
+    -- Note: The upper bound (m - 1) was previously commented out. I
+    -- restored it in order to make the substitution finite.
+    sub rho wild = parallelS $ map term [0 .. m - 1]
       where
         -- thinking required.. but ignored
         -- dropping the reverse seems to work better
         -- Andreas, 2010-09-09: I DISAGREE.
         -- Ulf, 2011-09-02: Thinking done. Neither was correct.
         -- We had the wrong permutation and we used it incorrectly. Should work now.
-        term i = case findIndex (Just i ==) rho of
-          Nothing -> wild
-          Just j  -> Var (fromIntegral j) []
+        term i = maybe wild var $ findIndex (Just i ==) rho
 
 patsToTerms :: [Arg Pattern] -> [Arg DisplayTerm]
 patsToTerms ps = evalState (toTerms ps) 0
   where
     mapMr f xs = reverse <$> mapM f (reverse xs)
 
-    var :: State Nat Nat
-    var = do
+    nextVar :: State Nat Nat
+    nextVar = do
       i <- get
       put (i + 1)
       return i
@@ -270,7 +281,7 @@ patsToTerms ps = evalState (toTerms ps) 0
 
     toTerm :: Pattern -> State Nat DisplayTerm
     toTerm p = case p of
-      VarP _      -> var >>= \i -> return $ DTerm (Var i [])
+      VarP _      -> nextVar >>= \i -> return $ DTerm (var i)
       DotP t      -> return $ DDot t
       ConP c _ ps -> DCon c <$> toTerms ps
       LitP l      -> return $ DTerm (Lit l)
