@@ -7,13 +7,14 @@ module Agda.Utils.FileName
   , mkAbsolute
   , absolute
   , (===)
+  , doesFileExistCaseSensitive
   , tests
   ) where
 
 import Agda.Utils.TestHelpers
 import Agda.Utils.QuickCheck
 import Data.Function
-import Data.Generics (Typeable, Data)
+import Data.Typeable (Typeable)
 import Data.List
 import Data.Maybe
 import Control.Applicative
@@ -21,6 +22,11 @@ import Control.Monad
 import System.Directory
 import System.FilePath
 
+#if mingw32_HOST_OS
+import Control.Exception (bracket)
+import System.Win32 (findFirstFile, findClose, getFindDataFileName)
+#endif
+
 #include "../undefined.h"
 import Agda.Utils.Impossible
 
@@ -30,7 +36,10 @@ import Agda.Utils.Impossible
 -- paths point to the same files or directories.
 
 newtype AbsolutePath = AbsolutePath { filePath :: FilePath }
-  deriving (Show, Eq, Ord, Typeable, Data)
+  deriving (Eq, Ord, Typeable)
+
+instance Show AbsolutePath where
+  show = filePath
 
 -- | The paths have to be absolute, valid and normalised, without
 -- trailing path separators.
@@ -94,6 +103,22 @@ infix 4 ===
 (===) :: AbsolutePath -> AbsolutePath -> Bool
 (===) = equalFilePath `on` filePath
 
+-- | Case-sensitive doesFileExist for Windows.
+-- This is case-sensitive only on the file name part, not on the directory part.
+-- (Ideally, path components coming from module name components should be
+--  checked case-sensitively and the other path components should be checked
+--  case insenstively.)
+doesFileExistCaseSensitive :: FilePath -> IO Bool
+#if mingw32_HOST_OS
+doesFileExistCaseSensitive f = do
+  ex <- doesFileExist f
+  if ex then bracket (findFirstFile f) (findClose . fst) $
+               fmap (takeFileName f ==) . getFindDataFileName . snd
+        else return False
+#else
+doesFileExistCaseSensitive f = doesFileExist f
+#endif
+
 ------------------------------------------------------------------------
 -- Generators
 
