@@ -22,6 +22,33 @@ mhead :: [a] -> Maybe a
 mhead []    = Nothing
 mhead (x:_) = Just x
 
+-- | Opposite of cons @(:)@, safe.
+uncons :: [a] -> Maybe (a, [a])
+uncons []     = Nothing
+uncons (x:xs) = Just (x,xs)
+
+-- | downFrom n = [n-1,..1,0]
+downFrom :: Integral a => a -> [a]
+downFrom n | n <= 0     = []
+           | otherwise = let n' = n-1 in n' : downFrom n'
+
+-- | Update the last element of a list, if it exists
+updateLast :: (a -> a) -> [a] -> [a]
+updateLast f [] = []
+updateLast f [a] = [f a]
+updateLast f (a : as@(_ : _)) = a : updateLast f as
+
+-- | A generalized version of @partition@.
+--   (Cf. @mapMaybe@ vs. @filter@).
+mapEither :: (a -> Either b c) -> [a] -> ([b],[c])
+{-# INLINE mapEither #-}
+mapEither f xs = foldr (deal f) ([],[]) xs
+
+deal :: (a -> Either b c) -> a -> ([b],[c]) -> ([b],[c])
+deal f a ~(bs,cs) = case f a of
+  Left  b -> (b:bs, cs)
+  Right c -> (bs, c:cs)
+
 -- | Sublist relation.
 isSublistOf :: Eq a => [a] -> [a] -> Bool
 isSublistOf []       ys = True
@@ -67,6 +94,14 @@ holes :: [a] -> [(a, [a])]
 holes []     = []
 holes (x:xs) = (x, xs) : map (id -*- (x:)) (holes xs)
 
+-- | Check whether a list is sorted.
+--
+-- Assumes that the 'Ord' instance implements a partial order.
+
+sorted :: Ord a => [a] -> Bool
+sorted [] = True
+sorted xs = and $ zipWith (<=) (init xs) (tail xs)
+
 -- | Check whether all elements in a list are distinct from each
 -- other. Assumes that the 'Eq' instance stands for an equivalence
 -- relation.
@@ -169,6 +204,13 @@ prop_zipWith' f =
     forAll (two $ vector n) $ \(xs, ys) ->
       zipWith' f xs ys == zipWith f xs ys
 
+-- | Like zipWith, but returns the leftover elements of the input lists.
+zipWithTails :: (a -> b -> c) -> [a] -> [b] -> ([c], [a] , [b])
+zipWithTails f xs       []       = ([], xs, [])
+zipWithTails f []       ys       = ([], [] , ys)
+zipWithTails f (x : xs) (y : ys) = (f x y : zs , as , bs)
+  where (zs , as , bs) = zipWithTails f xs ys
+
 -- | Efficient version of nub that sorts the list first. The tag function is
 --   assumed to be cheap. If it isn't pair up the elements with their tags and
 --   call uniqBy fst (or snd).
