@@ -1,16 +1,20 @@
 {-# LANGUAGE DeriveDataTypeable, CPP #-}
 module Agda.Utils.Permutation where
 
-import Data.Generics (Typeable, Data)
+import Data.Typeable (Typeable)
 import Data.List
 import Agda.Utils.Size
 import Agda.Utils.Impossible
 
 #include "../undefined.h"
 
--- | @permute [2,3,1] [x,y,z] = [y,z,x]@
-data Permutation = Perm Integer [Integer]
-  deriving (Eq, Data, Typeable)
+-- | @permute [1,2,0] [x0,x1,x2] = [x1,x2,x0]@
+--
+--   Agda typing would be:
+--   @Perm : {m : Nat}(n : Nat) -> Vec (Fin n) m -> Permutation@
+--   @m@ is the 'size' of the permutation.
+data Permutation = Perm { permRange :: Int, permPicks :: [Int] }
+  deriving (Eq, Typeable)
 
 instance Show Permutation where
   show (Perm n xs) = showx [0..n - 1] ++ " -> " ++ showx xs
@@ -23,10 +27,13 @@ instance Show Permutation where
 instance Sized Permutation where
   size (Perm _ xs) = size xs
 
--- | @permute [2,3,1] [x1,x2,x3] = [x2,x3,x1]@
+-- | @permute [1,2,0] [x0,x1,x2] = [x1,x2,x0]@
 --   More precisely, @permute indices list = sublist@, generates @sublist@
 --   from @list@ by picking the elements of list as indicated by @indices@.
---   @permute [2,4,1] [x1,x2,x3,x4] = [x2,x4,x1]@
+--   @permute [1,3,0] [x0,x1,x2,x3] = [x1,x3,x0]@
+--
+--   Agda typing:
+--   @permute (Perm {m} n is) : Vec A m -> Vec A n@
 permute :: Permutation -> [a] -> [a]
 permute (Perm _ is) xs = map (xs !!!) is
   where
@@ -34,12 +41,19 @@ permute (Perm _ is) xs = map (xs !!!) is
     (x:xs) !!! 0 = x
     (x:xs) !!! n = xs !!! (n - 1)
 
-idP :: Integer -> Permutation
+idP :: Int -> Permutation
 idP n = Perm n [0..n - 1]
 
-takeP :: Integer -> Permutation -> Permutation
+takeP :: Int -> Permutation -> Permutation
 takeP n (Perm m xs) = Perm n $ filter (< n) xs
 
+-- | @liftP k@ takes a @Perm {m} n@ to a @Perm {m+k} (n+k)@.
+--   Analogous to 'Agda.TypeChecking.Substitution.liftS',
+--   but Permutations operate on de Bruijn LEVELS, not indices.
+liftP :: Int -> Permutation -> Permutation
+liftP n (Perm m xs) = Perm (n + m) $ xs ++ [m..m+n-1]
+-- liftP n (Perm m xs) = Perm (n + m) $ [0..n-1] ++ map (n+) xs -- WRONG, works for indices, but not for levels
+
 -- | @permute (compose p1 p2) == permute p1 . permute p2@
 composeP :: Permutation -> Permutation -> Permutation
 composeP p1 (Perm n xs) = Perm n $ permute p1 xs
@@ -71,11 +85,24 @@ compactP (Perm n xs) = Perm m $ map adjust xs
     holesBelow k = genericLength $ filter (< k) missing
     adjust k = k - holesBelow k
 
+-- | @permute (reverseP p) xs ==
+--    reverse $ permute p $ reverse xs@
+--
+--   Example:
+--   @
+--        permute (reverseP (Perm 4 [1,3,0])) [x0,x1,x2,x3]
+--     == permute (Perm 4 $ map (3-) [0,3,1]) [x0,x1,x2,x3]
+--     == permute (Perm 4 [3,0,2]) [x0,x1,x2,x3]
+--     == [x3,x0,x2]
+--     == reverse [x2,x0,x3]
+--     == reverse $ permute (Perm 4 [1,3,0]) [x3,x2,x1,x0]
+--     == reverse $ permute (Perm 4 [1,3,0]) $ reverse [x0,x1,x2,x3]
+--   @
 reverseP :: Permutation -> Permutation
 reverseP (Perm n xs) = Perm n $ map ((n - 1) -) $ reverse xs
 
 -- | @expandP i n π@ in the domain of @π@ replace the /i/th element by /n/ elements.
-expandP :: Integer -> Integer -> Permutation -> Permutation
+expandP :: Int -> Int -> Permutation -> Permutation
 expandP i n (Perm m xs) = Perm (m + n - 1) $ concatMap expand xs
   where
     expand j
