@@ -280,9 +280,11 @@ a solution CANNOT exist if
 
   v < v  for a rigid variable v
 
+{- Andreas, 2012-09-19 OUTDATED
   v <= v' for rigid variables v,v'
 
   x < v   for a flexible variable x and a rigid variable v
+-}
 
 thus, for each flexible x, only one of the following cases is possible
 
@@ -325,21 +327,27 @@ solve cs = -- trace (show cs) $
          -- compute the sets of flexible and rigid node numbers
          ns  = Map.keys (nodeMap gr)
          -- a set of flexible variables
-         flexs  = foldl (\ l k -> case k of (Flex i) -> i : l
-                                            (Rigid _) -> l) [] ns
+         flexs  = foldl' (\ l k -> case k of (Flex i) -> i : l
+                                             (Rigid _) -> l) [] ns
          -- a set of rigid variables
-         rigids = foldl (\ l k -> case k of (Flex _) -> l
-                                            (Rigid i) -> i : l) [] ns
+         rigids = foldl' (\ l k -> case k of (Flex _) -> l
+                                             (Rigid i) -> i : l) [] ns
 
          -- rigid matrix indices
-         rInds = foldl (\ l r -> let Just i = Map.lookup (Rigid r) (nodeMap gr)
-                                 in i : l) [] rigids
+         rInds = foldl' (\ l r -> let Just i = Map.lookup (Rigid r) (nodeMap gr)
+                                  in i : l) [] rigids
 
          -- check whether there is a solution
          -- d   = [ m!(i,i) | i <- [0 .. (n-1)] ]  -- diagonal
 -- a rigid variable might not be less than it self, so no -.. on the
 -- rigid part of the diagonal
-         solvable = all (\ x -> x >= Finite 0) [ m!(i,i) | i <- rInds ] &&
+         solvable = all (\ x -> x >= Finite 0) [ m!(i,i) | i <- rInds ] && True
+
+{-  Andreas, 2012-09-19
+    We now can have constraints between rigid variables, like i < j.
+    Thus we skip the following two test.  However, a solution must be
+    checked for consistency with the constraints on rigid vars.
+
 -- a rigid variable might not be bounded below by infinity or
 -- bounded above by a constant
 -- it might not be related to another rigid variable
@@ -357,7 +365,7 @@ solve cs = -- trace (show cs) $
                     edge = m!(row,col)
                 in  edge >= Finite 0)
              [ (x,v) | x <- flexs, (RVar v) <- rigids ]
-
+-}
 
          inScope :: FlexId -> Rigid -> Bool
          inScope x (RConst _) = True
@@ -378,7 +386,7 @@ while flexible variables and rigid rows left
          loop1 flxs (r:rgds) subst =
             let row = fromJust $ Map.lookup (Rigid r) (nodeMap gr)
                 (flxs',subst') =
-                  foldl (\ (flx,sub) f ->
+                  foldl' (\ (flx,sub) f ->
                           let col = fromJust $ Map.lookup (Flex f) (nodeMap gr)
                           in  case (inScope f r, m!(row,col)) of
 --                                Finite z | z <= 0 ->
@@ -518,9 +526,4 @@ prop_stable n' =
     g =~= g' = sort (edges g) == sort (edges g')
 
 tests :: IO Bool
-tests = runTests "Agda.Utils.Warshall"
-  [ quickCheck' prop_smaller
-  , quickCheck' prop_path
-  , quickCheck' prop_disjoint
-  , quickCheck' prop_stable
-  ]
+tests = runTests "Agda.Utils.Warshall" []
